/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Original Code is the Sunlabs Mobile IP.
 The Initial Developer of the Original Code is Vipul Gupta.
 Portions created by Vipul Gupta are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): Niko A. Fikouras (niko@comnets.uni-bremen.de)
		 Asanga D. Udugama (adu@comnets.uni-bremen.de)

*/

/*
 * An implementation of Mobile-IP as described in RFC 2002.
 * Author: Vipul Gupta (vgupta@eng.sun.com)
 * agent.h : Data structures and constants used by a Mobile IP agent.
 */

#ifndef _AGENT_H
#define _AGENT_H

#pragma ident "@(#)agent.h 1.3 97/09/26"

#ifdef __cplusplus
extern "C" {
#endif

#include "mip.h"

/*
 * This file contains definitions for structures used by
 * Mobility Agents (either Home Agents or Foreign Agents)
 */

#define MAX_SIMULTANEOUS_BINDINGS    7
#define DEFAULT_MAX_INTERFACES       4
#define DEFAULT_MAX_REG_TIME       300
#define DEFAULT_MAX_ADV_TIME       300
#define DEFAULT_MAX_INTERVAL        10
#define DEFAULT_MIN_INTERVAL         1

#define DEFAULT_PERIODIC_INTERVAL    1
#define DEFAULT_GRANULARITY     100000
#define DEFAULT_FRESHNESS_SLACK    300

#define MIP_PORT                   434

//#define CONF_FILE_NAME             "/etc/opt/SUNWmipma/mipagent.conf"
#define CONF_FILE_NAME             "/etc/mipagent.conf"
//#define CONF_FILE_NAME             "/home/adu/mipagent.conf"

/* Maximum number of mobility bindings maintained at a Home Agent */
#define DEFAULT_MAX_BINDINGS        32

/* Maximum number of mobile nodes supported by a Home Agent */
#define DEFAULT_MAX_MOBILE_NODES    32

/*
 * Maximum number of other nodes with which this mobility agent shares
 * a mobile security association.
 */
#define DEFAULT_MAX_OTHER_NODES     16

/*
 * Unreachability info is kept only for the last DEFAULT_MAX_UNREACHABLES
 * nodes and such info expires in DEFAULT_UNREACHABLE_TIME.
 */
#define DEFAULT_MAX_UNREACHABLES    16
#define DEFAULT_UNREACHABLE_TIME    30

/*
 * Maximum number of visitor entries (accepted + pending) maintained
 * at a Foreign Agent. A foreign agent sets the busy bit 'B' in its
 * advertisements when number of visitors reaches DEFAULT_HIGH_VISITORS
 * and does not reset it till it drops below DEFAULT_LOW_VISITORS.
 * A pending entry is kept in the visitor table for at most
 * DEFAULT_VISITOR_EXPIRY seconds.
 */
#define DEFAULT_MAX_VISITORS        32
#define DEFAULT_HIGH_VISITORS       28
#define DEFAULT_LOW_VISITORS        24
#define DEFAULT_VISITOR_EXPIRY      30



/* Max number of address intervals for specifying protected domain */
#define MAX_ADDR_INTERVALS             6

/* Max. number of firewalls */
#define MAX_FIREWALLS                  3


#define MAX_IFNAME_LEN               8
#define MAX_HWADDR_LEN               8
/* One such entry for each mobility-supporting interface */
typedef struct mace  {
    u32             maIfaceAddr;
    u32             maIfaceNetmask;
    unsigned char   maIfaceHWaddr[MAX_HWADDR_LEN]; /* used in proxy-ARPs */
    char            maIfaceName[MAX_IFNAME_LEN];
    int             maIfaceUnicastSock; /* UDP socket bound to IfaceAddr */
    int             maIfaceBcastSock;   /* ... bound to subnet bcastAddr */
    u32             maAdvMaxRegLifetime;
    u32             maAdvAddr;
    u32             maAdvMaxInterval;
    u32             maAdvMinInterval;
    u32             maAdvMaxAdvLifetime;
/*  boolean         maAdvResponseSolicitationOnly; */
    unsigned short  maAdvSeqNum;
    char            maAdvServiceFlags;  /* H,F,G,M,V flags */
    boolean         maAdvPrefixLenInclusion;
} MaAdvConfigEntry;

/* One entry for each visiting mobile node at a foreign agent. */
typedef struct fave {
    u32             faVisitorAddr;
    u32             faVisitorIfaceAddr;       /* interface addr through
					       * which visitor is reachable
					       */
    boolean         faVisitorRegIsAccepted;
    char	    faVisitorRegFlags;
    unsigned short  faVisitorPort;
    u32             faVisitorHomeAddr;
    u32             faVisitorHomeAgentAddr;
    u32             faVisitorCOAddr;          /* COaddr field in request */
    u32             faVisitorTimeGranted;     /* only valid if IsAccepted */
    u32             faVisitorTimeRemaining;   /* only valid if IsAccepted */
    u32             faVisitorRegIDHigh;
    u32             faVisitorRegIDLow;
} FaVisitorEntry;

/* Entry for a currently unreachable node maintained at a foreign agent */
typedef struct faue {
    u32             faUnreachableAddr;
    u32             faUnreachableTimeRemaining; /* Entries with zero time
						 * are empty/unused */
} FaUnreachableEntry;

/*
 * The Home Agent maintains one such entry for each mobility binding.
 * A single mobile node may have multiple bindings.
 */
typedef struct hambe {
     u32            haBindingMN;   /* Mobile node's home address */
     u32            haBindingCOA;  /* Mobile node's care-of address */
     u32            haBindingSrcAddr;
/*   u32            haBindingRegIDLow;   This is not binding specific */
/*   u32            haBindingRegIDHigh;  but rather MN specific       */
     u32            haBindingTimeGranted;
     u32            haBindingTimeRemaining;
     unsigned short haBindingSrcPort;
     char           haBindingRegFlags;
} HaBindingEntry;

/*
 * The Home Agent maintains one such entry for each supported Mobile
 * Node. In addition, the Home Agent keeps a MipSecAssocEntry for
 * for every other node with which it shares a mobility security
 * association. By using the interface information, the home agent
 * can filter out broadcast packets that it picks up from subnets
 * other than the mobile node's home subnet.
 */
typedef struct hamne {
    u32             haMnAddr;            /* Mobile Node's IP address */
    u32             haBindingIfaceAddr;  /* interface on the mobile
					  * node's home subnet */
    u32             haMnRegIDHigh;       /* Stored ID used in replay
				          * protection */
    u32             haMnRegIDLow;
    MipSecAssocEntry haMnMSA;
    unsigned char   haMnBindingCnt;      /* Number of current bindings */
} HaMobileNodeEntry;

/* Data structure to store the information regarding protected domain */
typedef struct domaininfo {
    u32 addr[MAX_ADDR_INTERVALS];
    u32 netmask[MAX_ADDR_INTERVALS];
    u32 addrIntervalCnt;
    u32 fwAddr[MAX_FIREWALLS];
    u32 firewallCnt;
} DomainInfo;


/* Data Structures for filters */
#ifndef MAX_FILTERS
#define MAX_FILTERS			10
#endif

#define MAX_FILTER_MODULES_PER_FILTER	5
#define DEFAULT_FILTER_SEQ		255

// structure for a filter module
typedef struct filtermodule {
	int filterType; 					   // type of the filter : as defined in FLT_ flags
	char *filterPtr; 					   // pointer to the filter data : can be DSCP type,
								   // port type, src addr type, src net type,
								   // src port type, src range type, load balancing
								   // type or free form type : would be one, from
			 					   // the following structures
} FilterModule;

// structure to hold distributions for
// eacf CoA
typedef struct controlmodule {
	u32 boundCoA;						   // currently bound CoA, not valid
								   // when FA
	int weight; 						   // indicates what the amount of
								   // packets distributed
}  ControlModule;


// structure for one filter
typedef struct filter {
	int filterModuleCount; 					   // filter count per module
	FilterModule filterModules[MAX_FILTER_MODULES_PER_FILTER]; // filter modules of each filter,
								   // each filter module is
								   // interpretted as AND
								   // against each other
	ControlModule controlModules[MAX_FILTER_MODULES_PER_FILTER]; // info of distributions per
								     // per each given CoA
	int seq;						   // the sequence number assigned
								   // to the filter
	int markValue;						   // value used to mark packets
} Filter;

// structure to specify the filter of each connected MN
typedef struct boundmn {
	u32 mnAddr;						   // Address of the MN
	int mnFilterCount; 					   // the number of filters
	Filter mnFilters[MAX_FILTERS]; 				   // each filter
	int filtersChanged; 					   // TRUE or FALSE, to indicate whether
								   // any filter data changed (so MN can
			  					   // retransmit)
} BoundMN;

// structures that hold filter data
typedef struct bafextdatadata {
	int DSCP;
} BAFExtData;

typedef struct protoextdata {
	int protoNum;
} ProtoExtData;

typedef struct srcaddrextdata {
	u32 srcAddr;
} SrcAddrExtData;

typedef struct srcnetextdata {
	u32 srcAddr;
	u32 srcNetMask;
} SrcNetExtData;

typedef struct srcportextdata {
	short int portNum;
} SrcPortExtData;

typedef struct srcportrangeextdata {
	short int minPortNum;
	short int maxPortNum;
} SrcPortRangeExtData;

typedef struct freeformextdata {
	short int offset;
	int length; // in nibbles
	char value[MAX_FREEFORM_VAL_SIZE + 1];
	char mask[MAX_FREEFORM_MASK_SIZE + 1];
} FreeFormExtData;

// new filters

typedef struct dstaddrextdata {
	u32 dstAddr;
} DstAddrExtData;

typedef struct dstnetextdata {
	u32 dstAddr;
	u32 dstNetMask;
} DstNetExtData;

typedef struct dstportextdata {
	short int portNum;
} DstPortExtData;

typedef struct dstportrangeextdata {
	short int minPortNum;
	short int maxPortNum;
} DstPortRangeExtData;



// structure to hold a mark assigned to a mnaddr & coaddr
// used in iptables
typedef struct bindingmark {
	u32 mnaddr;
	int fltSeq;
} BindingMark;

// structures to filter operation info until they are
// applied
#define FLTOP_DELETE	1
#define FLTOP_ADD	2
#define FLTOP_SHARE	3

typedef struct filteropentry {
	int opCode;         // 3 = SHARE the flow, 2 = APPLY filter, 1 = DELETE filter
	u32 boundCoA;
	int seq;
	int markValue;
} FltOpEntry;

typedef struct mnfltopentry {
	u32 mnaddr;
	FltOpEntry fltOpTable[MAX_FILTERS];
} MNFltOpEntry;

#ifdef __cplusplus
}
#endif

#endif /* _AGENT_H */
