/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Original Code is the Sunlabs Mobile IP.
 The Initial Developer of the Original Code is Vipul Gupta.
 Portions created by Vipul Gupta are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): Niko A. Fikouras (niko@comnets.uni-bremen.de)
		 Asanga D. Udugama (adu@comnets.uni-bremen.de)

*/

/*
 * An implementation of Mobile-IP as described in RFC 2002.
 * Author: Vipul Gupta (vgupta@eng.sun.com)
 * agentNet.c : Networking-related routines used by a mobility agent. 
 *              These may need be re-implemented for non-UNIX platforms
 *              but should require little work to port to different 
 *              flavors of UNIX.
 */

#pragma ident	"@(#)agentNet.c 1.9 97/12/19"

#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#include "mip.h"
#include "agent.h"

/*
 * Maximum anticipated pkt size for any registration/reply/icmp pkt
 * of interest to a mobility agent.
 */
#define  MAX_PKT_SIZE     2048
#define  ON_UNICAST_SOCK     0 
#define  ON_BCAST_SOCK       1
#define  MIN_IP_HDR_LEN     20
 
extern int  maAdvConfigTableSize;
extern u32  maAdvSentCnt;
extern u32  maAdvSentForSolicitationsCnt;
extern u32  maSolicitationsRecvdCnt;
extern int  ICMPsock;
extern int  BcastSock;
extern int  mipdebug;
extern char msg[], addr1[], addr2[];
extern int  errno;                      /* from <errno.h> */
extern MaAdvConfigEntry maAdvConfigTable[DEFAULT_MAX_INTERFACES];
extern boolean faBusy;

extern char *ntoa(u32, char *);
extern void maSendAdvertisements(MaAdvConfigEntry *, int, boolean);
extern FAprocessRegRequest(unsigned char *, int, u32, unsigned short, int);
extern HAprocessRegRequest(unsigned char *, int, u32, unsigned short, int);
extern FAprocessRegReply(unsigned char *, int, u32, unsigned short, int);

/* 
 * Send a UDP message from sock, to dst address and port dstPort. The pkt
 * is contained in pkt and is of length pktLen. dst is already in network
 * byte order.
 */
int sendUDPmessage(int sock, unsigned char *pkt,  int pktLen,
		   u32 dst, unsigned short dstPort)
{
    struct sockaddr_in sa;
    
    sa.sin_family = AF_INET;
    sa.sin_port = htons(dstPort);
    sa.sin_addr.s_addr = dst;

    /* send the message */
    if (sendto(sock, (char *) pkt, pktLen, 0, (struct sockaddr *) &sa,
	       sizeof(struct sockaddr)) < 0) {
        perror("sendto()");
        mip0dbg(("Sendto failed in sendUDPmessage.\n"));
	return (-1);
    }   

    return(0);
}

/* Send an ICMP message contained in data to dst from src */
void sendICMPmessage(int s, u32 src, u32 dst, unsigned char data[], int len)
{
    struct sockaddr_in sa;
    struct in_addr ifaddr;

    sa.sin_family = AF_INET;
    sa.sin_port = 0;
    sa.sin_addr.s_addr = dst;

    /* 
     * Make sure pkts to 224.0.0.0 are sent through the interface
     * represented by the src address.
     */

    if (dst == inet_addr("224.0.0.1")) {
        /*
	 * On Linux:
	 * sprintf(msg, "route add 224.0.0.0 %s 0", ntoa(src, addr1));
	 * system(msg);
	 */

        ifaddr.s_addr = src;
	if (setsockopt(s, IPPROTO_IP, IP_MULTICAST_IF,
		       (char *)&ifaddr, sizeof (ifaddr)) == -1) {
	    perror("setsockopt()");
	    mip0dbg(("Couldn't control iface for multicast addr.\n"));
	}
    }

    /* send the message */
    if (sendto(s, (char *) data, len, 0, (struct sockaddr *) &sa,
	       sizeof(struct sockaddr)) < 0) {
        perror("sendto()");
        mip0dbg(("Sendto failed in sendICMPmessage.\n"));
    }

    if (dst == inet_addr("224.0.0.1")) {
        /*
	 * On Linux:
	 * sprintf(msg, "route delete 224.0.0.0 %s", ntoa(src, addr1));
	 * system(msg);
	 */
    }
}

/* 
 * Pkt contains a complete IP datagram (including IP header) received
 * on a socket monitoring ICMP packets.
 */
void screenICMPpkt(unsigned char pkt[], int pktLen, u32 src)
{
    icmph *icmpPtr;
    unsigned char ipByte;
    int ipHdrLen;
    u32 *ipDst;

    if (pktLen < MIN_IP_HDR_LEN + sizeof(icmph)) return;

    ipByte = pkt[0];
    ipHdrLen = (ipByte & 0x0f) << 2;
    if (pktLen < ipHdrLen + sizeof(icmph)) return;
    icmpPtr = (icmph *) (pkt + ipHdrLen);

    /* Too much chatter ...  */

     ipDst = (u32 *) &pkt[16];
     if (icmpPtr->type == 10) {
         mip3dbg(("Got ICMP solicitation <type %d, code %d> from %s to %s.\n",
		  icmpPtr->type, icmpPtr->code, ntoa(src, addr1),
		  ntoa(*ipDst, addr2))); 
     }

    switch (icmpPtr->type) {
    case ICMP_ROUTER_SOLICITATION:
        if (icmpPtr->code == 0) {
	    maSolicitationsRecvdCnt++;
	    /* TO DO:
	     * Restrict advertisement to the particular interface on 
	     * which the solicitation was received. Since there is
	     * no easy way of figuring out this interface, advertise
	     * on all mobility supporting interfaces for now.
	     */
	    maSendAdvertisements(maAdvConfigTable, maAdvConfigTableSize, faBusy);
	    /* should actually be manipulated inside maSendAdvertisements() */
	    maAdvSentForSolicitationsCnt++;
	}
	break;

    case ICMP_DEST_UNREACHABLE:
        /* TO DO:
	 * Make a note of the unreachable node in faUnreachableTable
	 * and send registration denials for pending visitor entries
	 * waiting to use the unreachable node as their home agent.
	 */
        break;

    default:
        break;
    }
}

void screenUDPpkt(unsigned char pkt[], int pktLen, u32 src,
		  unsigned short srcPort, int ifaceIdx, int addrFlag)
{
    regRequest *regReqPtr;
    regReply *regRepPtr;
    unsigned char *cp;
    u32    inAddr;
    int i = ifaceIdx;

    if (addrFlag == ON_UNICAST_SOCK) 
        inAddr = maAdvConfigTable[i].maIfaceAddr;
    else {
        inAddr = (maAdvConfigTable[i].maIfaceAddr &
		  maAdvConfigTable[i].maIfaceNetmask) |
	         ~maAdvConfigTable[i].maIfaceNetmask;
    }

    cp = pkt;
    switch (*cp) {
    case REG_REQUEST_TYPE:
        if (pktLen < sizeof(regRequest)) {
	    mip3dbg(("Received registration request is too short.\n"));
	    return;
	}
        regReqPtr = (regRequest *) pkt;

	if ((maAdvConfigTable[i].maAdvServiceFlags & ADV_IS_HOME_AGENT) &&
	    (regReqPtr->haAddr == inAddr)) {
	    /* process as HA */
	    HAprocessRegRequest(pkt, pktLen, src, srcPort, i);
	} else if ((maAdvConfigTable[i].maAdvServiceFlags & 
		    ADV_IS_FOREIGN_AGENT) && (regReqPtr->COAddr == inAddr) &&
		   (addrFlag == ON_UNICAST_SOCK)) {
	    /* process as FA */
	    FAprocessRegRequest(pkt, pktLen, src, srcPort, i);
	} else {
	    mip2dbg(("Not configured to handle this reg req on addr %s" \
		     " from %s.\n", ntoa(inAddr, addr1), ntoa(src, addr2)));
	}
	
	break;

    case REG_REPLY_TYPE:
        if (pktLen < sizeof(regReply)) {
	    mip3dbg(("Received registration reply is too short.\n"));
	    return;
	}
        regRepPtr = (regReply *) pkt;

	/* Too much chatter ...
	 * mip2dbg(("Got reg reply on addr %s from %s.\n", 
	 *          ntoa(inAddr, addr1), ntoa(src, addr2)));
	 */

	FAprocessRegReply(pkt, pktLen, src, srcPort, i);
	break;

    default:
        mip3dbg(("Unknown UDP message (first byte 0x%x) from %s.\n",
		*cp, ntoa(src, addr1)));
    } 
}

void getAndDispatchMessages()
{
    int i;
    int pktLen;
    int tempSock;
    static char pkt[MAX_PKT_SIZE];
    static fd_set fdvec;
    static struct timeval tv;
    int fromLen;
    struct sockaddr_in from;
    u32 src;
    unsigned short srcPort;
    regRequest *regReqPtr;

    /* prepare the vector of file descriptors we'd like to monitor */
    FD_ZERO(&fdvec);
    FD_SET(ICMPsock, &fdvec);
    FD_SET(BcastSock, &fdvec);
    for (i = 0; i < maAdvConfigTableSize; i++) {
        FD_SET(maAdvConfigTable[i].maIfaceUnicastSock, &fdvec);
        FD_SET(maAdvConfigTable[i].maIfaceBcastSock, &fdvec);
    }

    tv.tv_sec = tv.tv_usec = 0;

    /* "if" changed to "while" based on testing under Linux */
    while (select(FD_SETSIZE, &fdvec, NULL, NULL, &tv) < 0) {
        perror("select()");
        if ((errno != EINTR) && (errno != ERESTART)) {
	    mip0dbg(("select failed but not due to EINTR or ERESTART.\n"));
	}
    }

    /* process data received on ICMP socket */
    if (FD_ISSET(ICMPsock, &fdvec)) {
	pktLen = MAX_PKT_SIZE;
	fromLen = sizeof(struct sockaddr_in);
	if ((pktLen = recvfrom(ICMPsock, pkt, pktLen, 0, 
			       (struct sockaddr *) &from, &fromLen)) < 0) {
	    perror("recvfrom failed for ICMP msg");
	} else {
	    src = from.sin_addr.s_addr;
	    screenICMPpkt((unsigned char *)pkt, pktLen, src);
	}
    }

    /* process data on BcastSock bound to 255.255.255.255, port 434 */
    if (FD_ISSET(BcastSock, &fdvec)) {
        pktLen = MAX_PKT_SIZE;
	fromLen = sizeof(struct sockaddr_in);
	if ((pktLen = recvfrom(BcastSock, pkt, pktLen, 0, 
			       (struct sockaddr *) &from, &fromLen)) < 0) {
	    perror("recvfrom failed for UDP on bcast socket");
	} else {
	    src = from.sin_addr.s_addr;
	    srcPort = ntohs(from.sin_port);
	    if (pktLen < sizeof(regRequest)) {
	        mip3dbg(("Packet on 255.255.255.255 is too short for a " \
			 "registration request ... ignored.\n"));
	    }

	    /* 
	     * TODO: figure out the interface on which this packet arrived.
	     *       For now, we look at the HAaddr field and match that
	     *       against the broadcast address of all of our mobility
	     *       supporting interfaces. The iface that matches is the
	     *       one we assume it came in on.
	     */

	    regReqPtr = (regRequest *) pkt;
	    for (i = 0; i < maAdvConfigTableSize; i++) {
	        if (regReqPtr->haAddr == 
		    ((maAdvConfigTable[i].maIfaceAddr &
		      maAdvConfigTable[i].maIfaceNetmask) |
		     ~maAdvConfigTable[i].maIfaceNetmask)) 
		    break;
	    }

	    if (i < maAdvConfigTableSize) {
	        mip0dbg(("Warning: assuming pkt on 255.255.255.255 " \
			 "port 434 came in through %s.\a\n", 
			 maAdvConfigTable[i].maIfaceName));
		screenUDPpkt((unsigned char *)pkt, pktLen, src, srcPort,
			     i, ON_BCAST_SOCK);
	    } else {
	        mip0dbg(("Warning: assuming pkt on 255.255.255.255 " \
			 "port 434 came in through %s.\a\n", 
			 maAdvConfigTable[0].maIfaceName));
		screenUDPpkt((unsigned char *)pkt, pktLen, src, srcPort,
			     0, ON_BCAST_SOCK);
	    }
	}
    }

    /* for each mobility supporting interface ... */
    for (i = 0; i < maAdvConfigTableSize; i++) {
        /* ... process data on UDP socket bound to unicast address */
        if (FD_ISSET(maAdvConfigTable[i].maIfaceUnicastSock, &fdvec)) {
	    tempSock = maAdvConfigTable[i].maIfaceUnicastSock;
	    pktLen = MAX_PKT_SIZE;
	    fromLen = sizeof(struct sockaddr_in);
	    if ((pktLen = recvfrom(tempSock, pkt, pktLen, 0, 
				   (struct sockaddr *) &from, &fromLen)) < 0) {
	        perror("recvfrom failed for UDP on unicast socket");
	    } else {
	        src = from.sin_addr.s_addr;
	        srcPort = ntohs(from.sin_port);
	        screenUDPpkt((unsigned char *)pkt, pktLen, src, srcPort,
			     i, ON_UNICAST_SOCK);
	    }
	}

        /* ... and UDP socket bound to a broadcast address */
        if (FD_ISSET(maAdvConfigTable[i].maIfaceBcastSock, &fdvec)) {
	    tempSock = maAdvConfigTable[i].maIfaceBcastSock;
	    pktLen = MAX_PKT_SIZE;
	    fromLen = sizeof(struct sockaddr_in);
	    if ((pktLen = recvfrom(tempSock, pkt, pktLen, 0, 
				   (struct sockaddr *) &from, &fromLen)) < 0) {
	        perror("recvfrom failed for UDP on bcast socket");
	    } else {
	        src = from.sin_addr.s_addr;
	        srcPort = ntohs(from.sin_port);
	        screenUDPpkt((unsigned char *)pkt, pktLen, src, srcPort,
			     i, ON_BCAST_SOCK);
	    }
	}
    }
}
