/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Original Code is the Sunlabs Mobile IP.
 The Initial Developer of the Original Code is Vipul Gupta.
 Portions created by Vipul Gupta are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): Niko A. Fikouras (niko@comnets.uni-bremen.de)
		 Asanga D. Udugama (adu@comnets.uni-bremen.de)

*/

/*
 * An implementation of Mobile-IP as described in RFC 2002.
 * Author: Vipul Gupta (vgupta@eng.sun.com)
 * agentPeriodic.c : Routines implementing periodic chores carried out by
 *                   a mobility agent.
 */

#pragma ident   "@(#)agentPeriodic.c 1.5 99/04/16"

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#include "mip.h"
#include "agent.h"

/* global variables declared here */
extern boolean faBusy;

/* Common to all mobility agents */
extern MaAdvConfigEntry maAdvConfigTable[DEFAULT_MAX_INTERFACES];
extern int maAdvConfigTableSize;

extern u32  maAdvSentCnt;
extern u32  maAdvSentForSolicitationsCnt;
extern u32  maSolicitationsRecvdCnt;
extern int  ICMPsock;

/* Foreign Agent specific data structures. */

extern FaVisitorEntry faVisitorTable[DEFAULT_MAX_VISITORS];
extern int faVisitorTableSize;

extern FaUnreachableEntry faUnreachableTable[DEFAULT_MAX_UNREACHABLES];
extern int faUnreachableTableSize;

/* Home Agent specific data structures. */

extern HaMobileNodeEntry haMobileNodeTable[DEFAULT_MAX_MOBILE_NODES];
extern int haMobileNodeTableSize;

extern HaBindingEntry haBindingTable[DEFAULT_MAX_BINDINGS];
extern int haBindingTableSize;

extern DomainInfo domainInfo;

/* Other external declarations */
extern int  mipdebug;
extern int  regLifetime;
extern int  advLifetime;
extern int  periodicInterval;
extern char msg[], addr1[], addr2[];
extern unsigned short inChecksum(unsigned short *, int);
extern char * ntoa(u32, char *);
extern void sendICMPmessage(int, u32, u32, unsigned char *, int);
extern char *err2str(int);
extern boolean isOutsideProtectedDomain(u32);

/* to store the filters that have been defined for each connected mobile node */
extern BoundMN boundMNTable[DEFAULT_MAX_BINDINGS];			   // one for each MN bound
extern int boundMNTableSize;

/* OS dependent function prototypes */
extern int encapadd(u32, u32, u32, unsigned short);
extern int encaprem(u32, u32 );
extern int decapadd(u32, u32, u32, char *);
extern int decaprem(u32, u32, u32);
extern int routeadd(u32, u32, int);
extern int routedel(u32, u32);
extern int arpadd(u32, u_char *, char *);
extern int arpdel(u32);
extern int arprefresh(u32);
extern int mkroutethroughtunnel(u32, u32);
extern int rmroutethroughtunnel(u32, u32);
extern int removeSelectedFilters(BoundMN [], int *, u32, u32);


/* Prepare and send agent advertisement on each supported interface. */
void maSendAdvertisements(MaAdvConfigEntry maceTable[],
			  int maceTableSize, boolean faBusy)
{
    int     i, j;
    static unsigned char advPkt[256];
    int     len;
    icmph   *icmpPtr;
    u32     *addrPtr;
    advExt  *advExtPtr;
    unsigned char *cp;

    /* Fill out the ICMP header for a router advertisement */
    icmpPtr = (icmph *) advPkt;
    icmpPtr->type = ICMP_ROUTER_ADVERTISEMENT;
    icmpPtr->code = 16;
    icmpPtr->icmpAdvNumAddr = 0;
    icmpPtr->icmpAdvAddrEntrySize = 2;
    icmpPtr->icmpAdvLifetime = htons(advLifetime);

    for (i = 0; i < maceTableSize; i++) {
        /* For each mobility supporting interface, create advertisement ... */
        len = sizeof(icmph);
	addrPtr = (u32 *) (advPkt + sizeof(icmph));

	/*
	 * ... fill (optional) addresses in the RFC 1256 portion 
	 * we only fill out one address.
	 */
	icmpPtr->icmpAdvNumAddr = 1;
	icmpPtr->icmpAdvAddrEntrySize = 2;
	*addrPtr = maceTable[i].maIfaceAddr;
	addrPtr++;
	len += sizeof(u32);
	*addrPtr = htonl(0x80000000);
	addrPtr++;
	len += sizeof(u32);

	/*
	 * ... fill out the mobility agent advertisement extension 
	 * we only advertise one address
	 */
	advExtPtr = (advExt *) (advPkt + len);
	advExtPtr->type = ADV_EXT_TYPE;
	advExtPtr->length = 10;
	advExtPtr->seqNum = htons(maceTable[i].maAdvSeqNum);
	if (++maceTable[i].maAdvSeqNum == 0)
	    maceTable[i].maAdvSeqNum = 256;
	advExtPtr->regLifetime = htons(maceTable[i].maAdvMaxRegLifetime);
	advExtPtr->advFlags = maceTable[i].maAdvServiceFlags;

	/* set the busy bit, if appropriate */
	if (faBusy && (advExtPtr->advFlags & ADV_IS_FOREIGN_AGENT))
	    advExtPtr->advFlags |= ADV_IS_BUSY;

	advExtPtr->reserved = 0;
	len += sizeof(advExt);

	/* fill out the advertised care-of address */
	addrPtr = (u32 *) (advPkt + len);
	*addrPtr = maceTable[i].maIfaceAddr;
	addrPtr++;
	len += sizeof(u32);

	/* fill out prefix length extension, if required */
	if ((maceTable[i].maAdvPrefixLenInclusion == TRUE) &&
	    icmpPtr->icmpAdvNumAddr) {
	    cp = advPkt + len;
	    *cp = ADV_PREFIX_EXT_TYPE;
	    cp++;
	    *cp = icmpPtr->icmpAdvNumAddr;
	    cp++;
	    len += 2;
	    /* fill out prefix length for each addr in RFC1256 portion */
	    for (j = 0; j < icmpPtr->icmpAdvNumAddr; j++) {
	        *cp = prefixLen(maceTable[i].maIfaceNetmask);
		cp++;
		len++;
	    }
	}

	/* pad ICMP message to even length, if required */
	if (len % 2) {  
	    *cp = ADV_PADDING_EXT_TYPE;
	    cp++;
	    len++;
	}

	/* fill out ICMP checksum */
	icmpPtr->checksum = 0;
	icmpPtr->checksum = inChecksum((unsigned short *) icmpPtr, len);

	/* ship it! */
	maAdvSentCnt++;
	sendICMPmessage(ICMPsock, maceTable[i].maIfaceAddr, 
			maceTable[i].maAdvAddr, advPkt, len);
    }
}

/* find a haMobileNodeEntry in hamnTable with addr equal to mnAddr */
HaMobileNodeEntry * findHAMNE(HaMobileNodeEntry hamnTable[],
			    int hamnTableSize, u32 mnAddr)  {
    int toSee = hamnTableSize;
    int seen = 0;			       
    int i;

    for (i = 0; (i < DEFAULT_MAX_MOBILE_NODES) && (seen < toSee); i++) {
        if (hamnTable[i].haMnAddr) {
	    /* Entry in use. */
	    seen++;
	    if (hamnTable[i].haMnAddr == mnAddr)
	        return(&hamnTable[i]);
	}
    }

    return(NULL);
}

/* Delete the binding entry at index idx in habeTable */
void delHABEidx(HaBindingEntry habeTable[], int *habeTableSize, int idx)
{
    u32 mnAddr, coAddr;
    HaMobileNodeEntry *mnePtr;
    int val;

    mnAddr = habeTable[idx].haBindingMN;
    coAddr = habeTable[idx].haBindingCOA;

    /*
     * Disable encapsulation to this coaddr. Decrease
     * binding count for MN. If zero, cancel proxy ARP & route
     */
    mnePtr = findHAMNE(haMobileNodeTable, haMobileNodeTableSize, mnAddr);

    if (mnePtr == NULL) {
        mip0dbg(("Error: No MN entry ... in delHABEidx().\n"));
	exit(-1);
    } else  {
        /*
	 * If the care-of-address was outside protected domain remove the
	 * encapsulation through the firewall's internal address.
	 * TODO: This works only if we assume that a single external COAddr
	 *       is never shared by multiple mobile nodes.
         */

        if (isOutsideProtectedDomain(coAddr)) {
            mip3dbg(("Removing tunnel for %s through FW %s\n",
		     ntoa(coAddr, addr1),
		     ntoa(domainInfo.fwAddr[0], addr2)));
            if ((val = encaprem(coAddr, domainInfo.fwAddr[0])) < 0) {
	        mip0dbg(("encaprem failed ... %s\n", err2str(val)));
	    }

	    if ((val = rmroutethroughtunnel(coAddr,
			mnePtr->haBindingIfaceAddr)) < 0) {
		        mip0dbg(("rmroutethroughtunnel failed ... %s\n",
			 	err2str(val)));
	    }
        }


	/* Terminate encapsulation of MN's packets to COA */
        mip3dbg(("Removing tunnel for %s through %s\n",
		 ntoa(mnAddr, addr1), ntoa(coAddr, addr2)));
        if ((val = encaprem(mnAddr, coAddr)) < 0) {
            sprintf(msg, "encaprem failed ... %s\n", err2str(val));
            mip0dbg((msg));
        }

	if (--mnePtr->haMnBindingCnt == 0) {
	    /* Terminate decapsulation service */
	    mip3dbg(("Terminating decapsulation service for %s\n",
		     ntoa(mnAddr, addr1)));
	    if ((val = decaprem(coAddr, mnePtr->haBindingIfaceAddr, mnAddr)) < 0)
 	        mip0dbg(("decaprem failed ... %s\n", err2str(val)));

	    /*
	     * TODO:
	     * Solaris gets confused if it has two ARP entries
	     * for the same host on different interfaces. This
	     * needs further looking into (atleast it tries to
	     * keep a separate ARP table for each interface).
	     */
	    sprintf(msg, "Removing proxy ARP entry for %s\n",
		    ntoa(mnAddr, addr1));
	    mip3dbg((msg));
	    if ((val = arpdel(mnAddr)) < 0) {
	        sprintf(msg, "arpdel failed ... %s\n", err2str(val));
	        mip0dbg((msg));
	    }

	    sprintf(msg, "Disabling tunneling for %s\n", ntoa(mnAddr, addr1));
	    mip3dbg((msg));
	    if ((val = rmroutethroughtunnel(mnAddr,
		mnePtr->haBindingIfaceAddr)) < 0) {
	        sprintf(msg, "rmroutethroughtunnel failed ... %s\n",
			err2str(val));
	        mip0dbg((msg));
	    }
	}
    }

    sprintf(msg, "HA expired binding for %s@%s at pos %d (Binding Cnt %d).\n",
	    ntoa(mnAddr, addr1), ntoa(coAddr, addr2), idx,
	    mnePtr->haMnBindingCnt);
    mip2dbg((msg));

    habeTable[idx].haBindingTimeRemaining = 0;
    habeTable[idx].haBindingTimeGranted = 0;
    *habeTableSize -= 1;

    // when binding is deleted, all filters related to the CoA are removed
    // and the filters are reapplied
    removeSelectedFilters(boundMNTable, &boundMNTableSize, mnAddr, coAddr);
    HAapplyFilters(boundMNTable, haBindingTable, mnAddr);
}

void haAgeBindings(HaBindingEntry habeTable[], int *habeTableSize,
		   int period)
{
    int i;
    int toSee = *habeTableSize;
    int seen = 0;

    for (i = 0; (i < DEFAULT_MAX_BINDINGS) && (seen < toSee); i++) {
        if (habeTable[i].haBindingTimeRemaining) {
	    /* Entry in use. */
	    seen++;
	    if (habeTable[i].haBindingTimeRemaining > period)
	        habeTable[i].haBindingTimeRemaining -= period;
	    else {
	        /* Expired */
	        delHABEidx(habeTable, habeTableSize, i);
		sendInfoToAllGUIs();

	    }
	}
    }
}

/* Delete visitor entry pointed by favep idx in faveTable */
void delFAVEptr(FaVisitorEntry faveTable[], int *faveTableSize, 
		FaVisitorEntry *favep)
{
    int val;

    sprintf(msg, "FA expired %s visitor entry for %s\n",
	    (favep->faVisitorRegIsAccepted ? 
	     "accepted" : "pending"), 
	    ntoa(favep->faVisitorHomeAddr, addr1));
    mip2dbg((msg));

    /* For accepted requests, kill route and disable decapsulation. */

    if (favep->faVisitorRegIsAccepted == TRUE) {
        /* 
	 * TODO: this behavior is incorrect if multiple coaddrs can be
	 * registered for a mobile node at a FA. Decaprem should occur
	 * ONLY after a mobile node has NO accepted registrations.
	 */

        sprintf(msg, "Disabling decapsulation of inner pkts sent to %s\n", 
		ntoa(favep->faVisitorHomeAddr, addr1));
	mip3dbg((msg)); 
	if((val = decaprem(favep->faVisitorHomeAgentAddr,
		favep->faVisitorCOAddr,
		favep->faVisitorHomeAddr)) < 0) {
	    sprintf(msg, "decaprem failed ... %s\n", err2str(val));
	    mip0dbg((msg));
	}
            
	sprintf(msg, "Removing direct, local route for visitor %s thru %s\n",
		ntoa(favep->faVisitorHomeAddr, addr1),
		ntoa(favep->faVisitorIfaceAddr, addr2));
	mip3dbg((msg));
	if((val = routedel(favep->faVisitorHomeAddr, 
			   favep->faVisitorIfaceAddr)) < 0) {
	    sprintf(msg, "routedel failed ... %s\n", err2str(val));
	    mip0dbg((msg));
	}
    }

    favep->faVisitorTimeRemaining = 0;
    *faveTableSize -= 1;
}

void faAgeVisitors(FaVisitorEntry faveTable[], int *faveTableSize,
		   int period)
{
    int i;
    int toSee = *faveTableSize;
    int seen = 0;

    for (i = 0; (i < DEFAULT_MAX_VISITORS) && (seen < toSee); i++) {
        if (faveTable[i].faVisitorTimeRemaining) {
	    /* Entry in use. */
	    seen++;
	    if (faveTable[i].faVisitorTimeRemaining > period)
	        faveTable[i].faVisitorTimeRemaining -= period;
	    else {
	        /* Expired */
	        delFAVEptr(faveTable, faveTableSize, &faveTable[i]);
		sendInfoToAllGUIs();

	    }
	}
    }
}


void faAgeUnreachables(FaUnreachableEntry faueTable[], int *faueTableSize,
		       int period)
{
    int i;
    int toSee = *faueTableSize;
    int seen = 0;

    for (i = 0; (i < DEFAULT_MAX_UNREACHABLES) && (seen < toSee); i++) {
        if (faueTable[i].faUnreachableTimeRemaining) {
	    /* Entry in use. */
	    seen++;
	    if (faueTable[i].faUnreachableTimeRemaining > period)
	        faueTable[i].faUnreachableTimeRemaining -= period;
	    else {
	        /* Expired */
	        sprintf(msg, "FA forgetting unreachability of %s.\n",
			ntoa(faueTable[i].faUnreachableAddr, addr1));
	        mip2dbg((msg));
	        faueTable[i].faUnreachableTimeRemaining = 0;
	        faueTable[i].faUnreachableAddr = 0;
		*faueTableSize -= 1;

		sendInfoToAllGUIs();

	    }
	}
    }
}


/* Perform periodic chores assuming t seconds have elapsed since last call. */
void doPeriodicTask(int t)
{
    mip3dbg(("*"));
    fflush(stdout);

    if (faVisitorTableSize <= DEFAULT_LOW_VISITORS)
        faBusy = FALSE;
    if (faVisitorTableSize >= DEFAULT_HIGH_VISITORS)
        faBusy = TRUE;
/*    maSendAdvertisements(maAdvConfigTable, maAdvConfigTableSize, faBusy);*/
    haAgeBindings(haBindingTable, &haBindingTableSize, t);
    faAgeVisitors(faVisitorTable, &faVisitorTableSize, t);
    faAgeUnreachables(faUnreachableTable, &faUnreachableTableSize, t);
}


