/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Original Code is the Sunlabs Mobile IP.
 The Initial Developer of the Original Code is Vipul Gupta.
 Portions created by Vipul Gupta are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): Niko A. Fikouras (niko@comnets.uni-bremen.de)
		 Asanga D. Udugama (adu@comnets.uni-bremen.de)

*/

/*
 * Copyright (c) 1997 by Sun Microsystems, Inc.
 */

#pragma ident "@(#)mipmodlinux.c 1.9 99/04/27" 

/*
 * Author: sunil.madhani@eng.sun.com
 * Major modifications and cleanup: vgupta@eng.sun.com
 *
 * Routines in this program implement  OS-specific portions of 
 * a Mobile-IP agent (RFC 2002).
 *
 * void InitNet(): 
 *
 *     Includes initialization, e.g. in solaris, we initialize the
 * vtunl module here configuring it as a point to point interface
 * and also plumb vtunl on each of the mobility supporting interfaces
 * and create a stream for vtunl-related ioctl processing. We also
 * enable IP forwarding and disable ICMP re-directs. For Linux, 
 * we load the IP-in-IP handler.
 *
 * int encapadd(u32 target_addr, u32 haAddr, u32 tunel_exit_addr,
 *          unsigned short encap_flags):
 *
 *     Enables encapsulation of pkts meant for target_addr to the
 * tunnel_exit_addr at the specified home agent. Encap_flags is either
 * 0 or 80 (hex). The latter indicates simultaneous bindings. Currently,
 * simultaneous bindings are not supported under Linux. A negative
 * returned value indicates an error and the routine err2str is expected
 * to return a meaningful message corresponding to the error number
 * returned. Encapadd must be followed by a call to mkrtthroughtunnel
 * with the target_addr as argument. The latter is responsible for ensuring
 * that pkts meant for target_addr are routed through the kernel's
 * encapsulating module.
 *
 * NOTE: The argument haAddr is not used for Solaris.
 *
 * int encaprem (u32 target_addr, u32 tunnel_exit_addr)
 *
 *      Terminates encapsulation service for target_addr to the given
 * tunnel exit. If simultaneous bindings are supported, other tunnels
 * are unaffected.
 * 
 * int decapadd(u32 outersrc, u32 outerdst, u32 inneraddr, char *where)
 *          where = "dst"|"src"|"either"
 * 
 *      Enable decapsulation service for inneraddr if it appears
 * in the inner datagram as specified by the last argument.
 * If the target_addr is zero, all encapsulated packets are decapsulated 
 * and forwarded (the last argument is ignored). Under Linux, decapadd 
 * and decaprem are null functions; *all* IP-in-IP packets are automatically 
 * decapsulated and forwarded. 
 * 
 * int decaprem(u32 outersrc, u32 outerdst, u32 inneraddr) 
 *
 *      Terminate decapsulation service for inneraddr. One may use 
 * a 0.0.0.0 address to cancel ALL decapsulation services.
 *
 * int routeadd(unsigned long dst,unsigned long gw, int metric)
 *   
 *      Add route for destination "dst" with gateway "gw" and sets 
 * the appropriate metric specified in argment metric..
 *
 * int routedel(unsigned long dst,unsigned long gw)
 *
 *      Delete route for specified destination through gateway.
 * 
 * int mkroutethroughtunnel(u32 target_addr, u32 tsrc)
 *     
 *      Is responsible for ensuring that pkts to target_addr are routed
 * through the kernel's routing module. 
 *
 * int rmroutethroughtunnel(u32 target_addr, u32 tsrc)
 *
 *      Ensures that pkts for target_addr are no longer sent through
 * the kernel's routing module. Under Soalris, this calls 
 * routedel(target_addr,inet_addr(255.255.255.254)).
 *
 * int arpadd(unsigned long host, unsigned char *eaddr, char *flags)
 *
 *      Adds an arp entry with ip address set to host and hardware
 * address set to eaddr with approriate flags specified by flag
 * argument, e.g. arpadd(inet_addr("129.146.122.121"), "08:20:AB:FE:33:11",
 * "pub") creates a proxy ARP entry for 129.146.122.121. A gratuitous
 * ARP is also sent whenever a proxy entry is inserted.
 *
 * int arpdel(unsigned long host)
 *
 *      Deletes an arp entry for ip address set to host
 *
 * int arprefresh(unsigned long mnaddr)
 * 
 *      Sends gratuitous arp for the mnaddr using the hardware address
 * found for mnaddr in its own ARP cache (making sure that the hardware
 * address is not HA's own). The home agent calls arprefresh when
 * a mobile node returns home and successfully deregisters itself.
 *
 * char *err2str(int errval)
 *     Converts an error value into a user-friendly error message string.
 */

#include <stdio.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/ioctl.h>

#ifdef OLD_LIBC
#include <linux/if.h>
#include <linux/if_ether.h>
#include <linux/if_arp.h>
#include <linux/ip.h>
#include <linux/route.h>
#else  /* for GLIBC, avoid the kernel include files */
#include <net/ethernet.h>
#include <net/if.h>
#include <net/route.h>
#endif /* OLD_LIBC */

#include <netinet/in.h>
#include <net/if_arp.h>

#include <arpa/inet.h>
#include "mip.h"
#include "agent.h"

/* Tunnel related definitions */
#define MAX_TUNNEL_SUPPORTED    32
#define MAX_ERR 1024
#define NO_FREE_TUNNEL     (MAX_ERR + 1)
#define TUNNEL_NOT_FOUND   (MAX_ERR + 2)
#define DEV_NAME_NOT_FOUND ENXIO
#define INVALID_IP_ADDR    ENXIO
#define MN_ENTRY_ABSENT         (MAX_ERR + 11)
#define NO_MAPPING              (MAX_ERR + 12)

// filename of the iproute2 table name definition file
#define ROUTE2_TABLE_FILE 	"/etc/iproute2/rt_tables"


/*
 * For Linux 2.2 we need to avoid dealing with tunl0
 * which is special ... cannot be deleted.
 */
#ifdef LINUX_2_0
#define FIRST_TUNNEL            0
#else
#include <netinet/ip.h>
#include <bits/ioctls.h>
#define SIOCADDTUNNEL   (SIOCDEVPRIVATE + 1)
#define SIOCDELTUNNEL   (SIOCDEVPRIVATE + 2)
struct ip_tunnel_parm {
           char     name[IFNAMSIZ];
           int      link;
           unsigned short i_flags;
           unsigned short o_flags;
           unsigned long  i_key;
           unsigned long  o_key;
           struct iphdr   iph;
};
#define FIRST_TUNNEL            1
#endif

extern int errno;
static int ioctl_sockid;

/* Common to all mobility agents */
extern HaMobileNodeEntry haMobileNodeTable[];
extern int haMobileNodeTableSize ;
extern MaAdvConfigEntry maAdvConfigTable[];
extern int maAdvConfigTableSize ;
extern int mipdebug;
extern char msg[];
extern char addr1[];

/* Table for mapping tunnelno with mobile-node address */
//static unsigned long mnaddr_tunl[MAX_TUNNEL_SUPPORTED + FIRST_TUNNEL + 1];
//static unsigned long coaddr_tunl[MAX_TUNNEL_SUPPORTED + FIRST_TUNNEL + 1];
unsigned long mnaddr_tunl[MAX_TUNNEL_SUPPORTED + FIRST_TUNNEL + 1];
unsigned long coaddr_tunl[MAX_TUNNEL_SUPPORTED + FIRST_TUNNEL + 1];
int tunlInit = TRUE;

typedef struct arhdr{
    unsigned short  ar_hrd;         /* format of hardware address   */
    unsigned short  ar_pro;         /* format of protocol address   */
    unsigned char   ar_hln;         /* length of hardware address   */
    unsigned char   ar_pln;         /* length of protocol address   */
    unsigned short  ar_op;          /* ARP opcode (command)         */
    unsigned char ar_sha[ETH_ALEN]; /* sender hardware address      */
    unsigned char ar_sip[4];        /* sender IP address            */
    unsigned char ar_tha[ETH_ALEN]; /* target hardware address      */
    unsigned char ar_tip[4];        /* target IP address            */
} Arphdr;

extern int numBitsInMask(u32);
extern int isWordIn(char *, char *);

/*
 * Returns the available tunnel number in the range 0-31. If not available
 * it returns -(NO_FREE_TUNNEL)
 */
int newtunnel(u32 addr, u32 coaddr) // changed for MSB
{
    int i;
    char cmd[50];

    if(tunlInit) {
    	tunlInit = FALSE;
    	for (i = FIRST_TUNNEL; i < FIRST_TUNNEL + MAX_TUNNEL_SUPPORTED; i++) {
		mnaddr_tunl[i] = 0;
		coaddr_tunl[i] = 0;
	}
    }

    for (i = FIRST_TUNNEL; i < FIRST_TUNNEL + MAX_TUNNEL_SUPPORTED; i++) {
        if ((mnaddr_tunl[i] == 0) && (coaddr_tunl[i] == 0)) {  // changed for MSB
#ifdef LINUX_2_0

            sprintf(cmd, "/sbin/insmod -o tunl%d new_tunnel", i);
	    mip3dbg(("Loading tunnel module (%s) ... \n", cmd));
	    system(cmd);
	    /*
	     * TODO: check for error from previous step and other system()
	     * calls.
	     */
#endif
	    mnaddr_tunl[i] = addr;
	    coaddr_tunl[i] = coaddr; // added for MSB
            return(i);
        }
    }

    return(-1 * NO_FREE_TUNNEL);
}

void freetunnel(int tnum)
{
    char cmd[50];

    mnaddr_tunl[tnum] = 0;
    coaddr_tunl[tnum] = 0; // added for MSB
#ifdef LINUX_2_0
    sprintf(cmd, "/sbin/rmmod tunl%d", tnum);
    mip3dbg(("Freeing tunnel module (%s) ... \n", cmd));
    system(cmd);
#endif
}

/*
 * Gets the tunnel number corresponding to given
 * mobile node address .
 * If the corresponding tunnel number exist then it returns the tunnel no.
 * else returns -(TUNNEL_NOT_FOUND)
 */
int gettunnelno(u32 mnaddr, u32 coaddr) // changed for MSB
{
    int i;
    for (i= FIRST_TUNNEL; i < FIRST_TUNNEL + MAX_TUNNEL_SUPPORTED; i++) {
        if ((mnaddr_tunl[i] == mnaddr) && (coaddr_tunl[i] == coaddr))
	   return(i);
    }
    return(-1 * TUNNEL_NOT_FOUND);
}

void printtunnel()
{
    char mnaddr[20], coaddr[20];
    int i;

    for (i = FIRST_TUNNEL; i < FIRST_TUNNEL + MAX_TUNNEL_SUPPORTED; i++) {
        if (mnaddr_tunl[i] != 0) {
	    mip3dbg(("Tunnel %d is for %s with COA %s\n",
		     i, ntoa(mnaddr_tunl[i], mnaddr), ntoa(coaddr_tunl[i], coaddr)));
	}
    }
}

/* Process encapadd command */
int encapadd(u32 mnaddr, u32 haaddr, u32 coaddr, int flags)
{
    char ifname[10];
    struct sockaddr_in *sock;
    struct ifreq ifr;
    int tnum, val;
#ifndef LINUX_2_0
    struct ip_tunnel_parm tp;
#endif

    if ((tnum = newtunnel(mnaddr, coaddr)) < 0) { // changed for MSB
        return(tnum);
    }

    sprintf(ifname, "tunl%d", tnum);

#ifndef LINUX_2_0
    memset(&tp, 0, sizeof(tp));
    strncpy(tp.name, ifname, strlen(ifname) + 1);
    tp.iph.version = 4;
    tp.iph.ihl = 5;
    tp.iph.frag_off = htons(0x4000);
    tp.iph.protocol = 4;
    tp.iph.ttl = 0;
    tp.iph.daddr = coaddr;

    strncpy(ifr.ifr_name, "tunl0", strlen("tunl0")+1);
    ifr.ifr_ifru.ifru_data = (void *) &tp;

    /*
     * The following accomplishes the equivalent of
     *     ip tunnel add ifname mode ipip remote coaddr
     */
    if ((val = ioctl(ioctl_sockid, SIOCADDTUNNEL, &ifr)) < 0) {
        goto failure;
    }
#endif

    memset(&ifr, 0, sizeof(struct ifreq));
    strncpy(ifr.ifr_name, ifname, strlen(ifname)+1);

    /*
     * create a pt-to-pt iface from haaddr to coaddr
     * with the given name and manipulate tunnelbitvec
     * appropriately
     */

    sock = (struct sockaddr_in *) &ifr.ifr_addr;
    sock->sin_family = AF_INET;
    sock->sin_addr.s_addr = haaddr;
    if (ioctl(ioctl_sockid, SIOCSIFADDR, (char *) &ifr) < 0)  {
	goto failure;
    }

#ifdef LINUX_2_0
    sock = (struct sockaddr_in *) &ifr.ifr_dstaddr;
    sock->sin_family = AF_INET;
    sock->sin_addr.s_addr = coaddr;
    if (ioctl(ioctl_sockid, SIOCSIFDSTADDR, (char *) &ifr) < 0)  {
	goto failure;
    }
#endif

    ifr.ifr_flags = (IFF_UP | IFF_NOARP | IFF_POINTOPOINT);

    if (val = ioctl(ioctl_sockid, SIOCSIFFLAGS, (char *) &ifr) < 0)  {
	goto failure;
    }

    printtunnel();
    return(0);

failure:
    val = (-1) * errno;
    freetunnel(tnum);
    return(val);
}

/* Process encaprem command. Bring the tunnel down */
int
encaprem(u32 mnaddr, u32 coaddr)
{
    char ifname[10];
    struct ifreq ifr;
    int tnum;
#ifndef LINUX_2_0
    struct ip_tunnel_parm tp;
#endif

    if ((tnum = gettunnelno(mnaddr, coaddr)) <  0) { // changed for MSB
       return(tnum);
    }

    sprintf(ifname, "tunl%d",tnum);
    memset(&ifr, 0, sizeof(struct ifreq));
    strncpy(ifr.ifr_name, ifname, strlen(ifname));

    /* bring down the tunnel interface and free the corresponding tunnel */
    if (ioctl(ioctl_sockid, SIOCGIFFLAGS, (char *) &ifr) < 0) {
        return(-1 * errno);
    }

    ifr.ifr_flags &= ~IFF_UP;

    if (ioctl(ioctl_sockid, SIOCSIFFLAGS, (char *) &ifr) < 0)  {
        return(-1 * errno);
    }

#ifndef LINUX_2_0
    memset(&tp, 0, sizeof(tp));
    strncpy(tp.name, ifname, strlen(ifname) + 1);
    tp.iph.version = 4;
    tp.iph.ihl = 5;
    tp.iph.frag_off = htons(0x4000);
    tp.iph.protocol = 4;
    tp.iph.ttl = 0;
    tp.iph.daddr = coaddr;

    memset(&ifr, 0, sizeof(struct ifreq));
    strncpy(ifr.ifr_name, ifname, strlen(ifname)+1);
    ifr.ifr_ifru.ifru_data = (void *) &tp; 
    /* Do the equivalent of "ip tunnel del ifname" */ 
    if (ioctl(ioctl_sockid, SIOCDELTUNNEL, &ifr) < 0) { 
       mip0dbg(("SIOCDELTUNNEL failed for %s in encaprem.", ifname)); 
    } 
#endif    
    
    freetunnel(tnum);
    printtunnel();
    return(0);
}

/* Process decapadd command */
int
decapadd(u32 outersrc, u32 outerdst, u32 inneraddr, char*flags)
{
    return(0);
}

/* Process decaprem command */
int
decaprem(u32 outersrc, u32 outerdst, u32 innerdst)
{
    return(0);
}

/* Add routing entry for the tunnel */
int mkroutethroughtunnel(u32 mnaddr, u32 haAddr, u32 coaddr)
{
    struct rtentry rtent;
    int count;
    char ifname[16];
    struct sockaddr_in *sock;
    int tnum;

    /* Add the corresponding route */
    memset(&rtent, 0, sizeof(struct rtentry));
    sock = (struct sockaddr_in *) &rtent.rt_dst;
    sock->sin_family = AF_INET;
    sock->sin_addr.s_addr = mnaddr;
    rtent.rt_flags = (RTF_UP | RTF_HOST);

    if ((tnum = gettunnelno(mnaddr, coaddr)) < 0 ) // changed for MSB
        return(tnum);
    count = sprintf(ifname, "tunl%d", tnum);
    ifname[count] = '\0';

    rtent.rt_dev = ifname;
    errno = 0;
    if (ioctl(ioctl_sockid, SIOCADDRT, (char *) &rtent) < 0)  {
        if (errno != EEXIST) {
            return(-1 * errno);
	}
    }

    return(0);
}

/* Delete the routing entry for the tunnel */
int rmroutethroughtunnel(u32 mnaddr, u32 haAddr, u32 coaddr) // changed for MSB
{
    struct rtentry rtent;
    struct sockaddr_in *sock;
    int tnum;
    char addr[30];

    if ((tnum = gettunnelno(mnaddr, coaddr)) < 0) { // changed for MSB
        /* Hmmm ... tunnel may have been removed as part of encaprem */
        return(0);
    }

    memset(&rtent, 0, sizeof(struct rtentry));
    sock = (struct sockaddr_in *) &rtent.rt_dst;
    sock->sin_family = AF_INET;
    sock->sin_addr.s_addr = mnaddr;

    errno = 0;
    if (ioctl(ioctl_sockid, SIOCDELRT, (char *) &rtent) < 0) {
	if (errno != ENXIO) {
	    return(-1 * errno);
	}
    }

    return(0);
}

/* Fetch the interface name corresponding to the given interface address. */
int getifacename(u32 ifaddr, char *ifname)
{
    int i;

    /* Get the matching maIfaceName from mnAdvConfigTable */
    for (i = 0; i < maAdvConfigTableSize; i++) {
       if (ifaddr == maAdvConfigTable[i].maIfaceAddr) {
           strcpy(ifname, maAdvConfigTable[i].maIfaceName);
	   return(0);
       }
    }
    return(-1 * MN_ENTRY_ABSENT); /* Unlikely to occur */
}

/* Add/Delete the route depending on argument cmd */
routemodify(unsigned long dst, unsigned long  gw, int metric, char cmd)
{
    int tmp;
    struct rtentry route;
    struct sockaddr_in *sin;
    char a1[20],a2[20];
    char ifname[MAX_IFNAME_LEN];
    bzero((char *)&route, sizeof(route));

    /* Set route destination */
    sin = (struct sockaddr_in *) &route.rt_dst;
    sin->sin_family = AF_INET;
    sin->sin_addr.s_addr = dst ;

    /* Set route gateway */
    if (cmd == 'a'){
       if ((tmp = getifacename(gw, ifname)) < 0)
         return(tmp);

       route.rt_dev = ifname;
       route.rt_flags = RTF_UP | RTF_HOST ;
  
       if ((metric > 0))
         route.rt_flags |= RTF_GATEWAY ;
    }

    errno = 0;

    if (ioctl(ioctl_sockid, cmd == 'a' ? SIOCADDRT : SIOCDELRT ,
	      (char *)&route) < 0) {
        /* 
	 * SIOCADDRT does not return ENXIO & 
	 * SIOCDELRT never returns EEXIST
	 */
        if ((errno != ENXIO) && (errno != EEXIST)) {
            return(-1 * errno);
	}
    }

    return(0);
}
 
/* Delete the route */ 
int routedel(unsigned long dst, unsigned long  gw) 
{ 
    char cmd[120], addr1[20]; 
    
    sprintf(cmd, "/sbin/route del %s", ntoa(dst, addr1));
    mip3dbg(("Deleting route (%s) ... \n", cmd));
    system(cmd);
    /* return(routemodify(dst, gw, 0, 'd')); */
} 

/* Add the route */
int routeadd(unsigned long dst, unsigned long  gw, int metric)
{
    return(routemodify(dst, gw, metric, 'a'));
}

/*
 * Send out a gratuitous ARP packet through the interface "iface"
 * with hardware address "ifaceaddr" and that maps IP address
 * "ipaddr" to hardware address "ethaddr".
 * The complete packet contents are as follows:
 * 
 * Ethernet hdr
 *        src = ifaceaddr  dst = ff:ff:ff:ff:ff:ff
 *        frame type = 0x0806 (ETH_P_ARP)
 * 
 * ARP hdr
 *        hardware addr type = 1 (ARPHRD_ETHER)
 *        protocol addr type = 0x0800 (ETH_P_IP)
 *        hardware addr len = 6 (ETH_ALEN)
 *        protocol addr len = 4
 *        op code = 1 (ARPOP_REQUEST)
 *
 * ARP payload
 *        sender hardware addr = ethaddr
 *        sender IP addr = ipaddr
 *        dst hardware addr = don't care
 *        target IP addr = don't care
 *
 * NOTE: When this function is called, ipaddr is in network-byte order.
 *
 */
int sendGARPpkt(char *iface, unsigned char *ifaceaddr, 
	       u32 ipaddr, unsigned char *ethaddr)
{
    char frame[80];
    struct ethhdr *ehdr;
    Arphdr *arphd;
    int sockid;
    struct sockaddr sa;

    errno = 0;
    if ((sockid = socket(AF_INET, SOCK_PACKET, htons(ETH_P_802_3))) < 0)
        return(-1 * errno);

    memset((void*)frame, 0, sizeof(frame));

    ehdr = (struct ethhdr*) frame;
    memcpy(ehdr->h_source, ifaceaddr, ETH_ALEN);
    memset(ehdr->h_dest, 0xff, ETH_ALEN);
    ehdr->h_proto = htons(ETH_P_ARP);

    arphd = (Arphdr*)(frame + sizeof(struct ethhdr));
    arphd->ar_hrd = htons(ARPHRD_ETHER);
    arphd->ar_pro = htons(ETH_P_IP);
    arphd->ar_hln = ETH_ALEN;
    arphd->ar_pln = 4;
    arphd->ar_op  = htons(ARPOP_REQUEST);
    memcpy(arphd->ar_sha, ethaddr, ETH_ALEN);

    ipaddr = ntohl(ipaddr);
    arphd->ar_sip[0] = (unsigned char)((ipaddr >> 24) & 0xff);
    arphd->ar_sip[1] = (unsigned char)((ipaddr >> 16) & 0xff);
    arphd->ar_sip[2] = (unsigned char)((ipaddr >> 8 ) & 0xff);
    arphd->ar_sip[3] = (unsigned char)(ipaddr & 0xff);
    memset(arphd->ar_tha, 0xff, ETH_ALEN);
    memcpy(arphd->ar_tip, arphd->ar_sip, 4);

    sa.sa_family = AF_INET;
    strcpy(sa.sa_data, iface);

    /* TODO: figure out the correct length */
    if (sendto(sockid, frame, sizeof(struct ethhdr) + sizeof(Arphdr),
	       0, &sa, sizeof(sa)) < 0) {
        close(sockid);
        return(-1 * errno);
    }

    close(sockid);
    return(0);
}

/* Get the hardware address corresponding to addr from the arp table */
int arpgetHWaddr(u32 addr, unsigned char *etheraddr)
{
    struct arpreq ar;
    struct sockaddr_in *sin;
  
    bzero((caddr_t)&ar, sizeof ar);
    ar.arp_pa.sa_family = AF_INET;
    sin = (struct sockaddr_in *)&ar.arp_pa;
    sin->sin_family = AF_INET;
    sin->sin_addr.s_addr = addr;

    errno = 0;
    if (ioctl(ioctl_sockid, SIOCGARP, (caddr_t)&ar) < 0) {
        return(-1 * (errno));
    } else {
	memcpy(etheraddr, ar.arp_ha.sa_data, ETH_ALEN);
	return(0);
    }
}


/* Get the ethernet address for the given device name */
int getEthernetAddr(char *dev, unsigned char *addr)
{
    struct ifreq ifr;
    char *buf;
  
    bzero((char *)&ifr, sizeof(ifr));
    strncpy(ifr.ifr_name, dev, strlen(dev)); 

    errno = 0;
    if (ioctl(ioctl_sockid, SIOCGIFHWADDR, (char *)&ifr) < 0) 
        return(-1 * errno);

    memcpy(addr, (unsigned char *) &ifr.ifr_hwaddr.sa_data, ETH_ALEN);
    return(0);
}

/* 
 * Get the device name corresponding to the given hardware address.
 * Currently, we get a list of interfaces available
 * and their corresponding ethernet addresses are compared
 * with eaddr. The one which matches is returned.
 * TODO: Find a better way. Is there something like SIOCGIFNAME :) ?
 */
int hwaddr2devname(unsigned char *eaddr, char *dev)
{
    char buff[1024];
    struct ifconf ifc;
    struct ifreq *ifr;
    int i;
    unsigned char addr[ETH_ALEN];

    ifc.ifc_len = sizeof(buff);
    ifc.ifc_buf = buff;

    errno = 0;
    if (ioctl(ioctl_sockid, SIOCGIFCONF, &ifc) < 0) 
        return(-1 * errno);

    ifr = ifc.ifc_req;
    for (i = ifc.ifc_len / sizeof(struct ifreq); --i >= 0; ifr++) {
        if (getEthernetAddr(ifr->ifr_name, addr) == 0) {
            if (memcmp(addr, eaddr, ETH_ALEN) == 0) {
	       memcpy(dev, ifr->ifr_name, ETH_ALEN);
	       return(0);
            }
        }
    }

    return(-1 * DEV_NAME_NOT_FOUND);
}

/* Insert an IP addr and ethernet addr mapping in the ARP cache. */
int arpadd(unsigned long host, unsigned char *eaddr, char *flag)
{
    struct arpreq ar;
    struct sockaddr_in *sock;
    char iface[10];
    int val;

    bzero((char *)&ar, sizeof(ar));
    sock = (struct sockaddr_in *) &ar.arp_pa;
    sock->sin_family = AF_INET;
    sock->sin_addr.s_addr = host;
 
    sock = (struct sockaddr_in *) &ar.arp_ha;
    sock->sin_family = ARPHRD_ETHER;
    memcpy(ar.arp_ha.sa_data, eaddr, ETH_ALEN);
  
    if (strncmp(flag, "temp", 4) == 0)
        ar.arp_flags &= ~ATF_PERM;
    else if (strncmp(flag, "pub", 3) == 0)
        ar.arp_flags = (ATF_PUBL | ATF_PERM |  ATF_COM);
    else if (strncmp(flag, "trail", 5) == 0)
        ar.arp_flags |= ATF_USETRAILERS;

    errno = 0;
    if (ioctl(ioctl_sockid, SIOCSARP, (char *) &ar) < 0) {
        if (errno != EEXIST) {
            return(-1 * errno);
        }
    }

    if (strncmp(flag, "pub", 3) == 0) {
	/* we must send out a gratuitous ARP packet */
	if ((val = hwaddr2devname(eaddr, iface)) < 0) {
	    return(val);
	} else {
	    return(sendGARPpkt(iface, eaddr, host, eaddr));
	}
    }

    return(0);
}

/* 
 * Delete the arp entry for the specified host. Th host address
 * is in network byte order at this point.
 */
int arpdel(unsigned long host)
{
    struct arpreq ar;
    struct sockaddr_in *sin;
  
    bzero((char *)&ar, sizeof(ar));
    ar.arp_pa.sa_family = AF_INET;
    sin = (struct sockaddr_in *) &ar.arp_pa;
    sin->sin_family = AF_INET;
    sin->sin_addr.s_addr = host;

    errno = 0;
    if (ioctl(ioctl_sockid, SIOCDARP, (char *) &ar) < 0){
         if (errno != ENXIO) {
             return(-1 *  errno);
	 }
    }

    return(0);
}

/* OS-specific initialization. */
void InitNet()
{
    char cmd[50];

    /* Make a socket for the various SIOCxxxx ioctl calls */
    if ((ioctl_sockid = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
        mip0dbg(("Error: could not create socket for SIOCxxxx commands.\n"));
	exit(-1);
    }

    /*
     * Enable decapsulation service for mobile node's home address if it
     * appears in the inner datagram as dst.
     * For LINUX 2.0, simply load the ipip handler and enable ip forwarding.
     * For Linux 2.2, we only need to enable ip_forwarding (IPinIP support
     * should be compiled into the kernel).
     */
//#ifdef LINUX_2_0
    sprintf(cmd, "/sbin/insmod ipip");
    mip0dbg(("Loading IPIP handler (%s) ... \n", cmd));
    system(cmd);
//#else
    sprintf(cmd, "/sbin/ifconfig tunl0 0.0.0.0 up");
    mip0dbg(("Enabling IPIP decapsulation (%s) ... \n", cmd));
    system(cmd);
//#endif

   sprintf(cmd, "echo 1 > /proc/sys/net/ipv4/ip_forward");
   mip0dbg(("Enabling IP forwarding (%s) ... \n", cmd));
   system(cmd);

}

/* Counterpart to InitNet() */
void OScleanup()
{
    char cmd[50];

    close(ioctl_sockid);

#ifdef LINUX_2_0
    sprintf(cmd, "/sbin/rmmod ipip");
    mip0dbg(("Unloading IPIP handler (%s) ... \n", cmd));
    system(cmd);
#endif
}

/*
 * Send out a gratuitous ARP to reestablish an ARP mapping between
 * the mobile node's IP and hardware addresses. The mobile node's
 * hardware address is picked up from the agent's ARP cache.
 */
arprefresh(u32 mnaddr)
{
    int i, ret;
    int found;
    u32 ifaceaddr;
    char devname[MAX_IFNAME_LEN + 1];
    unsigned char hainterfaceaddr[ETH_ALEN + 1];
    unsigned char mnetheraddr[ETH_ALEN + 1];
    unsigned char zero_addr[ETH_ALEN + 1] = {0x00, 0x00, 0x00, 
	  				     0x00, 0x00, 0x00};

    mip3dbg(("Arp refresh called for %s\n", ntoa(mnaddr, addr1)));

    /* Get the MN's entry in haMobileNodeTable */
    found = 0 ;
    for (i = 0; i < haMobileNodeTableSize; i++) {
        if (mnaddr == haMobileNodeTable[i].haMnAddr) {
            found = 1;
            ifaceaddr = haMobileNodeTable[i].haBindingIfaceAddr;
            break;
        }  
    }

    if (!found)
        return(-1 * MN_ENTRY_ABSENT); /* Unlikely to occur */


    /* Get the matching maIfaceAddr from mnAdvConfigTable */
    found = 0;
    for (i = 0; i < maAdvConfigTableSize; i++) {
        if (ifaceaddr == maAdvConfigTable[i].maIfaceAddr) {
            found = 1;
            strcpy(devname, maAdvConfigTable[i].maIfaceName);
            memcpy(hainterfaceaddr, maAdvConfigTable[i].maIfaceHWaddr, 
		   ETH_ALEN);
            break;
        }
    }
 
    if (!found)
        return(-1 * MN_ENTRY_ABSENT); /* Unlikely to occur */
 
    if ((ret = arpgetHWaddr(mnaddr, mnetheraddr)) < 0) { 
        if (ret == (-1)*ENODEV) { 
	   mip3dbg(("warning: arprefresh did not find hardware address in " 
                    "arp cache <no big deal>\n")); 
	   return (0); 
	} else {
	   return(ret); 
	}
    }

    if((memcmp(hainterfaceaddr, mnetheraddr, ETH_ALEN) == 0) ||
       (memcmp(zero_addr, mnetheraddr, ETH_ALEN) == 0))
        return(-1 * NO_MAPPING);  
    else
        return(sendGARPpkt(devname, mnetheraddr, mnaddr, mnetheraddr));
}

/* Convert an error value into a human readable error message */
char *err2str(int errval)
{
    int err =  (-1 * errval);
    static char tmp[30];

    switch (err) {

      /* Errors related to tunnel */
      case NO_FREE_TUNNEL:
	  return("No free tunnel available");

      case TUNNEL_NOT_FOUND:
	  return("Tunnel for the corresponding mobile node not found");

      /* Errors related to socket */
      case EPROTONOSUPPORT :
	  return("Protocol or specified type is not supported within the domain");

      case EMFILE :
	  return("Per process descriptor table is full");

      case ENFILE :
	  return("System file table is full");

      case ENOBUFS :
	  return("Insufficient buffer space\n");

      /* Errors related to ioctls */
      case EBADF :
	  return("Invaild descriptor");

      case ENOTTY :
	  return("descriptor not associated with character device");

      case EINVAL:
          return("Invalid arguments");

      case ENXIO:
      case ENODEV:
          return("No such device or address");

      case EEXIST:
          return("Already exists");

      case MN_ENTRY_ABSENT:
          return("Missing Mobile node entry");

      case NO_MAPPING:
          return("Bad ARP mapping for mobile node");

      default:
          sprintf(tmp, "Reason unclear : %d", err);
          return(tmp);
    }
}

/*------ Filter Management ------------*/


/* Creates or deletes a filter set at the kernel level. Does the
 * following
 * if delete ; 	iptables -D
 *		ip rule del
 *		ip route del
 *
 * if add ;     iptables -I
 *              ip rule add
 *              ip route add
 */
//int applyFilter(u32 mnaddr, u32 coaddr, char *fltStr) {
int applyFilter(u32 mnaddr, int seq, char *fltStr) {
	char tblNameStr[64], tnline[256], cmd[256], cmd2[256], rlline[256];
	char addr1[32], addr2[32], str1[32], word[64], actStr[32],
		startportStr[32], fwmarkStr[256], prioStr[64], mnaddrStr1[32],
		mnaddrStr2[32], rulenumStr[32], *sptr, coaStr[32], weightStr[32];
	FILE *rtnfp, *irlfp;
	int found, maxTblID, fltCnt, fltType, fltTarget;
	int int1, i, j, need, markValue, used, fltAction, fltMark, more, coaCount;
	u32 ipaddr;

	sprintf(tblNameStr, "mn%s-%03d", ntoa(mnaddr, addr1), seq);
	//sprintf(tblNameStr, "mn%s-%s", ntoa(mnaddr, addr1), ntoa(coaddr, addr2));
	sscanf(fltStr, " %d ", &fltAction);
	switch(fltAction) {
		case FLTOP_DELETE:
			sscanf(fltStr, " %d %d ", &fltAction, &fltMark);

			// delete the rules that have the table name
			more = TRUE;
			while(more) {
				more = FALSE;
				sprintf(cmd, "ip rule list");
				irlfp = popen(cmd, "r");
				while(fgets(rlline, 256, irlfp) != NULL) {
					if(strlen(rlline) == 0 || *rlline == '#')
						continue;

					if(!isWordIn(rlline, tblNameStr))
						continue;

					sptr = (char *) strstr(rlline, "fwmark");
					if(sptr == NULL)
						continue;

					sptr = sptr + strlen("fwmark");

					sscanf(sptr, "%d", &markValue);
					if(markValue != fltMark)
						continue;

					strcpy(cmd2, "ip rule del ");
					getWord(rlline, 1, prioStr);

					replaceStr(rlline, "from all", "        ");
					replaceStr(rlline, "lookup", "table ");

					strcat(cmd2, (rlline + strlen(prioStr)));
					more = TRUE;
					break;
				}
				pclose(irlfp);
				if(!more)
					break;
				printf("%s \n", cmd2);
				system(cmd2);
			}

			// delete all iptable markings
			ntoa(mnaddr, mnaddrStr1);
			more = TRUE;
			while(more) {
				more = FALSE;
				sprintf(cmd, "iptables --list PREROUTING -v -t mangle  --line-numbers -n");
				irlfp = popen(cmd, "r");
				while(fgets(rlline, 256, irlfp) != NULL) {
					if(strlen(rlline) == 0 || *rlline == '#')
						continue;

					memset(mnaddrStr2, (char) NULL, sizeof(mnaddrStr2));
					getWord(rlline, 10, mnaddrStr2);

					if(strncmp(mnaddrStr1, mnaddrStr2, strlen(mnaddrStr1)) != 0)
						continue;


					sptr = (char *) strstr(rlline, "MARK set");
					if(sptr == NULL)
						continue;

					sptr = sptr + strlen("MARK set");

					sscanf(sptr, "%x", &markValue);
					if(markValue != fltMark)
						continue;

					getWord(rlline, 1, rulenumStr);
					sprintf(cmd2, "iptables -t mangle -D PREROUTING %s ", rulenumStr);

					more = TRUE;
					break;
				}
				pclose(irlfp);
				if(!more)
					break;
				printf("%s \n", cmd2);
				system(cmd2);
			}

			// check if routing table being used
			used = FALSE;
			sprintf(cmd, "ip rule list");
			irlfp = popen(cmd, "r");
			while(fgets(rlline, 256, irlfp) != NULL) {
				if(strlen(rlline) == 0 || *rlline == '#')
					continue;

				if(isWordIn(rlline, tblNameStr)) {
					used = TRUE;
					break;
				}
			}
			pclose(irlfp);

			// if routing table not used, then delete
			// any routing entries related to the table name
			if(!used) {
				more = TRUE;
				while(more) {
					more = FALSE;
					*rlline = (char) NULL;
					sprintf(cmd, "ip route list table %s", tblNameStr);
					irlfp = popen(cmd, "r");
					fgets(rlline, 256, irlfp);
					if(isWordIn(rlline, "default") || isWordIn(rlline, "tunl")) {
						more = TRUE;
						sprintf(cmd2, "ip route flush table %s", tblNameStr);
					}
					pclose(irlfp);
					if(more) {
						printf("%s \n", cmd2);
						system(cmd2);
					}
				}
			}
			break;

		case FLTOP_SHARE:
		case FLTOP_ADD:
			maxTblID = 0;
			found = FALSE;

			printf("ADDING FILTER \n");

			// check the route table name file to see if an entry
			// exist, if not insert one
			rtnfp = fopen(ROUTE2_TABLE_FILE, "r");
			while(fgets(tnline, 256, rtnfp) != NULL) {
				if(strlen(tnline) == 0 || *tnline == '#')
					continue;

				if(sscanf(tnline, " %d %s", &int1, str1) != 2)
					continue;

				if(strncmp(tblNameStr, str1, strlen(tblNameStr)) == 0) {
					found = TRUE;
					break;
				}

				// also, figure out the highest table ID
				if(maxTblID < int1)
					maxTblID = int1;
			}
			fclose(rtnfp);

			if(!found) {
				sprintf(cmd, "echo %d %s >> %s", (maxTblID + 1), tblNameStr, ROUTE2_TABLE_FILE);
				system(cmd);
			}

			// extract the initial info in the str
			sscanf(fltStr, " %d %d %d %d %d ", &fltAction, &fltCnt, &fltTarget, &markValue, &coaCount);

			// set the route entry
			sprintf(cmd, "ip route replace default table %s ", tblNameStr);
			strcat(cmd, " equalize ");
			j = 5;
			for(i = 0; i < coaCount; i++) {
				j++;
				if(getWord(fltStr, j, coaStr) != 0)
					break;
				j++;
				if(getWord(fltStr, j, weightStr) != 0)
					break;
				sprintf(cmd + strlen(cmd), " nexthop dev tunl%d weight %s ",
										gettunnelno(mnaddr, inet_addr(coaStr)),
										weightStr);
				//sprintf(cmd + strlen(cmd), " nexthop via %s dev tunl%d weight %s ",
				//						coaStr,
				//						gettunnelno(mnaddr, inet_addr(coaStr)),
				//						weightStr);
			}
			printf("%s\n", cmd);
			system(cmd);

			// if the filter update is due to new weight
			// added to an existing filter, then simple
			// set the route entry, nothing else, as other
			// things are all done
			if(fltAction == FLTOP_SHARE)
				break;
			// create the routing table, if not available
			//need = TRUE;
			//*rlline = (char) NULL;
			//sprintf(cmd, "ip route list table %s", tblNameStr);
			//irlfp = popen(cmd, "r");
			//while(fgets(rlline, 256, irlfp) != NULL) {
			//	if(isWordIn(rlline, "default") || isWordIn(rlline, "tunl")) {
			//		need = FALSE;
			//		break;
			//	}
			//}
			//pclose(irlfp);
			//if(need) {
			//	sprintf(cmd, "ip route add default dev tunl%d table %s",
			//			gettunnelno(mnaddr, coaddr), tblNameStr);
			//	printf("%s\n", cmd);
			//	system(cmd);
			//}



			//build the ip rule for the mark
			sprintf(cmd, "ip rule add fwmark %d ", markValue);
			switch(fltTarget) {
				case TG_DROP:
					strcpy(actStr, "reject");
					break;
				case TG_ACCEPT:
					strcpy(actStr, " ");
					break;
			}
			sprintf(str1, " to %s table %s %s ", ntoa(mnaddr, addr1), tblNameStr, actStr);
			strcat(cmd, str1);
			printf("%s\n", cmd);
			system(cmd);




			// build the iptables rule
			if(fltCnt < 0) {
				// the default route should be at the top ; fltCnt == -1 means default route
				sprintf(cmd, "iptables -t mangle -I PREROUTING 1 -v -d %s ", ntoa(mnaddr, addr1));
			} else {
				sprintf(cmd, "iptables -t mangle -A PREROUTING  -v -d %s ", ntoa(mnaddr, addr1));
			}
			j = 5 + (coaCount * 2);
			for(i = 0; i < fltCnt; i++) {
				j++;
				if(getWord(fltStr, j, word) != 0)
					break;

				if(sscanf(word, "%d", &fltType) != 1)
					break;

				switch(fltType) {
					case FLT_BAF_EXT:
						j++;
						if(getWord(fltStr, j, word) != 0)
							break;
						sscanf(word, "%d", &int1);
						sprintf((cmd + strlen(cmd)), " -m tos --tos 0x%02x", int1);
						break;
					case FLT_PROTO_EXT:
						j++;
						if(getWord(fltStr, j, word) != 0)
							break;
						sprintf((cmd + strlen(cmd)), " -p %s", word);
						break;
					case FLT_SRCADR_EXT:
						j++;
						if(getWord(fltStr, j, word) != 0)
							break;
						sprintf((cmd + strlen(cmd)), " -s %s", word);
						break;
					case FLT_SRCNET_EXT:
						j++;
						if(getWord(fltStr, j, word) != 0)
							break;
						strcpy(addr1, word);

						j++;
						if(getWord(fltStr, j, word) != 0)
							break;
						ipaddr = inet_addr(word);
						sprintf((cmd + strlen(cmd)), " -s %s/%d", addr1, numBitsInMask(ipaddr));
						break;
					case FLT_SRCPORT_EXT:
						j++;
						if(getWord(fltStr, j, word) != 0)
							break;
						sprintf((cmd + strlen(cmd)), " --source-port %s", word);
						break;
					case FLT_SRCPORTRNG_EXT:
						j++;
						if(getWord(fltStr, j, word) != 0)
							break;
						strcpy(startportStr, word);

						j++;
						if(getWord(fltStr, j, word) != 0)
							break;

						sprintf((cmd + strlen(cmd)), " --source-port %s:%s", startportStr, word);
						break;
					case FLT_FRFORM_EXT:
						break;

					// new filters

					case FLT_DSTPORT_EXT:
						j++;
						if(getWord(fltStr, j, word) != 0)
							break;
						sprintf((cmd + strlen(cmd)), " --dport %s", word);
						break;
					case FLT_DSTPORTRNG_EXT:
						j++;
						if(getWord(fltStr, j, word) != 0)
							break;
						strcpy(startportStr, word);

						j++;
						if(getWord(fltStr, j, word) != 0)
							break;

						sprintf((cmd + strlen(cmd)), " --destination-port %s:%s", startportStr, word);
						break;

					default:
						break;
				}
			}
			sprintf(str1, " -j MARK --set-mark %d ", markValue);
			strcat(cmd, str1);
			printf("%s\n", cmd);
			system(cmd);



			break;
	}
	flushRouteCache();
}

int flushRouteCache() {
//	char cmd[128];

//	sprintf(cmd, "ip route flush cache ");
//	printf("%s \n", cmd);
//	system(cmd);

	return TRUE;
}

