/*
 * Copyright (c) 1997 by Sun Microsystems, Inc.
 */

#pragma ident "@(#)mipmodsolaris.c 1.11 98/12/01"

/*
 * Author: sunil.madhani@eng.sun.com
 * Major modifications: vgupta@eng.sun.com
 *
 * Routines in this program implement  OS-specific portions of 
 * a Mobile-IP agent (RFC 2002).
 *
 * void InitNet(): 
 *
 *     Includes initialization, e.g. in solaris, we initialize the
 * vtunl module here configuring it as a point to point interface
 * and also plumb vtunl on each of the mobility supporting interfaces
 * and create a stream for vtunl-related ioctl processing. We also
 * enable IP forwarding and disable ICMP re-directs.
 *
 * int encapadd(u32 target_addr, u32 haAddr, u32 tunel_exit_addr, 
 *          unsigned short encap_flags):
 *
 *     Enables encapsulation of pkts meant for target_addr to the
 * tunnel_exit_addr at the specified home agent. Encap_flags is either
 * 0 or 80 (hex). The latter indicates simultaneous bindings. A negative
 * returned value indicates an error and the routine err2str is expected
 * to return a meaningful message corresponding to the error number
 * returned. Encapadd must be followed by a call to mkrtthroughtunnel
 * with the target_addr as argument. The latter is responsible for ensuring
 * that pkts meant for target_addr are routed through the kernel's
 * encapsulating module.
 *
 * NOTE: The argument haAddr is not used for Solaris.
 *
 * int encaprem (u32 target_addr, u32 tunnel_exit_addr)
 *
 *      Terminates encapsulation service for target_addr to the given
 * tunnel exit. If simultaneous bindings are supported, other tunnels
 * are unaffected. 
 * 
 * int decapadd(u32 target_addr, char *where)  
 *          where = "dst"|"src"|"either"
 *
 *      Enable decapsulation service for target_addr if it appears
 * in the inner datagram as specified by the last argument.
 * If the target_addr is zero, all encapsulated packets are decapsulated
 * and forwarded (the last argument is ignored).
 *
 * int decaprem(u32 target_addr)
 *
 *      Terminate decapsulation service for target_addr. One may use
 * a 0.0.0.0 address to cancel ALL decapsulation services. 
 *
 * int routeadd(unsigned long dst,unsigned long gw, int metric)
 *   
 *      Add route for destination "dst" with gateway "gw" and sets 
 * the appropriate metric specified in argment metric..
 *
 * int routedel(unsigned long dst,unsigned long gw)
 *
 *      Delete route for specified destination through gateway.
 * 
 * int mkroutethroughtunnel(u32 target_addr)
 *     
 *      Is responsible for ensuring that pkts to target_addr are routed
 * through the kernel's routing module. Under Soalris, this calls
 * routeadd(target_addr, inet_addr("255.255.255.254"), 1).
 *
 * int rmroutethroughtunnel(u32 target_addr)
 *
 *      Ensures that pkts for target_addr are no longer sent through
 * the kernel's routing module. Under Soalris, this calls 
 * routedel(target_addr,inet_addr(255.255.255.254)).
 *
 * int arpadd(unsigned long host, unsigned char *eaddr, char *flags)
 *
 *      Adds an arp entry with ip address set to host and hardware  
 * address set to eaddr with approriate flags specified by flag
 * argument, e.g. arpadd(inet_addr("129.146.122.121"), "08:20:AB:FE:33:11",
 * "pub") creates a proxy ARP entry for 129.146.122.121.
 *
 * int arpdel(unsigned long host)
 *
 *      Deletes an arp entry for ip address set to host  
 *
 * int arprefresh(unsigned long mnaddr)
 * 
 *      Sends gratuitous arp for the mnaddr using the hardware address
 * found for mnaddr in its own ARP cache (making sure that the hardware
 * address is not HA's own). The home agent calls arprefresh when
 * a mobile node returns home and successfully deregisters itself.
 *
 * char *err2str(int errval)
 *     Converts an error value into a user-friendly error message string.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <netdb.h>
#include <signal.h>
#include <setjmp.h>
#include <errno.h>

#include <sys/dlpi.h>
#include <sys/types.h>
#include <sys/stream.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/sockio.h>
#include <sys/errno.h>
#include <sys/param.h>
#include <sys/conf.h>
#include <sys/stropts.h>
#include <sys/tihdr.h>
#include <sys/tiuser.h>
#include <sys/timod.h>
#include <sys/stropts.h>
#include <sys/signal.h>

#include <inet/common.h> 
#include <inet/ip.h> 
#include <inet/common.h>
#include <inet/ip.h>

#include <net/if.h>
#include <net/route.h>
#include <net/if_arp.h>

#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/ip_mroute.h>

#include "vtunl.h"
#include "agent.h"
#include "mip.h"

/* Common to all mobility agents */
extern HaMobileNodeEntry haMobileNodeTable[];
extern int haMobileNodeTableSize ;
extern MaAdvConfigEntry maAdvConfigTable[];  
extern int maAdvConfigTableSize ;
extern int mipdebug;
extern char msg[];
extern char addr1[];

extern int errno;
static int ioctl_fd;
static int ioctl_sockid;
static int stream_id[16];  

char *err2str(int);

#define DEVICEDIR   "/dev/"
#ifndef ETH_ALEN
#define ETH_ALEN 6
#endif

/* Definitions needed for refresh_mn_arp() module */
#define	ABS(x)		((x) < 0 ? -(x) : (x))
#define	BITSPERBYTE	8
#define ETHERADDRL      sizeof(ether_addr_t)
#define IPADDRL         sizeof(struct in_addr)
#define DEVICEDIR       "/dev/"
#define BCAST_HW_ADDR   "ff:ff:ff:ff:ff:ff"

#define MAX_ERR                 2048
#define INVALID_PPA             (MAX_ERR + 1)
#define INVALID_STRING          (MAX_ERR + 2)
#define DLOKACK_SHORT_RESPONSE  (MAX_ERR + 3)
#define DLOKACK_NOT_M_PCPROTO   (MAX_ERR + 4)
#define ERR_MORECTL             (MAX_ERR + 5)
#define ERR_MOREDATA            (MAX_ERR + 6)
#define ERR_MORECTLDATA         (MAX_ERR + 7)
#define DL_PRIMITIVE_ERROR      (MAX_ERR + 8)
#define SHORT_CONTROL_PORTION   (MAX_ERR + 9)
#define INVALID_ADDR            (MAX_ERR + 10)
#define MN_ENTRY_ABSENT         (MAX_ERR + 11)
#define NO_MAPPING              (MAX_ERR + 12)

/*  Maximum control/data buffer size (in long's !!) for getmsg() */
#define         MAXDLBUF        8192
#define         MAXWAIT         15
/* Maximum address buffer length */
#define         MAXDLADDR       1024
/* Handy macro.*/
#define         OFFADDR(s, n)   (u_char*)((char*)(s) + (int)(n))
#define         DEFAULT_MIN_MTU 1250

int getMinMTU();
int setMTU(char *, int);
int garp(char *, u32, unsigned char *);

/* Convert a negative error value into a human readable error message */
char *err2str(int errval)
{
    int err =  (-1 * errval);
    static char tmp[30]; 

    switch (err) {
      case EMFILE:
	  return("Descriptor table full");

      case EBADF:
	  return("Not a valid open file descriptor");

      case EACCES:
	  return("Permission denied");

      case ENOMEM:
	  return("Insufficient memory");

      case ENOENT:
          return("No such device/file/directory");
       
      case ENXIO:
	  return("No such device or address");
	  
      case EEXIST:
	  return("Already exists");

      case EINVAL:
	  return("Invalid request or arguments");
      
      case INVALID_PPA: 
	  return("Invalid PPA") ;

      case INVALID_STRING:
          return("Invalid input string");
      
      case DLOKACK_SHORT_RESPONSE:
          return("Short DLOKACK response");

      case DLOKACK_NOT_M_PCPROTO: 
          return("DLOKACK is not M_PCPROTO");

      case ERR_MORECTL: 
          return("MORECTL");

      case ERR_MOREDATA: 
          return("MOREDATA");

      case ERR_MORECTLDATA: 
          return("MORECTL or MOREDATA");

      case SHORT_CONTROL_PORTION: 
          return("Control portion is short");

      case DL_PRIMITIVE_ERROR: 
          return("DL primitive error");

      case EAGAIN: 
          return("A non-prioriy message specified");

      case EFAULT: 
          return("ctlptr or dataptr points to an illegal address");

      case EINTR: 
          return("Caught a signal");

      case ENOSR: 
          return("Insufficient stream resources");

      case ERANGE: 
          return("Packet size is out of range");

      case INVALID_ADDR: 
          return("Invalid Address");

      case EBADMSG: 
          return("Bad queued message");

      case MN_ENTRY_ABSENT: 
          return("Missing Mobile node entry");

      case NO_MAPPING:
          return("Bad ARP mapping for mobile node");

      default:
          sprintf(tmp, "Reason unclear : %d", err);
          return(tmp);
    }
}

/*
 * Parse the interface name (e.g. "le0" or "hme1") and place the
 * corresponding device name (e.g. "/dev/le", "/dev/hme") in
 * devname and the ppa (e.g. 0 or 1 for the examples above) in ppa.
 */
int ifname2devppa(char *ifname, char *devname, int *ppa)
{
    char *p;
    int  i, j, val;

    val = 0;
    j = strlen(DEVICEDIR);
    strncpy(devname, DEVICEDIR, j);
    for (i = 0, p = ifname; i < strlen(ifname); i++, p++) {
	if (*p >= '0' && *p <= '9') 
	    val = 10*val + (*p - '0');
	else 
	    devname[j++] = *p;
     }
     devname[j] = '\0';
     *ppa = val;
}

/*
 *  Configure the point-to-point interface named ifname with the
 *  given addresses.
 *
 *  If you wish to look at ip0 using ifconfig, always specify the
 *  inet option, e.g.
 *      ifconfig -a inet
 *  or
 *      ifconfig ip0 inet
 *
 *  If this option is not specified, ifconfig may attempt to print
 *  ethernet info for ip0 and hang.
 */
int mkpt2pt(char *ifname, long srcaddr, long dstaddr)
{   
    struct ifreq ifr;
    struct sockaddr_in sin;
    short savedflags;

    errno = 0 ;
       
    memset((char *) &ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, ifname, sizeof ifr.ifr_name);
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = srcaddr;
    ifr.ifr_addr = *(struct sockaddr *) &sin;
    if (ioctl(ioctl_sockid, SIOCSIFADDR, (caddr_t)&ifr) < 0) 
        return(-1 * (errno));

    memset((char *) &ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, ifname, sizeof ifr.ifr_name);
    if (ioctl(ioctl_sockid, SIOCGIFFLAGS, (caddr_t)&ifr) < 0) 
	return(-1 * (errno));
    savedflags = ifr.ifr_flags;

    memset((char *) &ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, ifname, sizeof ifr.ifr_name);
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = dstaddr;
    ifr.ifr_addr = *(struct sockaddr *)&sin;
    if (ioctl(ioctl_sockid, SIOCSIFDSTADDR, (caddr_t)&ifr) < 0)
	return(-1 * (errno));

    memset((char *) &ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, ifname, sizeof ifr.ifr_name);
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = inet_addr("255.255.255.255");
    ifr.ifr_addr = *(struct sockaddr *)&sin;
    if (ioctl(ioctl_sockid, SIOCSIFNETMASK, (caddr_t)&ifr) < 0)
	return(-1 * (errno));

     /*
      * By bringing UP this interface, we force the creation of
      * a routing table entry for srcaddr through ip<ppa>.
      */

    memset((char *) &ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, ifname, sizeof ifr.ifr_name);
    ifr.ifr_flags = savedflags | IFF_UP | 
	        IFF_PRIVATE | IFF_POINTOPOINT ;
    if (ioctl(ioctl_sockid, SIOCSIFFLAGS, (caddr_t)&ifr) < 0) 
        return(-1 * (errno));

    return(0);
}

/*
 * Issue an ioctl for the vtunl module using the I_STR ioctl call 
 * (refer to man streamio).
 */
int 
vtunl_ctl(int s, int cmd, void *arg, int arglen)
{
    struct strioctl ioc;

    memset((char *) &ioc, 0, sizeof(ioc));
    ioc.ic_cmd = cmd;
    ioc.ic_timout = 0;
    ioc.ic_len = (size_t) arglen;
    ioc.ic_dp = (char *) arg;

    errno = 0;
    ioctl(s, I_STR, (char *) &ioc);
    return(errno);
}

/*
 * Create a stream containing the named device, and vtunl. Then
 * configure it with the information provided in conf.
 * Finally, push "ip" and return the stream descriptor.
 */
int plumb(char *devname, struct vtunl_req *conf)
{
    int fd;
    struct vtunl_conf vc;
       
    errno = 0;
    if ((fd = open(devname, 0)) < 0) 
	return(-1 * (errno));

    if (ioctl(fd, I_PUSH, "vtunl") < 0) {
	close(fd);
	return(-1 * (errno));
    }

    if (strncmp("/dev/ip", devname, strlen("/dev/ip")) == 0) {
	if (vtunl_ctl(fd, VTUNL_SETCONF_TRANSPORT, conf, 
		     sizeof(struct vtunl_req))) {
	    close(fd);
	    return(-1 * (errno));
	}
	if (vtunl_ctl(fd, VTUNL_GETCONF, &vc, 
		sizeof(struct vtunl_conf))){
	    close(fd);
	    return(-1 * (errno));
	 }
	 mip0dbg(("Configured vtunl type %d proto %d addr %x (%d)",
	     vc.type, vc.ip_protocol, ntohl(vc.address), vc.state));
    } else {
	if (vtunl_ctl(fd, VTUNL_SETCONF_DLSUSER, conf, 
			     sizeof(struct vtunl_req))){
	   close(fd);
	   return(-1 * (errno));
	}

	if (vtunl_ctl(fd, VTUNL_GETCONF, &vc, 
		sizeof(struct vtunl_conf))){
	    close(fd);
	    return(-1 * (errno));
	}
	 
	mip0dbg(("Configured vtunl type %d ppa %d addr %x mask %x (%d)",
		 vc.type, vc.ip_protocol, ntohl(vc.address),
		 ntohl(vc.addr_mask), vc.state));
    }

    mip0dbg((" on %s.\n", devname));

    if (ioctl(fd, I_PUSH, "ip") < 0) {
	close(fd);
	return(-1 * (errno));
    }

    return(fd);
}


/*
 * Network-related initialization, e.g. loading tunneling module etc.
 * Information about each mobility supporting interface is available
 * in maAdvConfigTable whit size specified in maAdvConfigTableSize.
 */
int InitNet()
{
    int i;
    int ppa;
    int val;
    char devname[20];
    struct vtunl_req vrq;
    int MinMTU;

    /* 
     * Enable forwarding, disable ICMP redirects in kernel, e.g. FA
     * should not redirect visiting mobile nodes to another router).
     */
    system("/usr/sbin/ndd -set /dev/ip ip_forwarding 1 > /dev/null 2>&1\n");
    system("/usr/sbin/ndd -set /dev/ip ip_send_redirects 0 > /dev/null 2>&1\n");
    mip2dbg(("IP forwarding on, ICMP redirects off.\n"));
    

    /* 
     * We do this as a guaranteed means of ensuring that vtunl
     * is correctly initialized from scratch.
     */
    mip2dbg(("rem_drv followed by add_drv vtunl\n"));
    system("/usr/sbin/rem_drv vtunl > /dev/null 2>&1");
    system("/usr/sbin/add_drv vtunl > /dev/null 2>&1");

    /* Make a socket for the various SIOCxxxx ioctl commands */
    if ((ioctl_sockid = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
        mip0dbg(("Error: could not create socket for SIOCxxxx commands.\n"));
	exit(-1);
    }

    /* Create a stream that encapsulates unicasts and decapsulates */
    memset((char *) &vrq, 0, sizeof(struct vtunl_req));
    vrq.vr_address = maAdvConfigTable[0].maIfaceAddr;
    vrq.vr_ip_protocol = 4;

    stream_id[0] = plumb("/dev/ip", &vrq);
    if (((val = stream_id[0]) < 0) || 
	((val = mkpt2pt("ip0", maAdvConfigTable[0].maIfaceAddr, 
			inet_addr("255.255.255.254"))) < 0)) {
        mip0dbg(("Could not create stream for encap/decap ... %s\n",
		 err2str(val)));

	mip0dbg(("\nEnsure that vtunl is installed properly on your " \
		 " system, you are\n"));
	mip0dbg(("logged in as root, and another  process hasn't " \
		 "already created a\n"));
	mip0dbg(("stream with \"vtunl\" (tunneling driver) plumbed " \
		 "between \"/dev/ip\"\n"));
	mip0dbg(("(IP driver) and \"ip\" (IP module).\n\n"));

	exit(-1);
    }

    mip0dbg(("Tunneling driver (vtunl) installed successfully.\n"));

    /* Create streams on each specified network device */
    for (i = 0; i < maAdvConfigTableSize; i++) {
        ifname2devppa(maAdvConfigTable[i].maIfaceName, devname, &ppa);
	if (ppa < 0 || ppa > 255) continue;

	memset((char *) &vrq, 0, sizeof(struct vtunl_req));
	vrq.vr_address = maAdvConfigTable[i].maIfaceAddr;
	vrq.vr_addr_mask = maAdvConfigTable[i].maIfaceNetmask;
	vrq.vr_ppa = ppa;
	if ((val = plumb(devname, &vrq)) < 0) {
	    mip0dbg(("Could not plumb vtunl over %s ... %s\n",
		     maAdvConfigTable[i].maIfaceName, err2str(val)));
	    exit(-1);
	}
	stream_id[i+1] = val;
    }

    /* 
     * Create a stream for ioctls. This one has vtunl over /dev/ip
     * with nothing else between vtunl and the stream head. This
     * ensures that vtunl related ioctls will not be mis-interpreted
     * by any intermediate modules.
     */
  
    if (((ioctl_fd = open("/dev/ip", 0)) < 0) ||
	(ioctl(ioctl_fd, I_PUSH, "vtunl") < 0)) {
        mip0dbg(("Error: could not push vtunl on /dev/ip for ioctls.\n"));
	exit(-1);
    }

    /* 
     * Change the ip0 MTU to minimum of the MTUs on all active interfaces
     * (not just those that support mobility).
     */

    MinMTU = getMinMTU();
    mip0dbg(("Setting ip0 MTU to %d.\n", MinMTU));
    if (setMTU("ip0", MinMTU) < 0) {
        mip0dbg(("setMTU failed\n"));
	exit(-1);
    }	
    
    if ((val = pmtubound(0, MinMTU)) < 0) {    
        mip0dbg(("Error while setting default PMTU %d ... %s\n", 
		 MinMTU, err2str(val)));
	exit(-1);
    }

}

/* 
 * Sets an upper bound on the path-mtu for the
 * given tunnel_exit.
 */
int pmtubound(u32 t_exit, int pmtu)
{
    struct vtunl_req vrq;
    int err;
           
    memset((char *) &vrq, 0, sizeof(struct vtunl_req));
    vrq.vr_address = t_exit;
    vrq.vr_mtu_bound = (u16) pmtu;

    if (err = vtunl_ctl(ioctl_fd, VTUNL_MAXPMTU_SET, &vrq, sizeof(vrq))) 
        return(-1 * err);
    else
        return(0);
}

/* Process encapadd command */
int
encapadd(u32 target, u32 haAddr, u32 t_exit, unsigned short flags) 
{
   int  err;
   struct vtunl_req vrq;

    memset((char *) &vrq, 0, sizeof(struct vtunl_req));
    vrq.vr_target_addr = target;
    vrq.vr_t_exit_addr = t_exit ;
    vrq.vr_encap_flags = (u16) flags;

    if (err = vtunl_ctl(ioctl_fd, VTUNL_ENCAP_ADD, &vrq, sizeof(vrq))) 
        return(-1 * err);
    else {
        return(0);
    }
}
	
/* Process encaprem */
int
encaprem(u32 target, u32 t_exit)
{
    int  err;
    struct vtunl_req vrq;

    memset((char *) &vrq, 0, sizeof(struct vtunl_req));
    vrq.vr_target_addr = target;
    vrq.vr_t_exit_addr = t_exit ;

    if (err = vtunl_ctl(ioctl_fd, VTUNL_ENCAP_REM, &vrq, sizeof(vrq))) 
        return(-1 * (err));
    else 
        return(0);
}

/* Process decapadd command */
int
decapadd(u32 target, char *flags)
{
    int err;
    struct vtunl_req vrq;

    memset((char *) &vrq, 0, sizeof(struct vtunl_req));
    vrq.vr_target_addr = target ;
    if (strncmp(flags, "dst", strlen("dst")) == 0) 
        vrq.vr_decap_flags = DECAP_IF_DST;
    else if (strncmp(flags, "src", strlen("src")) == 0) 
	vrq.vr_decap_flags = DECAP_IF_SRC;
    else if (strncmp(flags, "either", strlen("either")) == 0) 
        vrq.vr_decap_flags = DECAP_IF_EITHER;
    else 
       return(-1*EINVAL);

    if (err = vtunl_ctl(ioctl_fd, VTUNL_DECAP_ADD, &vrq, sizeof(vrq)))
        return(-1 * (err));
    else 
        return(0);
}

/* Process decaprem command */
int
decaprem(u32 target)
{
   int  err;
   struct vtunl_req vrq;

   memset((char *) &vrq, 0, sizeof(struct vtunl_req));
   vrq.vr_target_addr = target;
	   
   if (err = vtunl_ctl(ioctl_fd, VTUNL_DECAP_REM, &vrq, sizeof(vrq)))
        return(-1 * (err));
   else 
	return(0);
}

/*
 * Set an individual arp entry 
 */
int arpadd(unsigned long host, unsigned char *eaddr, char *flag)
{
    struct arpreq ar;
    struct sockaddr_in *sin;

    bzero((caddr_t)&ar, sizeof ar);
    sin = (struct sockaddr_in *)&ar.arp_pa;
    sin->sin_family = AF_INET;
    sin->sin_addr.s_addr = (host);

     memcpy(ar.arp_ha.sa_data, eaddr, ETH_ALEN);

     ar.arp_flags = ATF_PERM;

     if (strncmp(flag, "temp", 4) == 0)
	 ar.arp_flags &= ~ATF_PERM;
     if (strncmp(flag, "pub", 3) == 0)
	 ar.arp_flags |= ATF_PUBL;
     if (strncmp(flag, "trail", 5) == 0)
	 ar.arp_flags |= ATF_USETRAILERS;

     errno = 0;		
     if (ioctl(ioctl_sockid, SIOCSARP, (caddr_t)&ar) < 0) {
         if (errno != EEXIST)
             return((-1) * errno);
     }

     return (0);
}

/* 
 * Delete an arp entry  
 */
int arpdel(unsigned long host)
{
    struct arpreq ar;
    struct sockaddr_in *sin;

    bzero((caddr_t)&ar, sizeof ar);
    ar.arp_pa.sa_family = AF_INET;
    sin = (struct sockaddr_in *)&ar.arp_pa;
    sin->sin_family = AF_INET;
    sin->sin_addr.s_addr = (host);

    errno = 0;
    if (ioctl(ioctl_sockid, SIOCDARP, (caddr_t)&ar) < 0) {
        if (errno != ENXIO)
            return((-1) * errno);
    }
    return(0);
}

/* Get the hardware address corresponding to mnaddr from the arp table */ 
int arpgetHWaddr(u32 mnaddr, unsigned char *mnetheraddr)
{
    struct arpreq ar;
    struct sockaddr_in *sin;

    bzero((caddr_t)&ar, sizeof ar);
    ar.arp_pa.sa_family = AF_INET;
    sin = (struct sockaddr_in *)&ar.arp_pa;
    sin->sin_family = AF_INET;
    sin->sin_addr.s_addr = mnaddr;

    errno = 0;
    if (ioctl(ioctl_sockid, SIOCGARP, (caddr_t)&ar) < 0) 
        return(-1 * (errno));
    else {
      memcpy(mnetheraddr, ar.arp_ha.sa_data, ETH_ALEN);
      return(0);
    }
} 


/*
 * Send out a grarp to reestablish ARP mapping between MN and
 * MN's own hardware address.
 */
arprefresh(u32 mnaddr)
{
    int i, ret;
    int found;
    u32 ifaceaddr;
    char devname[MAX_IFNAME_LEN + 1];
    unsigned char hainterfaceaddr[ETH_ALEN + 1];
    unsigned char mnetheraddr[ETH_ALEN + 1];
    unsigned char zero_addr[ETH_ALEN + 1] = {0x00, 0x00, 0x00, 
					   0x00, 0x00, 0x00};  

    mip3dbg(("Arp refresh called for %d.%d.%d.%d\n", 
	   (unsigned char) ((ntohl(mnaddr) >> 24) & 0xff),
	   (unsigned char) ((ntohl(mnaddr) >> 16) & 0xff),
	   (unsigned char) ((ntohl(mnaddr) >> 8) & 0xff),
	   (unsigned char) (ntohl(mnaddr) & 0xff)));

    /* Get the MN's entry in haMobileNodeTable */
    found = 0 ;
    for (i = 0; i < haMobileNodeTableSize; i++) {
        if (mnaddr == haMobileNodeTable[i].haMnAddr) { 
	    found = 1;
	    ifaceaddr = haMobileNodeTable[i].haBindingIfaceAddr;
	    break;
        }
    }

    if (!found)
       return(-1 * MN_ENTRY_ABSENT); /* Unlikely to occur */
  
    /* Get the matching maIfaceAddr from mnAdvConfigTable */
    found = 0;
    for (i = 0; i < maAdvConfigTableSize; i++) {
        if (ifaceaddr == maAdvConfigTable[i].maIfaceAddr) {
	   found = 1;
	   strcpy(devname, maAdvConfigTable[i].maIfaceName);
	   memcpy(hainterfaceaddr, maAdvConfigTable[i].maIfaceHWaddr, ETH_ALEN);
	   break;
        }
    }

    if (!found)
        return(-1 * MN_ENTRY_ABSENT); /* Unlikely to occur */

    if ((ret = arpgetHWaddr(mnaddr, mnetheraddr)) < 0)
        return(ret);

    if((memcmp(hainterfaceaddr, mnetheraddr, ETH_ALEN) == 0) ||
       (memcmp(zero_addr, mnetheraddr, ETH_ALEN) == 0))
        return(-1 * NO_MAPPING);  
    else
        return(garp(devname, mnaddr, mnetheraddr));
}

/*
 * Add or Delete the route depending on argument cmd 
 */
int routemodify(unsigned long dst, unsigned long gw, int metric, char cmd)
{
    struct rtentry route;
    struct sockaddr_in *sin;
    int ret;
    int val = 0;

    /*
     * Workaround for Solaris 2.6 which seems to use an ARP
     * entry for x even when no local route exists for x.
     */
    if (cmd == 'd') {
        val = arpdel(dst);
    }

    /* Now do the actual route modification */
    sin = (struct sockaddr_in *)&route.rt_dst;
    sin->sin_addr.s_addr = dst;

    sin = (struct sockaddr_in *)&route.rt_gateway;
    sin->sin_addr.s_addr = gw;

    route.rt_flags = RTF_UP | RTF_HOST;
	
    if ((metric > 0) && (cmd == 'a'))
	route.rt_flags |= RTF_GATEWAY;

    errno = 0;
    if ((ret = ioctl(ioctl_sockid, cmd == 'a' ? SIOCADDRT : SIOCDELRT,
	  (caddr_t)&route)) < 0){
        /* 
	 * SIOCADDRT does not return ENXIO & 
	 * SIOCDELRT never returns EEXIST
	 */
        if (!((errno == ENXIO) || (errno == EEXIST))) 
           return(-1 * (errno));
    } 

    return (val);
}

/* Add route */
int routeadd(unsigned long dst, unsigned gw,int metric)
{
   return(routemodify(dst, gw, metric, 'a'));
}

/* Delete the route */
int routedel(unsigned long dst, unsigned gw)
{
   return(routemodify(dst, gw , 0, 'd'));
}

/*
 * Add routing entry for the tunnel 
 */
int mkroutethroughtunnel(u32 target_addr)
{
    return(routeadd(target_addr, inet_addr("255.255.255.254"), 1));  
}

/*
 * Delete the routing entry corresponding to the tunnel 
 */ 
int rmroutethroughtunnel(u32 target_addr)
{
    return(routedel(target_addr, inet_addr("255.255.255.254")));
} 


/* Routines for refresh_mn_arp() */
/* -------- Start of dlcommon routines */
/*
 * Common (shared) DLPI test routines.
 * Mostly pretty boring boilerplate sorta stuff.
 * These can be split into individual library routines later
 * but it's just convenient to keep them in a single file
 * while they're being developed.
 *
 * Not supported:
 *   Connection Oriented stuff
 *   QOS stuff
 */


int
dlinforeq(fd)
int	fd;
{
    dl_info_req_t	info_req;
    struct	strbuf	ctl;
    int	flags;

    info_req.dl_primitive = DL_INFO_REQ;

    ctl.maxlen = 0;
    ctl.len = sizeof (info_req);
    ctl.buf = (char *) &info_req;

    flags = RS_HIPRI;
    errno = 0;
    if (putmsg(fd, &ctl, (struct strbuf*) NULL, flags) < 0){
	return(-1 * errno);
    }
    return(0);
}

int
dlinfoack(fd, bufp)
int	fd;
char	*bufp;
{
    union DL_primitives	*dlp;
    struct strbuf ctl;
    int flags;
    int ret ;

    ctl.maxlen = MAXDLBUF;
    ctl.len = 0;
    ctl.buf = bufp;

    strgetmsg(fd, &ctl, (struct strbuf*)NULL, &flags, "dlinfoack");

    dlp = (union DL_primitives *) ctl.buf;

    if ((ret = expecting(DL_INFO_ACK, dlp)) < 0) {
       close(fd);
       return(ret);
    }
	 
    if (ctl.len < sizeof (dl_info_ack_t)){
        return(-1 * DLOKACK_SHORT_RESPONSE);
    }

    if (flags != RS_HIPRI) {
	return(-1 * DLOKACK_NOT_M_PCPROTO);
    }

    if (ctl.len < sizeof (dl_info_ack_t)){
	return(-1 * DLOKACK_SHORT_RESPONSE);
    }

    return(0);
}

int
dlattachreq(fd, ppa)
int	fd;
u_long	ppa;
{
    dl_attach_req_t attach_req;
    struct strbuf ctl;
    int	flags;

    attach_req.dl_primitive = DL_ATTACH_REQ;
    attach_req.dl_ppa = ppa;

    ctl.maxlen = 0;
    ctl.len = sizeof (attach_req);
    ctl.buf = (char *) &attach_req;

    flags = 0;
    errno = 0;
    if (putmsg(fd, &ctl, (struct strbuf*) NULL, flags) < 0)
	return(-1 * errno);
    return(0);
}


int
dlbindreq(fd, sap, max_conind, service_mode, conn_mgmt, xidtest)
int	fd;
u_long	sap;
u_long	max_conind;
u_long	service_mode;
u_long	conn_mgmt;
u_long	xidtest;
{
    dl_bind_req_t bind_req;
    struct strbuf ctl;
    int flags;

    bind_req.dl_primitive = DL_BIND_REQ;
    bind_req.dl_sap = sap;
    bind_req.dl_max_conind = max_conind;
    bind_req.dl_service_mode = service_mode;
    bind_req.dl_conn_mgmt = conn_mgmt;
    bind_req.dl_xidtest_flg = xidtest;

    ctl.maxlen = 0;
    ctl.len = sizeof (bind_req);
    ctl.buf = (char *) &bind_req;

    flags = 0;
	
    errno = 0;
    if (putmsg(fd, &ctl, (struct strbuf*) NULL, flags) < 0)
	return(-1 * errno);
    return(0);
}

int
dlunitdatareq(fd, addrp, addrlen, minpri, maxpri, datap, datalen)
int	fd;
u_char	*addrp;
int	addrlen;
u_long	minpri, maxpri;
u_char	*datap;
int	datalen;
{
    long buf[MAXDLBUF];
    union DL_primitives	*dlp;
    struct strbuf data, ctl;

    dlp = (union DL_primitives*) buf;

    dlp->unitdata_req.dl_primitive = DL_UNITDATA_REQ;
    dlp->unitdata_req.dl_dest_addr_length = addrlen;
    dlp->unitdata_req.dl_dest_addr_offset = sizeof (dl_unitdata_req_t);
    dlp->unitdata_req.dl_priority.dl_min = minpri;
    dlp->unitdata_req.dl_priority.dl_max = maxpri;

    (void) memcpy(OFFADDR(dlp, sizeof (dl_unitdata_req_t)), addrp, addrlen);

    ctl.maxlen = 0;
    ctl.len = sizeof (dl_unitdata_req_t) + addrlen;
    ctl.buf = (char *) buf;

    data.maxlen = 0;
    data.len = datalen;
    data.buf = (char *) datap;
	
    errno = 0;
    if (putmsg(fd, &ctl, &data, 0) < 0) 
	return(-1 * errno);
    return(0);
}


int
dlokack(fd, bufp)
int	fd;
char	*bufp;
{
    union DL_primitives	*dlp;
    struct	strbuf	ctl;
    int	flags;
    int ret; 

    ctl.maxlen = MAXDLBUF;
    ctl.len = 0;
    ctl.buf = bufp;

    strgetmsg(fd, &ctl, (struct strbuf*)NULL, &flags, "dlokack");

    dlp = (union DL_primitives *) ctl.buf;

    if ((ret = expecting(DL_OK_ACK, dlp)) < 0)
	return(ret);

    if (ctl.len < sizeof (dl_ok_ack_t))
	return(-1 * DLOKACK_SHORT_RESPONSE);

    if (flags != RS_HIPRI)
	return(-1 * DLOKACK_NOT_M_PCPROTO);

    if (ctl.len < sizeof (dl_ok_ack_t))
	return(-1 * DLOKACK_SHORT_RESPONSE);

    return(0);
}

int
dlbindack(fd, bufp)
int	fd;
char	*bufp;
{
    union DL_primitives	*dlp;
    struct strbuf	ctl;
    int	flags;
    int ret ;

    ctl.maxlen = MAXDLBUF;
    ctl.len = 0;
    ctl.buf = bufp;

    strgetmsg(fd, &ctl, (struct strbuf*)NULL, &flags, "dlbindack");

    dlp = (union DL_primitives *) ctl.buf;

    if ((ret=expecting(DL_BIND_ACK, dlp)) < 0)
	return(ret);

    if (flags != RS_HIPRI)
	return(-1 * DLOKACK_NOT_M_PCPROTO);
	
    if (ctl.len < sizeof (dl_bind_ack_t))
	return(-1 * DLOKACK_SHORT_RESPONSE);

    return(0);
}

int
strgetmsg(fd, ctlp, datap, flagsp, caller)
int	fd;
struct	strbuf	*ctlp, *datap;
int	*flagsp;
char	*caller;
{
    int	rc;

    /*
     * Set flags argument and issue getmsg().
     */
    *flagsp = 0;
    errno = 0;
    if ((rc = getmsg(fd, ctlp, datap, flagsp)) < 0) {
	return(-1 * errno);
    }

    /*
     * Check for MOREDATA and/or MORECTL.
     */
    if ((rc & (MORECTL | MOREDATA)) == (MORECTL | MOREDATA))
	return(-1 * ERR_MORECTLDATA);
	
    if (rc & MORECTL)
	return(-1 * ERR_MORECTL);

    if (rc & MOREDATA)
	return(-1 * ERR_MOREDATA);
    
    /*
     * Check for at least sizeof (long) control data portion.
     */
    if (ctlp->len < sizeof (long))
	return(-1 * SHORT_CONTROL_PORTION);

   return(0);
}

int expecting(prim, dlp)
int	prim;
union	DL_primitives	*dlp;
{
    if (dlp->dl_primitive != (u_long)prim)
	return(-1 * DL_PRIMITIVE_ERROR);

    return(0);
}


/*
 * Return string.
 */
void
addrtostring(addr, length, s)
u_char	*addr;
u_long	length;
u_char	*s;
{
    int	i;

    for (i = 0; i < length; i++) {
	(void) sprintf((char*) s, "%x:", addr[i] & 0xff);
	s = s + strlen((char*)s);
    }
    if (length)
	*(--s) = '\0';
}

/*
 * Return length
 */

int
stringtoaddr(sp, addr)
char	*sp;
char	*addr;
{
    int	n = 0;
    char *p;
    int	val;

    p = sp;
    while (p = strtok(p, ":")) {
	if (sscanf(p, "%x", &val) != 1)
	    return(-1 * INVALID_STRING);
	if (val > 0xff)
	    return(-1 * INVALID_STRING);
	*addr++ = val;
	n++;
	p = NULL;
    }

    return (n);
}

  /*-------- End of dlcommon routines    */

/*
 * The parms are as follows:
 * device	String definition of the device (e.g. "/dev/le").
 * ppa		Int for the instance of the device (e.g. 3, for "le3").
 * phys		Byte String for the dst MAC address of this packet.
 *		This is just the bcast address ("ff:ff:ff:ff:ff:ff") 
 * physlen	Length (int) of the mac address (not the string). E.g:
 *		6 for enet.
 * ipaddr	long for the ip address of the host we are advertising
 *		our mac address for.
 * ether_addr   Ether address we want to set for the ipaddress we are
 *              impersonating
 * NOTE: The mac addr of this system is not passed in, it is obtained 
 * 	directly from the dlpi info ack structure.
 * Steps executed :-
 * -----------------
 * Open datalink provider.
 * Attach to PPA.
 * Bind to sap
 * Send arp request as DL_UNITDATA_REQ msg
 * 
 */
int send_gratuitous_arp(char *device, int ppa, u_char *phys, int physlen,
		u_long ipaddr, u_char *ether_addr)
{
    int saplen;
    int	size = sizeof(struct ether_arp);
    int	sapval = ETHERTYPE_ARP;
    int	localsap = ETHERTYPE_ARP;
    int	fd;
    long buf[MAXDLBUF];
    u_char sap[MAXDLADDR];
    u_char addr[MAXDLADDR];
    int	addrlen;
    struct ether_arp req;
    union DL_primitives	*dlp;
    int	i, ret;
    u_long target_ipaddr = ipaddr;


    /* initialize buf[] */
    for (i = 0; i < MAXDLBUF; i++)
	buf[i] = (unsigned char) i & 0xff;

    errno = 0;
    /* Open the device.*/
    if ((fd = open(device, 2)) < 0)
	return(-1 * errno);

    /* Attach. */
    if ((ret = dlattachreq(fd, ppa)) < 0) {
        close(fd);
	return(ret);
    }

    if ((ret = dlokack(fd, buf)) < 0) {
	close(fd);
	return(ret);
    }

    /* Bind.*/
    if ((ret = dlbindreq(fd, localsap, 0, DL_CLDLS, 0, 0)) < 0) {
	close(fd);
	return(ret);
    }

    if ((ret = dlbindack(fd, buf)) < 0) {
	close(fd);
	return(ret);
    }

    /* Get info */
    if ((ret = dlinforeq(fd)) < 0) {
	close(fd);
	return(ret);
    }

    if ((ret = dlinfoack(fd, buf)) < 0){
	close(fd);
	return(ret);
    }

    /*
     * Verify sap and phys address lengths.
     */

    dlp = (union DL_primitives*) buf;

    addrtostring(OFFADDR(dlp, dlp->info_ack.dl_addr_offset),
	dlp->info_ack.dl_addr_length, addr);

    saplen = ABS(dlp->info_ack.dl_sap_length);

    /*   
     * Convert destination address string to address.
     */
    for (i = 0; i < saplen; ++i) {
	int rev_index = saplen - 1 -i;
	sap[i] = (sapval >> (rev_index * BITSPERBYTE)) & 0xff;
    }

    /*
     printdlprim(dlp);
     */
    if (physlen != (dlp->info_ack.dl_addr_length - saplen)) {
	close(fd);
	return(-1 * INVALID_ADDR);
    }
       
    addrlen = saplen + physlen;

    /*
     * Construct destination address.
     */
    if (dlp->info_ack.dl_sap_length > 0) {	/* order is sap+phys */
	(void) memcpy((char*)addr, (char*)sap, saplen);
	(void) memcpy((char*)(addr+saplen), (char*)phys, physlen);

	/* obtain our MAC address */
	/* memcpy((char *)my_etheraddr,
		(char *)OFFADDR(dlp, 
			dlp->info_ack.dl_addr_offset + saplen),
			physlen);
	*/
    } else {	/* order is phys+sap */
	(void) memcpy((char*)addr, (char*)phys, physlen);
	(void) memcpy((char*)(addr+physlen), (char*)sap, saplen);
	
	/* Obtain our MAC address */
	/*
	memcpy((char *)my_etheraddr,
	(char *)OFFADDR(dlp, dlp->info_ack.dl_addr_offset), physlen);
	*/
	}
	
    /* create arp request */
    memset((char *)&req, 0, sizeof(req));
    req.arp_hrd = htons(ARPHRD_ETHER);
    req.arp_pro = htons(ETHERTYPE_IP);
    req.arp_hln = ETHERADDRL;
    req.arp_pln = IPADDRL;
    req.arp_op = htons(ARPOP_REQUEST);
    memcpy((char *)&req.arp_sha, (char *)ether_addr, ETHERADDRL);
    memcpy((char *)&req.arp_spa, (char *) &ipaddr, IPADDRL);
    memcpy((char *)&req.arp_tpa, (char *) &target_ipaddr, IPADDRL);

    /* Transmit it. */

    if ((ret = dlunitdatareq(fd, addr, addrlen, 0, 0, (char *)&req, size)) < 0){
        close(fd);
	return(ret);
    }

    close(fd);
    return(0);
} /* send_gratuitous_arp() */


/* SYNOPSIS:
 *	garp interface ipaddr ether_addr
 *
 * RETURN VALUE :
 *       Returns 0 on success else -(error code)
 * "interface" is the interface to send the grarp out on.
 * "interface" is expressed in ifname convention (e.g. "le0", "bf2"),
 * and it will be converted by ifname2device_ppa() into the dlpi convention
 * (e.g. "/dev/le" + "0" and "/dev/bf" + "2").
 * "ipaddr" is the ipaddr of the system we're "impersonating"
 * ether_addr is the ethernet address to which we want to be impersonated. 
 * Sends a gratuitous arp to hw addr ff:ff:ff:ff:ff:ff.
 *
 * The arp packet  fields are filled in as follows:
 * (To be in conformance gratuitous arp described in RFC2002)
 * In the gratuitous arp packet for Mobile - IP :
 * Ethernet header :
 * Source address      : Its own hardware address
 * Destination address : ff:ff:ff:ff:ff:ff (Broadcast address)
 * Frame Type          : 0x0806 Arp Request/Reply
 *
 * Arp Packet :
 * Format of hardware address : ARPHRD_ETHER   1
 * Format of protocol address :                0x0800
 * Length of hardware address : ETH_ALEN       6
 * Length of protocol address :                4
 * ARP opcode                 : ARPOP_REQUEST  1

 * Sender hardware address : Ethernet address to which to be updated.
 * Destination harware address : XXXX Don't Care
 * Sender IP address    : IP Address (Cache entry to be updated)
 * Target IP address    : Sender IP Address. *      For an ARP Request Packet :
 *
 *     Note : For ARP Reply packet target IP address should be set same
 *            as source IP address
 * 
 * Modified from gabriels's arp module which was
 * blatantly stolen from dlunitdatareq.c by Neal Nuckolls. Just added
 * stuff to compose an arp packet. 
 */
int		
garp(char *dev, u32 mnaddr, unsigned char *mnetheraddr)
{
    int	ppa;
    u_char phys[MAXDLADDR]; 
    char device[MAX_INPUT];	

    ifname2devppa(dev, device, &ppa);
    memset(phys, 0xff, ETH_ALEN);

    /* Validate arguments.*/
    if (ppa < 0)
        return(-1 * INVALID_PPA);

    return(send_gratuitous_arp(device, ppa, phys, ETH_ALEN,
			  mnaddr, mnetheraddr));
}

int setMTU(char *ifname, int mtu)
{
    struct ifreq ifr;

    memset((char *)&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, ifname, sizeof (ifr.ifr_name));
    ifr.ifr_metric = mtu;
    if (ioctl(ioctl_sockid, SIOCSIFMTU, (caddr_t)&ifr) < 0)
        return(-1 * errno);
    return(0); 
}


int getMinMTU()
{

    int i;
    int ifaceCnt;
    struct ifconf ifc;
    struct ifreq ifr;
    struct ifreq *ifrp;
    struct sockaddr_in *sin;  
    int bufsize;
    char *buf;
    int minMTU = 5000;     

    /* Get the number of interfaces */
    if (ioctl(ioctl_sockid, SIOCGIFNUM, (char *)&ifaceCnt) < 0) 
        goto usedefault;

    bufsize = ifaceCnt * sizeof(struct ifreq);
    if ((buf = (char *) malloc(bufsize)) == NULL) 
        goto usedefault;
    
    /* Get the list of interface names */
    memset((char *)&ifc, 0, sizeof(ifc));
    ifc.ifc_len = bufsize;
    ifc.ifc_buf = buf;
    if (ioctl(ioctl_sockid, SIOCGIFCONF, (char *) &ifc) < 0) {
        free(buf);	
	goto usedefault;
    }       

    /* 
     * Get the MTU of each interface, ifc_req points to the same
     * place as ifc_buf
     */
    ifrp = ifc.ifc_req;   
    for (i = 0; i < (ifc.ifc_len/sizeof(struct ifreq)); i++, ifrp++) {
        /* ... flags defined in <net/if.h> */
        bzero((char *)&ifr, sizeof(ifr));
	strncpy(ifr.ifr_name, ifrp->ifr_name, 8); 
	if ((strcmp(ifr.ifr_name, "lo") == 0) || 
	    (strcmp(ifr.ifr_name, "ip0") == 0)) 
	    continue;
	    
	/* Get the interface MTU */
	if (ioctl(ioctl_sockid, SIOCGIFMTU, (char *)&ifr) == 0) {
	    if (ifr.ifr_metric < minMTU) {
	        minMTU = ifr.ifr_metric; 
	    }
	}
    }

    free(buf);
    if (minMTU != 5000)
        return(minMTU);

usedefault:
    return(DEFAULT_MIN_MTU);
}

void OScleanup()
{
    close(ioctl_sockid);
    return;
}
