/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Original Code is the Sunlabs Mobile IP.
 The Initial Developer of the Original Code is Vipul Gupta.
 Portions created by Vipul Gupta are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): Niko A. Fikouras (niko@comnets.uni-bremen.de)
		 Asanga D. Udugama (adu@comnets.uni-bremen.de)

*/

/*
 * An implementation of Mobile-IP as described in RFC 2002.
 * Author: Vipul Gupta (vgupta@eng.sun.com)
 * utils.c : Miscellaneous utility routines some of which may need to be
 *           implemented in a platform-specific fashion.
 */

#pragma ident   "@(#)utils.c 1.8 98/12/16"

#include <stdio.h>
#include <sys/time.h>
#include <time.h>

#include "mip.h"
#include "agent.h"

#define SECONDS_IN_A_DAY     86400
#define NTP_ADJUSTMENT       (SECONDS_IN_A_DAY * (((u32) 365*70) + 17))

extern int mipdebug;

/*--------------------- Platform-specific utilities --------------------*/

/* initialize the random number generator */
void randomInit() 
{
    struct timeval time;

    if (gettimeofday(&time, 0) < 0) {
        mip0dbg(("Gettimeofday failed in randominit(), using fixed seed.\n"));
	srandom(181067);
    } else {
        srandom(time.tv_sec);
    }
}

/* returns an unsigned 32-bit random number */
unsigned long randomLong()
{
  return ((unsigned long) random());
}

/*
 * Returns the number of seconds elapsed since Jan 1, 1990 ie.
 * higher 32 bits of the 64-bit quantity representing current
 * time in NTP format.
 */

u32 CurrentTimeNTPSec()
{
    struct timeval time;

    if (gettimeofday(&time, 0) < 0) {
        mip0dbg(("Error: gettimeofday failed in CurrentTimeNTPSec()"));
	return(0);
    } else {
        return ((u32) (time.tv_sec + NTP_ADJUSTMENT));
    }
}

/* 
 * Print the current time in buffer. If buffer is large enough, 
 * the current time is printed in the format: 
 *                 Fri Sep 13 00:00:00 1986\0
 * Useful for debugging.
 */
char * sprintTime(char *buffer, int buflen)
{
    struct timeval clock;

    if (buflen >= 26) {
        gettimeofday(&clock, 0);
#ifdef _POSIX_THREAD_SAFE_FUNCTIONS       
        ctime_r((time_t *) &(clock.tv_sec), buffer);
#else       
        ctime_r((time_t *) &(clock.tv_sec), buffer, buflen);
#endif
	buffer[24] = '\0';
	return(buffer);
    } else {
        return("");
    }
}

/*
 * Print relative time since first call to this routine in the form
 *    seconds.xxxxxx\0
 */

char *sprintRelativeTime(char *buffer, int buflen)
{
    static int i = 0;
    static struct timeval starttime;
    static struct timeval now;
    long diff_sec, diff_usec;

    if (buflen < 20) {
        return ("");
    }

    if (i == 0) {
        gettimeofday(&starttime, 0);
	sprintf(buffer, "%d.%06d", 0, 0);
        i = 1;
    } else {
        gettimeofday(&now, 0);
	diff_sec = now.tv_sec - starttime.tv_sec;
	if (now.tv_usec < starttime.tv_usec) {
	    diff_usec = 1000000 + now.tv_usec - starttime.tv_usec;
	    diff_sec -= 1;
	} else {
	    diff_usec = now.tv_usec - starttime.tv_usec;
	}
	sprintf(buffer, "%d.%06d", diff_sec, diff_usec);
    }

    return(buffer);
}

        
/*--------------------- Platform-independent utilities --------------------*/

/* Compute the internet checksum for len number of bytes starting at addr. */
unsigned short inChecksum(unsigned short *addr, int len)
{
    register int nleft = len;
    register unsigned short *w = addr;
    register unsigned short answer;
    unsigned short odd_byte = 0;
    register int sum = 0;

    /*
     *  Our algorithm is simple, using a 32 bit accumulator (sum),
     *  we add sequential 16 bit words to it, and at the end, fold
     *  back all the carry bits from the top 16 bits into the lower
     *  16 bits.
     */
    while( nleft > 1 )  {
        sum += *w++;
	nleft -= 2;
    }

    /* mop up an odd byte, if necessary */
    if( nleft == 1 ) {
        *(unsigned char *)(&odd_byte) = *(unsigned char *)w;
	sum += odd_byte;
    }

    /*
     * add back carry outs from top 16 bits to low 16 bits
     */
    sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
    sum += (sum >> 16);			/* add carry */
    answer = ~sum;			/* truncate to 16 bits */
    return (answer);
}

/*
 * Converts the long value in addr_long to an internet address of
 * the type a.b.c.d in addr_string and returns a pointer to addr_string.
 * NOTE: addr_long is in network byte order and addr_string is assumed
 * to be long enough (atleast 16 bytes).
 *
 * By having the caller supply its own buffer, we try to avoid problems
 * associated with multiple threads writing the same buffer concurrently.
 */
char * ntoa(u32 addr_long, char * addr_string)
{
    u32    temp;

    temp = ntohl(addr_long);
    sprintf(addr_string, "%d.%d.%d.%d",
	    ((temp >> 24) & 0xff), ((temp >> 16) & 0xff),
	    ((temp >> 8) & 0xff), (temp & 0xff));
    return(addr_string);
}

/*
 * Converts an ethernet address of type aa:bb:cc:dd:ee:ff to a byte array 
 * specified in the first argument.
 */
int hwAddrRead(unsigned char s[], char *hwadr)  {
    int a, b, c, d, e, f;

    if (sscanf(hwadr,"%x:%x:%x:%x:%x:%x", &a, &b, &c, &d, &e, &f) != 6) 
     return(-1);
    else  {
        s[0] = (unsigned char) a;
	s[1] = (unsigned char) b;
	s[2] = (unsigned char) c;
	s[3] = (unsigned char) d;
	s[4] = (unsigned char) e;
	s[5] = (unsigned char) f;
	return(0);
    }
}

char * hwAddrWrite(unsigned char hwaddr[], char *hwStr)  {

    sprintf(hwStr,"%x:%x:%x:%x:%x:%x", 
	    (unsigned char) hwaddr[0],
	    (unsigned char) hwaddr[1],
	    (unsigned char) hwaddr[2],
	    (unsigned char) hwaddr[3],
	    (unsigned char) hwaddr[4],
	    (unsigned char) hwaddr[5]);
    return(hwStr);
}

boolean isBlank(char line[], int len)
{
    int  i;
    char j;

    for (i = 0; i < len ; i++) {
        j = line[i];
        if ((j >= '0' && j <= '9') ||
	    (j >= 'a' && j <= 'z') ||
	    (j >= 'A' && j <= 'Z')) {
	    return (FALSE);
	}
    }

    return (TRUE);
}

int hexDigit(int c)
{
    if (c >= '0' && c <= '9')
        return(c - '0');
    if (c >= 'a' && c <= 'f')
        return(c - 'a' + 10);
    if (c >= 'A' && c <= 'F')
        return(c - 'A' + 10);
    return (-1);
}

/* Convert a string of hexadecimal digits into an array of len bytes
 * each containing a byte represented by a pair of hex digits, e.g.
 * keystr="12abC4De"
 * is converted to key[]={0x12, 0xab, 0xC4, 0xDe}
 * key has space for len bytes only.
 */
int hexConvert(char *key, int len, char *keystr)
{
    int i, d1, d2;

    for (i = 0; i < len; ++i) {
        if ((d1 = hexDigit(*keystr++)) == -1)
	    return (-1);
	if ((d2 = hexDigit(*keystr++)) == -1)
	    return (-1);
	*key++ = d1*16 + d2;
    }
    
    return (0);
}

/* 
 * Computes prefix length of a network mask. Assumes it is the same
 * as the number of bits set to one.
 */
int prefixLen(u32 netmask)
{
    int len = 0;

    while (netmask) {
       len++;
       netmask &= (netmask - 1);
    }

    return(len);
}

/*
 * Get the idxth word from string str
 * if str is "123 245 dis" then getWord(str, 2, word) will
 * return 0 and word will cotain "245"
 * If the idxth word is not found then it will return -1
 */
int getWord(char *str, int idx, char *word)
{
    int i;
    int wordIdx = 0;
    int len = strlen(str);
    char str1[30];

    for(i = 0; i < len; i++){
        if (str[i] == ' ')
            continue;
        wordIdx++;
        if (sscanf((str+i), "%s", str1) != 1)
            return(-1);
        if (wordIdx == idx) {
            strcpy(word, str1);
            return(0);
        }
        i += strlen(str1);
    }
    return(-1);
}


/* Prints a buffer in hexadecimal. */
void printBuffer(unsigned char buffer[], int buflen)
{
    unsigned int i;

    if (buflen == 6) {
        /* Is likely an ethernet address */
        for (i = 0; i < buflen; i++) {
	    mip3dbg(("%02x", buffer[i]));
	    if (i < (buflen-1)) mip3dbg((":"));
	}
    } else {
        /* Short enough to be printed in one shot */
        if (buflen <= 16) {
	    for (i = 0; i < buflen; i++)
	        mip3dbg(("%02x", buffer[i]));
	} else {
	    /* Possibly a packet */
	    mip3dbg(("\t"));
	    for (i = 0; i < buflen; i++) {
	        mip3dbg(("%02x", buffer[i]));
		if ((i % 2) == 1) mip3dbg((" "));
		if ((i % 16) == 15) mip3dbg(("\n\t"));
	    }
	}
    }
}

/* Prints the contents of a MaAdvConfigEntry. */
void printMaAdvConfigEntry(MaAdvConfigEntry *macep)
{
    char addr[20];

    mip3dbg(("MaAdvConfigEntry contents:\n"));
    mip3dbg(("\tInterface: %s\n", macep->maIfaceName));
    mip3dbg(("\tAddress  : %s\n", ntoa(macep->maIfaceAddr, addr)));
    mip3dbg(("\tNetmask  : %s\n", ntoa(macep->maIfaceNetmask, addr)));
    mip3dbg(("\tHWaddr   : "));
    if (mipdebug > 2) printBuffer(macep->maIfaceHWaddr, 6);
    mip3dbg(("\n"));
    mip3dbg(("\tAdv seq# : %d\n", macep->maAdvSeqNum));
    mip3dbg(("\tFlags    : %x (RBHFMGV_)\n", macep->maAdvServiceFlags));
    mip3dbg(("\tPrefix   : %s\n", 
	     macep->maAdvPrefixLenInclusion ? "Yes" : "No"));
}

void printProtectedDomainInfo(DomainInfo domainInfo)
{
    int i;
    char addr1[20], addr2[20];
    
    if (domainInfo.addrIntervalCnt > 0) {
        mip3dbg(("Protected Domain Info:  Address - Netmask pairs\n"));
        for (i = 0; i < domainInfo.addrIntervalCnt; i++) {
            mip3dbg(("         %s/%s\n",
                     ntoa(domainInfo.addr[i], addr1),
                     ntoa(domainInfo.netmask[i], addr2)));
        }
    }

    if (domainInfo.firewallCnt > 0) {
        mip3dbg(("No. of firewalls : %d\n", domainInfo.firewallCnt));
        mip3dbg(("List of firewall addresses : "));
        for (i = 0; i < domainInfo.firewallCnt; i++) {
            mip3dbg(("%s ", ntoa(domainInfo.fwAddr[i], addr1)));
        }
    }

    mip3dbg(("\n"));
    fflush(stdout);
}

void printMaAdvConfigTable(MaAdvConfigEntry Table[], int TableSize)
{
    int i;

    mip3dbg(("------ MaAdvConfigTable contents ------\n"));
    for (i = 0; i < TableSize; i++) {
        mip3dbg(("%5d. ", i));
	printMaAdvConfigEntry(&Table[i]);
    }
}

/* Prints the contents of a FaVisitorEntry. */
void printFaVisitorEntry(FaVisitorEntry *favep)
{
    char addr[20];

    mip3dbg(("FaVisitorEntry contents:\n"));
    mip3dbg(("\tVisitor  : %s\n", ntoa(favep->faVisitorAddr, addr)));
    mip3dbg(("\tIface    : %s\n", ntoa(favep->faVisitorIfaceAddr, addr)));
    mip3dbg(("\tStatus   : %s\n", (favep->faVisitorRegIsAccepted ? 
				 "Accepted" : "Pending")));
    mip3dbg(("\tTimeGrant: %ld\n", favep->faVisitorTimeGranted));
    mip3dbg(("\tTimeLeft : %ld\n", favep->faVisitorTimeRemaining));
    mip3dbg(("\tHomeAddr : %s\n", ntoa(favep->faVisitorHomeAddr, addr)));
    mip3dbg(("\tHomeAgent: %s\n", ntoa(favep->faVisitorHomeAgentAddr, addr)));
    mip3dbg(("\tCOAddr   : %s\n", ntoa(favep->faVisitorCOAddr, addr)));
    mip3dbg(("\tReg Flag : %x (SBDMGV__)\n", favep->faVisitorRegFlags));
    mip3dbg(("\tID High  : %lx\n", favep->faVisitorRegIDHigh));
    mip3dbg(("\tID Low   : %lx\n", favep->faVisitorRegIDLow));
}

void printFaVisitorTable(FaVisitorEntry Table[], int TableSize)
{
    int i;

    mip3dbg(("------ FaVisitorTable contents ------\n"));
    for (i = 0; i < TableSize; i++) {
        if (Table[i].faVisitorTimeRemaining) {
	    mip3dbg(("%5d. ", i));
	    printFaVisitorEntry(&Table[i]);
	}
    }
}

/* Prints the contents of a FaUnreachableEntry. */
void printFaUnreachableEntry(FaUnreachableEntry *fauep)
{
    char addr[20];

    mip3dbg(("FaUnreachableEntry contents:\n"));
    mip3dbg(("\tAddr     : %s\n", ntoa(fauep->faUnreachableAddr, addr)));
    mip3dbg(("\tTimeLeft : %ld\n", fauep->faUnreachableTimeRemaining));
}

void printFaUnreachableTable(FaUnreachableEntry Table[], int TableSize)
{
    int i;

    mip3dbg(("------ FaUnreachableTable contents ------\n"));
    for (i = 0; i < TableSize; i++) {
        if (Table[i].faUnreachableTimeRemaining) {
	    mip3dbg(("%5d. ", i));
	    printFaUnreachableEntry(&Table[i]);
	}
    }
}

/* Prints the contents of a HaBindingEntry. */
void printHaBindingEntry(HaBindingEntry *habep)
{
    char addr[20];

    mip3dbg(("HaBindingEntry contents:\n"));
    mip3dbg(("\tMN addr  : %s\n", ntoa(habep->haBindingMN, addr)));
    mip3dbg(("\tCO addr  : %s\n", ntoa(habep->haBindingCOA, addr)));
    mip3dbg(("\tSrc addr : %s\n", ntoa(habep->haBindingSrcAddr, addr)));
    mip3dbg(("\tSrc port : %d\n", habep->haBindingSrcPort));
    mip3dbg(("\tTimeGrant: %ld\n", habep->haBindingTimeGranted));
    mip3dbg(("\tTimeLeft : %ld\n", habep->haBindingTimeRemaining));
    mip3dbg(("\tReg Flag : %x (SBDMGV__)\n", habep->haBindingTimeRemaining));
}

void printHaBindingTable(HaBindingEntry Table[], int TableSize)
{
    int i;

    mip3dbg(("------ HaBindingTable contents ------\n"));
    for (i = 0; i < TableSize; i++) {
        if (Table[i].haBindingTimeRemaining) {
	    mip3dbg(("%5d. ", i));
	    printHaBindingEntry(&Table[i]);
	}
    }
}

/* Prints a Mobility Security Association Entry */
void printMSAE(MipSecAssocEntry *msae)
{
    char addr[20];

    mip3dbg(("\tMipSecAssocEntry contents:\n"));
    mip3dbg(("\tIP addr %s\n", ntoa(msae->mipSecAddr, addr)));
    mip3dbg(("\tSPI %d\n", msae->mipSecSPI));
    mip3dbg(("\tAlgo type %d\n", msae->mipSecAlgorithmType));
    mip3dbg(("\tAlgo Mode %d\n", msae->mipSecAlgorithmMode));
    mip3dbg(("\tKey Len %d\n", msae->mipSecKeyLen));
    mip3dbg(("\tKey: 0x"));
    if (mipdebug > 2) printBuffer(msae->mipSecKey, msae->mipSecKeyLen);
    mip3dbg(("\n"));
    mip3dbg(("\tReplay method %d\n", msae->mipSecReplayMethod));
}

/* Prints a Mobile Node entry */
void printHaMobileNodeEntry(HaMobileNodeEntry *hamne)
{
    char addr[20];

    mip3dbg(("HaMobileNodeEntry contents:\n"));
    mip3dbg(("\tMN addr %s\n", ntoa(hamne->haMnAddr, addr)));
    mip3dbg(("\tHA's Iface Addr %s\n", ntoa(hamne->haBindingIfaceAddr, addr)));
    mip3dbg(("\tMN binding count %d\n", hamne->haMnBindingCnt));
    mip3dbg(("\tID High: 0x%x\n", hamne->haMnRegIDHigh));
    mip3dbg(("\tID Low : 0x%x\n", hamne->haMnRegIDLow));
    printMSAE(&(hamne->haMnMSA));
}

void printHaMobileNodeTable(HaMobileNodeEntry Table[], int TableSize)
{
    int i;

    mip3dbg(("------ HaMobileNodeTable contents ------\n"));
    for (i = 0; i < TableSize; i++) {
        mip3dbg(("%5d. ", i));
	printHaMobileNodeEntry(&Table[i]);
    }
}

/* Functions for Filters */

/* Converts a short int received as a char array to a
 * short int
*/
void cvtCharArrayToShortInt(short int *si, unsigned char *ca) {
	int i;

	*si = 0;
	for(i = 0; i < 2; i++) {
		*si = *si * 256;
		*si = *si + (unsigned char) *(ca + i);
	}
}

/* Converts an ip address received as a char array to an
 * ip address data type (u32)
*/
void cvtCharArrayTou32(u32 *li, unsigned char *ca) {
	int i;

	*li = 0;
	for(i = 3; i >= 0; i--) {
		*li = *li * 256;
		*li = *li + *(ca + i);
	}
}

/* returns the number of bits in a ip address
 * mask (u32)
 */
int numBitsInMask(u32 addr) {
	u32 invAddr;
	int count;

	invAddr = ~addr;
    	invAddr = ntohl(invAddr);
	count = 0;
	while(invAddr > 0) {
		invAddr = invAddr >> 1;
		count++;
	}
	return (32 - count);
}

/* Checks if a string (str) exist inside another
 * string (ln) (expects ln to be longer than str)
 */
int isWordIn(char *ln, char *str) {
	int i, len;

	len = strlen(ln) - strlen(str);
	for(i = 0; i < len; i++) {
		if(*(ln + i) == ' ' || *(ln + i) == '\t')
			continue;
		if(strncmp((ln + i), str, strlen(str)) == 0)
			return TRUE;

	}
	return FALSE;
}

/* Replaces a string if it is found on the line that is
 * passed. Both findStr and replaceStr should be equal
 * in size.
 */
int replaceStr(char *ln, char *findStr, char *replaceStr) {
	int i, len, foundPos;

	len = strlen(ln) - strlen(findStr);
	foundPos = (-1);
	for(i = 0; i < len; i++) {
		if(*(ln + i) == ' ' || *(ln + i) == '\t')
			continue;
		if(strncmp((ln + i), findStr, strlen(findStr)) == 0) {
			foundPos = i;
			break;
		}
	}

	if(foundPos < 0)
		return FALSE;

	// shift the string
//	*(ln + strlen(ln) + strlen(replaceStr)) = (char) NULL;
//	for(i = strlen(ln) - 1; i > (foundPos + strlen(findStr) - 1); i--) {
//		*(ln + i + strlen(replaceStr)) = *(ln + i);
//	}

	//copy the string
	for(i = foundPos; i < foundPos + strlen(replaceStr); i++) {
		*(ln + i) = *(replaceStr + i - foundPos);
	}

	return TRUE;
}
