/*
 * Copyright (c) 1997 by Sun Microsystems, Inc.
 */

#pragma ident	"@(#)vtconfig.c 1.1 97/04/09"

/* Author: vgupta@eng.sun.com
 * Change history :
 *        Modified vtunlconf.c as vtconfig.c
 *
 * Compile as:  cc vtconfig.c -lthread -lnsl -lsocket
 *
 * A program that is used to control the behavior of 
 * a vtunl module.
 *
 *
 * The program undesrtands the following arguments : 
 * 
 *      vtconfig  plumb vtunl_src  [list of configured interfaces]
 *
 *       where vtunl_src is an IP address belonging to one of the
 *   network interfaces.
 *
 *
 *    vtconfig encapadd  target_addr   tunnel_exit_addr   encap_flags
 *
 *         Provide encapsulation service for target_addr, sending 
 *     encapsulated packets to tunnel_exit_addr. The argument
 *     encap_flags is an unsigned short in which the least 
 *     significant byte contains the Mobile-IP registration flags
 *     SBDMGV--. The encapsulation module may not support all flags
 *   . If required, a new routing table entry for target_addr (with
 *     dummy acting as gateway) is automatically created. This 
 *     command may fail due to any of the following reasons:
 *          EINVAL          A zero address was specified or an 
 *                          unsupported flag bit was turned on.
 *          ENOMEM          There wasn't sufficient memory to
 *                          allocate necessary data structures.
 *          ENOSPC          The limit on maximum Simultaneous
 *                          tunnels for one target was exceeded.
 *          EEXIST          Request duplicates existing service.
 *
 *     NOTE: Instead of specifying a target_addr, the keyword "default"
 *           may be used. This has the effect of inserting a default
 *           encapsulation entry.
 *
 *     vtconfig encaprem  target_addr   tunnel_exit_addr
 *
 *         Terminates encapsulation service for target_addr. If 
 *     target_addr is zero, all encapsulation services are cancelled.
 *     If tunnel_exit is zero then all existing encapsulation
 *     services for the given target_addr are terminated. If both 
 *     are non-zero, then only the encapsulation service (identified
 *     by the target_addr-tunnel_exit pair) is terminated. The 
 *     command may fail due to the following:
 *          ENXIO           Encapsulation service already unavailable.
 *
 *     Note: If a zero address is specified, it is the user's 
 *           responsibility to ensure that the routing table is 
 *           appropriately updated. Some routing table entries may
 *           need to be deleted.
 *
 *           Instead of specifying a target_addr, the keyword "default"
 *           may be used. This has the effect of removing the default
 *           encapsulation entry.
 *
 *           "encaprem all" is equivalent to encaprem 0.0.0.0 0.0.0.0
 *           "encaprem target_addr all" is equivalent to 
 *            encaprem target_addr 0.0.0.0
 *
 *     vtconfig decapadd  target_addr {dst|src|either}
 *
 *         Provide decapsulation service for target_addr if it appears
 *     in the decapsulated datagram as specified by the last argument.
 *     If the last argument is missing, it is assumed to be dst.
 *     If the target_addr is zero, all encapsulated packets are decapsulated
 *     and forwarded (the last argument is ignored). The equivalent 
 *     effect is achieved by "decapadd all". This command may fail due 
 *     to any of the following reasons:
 *          ENOMEM          Insufficient memory to allocate
 *                          necessary data structures.
 *
 *
 *     vtconfig decaprem  target_addr
 *
 *         Terminate decapsulation service for target_addr. One may use
 *     the keyword "all" or a 0.0.0.0 address to cancel ALL decapsulation
 *     services. This command may fail due to the following reason:
 *          ENXIO           Decapsulation service already unavailable.
 *
 *
 *
 *     vtconfig dump
 *   
 *         Dumps internal (global) configuration information for vtunl.
 *
 *
 *     vtconfig unplumb [list of plumbed interfaces]
 *
 *         Unplimbs the configured interface.
 *         Note that you need to unplumb "ip0" at the end
 *
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stream.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/sockio.h>
#include <sys/errno.h>

#include <sys/conf.h>
#include <sys/stropts.h>

#include <netinet/in.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <inet/common.h> 
#include <inet/ip.h> 
#include <netdb.h>

#include "vtunl.h"

#define ENCAPIFACE  "ip0"
#define DEVICEDIR   "/dev/"

int ioctl_fd;

struct exit_fill {
       u32      addr;
       u16      curr_pmtu;
       u16      max_pmtu;
       u16      state;
};

struct decap_fill {
       u32     addr;
       u16     flags;
};

int printusage(char *progname)
{
  /*        fprintf(stderr, "\nUsage is:\n\n\t%s <taddr> [%s]\n",
		progname, "list of configured network devices e.g. le0");
	fprintf(stderr, "or\n\t%s -n\n\n", progname);
	fprintf(stderr, "Here <taddr> is an address assigned to"
		"one of the configured network\n");
	fprintf(stderr, "interfaces, e.g. %s 129.146.122.174 le0\n", progname);
	fprintf(stderr, "\nWhen -n is used, existing vtunl plumbing is NOT"
		" disturbed. This\n");
	fprintf(stderr,"mode is suitable for use in scripts.\n\n");
	fprintf(stderr, "Type help at the program prompt for more info.\n\n");

	*/
  fprintf(stderr,"\nIncorrect usage of %s\n",progname);

}

/*
 * Parse the interface name (e.g. "le0" or "hme1") and place the
 * corresponding device name (e.g. "/dev/le", "/dev/hme") in
 * devname and the ppa (e.g. 0 or 1 for the examples above) in ppa.
 */
int ifname2devppa(char *ifname, char *devname, int *ppa)
{
      char *p;
      int  i, j, val;

      val = 0;
      j = strlen(DEVICEDIR);
      strncpy(devname, DEVICEDIR, j);
      for (i = 0, p = ifname; i < strlen(ifname); i++, p++) {
	   if (*p >= '0' && *p <= '9') 
	           val = 10*val + (*p - '0');
	   else 
	           devname[j++] = *p;
      }
      devname[j] = '\0';
      *ppa = val;
}

/*
 * Convert a long value (an IP address in network-byte-order) into
 * the dotted-decimal notation.
 */
static char *
ntoa(u32 addr_long, char * addr_string)
{
        u32    temp;
	
	temp = ntohl(addr_long);
	sprintf(addr_string, "%d.%d.%d.%d", 
		((temp >> 24) & 0xff), ((temp >> 16) & 0xff),
		((temp >> 8) & 0xff), (temp & 0xff));
	return(addr_string);
}

/*
 * Issue an ioctl for the vtunl module using the I_STR ioctl call 
 * (refer to man streamio).
 */
int 
vtunl_ctl(int s, int cmd, void *arg, int arglen)
{
       struct strioctl ioc;

       memset((char *) &ioc, 0, sizeof(ioc));
       ioc.ic_cmd = cmd;
       ioc.ic_timout = 0;
       ioc.ic_len = (size_t) arglen;
       ioc.ic_dp = (char *) arg;

       errno = 0;
       ioctl(s, I_STR, (char *) &ioc);
       return(errno);
}

/*
 * Complain with a sensible message.
 */
complain(int err)
{
       switch (err) {
       case ENXIO:
	       fprintf(stderr, "Not found.\n");
	       break;
       case EEXIST:
	       fprintf(stderr, "Already so.\n");
	       break;
       case ENOMEM:
	       fprintf(stderr, "No memory for data structures.\n");
	       break;
       case ENOSPC:
	       fprintf(stderr, "Would exceed system limit.\n");
	       break;
       case EINVAL:
	       fprintf(stderr, "Invalid arguments.\n");
	       break;
       default:
	       fprintf(stderr, "Unknown error %d.\n", err);
	       break;
       }
}

/*
 * Print the error message and exit
 */
Perror(char *c) {
  perror(c);
  exit(-1);
}

/*
 *  Configure the point-to-point interface named ifname with the
 *  given addresses.
 *
 *  If you wish to look at ip0 using ifconfig, always specify the
 *  inet option, e.g.
 *      ifconfig -a inet
 *  or
 *      ifconfig ip0 inet
 *
 *  If this option is not specified, ifconfig may attempt to print
 *  ethernet info for ip0 and hang.
 */
void mkpt2pt(char *ifname, long srcaddr, long dstaddr)
{   
       int s;
       struct ifreq ifr;
       struct sockaddr_in sin;
       short savedflags;
       
       if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0) 
	       Perror("socket");
       
       memset((char *) &ifr, 0, sizeof(ifr));
       strncpy(ifr.ifr_name, ifname, sizeof ifr.ifr_name);
       sin.sin_family = AF_INET;
       sin.sin_addr.s_addr = srcaddr;
       ifr.ifr_addr = *(struct sockaddr *) &sin;
       if (ioctl(s, SIOCSIFADDR, (caddr_t)&ifr) < 0) 
	       Perror("SIOCSIFADDR");

       memset((char *) &ifr, 0, sizeof(ifr));
       strncpy(ifr.ifr_name, ifname, sizeof ifr.ifr_name);
       if (ioctl(s, SIOCGIFFLAGS, (caddr_t)&ifr) < 0) 
	       Perror("SIOCSIFFLAGS");
       savedflags = ifr.ifr_flags;

       memset((char *) &ifr, 0, sizeof(ifr));
       strncpy(ifr.ifr_name, ifname, sizeof ifr.ifr_name);
       sin.sin_family = AF_INET;
       sin.sin_addr.s_addr = dstaddr;
       ifr.ifr_addr = *(struct sockaddr *)&sin;
       if (ioctl(s, SIOCSIFDSTADDR, (caddr_t)&ifr) < 0)
	       Perror("SIOCSIFDSTADDR");

       memset((char *) &ifr, 0, sizeof(ifr));
       strncpy(ifr.ifr_name, ifname, sizeof ifr.ifr_name);
       sin.sin_family = AF_INET;
       sin.sin_addr.s_addr = inet_addr("255.255.255.255");
       ifr.ifr_addr = *(struct sockaddr *)&sin;
       if (ioctl(s, SIOCSIFNETMASK, (caddr_t)&ifr) < 0)
	       Perror("SIOCSIFNETMASK");

       /* By bringing UP this interface, we force the creation of
	* a routing table entry for argv[2] through ip<ppa>.
	*/

       memset((char *) &ifr, 0, sizeof(ifr));
       strncpy(ifr.ifr_name, ifname, sizeof ifr.ifr_name);
       ifr.ifr_flags = savedflags | IFF_UP | 
	               IFF_PRIVATE | IFF_POINTOPOINT ;
       if (ioctl(s, SIOCSIFFLAGS, (caddr_t)&ifr) < 0) 
	       Perror("SIOCGIFFLAGS");
}

/*
 * Create a stream containing the named device, and vtunl. Then
 * configure it with the information provided in conf.
 * Finally, push "ip" and return the stream descriptor.
 */
int plumb(char *devname, struct vtunl_req *conf)
{
       int fd;
       struct vtunl_conf vc;

       if ((fd = open(devname, 0)) == -1) 
	     Perror("Open device in plumb failed.\n");
  
       if (ioctl(fd, I_PUSH, "vtunl") == -1) 
             Perror("I_PUSH vtunl");

       if (strncmp("/dev/ip", devname, strlen("/dev/ip")) == 0) {
	       if (vtunl_ctl(fd, VTUNL_SETCONF_TRANSPORT, conf, 
			   sizeof(struct vtunl_req))) 
		       Perror("VTUNL_SETCONF_TRANSPORT");

	       if (vtunl_ctl(fd, VTUNL_GETCONF, &vc, 
			     sizeof(struct vtunl_conf)))
	               Perror("VTUNL_GETCONF");
	 
	       fprintf(stdout, 
		       "Configured vtunl type %d proto %d addr %x (%d) ",
		       vc.type, vc.ip_protocol, ntohl(vc.address), vc.state);
       } else {
	       if (vtunl_ctl(fd, VTUNL_SETCONF_DLSUSER, conf, 
			     sizeof(struct vtunl_req)))
	               Perror("VTUNL_SETCONF_DLSUSER");

	       if (vtunl_ctl(fd, VTUNL_GETCONF, &vc, 
			     sizeof(struct vtunl_conf)))
		       Perror("VTUNL_GETCONF");
	 
	       fprintf(stdout,
		       "Configured vtunl type %d ppa %d addr %x mask %x (%d) ",
		       vc.type, vc.ip_protocol, ntohl(vc.address),
		       ntohl(vc.addr_mask), vc.state);
       }

       fprintf(stdout, "on %s.\n", devname);

       if (ioctl(fd, I_PUSH, "ip") == -1) 
	       Perror("I_PUSH ip");

       return(fd);
}

/* Process encapadd command */
int
encapadd(char *command) 
{
        int  err;
	struct vtunl_req vrq;
	char cmdname[20], target[32], t_exit[32];
	int  flags;

        if (sscanf(command, "%s %s %s %x", cmdname, target, t_exit, 
		   &flags) != 4) {
	 	fprintf(stderr, "Insufficient or bad arguments.\n");
		return;
	}

	memset((char *) &vrq, 0, sizeof(struct vtunl_req));

	if (strncmp(target, "default", strlen("default")) == 0) 
	        vrq.vr_target_addr = DEFAULT_ADDR;
	else {
	        vrq.vr_target_addr = inet_addr(target);

	        if (vrq.vr_target_addr == -1 && 
		    strcmp(target, "255.255.255.255")) {
		        fprintf(stderr,"Bad target.\n");
			return;
		}
	}

	if ((vrq.vr_t_exit_addr = inet_addr(t_exit)) == -1) {
	        fprintf(stderr, "Bad tunnel exit.\n");
		return;
	}

	vrq.vr_encap_flags = (u16) flags;

	if (err = vtunl_ctl(ioctl_fd, VTUNL_ENCAP_ADD, &vrq, sizeof(vrq))) {
	        complain(err);
		return;
	}

	/* We don't add a route for multicast addresses */
	if ((vrq.vr_target_addr & htonl(0xe0000000)) == htonl(0xe0000000))
	        return;

	/* We must add a route through vtunl */
	if (strcmp(target, "default") == 0) {
		  fprintf(stderr, "Delete older default route.\n");
	          sprintf(command, "route add default 255.255.255.254 1\n");
		  system(command);
	} else {
		  sprintf(command, "route add %s 255.255.255.254 1\n", target);
		  system(command);
	}
}
	
/* Process "encaprem all" */
int
encaprem1(char *target)
{
        int  err;
	struct vtunl_req vrq;


	if (strncmp(target, "all", strlen("all")) != 0) {
	        fprintf(stderr, "Bad target or missing exit.\n");
		return;
	}

	memset((char *) &vrq, 0, sizeof(struct vtunl_req));

	if (err = vtunl_ctl(ioctl_fd, VTUNL_ENCAP_REM, &vrq, sizeof(vrq))) {
	        complain(err);
		return;
	}

	/*
	 * We flushed all targets but don't know which
	 * routes to delete, just tell the user.
	 */
	
	fprintf(stdout, "All encapsulation requests %s\n",
		"will be denied, update routing tables.");
}

/* Process the more general version */
int
encaprem2(char *target, char *t_exit)
{
        int  err;
	struct vtunl_req vrq;
	char command[256];

	memset((char *) &vrq, 0, sizeof(struct vtunl_req));

	if (strncmp(target, "default", strlen("default")) == 0) 
	        vrq.vr_target_addr = DEFAULT_ADDR;
	else {
	        vrq.vr_target_addr = inet_addr(target);
	        if (vrq.vr_target_addr == -1 && 
		    strcmp(target, "255.255.255.255")) {
		        fprintf(stderr,"Bad target.\n");
			return;
		}
	}

	if (strncmp(t_exit, "all", strlen("all")) == 0)
	        vrq.vr_t_exit_addr = 0;
	else {
	        if (((vrq.vr_t_exit_addr = inet_addr(t_exit)) == -1) &&
		    strcmp(t_exit, "255.255.255.255")) {
		        fprintf(stderr, "Bad address.\n");
			return;
		}
	}

	if (err = vtunl_ctl(ioctl_fd, VTUNL_ENCAP_REM, &vrq, sizeof(vrq))) {
	        complain(err);
		return;
	}

	/* We don't delete routes for multicast addresses */
	if ((vrq.vr_target_addr & htonl(0xe0000000)) == htonl(0xe0000000))
	        return;

	/* 
	 * We must delete a route through vtunl unless there is still
	 * an encapsulation entry for this target.
	 */

	err = vtunl_ctl(ioctl_fd, VTUNL_ENCAP_OK, &vrq, sizeof(vrq));
	if (err != ENXIO) return;

	/* Delete route */

	if (strcmp(target, "default") == 0) {
	          sprintf(command, "route delete default 255.255.255.254\n");
		  system(command);
		  fprintf(stderr, "Restore older default route.\n");
	} else {
		  sprintf(command, "route delete %s 255.255.255.254\n",target);
		  system(command);
	}

}


/* Process encaprem command */
int
encaprem(char *command)
{
	char cmdname[20], target[32], t_exit[32];

	if (sscanf(command, "%s %s %s", cmdname, target, t_exit) == 3) 
	        encaprem2(target, t_exit);
	else if (sscanf(command, "%s %s", cmdname, target) == 2) 
	        encaprem1(target);
	else {
	        fprintf(stderr, "Bad command.\n");
		return;
	}
}

/* Process the single argument decapadd command */
int
decapadd1(char *target)
{
        int  err;
	struct vtunl_req vrq;

	if (strncmp(target, "all", strlen("all")) == 0) {
	  memset((char *) &vrq, 0, sizeof(struct vtunl_req));
	  if (err = vtunl_ctl(ioctl_fd, VTUNL_DECAP_ADD, &vrq, sizeof(vrq)))
	        complain(err);
	  else 
	        fprintf(stderr, "Decapsulation will be provided for all.\n");
	  return;
	} 
	  
	decapadd2(target, "dst");
}

/* Process the two-argument decapadd command */
int
decapadd2(char *target, char *flags)
{
        int err;
	struct vtunl_req vrq;

	memset((char *) &vrq, 0, sizeof(struct vtunl_req));
	
	if ((vrq.vr_target_addr = inet_addr(target)) == -1 &&
	    strcmp(target, "255.255.255.255")) {
	        fprintf(stderr,"Bad target.\n");
		return;
	}

	if (strncmp(flags, "dst", strlen("dst")) == 0) 
	        vrq.vr_decap_flags = DECAP_IF_DST;
	else if (strncmp(flags, "src", strlen("src")) == 0) 
	        vrq.vr_decap_flags = DECAP_IF_SRC;
	else if (strncmp(flags, "either", strlen("either")) == 0) 
	        vrq.vr_decap_flags = DECAP_IF_EITHER;
	else {
	        fprintf(stderr, "Bad flag.\n");
		return;
	}

	if (err = vtunl_ctl(ioctl_fd, VTUNL_DECAP_ADD, &vrq, sizeof(vrq)))
	        complain(err);
	else 
		fprintf(stdout, "Decapsulation enabled for %s as %s\n",
			target, flags);
}


/* Process decapadd command */
int
decapadd(char *command) 
{
	char cmdname[20], target[32], flags[20];

	if (sscanf(command, "%s %s %s", cmdname, target, flags) == 3) 
	        decapadd2(target, flags);
	else if (sscanf(command, "%s %s", cmdname, target) == 2) 
	        decapadd1(target);
	else {
	        fprintf(stderr, "Bad command.\n");
		return;
	}
}

/* Process decaprem command */
int
decaprem(char *command)
{
        int  err;
	struct vtunl_req vrq;
	char cmdname[20], target[32], t_exit[32];

	if (sscanf(command, "%s %s", cmdname, target) != 2) {
	        fprintf(stderr, "Bad command.\n");
		return;
	}

	memset((char *) &vrq, 0, sizeof(struct vtunl_req));

	if (strncmp(target, "all", strlen("all")) == 0) 
	        vrq.vr_target_addr = 0;
	else {
	  	vrq.vr_target_addr = inet_addr(target);

		if (vrq.vr_target_addr == -1 &&
		    strcmp(target, "255.255.255.255")) {
		        fprintf(stderr,"Bad target.\n");
			return;
		}
	}
	   
	if (err = vtunl_ctl(ioctl_fd, VTUNL_DECAP_REM, &vrq, sizeof(vrq)))
	        complain(err);
	else
	        fprintf(stdout, "Decapsulation cancelled for %s\n",
			vrq.vr_target_addr ? target : "all");
}

/*
 * Print encapsulation related configuration information in user-friendly
 * form.
 */

int
prettyprint1(unsigned char *buffer)
{
        u32    *cpos;
	char   ipaddr[30];
	int    i, j, entries;
	struct exit_fill *efp;

	fprintf(stdout, "\t\tE N C A P S U L A T I O N      I N F O.\n\n");

	cpos = (u32 *) buffer;
	cpos++;  /* ignore the number of bytes filled */
	entries = *cpos;
	fprintf(stdout, "\tNumber of encapsulation targets = %d.\n", entries);
	cpos++;

	if (entries == 0) return;
	fprintf(stdout, "\n\tTarget\t\tExits\tAddresses of tunnel exits\n");
	fprintf(stdout,   "\t------\t\t-----\t-------------------------\n");
	for (i = 0; i < entries; i++) {
	         if (*cpos == DEFAULT_ADDR) 
		         fprintf(stdout, "\tdefault\t");
		 else
		         fprintf(stdout, "\t%s", ntoa(*cpos, ipaddr));
		 cpos++;
		 j = (int) *cpos;
	         fprintf(stdout, "\t%d\t", j);
		 cpos++;
		 while (j) {
		         fprintf(stdout, "%s,  ", ntoa(*cpos, ipaddr));
			 cpos++;
			 j--;
		 }
		 fprintf(stdout, "\n");
	}

	entries = *cpos;
	fprintf(stdout, "\n\tNumber of tunnel exits = %d.\n", entries);
	cpos++;
	fprintf(stdout, "\n\tTunnel\t\tCurr\tMax\t\n");
	fprintf(stdout,   "\texit\t\tPMTU\tPMTU\tState\n");
	fprintf(stdout,   "\t----\t\t----\t----\t-----\n");

	efp = (struct exit_fill *) cpos;
	for (i = 0; i < entries; i++) {
	         fprintf(stdout, "\t%s", ntoa(efp->addr, ipaddr));
		 cpos++;
	         fprintf(stdout, "\t%d\t%d\t0x%4x\n",
			 efp->curr_pmtu, efp->max_pmtu, efp->state);
		 efp++;
	}

	fprintf(stdout, "\n");
}

/*
 * Print decapsulation related configuration information in user-friendly
 * form.
 */

int
prettyprint2(unsigned char *buffer)
{
        u32    *cpos;
	char   ipaddr[30];
	int    i, entries;
	struct decap_fill *dfp;

	fprintf(stdout, "\n\t\tD E C A P S U L A T I O N      I N F O.\n\n");

	cpos = (u32 *) buffer;
	cpos++;  /* ignore the number of bytes filled */

	if (*cpos)
	  fprintf(stdout, "\tDecapsulation is ENABLED for ALL.\n");
	cpos++;

	entries = *cpos;
	fprintf(stdout, "\tNumber of decapsulation targets = %d.\n", entries);
	cpos++;

	if (entries == 0) return;

	fprintf(stdout, "\n\tTarget\t\tFlags\n");
	fprintf(stdout,   "\t------\t\t-----\n");
	dfp = (struct decap_fill *) cpos;
	for (i = 0; i < entries; i++) {
	         fprintf(stdout, "\t%s", ntoa(dfp->addr, ipaddr));
		 if ((dfp->flags & DECAP_IF_MASK) == DECAP_IF_DST) 
		         fprintf(stdout, "\tdst\n");
		 else if ((dfp->flags & DECAP_IF_MASK) == DECAP_IF_SRC) 
		         fprintf(stdout, "\tsrc\n");
		 else if ((dfp->flags & DECAP_IF_MASK) == DECAP_IF_EITHER) 
		         fprintf(stdout, "\teither\n");
		 else 
		         fprintf(stdout, "\t??\n");
		 dfp++;
	}

	fprintf(stdout, "\n");
}

/* 
 * Print vtunl's internal configuration information -- both for
 * encapsulation and decapsulation.
 */
int
dump()
{
        int           bufsize;
        unsigned char *buffer;
	int           err = 0;

	err = vtunl_ctl(ioctl_fd, VTUNL_ENCAP_SIZE, &bufsize, sizeof(bufsize));
	if (err) {
	        complain(err);
	} else {

	        /* fprintf(stderr, "Need %d bytes\n", bufsize); */
	        buffer = (unsigned char *)malloc(bufsize + 64);

		err = vtunl_ctl(ioctl_fd, VTUNL_ENCAP_GET, buffer,
				(int) (bufsize + 64));
		if (err) {
		        fprintf(stderr, "VTUNL_ENCAP_GET failed.\n");
			complain(err);
		} else {
		        prettyprint1(buffer);
		}
		free(buffer);
	}


	err = vtunl_ctl(ioctl_fd, VTUNL_DECAP_SIZE, &bufsize, sizeof(bufsize));
	if (err) {
	        complain(err);
	} else {

	        /* fprintf(stderr, "Need %d bytes\n", bufsize); */
	        buffer = (unsigned char *)malloc(bufsize + 64);

		err = vtunl_ctl(ioctl_fd, VTUNL_DECAP_GET, buffer, 
				(int) (bufsize + 64));
		if (err) {
		        fprintf(stderr, "VTUNL_DECAP_GET failed.\n");
			complain(err);
		} else {
		        prettyprint2(buffer);
		}
		free(buffer);
	}
}

/* Plumbs the stream permanently under UDP. */ 
int 
plumbudp(int ip_fd, char *device)
{
 
  int  mux_fd, ip_muxid, s;
  struct ifreq ifr;
  short savedflags;
  
  mux_fd = open("/dev/udp", 2);
  if (mux_fd == -1) {
    perror("mux_fd");
    exit(-1);
  }

  strncpy(ifr.ifr_name,device, sizeof(ifr.ifr_name));

  if (ioctl(ip_fd, SIOCGIFFLAGS, (caddr_t)&ifr) < 0) 
    perror("SIOCGIFFLAGS");

  
  if ((ip_muxid = ioctl(mux_fd, I_PLINK, ip_fd)) == -1) {
    perror("I_PLINK for ip");
  }

  
  /*
   * Tell IP the muxids of the PLINKed interface and ARP
   * streams so that they can be closed down at a later
   * time by an "ifconfig ... unplumb" command.
   */

  strncpy(ifr.ifr_name, device, sizeof (ifr.ifr_name));
  ifr.ifr_ip_muxid = ip_muxid;
  if (ioctl(mux_fd, SIOCSIFMUXID, (caddr_t)&ifr) < 0)
    perror("SIOCSIFMUXID");

  close(ip_fd); 
  close(mux_fd);
}

/* Unplumbs the device identified by <device>*/
int
unplumbit(int argc, char **argv)
{
  int  mux_fd, ip_muxid, s, i;
  struct ifreq ifr;

   mux_fd = open("/dev/udp", 2);
   if (mux_fd == -1) 
       perror("mux_fd");
   for (i = 2; i <= argc; i++) {
  
    if (i == argc) 
      /* Need to unplumb ip0(ENCAPIFACE) at end */
      strncpy(ifr.ifr_name, ENCAPIFACE, sizeof ifr.ifr_name);     
    else
      strncpy(ifr.ifr_name, argv[i], sizeof ifr.ifr_name);

    /* Get the stream associated with mux_fd & ifr_name */
    if (ioctl(mux_fd, SIOCGIFMUXID, (caddr_t)&ifr) < 0) 
      perror("SIOCGIFMUXID");

    ip_muxid = ifr.ifr_ip_muxid;
    if (ioctl(mux_fd, I_PUNLINK, ip_muxid) < 0)
      perror("I_PUNLINK");
  } 
  
}


/* Create an instance of stream but not pushed under IP.
 * Doing so prevents an intervening module (e.g.IP) from
 * misinterpreting vtunl-specific commands
 */
int
make_ioctl_fd()
{

       if ((ioctl_fd = open("/dev/ip", 0)) == -1) 
	       Perror("ioctl_fd");

       if (ioctl(ioctl_fd, I_PUSH, "vtunl") == -1) 
	       Perror("I_PUSH vtunl on ioctl_fd");
  
}

/* Initialize */
int
initialize(int argc, char **argv)
{
       int  stream_id[10];
       char devname[20];
       int  ppa, i;
       struct vtunl_req vrq;

       /* Create a stream that encapsulates unicasts and decapsulates */
       memset((char *) &vrq, 0, sizeof(struct vtunl_req));
       vrq.vr_address = inet_addr(argv[2]);
       vrq.vr_ip_protocol = 4;
       stream_id[0] = plumb("/dev/ip", &vrq);

       mkpt2pt(ENCAPIFACE, inet_addr(argv[2]), inet_addr("255.255.255.254"));
       plumbudp(stream_id[0],ENCAPIFACE); /* plumb it permanently */
 
       /* Create streams on each specified network device */
       for (i = 3; i < argc; i++) {
	       if (strlen(argv[i]) > 10) continue;
	       ifname2devppa(argv[i], devname, &ppa);
	       if (ppa < 0 || ppa > 255) continue;

	       memset((char *) &vrq, 0, sizeof(struct vtunl_req));
	       vrq.vr_address = inet_addr(argv[2]);
	       vrq.vr_addr_mask = inet_addr("255.255.255.0");
	       vrq.vr_ppa = ppa;
	       stream_id[i - 2] = plumb(devname, &vrq);
	       plumbudp(stream_id[i-2],argv[i]); /* Plumb it permanently */
       }
}

       
/* MAIN program */

main(int argc, char **argv)
{
       char command[256], cmdname[20];
       int i;
       
       if (argc < 2) {
	       printusage(argv[0]); 
	       exit(-1);
       }

       /* Enable forwarding in kernel. Turn off sending of
        * ICMP redirects (e.g. FA should not redirect visiting
        * mobile nodes to another router).
        */
       system("ndd -set /dev/ip ip_forwarding 1\n");
       system("ndd -set /dev/ip ip_send_redirects 0\n");
       
       /* Concatenate the command line arguments into
	  string command */
       command[0]='\0';
       for(i=1; i<argc; i++)
	 sprintf(command,"%s %s",command,argv[i]);


       if (!strncmp(argv[1],"plumb",strlen("plumb"))) {
	 if (inet_addr(argv[2]) == -1) {
		      printusage(argv[0]); 
		      exit(-1);
	 } else
		initialize(argc,argv);
       }
       else {
	 make_ioctl_fd();
	 if (!strncmp(argv[1], "encapadd", strlen("encapadd"))) 
	   encapadd(command);
	 else if (!strncmp(argv[1], "encaprem", strlen("encaprem"))) 
	   encaprem(command);
	 else if (!strncmp(argv[1], "decapadd", strlen("decapadd"))) 
	   decapadd(command);
	 else if (!strncmp(argv[1], "decaprem", strlen("decaprem"))) 
	   decaprem(command);
	 else if (!strncmp(argv[1], "unplumb", strlen("unplumb"))) 
	   unplumbit(argc,argv);
	 else if (!strncmp(argv[1], "dump", strlen("dump"))) 
	   dump();
	 else {
	   printusage(argv[0]);
	   exit(-1);
	 }
       }
}




