/*
 * Copyright (c) 1996 by Sun Microsystems, Inc.
 */

#pragma ident	"@(#)vtunl.c 1.2 97/08/29 SMI"

/* 
 *  Author: Vipul Gupta (vgupta@eng.sun.com)
 * 
 *  Based on the atun code in the IPv6 distribution. - V.G.
 *  At one point, I thought of retaining the IPv6 code but then                
 *  decided to focus solely on IPv4 since the IPv6 include files
 *  are not widely installed.
 * 
 *  Features include: participation in soft state management,
 *  encapsulation for multiple targets on the same interface,
 *  encapsulation of non-unicast datagrams.
 *
 *  Compile as:
 *      cc -g -D_KERNEL -I. -c vtunl.c vtunlddi.c
 *      ld -r -o vtunl vtunl.o vtunlddi.o
 */

#include <sys/types.h>
#include <sys/cred.h>
#include <sys/stream.h>
#include <sys/dlpi.h>
#include <sys/stropts.h>
#include <sys/strlog.h>
#include <sys/tihdr.h>
#include <sys/tiuser.h>
#include <sys/cmn_err.h>
#include <sys/debug.h>
#include <sys/kmem.h>
#include <sys/ksynch.h>    /* for reader-writer lock */

#include <sys/systm.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/vtrace.h>
#include <sys/isa_defs.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <net/route.h>
#include <sys/sockio.h>
#include <netinet/in.h>

#include <inet/common.h>
#include <inet/mi.h>
#include <inet/mib2.h>
#include <inet/nd.h>
#include <inet/arp.h>
#include <inet/snmpcom.h>

#include <netinet/igmp_var.h>

#include <inet/ip.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>

#include "vtunl_impl.h"

#define CHANGESTATE(state,correct,new,other)               \
        state = ((state)==(correct)) ? (new):(other);

static int vtunl_open(queue_t *, dev_t *, int, int, cred_t *);
static int vtunl_close(queue_t *);
static int vtunl_wput(queue_t *, mblk_t *);
static int vtunl_rput(queue_t *, mblk_t *);
static int vtunl_encap(queue_t *, mblk_t *);

/*
 *     write queue    |
 *                    V
 *                  +------------+
 *                  |   module   |
 *                  +------------+
 *                             ^
 *                             | read queue
 */

/*
 * module_info (9S) stores identification and limit values for a
 * STREAMS module or driver. Is READ-ONLY but flow control limits
 * mi_{hi,lo}wat and pkt size limits mi_{min,max}psz are copied to the 
 * QUEUE structure and may be modified there.
 */
/* streams linkages */
static struct module_info info = {
	0,          /* ushort   mi_idnum   */
	"vtunl",    /* char     *mi_idname */
	1,          /* long     mi_minpsz  */
	INFPSZ,     /* long     mi_maxpsz  */
	65536,      /* ulong    mi_hiwat   */
	1024        /* ulong    mi_lowat   */
};

/* 
 * qinit (9S) contains pointers to processing procedures for a QUEUE. 
 * The streamtab structure contains pointers to one queue structure for
 * both upstream and downstream processing.
 */

/* read side QUEUE */
static struct qinit vtunlrinit = {
	vtunl_rput,   /* int   (*qi_putp)(); put procedure */
	NULL,         /* int   (*qi_srvp)(); service procedure */
	vtunl_open,   /* int   (*qi_qopen)(); open procedure */
	vtunl_close,  /* int   (*qi_qclose)(); close procedure */
	NULL,         /* int   (*qi_qadmin)(); unused */
	&info,        /* struct module_info *qi_minfo; module params */
	NULL          /* struct module_stat *qi_mstat; modules stats */
};

/* write side QUEUE */
static struct qinit vtunlwinit = {
	vtunl_wput, NULL, NULL, NULL, NULL, &info, NULL
};

/* multiplexing drivers require both upper and lower qinit structures */
struct streamtab vtunlinfo = {
		&vtunlrinit, /* struct qinit *st_rdinit; read QUEUE */
		&vtunlwinit, /* struct qinit *st_wrinit; write QUEUE */
		NULL,        /* struct qinit *st_muxrinit; lower read QUEUE */
		NULL         /* struct qinit *st_muxwinit; lower write QUEUE */
};



/* Global data structures */

int	vtunldebug;	        /* debug flag */
int     vtunldecapall;          /* if set, decapsulate indiscriminately */
u16     vtunl_default_pmtu;     /* default PMTU bound for all exits */
queue_t  *timer_q;              /* read q of instance associated with
				 * the timer */
queue_t  *vtunl_transport_q;    /* write q of instance subscribed to IPIP */
int vtunl_timer_id;             /* value returned by last qtimeout() */
struct vtunl_inst *head_vip;    /* dummy header for the list of vtunl
				 * per-instance structures */
int total_exit_count;           /* Total number of tunnel exits in use. 
				 * An exit is counted as many times
				 * as the number of encapsulation targets
				 * that it serves as exit. Used to compute
				 * the memory required to dump encapsulation
				 * information. */

extern struct hash_table bcast_table, decap_table, encap_table, texit_table;

/* 
 * These buffers are passed to ntoa and since only one thread can execute
 * inside vtunl at any time, we shouldn't see any problems reusing these
 * buffers. Each procedure may need to allocate its own version of these
 * buffers should we increase multithreading support inside vtunl beyond
 * D_MTPERMOD.
 */
char   addr_buf1[20], addr_buf2[20], addr_buf3[20];


/* Definitions of functions that manipulate global data structures */
extern void hash_tbl_init(struct hash_table *htbl);
extern void hash_tbl_clean(struct hash_table *htbl, size_t entry_size);
extern struct hash_entry *hash_find(struct hash_table *htbl, u32 addr);
extern int hash_link(struct hash_table *htbl, struct hash_entry *entry) ;
extern struct hash_entry *hash_unlink(struct hash_table *htbl, u32 addr);
extern int vtunl_encap_add(struct vtunl_req *reqp);
extern int vtunl_encap_rem_target(u32 addr);
extern int vtunl_encap_rem_texit(u32 target_addr, u32 exit_addr);
extern int vtunl_encap_fill(unsigned char *);
extern int vtunl_decap_fill(unsigned char *);

/* 
 * The vtunl is a funny beast. It is plumbed as shown below and acts
 * both as a driver for the ip module above it (in this capacity,
 * it encapsulates outgoing packets before sending them to the ip
 * driver) and as a higher layer over /dev/ip that processes datagrams
 * with protocol number 4 (IPinIP). In this capacity, it performs
 * decapsulation before sending the packet to the ip module.
 * For this reason, vtunl has both write-side and read-side put
 * procedures. Since we expect to see different kinds of messages
 * in the two directions, we use distinct procedures for wput()
 * and rput().
 *
 *   This is how vtunl can be plumbed so that the plumbing lasts
 * even after the creating program has disappeared. /dev/udp can 
 * be used to PLINKs the stream that sandwiches vtunl between the
 * ip module and the ip driver.
 *
 *                   ---------------------------+
 *                     /dev/udp used as mux     |
 *                   ---------------------------+
 *                              |          ^
 *                              v          |
 *                           +----------------+
 *                           |   "ip" module  |
 *                           +----------------+
 *                              |          ^
 *       vtunl's write queue    v          |
 *                           +----------------+
 *                           |      vtunl     |
 *                           +----------------+
 *                              |          ^   vtunl's read queue
 *                              v          | 
 *                           +----------------+
 *                           | /dev/ip driver |
 *                           +----------------+
 *
 *
 *    Another way to visualize the vtunl's relationship with other
 *  pieces in the "traditional network layer model" is as follows:
 *
 *                   +-----------------+
 *                   |  vtunl (upper)  |
 *                   +-------------+   |
 *                   ----------+   |   |
 *                       ip    |   |   |
 *                   ----------+   |   |
 *                   +-------------+   |
 *                   | vtunl (lower)   |
 *                   +-----------------+
 *
 *     Also check out the tunneling module described in 
 *          draft-ietf-ipngwg-ipv6-tunnel-xx.txt
 *     by Conta and Deering.
 *
 *  In order to use vtunl, we plumb it between /dev/ip and ip. This
 *  stream appears to IP as "ip0". Since ip0 is a point-to-point
 *  interface, Solaris lets us assign it an IP address assigned to
 *  an existing network interface such as le0. The destination
 *  address assigned to this point-to-point interface is just
 *  a dummy filler and has no particular significance.
 *  For destinations that must be tunneled, we set routes for them
 *  through ip0.
 *  
 *  Since it is impossible to set routes for broadcast/multicast
 *  packets through ip0, we force those packets through vtunl by
 *  having a vtunl instance sit atop each network device (e.g. le0,
 *  hme0) and subscribing to all link-level non-unicast packets.
 *
 */
 
/*
 * Canned DL_INFO_ACK. This is what we send in response to DL_INFO_REQ.
 * TODO: Convince ifconfig not to try printing ether info for us.
 */
dl_info_ack_t infoack = {
	DL_INFO_ACK,	/* dl_primitive */
	VTUNL_MAXDLSDU, /* dl_max_dlsdu */
	1,		/* dl_min_sdu */
	0,              /* dl_addr_length */
	DL_OTHER,       /* dl_mac_type */
	0,		/* dl_reserved */
	DL_UNATTACHED,  /* dl_current_state */
	0,		/* dl_sap_length */
	DL_CLDLS,	/* dl_service_mode */
	0,		/* dl_qos_length */
	0,		/* dl_qos_offset */
	0,		/* dl_qos_range_length */
	0,		/* dl_qos_range_offset */
	DL_STYLE2,	/* dl_provider_style */
	0,	        /* dl_addr_offset */
	DL_VERSION_2,	/* dl_version */
	0,		/* dl_brdcast_addr_length, 0 means don't ARP */
	0,		/* dl_brdcst_addr_offset */
	0		/* dl_grow */
};

/* 
 * canned DL_BIND_ACK - IP doesn't use any of this info.
 */
dl_bind_ack_t bindack = {
	DL_BIND_ACK,	/* dl_primitive */
	0,		/* dl_sap */
	0,		/* dl_addr_length */
	0,		/* dl_addr_offset */
	0,		/* dl_max_conind */
	0		/* dl_xidtest_flg */
};

/*
 * Partial table from RFC 1191. The complete table looks like:
 * {32000, 17914, 8166, 4352, 2002, 1496, 1006, 508, 296, 68} 
 */
#define FRAGTBLSIZE  5
static u16 icmp_frag_size_table[] = {4352, 2002, 1496, 1006, 508};

/* 
 * Routines for Path MTU discovery. These could be made in-line.
 */
void mtu_up(u16 *pmtu, u16 bound) {
  int i = FRAGTBLSIZE-1;
  
  while (icmp_frag_size_table[i] <= *pmtu && i) i--;
  *pmtu = (bound < icmp_frag_size_table[i]) ? bound : icmp_frag_size_table[i];
}

void mtu_down(u16 *pmtu) {
  int i = 0;
    
  while (icmp_frag_size_table[i] >= *pmtu && (i < FRAGTBLSIZE-1)) i++;
  *pmtu = icmp_frag_size_table[i];
}

int relay_icmp(queue_t *, mblk_t *, int, int, u16);
char *parse_icmph(icmph_t *);

/*
 * Converts the long value in addr_long to an internet address of
 * the type a.b.c.d in addr_string and returns a pointer to addr_string.
 * NOTE: addr_string is assumed to be long enough (atleast 16 bytes).
 *
 * By having the caller supply its own buffer, we try to avoid problems
 * associated with multiple threads writing the same buffer concurrently.
 */
static char *
ntoa(u32 addr_long, char * addr_string)
{
        u32    temp;
	
	temp = ntohl(addr_long);
	sprintf(addr_string, "%d.%d.%d.%d", 
		((temp >> 24) & 0xff), ((temp >> 16) & 0xff),
		((temp >> 8) & 0xff), (temp & 0xff));
	return(addr_string);
}

static void
timer_handler(caddr_t q)
{
        int    i, done;
	u16    count;
	mblk_t *newmp;
	struct t_exit *exitp;
	struct vtunl_notification *notep;
	u16    old_pmtu;

	vtunl3dbg(("Timer handler called.\n"));
	
	for (i = 0, done = 0; 
	     i < HASH_TBL_SIZE && done < texit_table.size; i++) {
	  exitp = (struct t_exit *) texit_table.hash_entries[i];
	  while (exitp) {
  	      if (exitp->t_exit_state & 0x8000) {
		  exitp->t_exit_state &= ~0x8000;
		  vtunl2dbg(("No more ICMP unreachables for exit %s.\n",
			     ntoa(exitp->t_exit_addr, addr_buf1)));
	      }
	      count = exitp->t_exit_state & 0x0007;
	      if (count) {  
	          /* decrement 3-bit counter used for PMTU refresh */
		  exitp->t_exit_state &= ~0x0007;
		  exitp->t_exit_state |= ((--count) & 0x0007);
	      } else {
		  /* time to bump up the PMTU */
		  old_pmtu = exitp->t_exit_pmtu;
		  mtu_up((u16 *)&(exitp->t_exit_pmtu), exitp->t_exit_maxpmtu);
		  exitp->t_exit_state |= 0x0001;
		  vtunl2dbg(("Estimated PMTU for %s is %d (>= %d).\n",
			     ntoa(exitp->t_exit_addr, addr_buf1), 
			     exitp->t_exit_pmtu,
			     old_pmtu));
	      }
	      done++;     
	      exitp = exitp->t_exit_next;
	  }
	}

	vtunl_timer_id = qtimeout((queue_t *)q, timer_handler, q, 
				  VTUNL_TIMER_PERIOD);

	/* send notification to notify_on */

	/* Commented out till we figure out why SIGPOLL kills a user process
	 * even when it has installed a SIGPOLL handler
	 *     **************************

	 if (notify_on != NULL) {
	    newmp = allocb(sizeof(struct vtunl_notification), BPRI_LO);
	    if (newmp == (mblk_t *) 0) {
	       vtunl0dbg(("timer_handler: allocb failed.\n"));
	       return;
	    }
	    newmp->b_datap->db_type = M_PCPROTO;
	    newmp->b_wptr += sizeof(struct vtunl_notification);
	    notep = (struct vtunl_notification *) newmp->b_rptr;
	    notep->type = (int) newmp;
	    putnext(RD(notify_on), newmp);
	 }

	 *     **************************
	 *  Commented out for now.
	 */
}
   
static int
vtunl_open (queue_t *q, dev_t *devp, int flag, int sflag, cred_t *credp)
{
	mblk_t   *mp;
	queue_t  *tempq;
	char     *mname;
	struct vtunl_inst *vip;

        vtunl3dbg(("vtunl_open called with sflag=%d.\n", sflag));

	if (q->q_ptr) /* re-open of an already open instance */
	        return(0);

        /* If this is not a push of vtunl as a module, fail */
	if (sflag != MODOPEN)
		return(EINVAL);

	if (drv_priv(credp))     /* non-zero implies not suser */
	        return(EPERM);

	/* Allocate and initialize per-instance structure */
	vip = (struct vtunl_inst *) kmem_zalloc(sizeof(struct vtunl_inst),
						KM_NOSLEEP);
	if (vip == NULL) 
	        return(ENOMEM);
	vip->vcs.type = VTUNL_INST_UNKNOWN;
	vip->vcs.state = VTUNL_STATE_INIT;		 

	vip->prev = head_vip->prev;
	vip->next = head_vip;
	vip->prev->next = vip;
	vip->next->prev = vip;
	vip->q = q;

	q->q_ptr = WR(q)->q_ptr = vip;

        qprocson(q);    	/* enable put and service routines */

	/* Todo: Check if we have "ip" or "le" or "hme" below us. */
	/*
	 * This is what ill_init in ip_if.c does. Note that it is
	 * important to call qprocson(q) BEFORE walking down the
	 * stream or else WR(q)->q_next is null.
	 */

	if (WR(q)->q_next) {
	        tempq = WR(q);
		do {
		        tempq = tempq->q_next;
		} while (tempq->q_next);
		mname = tempq->q_qinfo->qi_minfo->mi_idname;
		if (mname && *mname) {
		        strncpy(vip->vcs.modname, mname, MAXMNAMESZ);
			vip->vcs.modname[MAXMNAMESZ] = '\0';
		}
	} 

	if (strncmp(vip->vcs.modname, "ip", strlen("ip")) == 0)
	        vip->vcs.type = VTUNL_INST_TRANSPORT;
	else /* if (strncmp(vip->vcs.modname, "le", strlen("le")) == 0 ||
		 strncmp(vip->vcs.modname, "hme", strlen("hme")) == 0) */
	        vip->vcs.type = VTUNL_INST_DLSUSER;

	vtunl3dbg(("vtunl_open: Opened instance type %d over %s.\n", 
		   vip->vcs.type, vip->vcs.modname));

	if (timer_q == NULL) {
	     /* Get the periodic timer going */
	     timer_q = q;
	     vtunl_timer_id = qtimeout(q, timer_handler, 
				       (caddr_t)q, VTUNL_TIMER_PERIOD);
	     vtunl3dbg(("vtunl_open: Periodic timer started.\n"));
	}

	return(0);   	/* success */
}

static int
vtunl_close (queue_t *q)
{
        struct vtunl_inst *vip;

	vtunl3dbg(("vtunl_close() called\n"));

	if (timer_q == q) {
	        quntimeout(q, vtunl_timer_id);
		timer_q = NULL; 
		/* 
		 * By doing this we allow instance besides the very
		 * first one (after a successful add_drv) to do 
		 * timer management.
		 */
	}

	if (vtunl_transport_q == q)
	        vtunl_transport_q = NULL;

	qprocsoff(q);   /* disable scheduling */

	/* Unlink and free per-instance structure */
	if (vip = (struct vtunl_inst *) q ->q_ptr) {
	        vip->prev->next = vip->next;
		vip->next->prev = vip->prev;
		kmem_free(vip, sizeof(struct vtunl_inst));
	}
	
	q->q_ptr = WR(q)->q_ptr = NULL;

	/* 
	 * While it may be tempting to clear global state when the
	 * last instance is closed, we shouldn't do so.
	 */

	return(0); 	/* success */
}


/* Send a DL_OK_ACK back upstream. */
static int
sendokack(queue_t *q, mblk_t *mp, u_long prim)
{
	mblk_t *newmp;
	dl_ok_ack_t *dlok;

	newmp = allocb(sizeof(dl_ok_ack_t), BPRI_HI);
	if (newmp == (mblk_t *) 0) {
 	        vtunl0dbg(("sendokack: couldn't allocate dl_ok_ack mblk\n"));
		return;
	}
	dlok = (dl_ok_ack_t *) newmp->b_rptr;
	dlok->dl_primitive = DL_OK_ACK;
	dlok->dl_correct_primitive = prim;
	newmp->b_datap->db_type = M_PCPROTO;
	newmp->b_wptr += sizeof(dl_ok_ack_t);
	qreply(q, newmp);
}

/*
 * All the DLPI message structures are defined in <sys/dlpi.h>
 * Here we respond to all the DLPI messages that the ip module
 * above sends us. We look like a DLPI driver to that IP module.
 * qreply(9F) sends a message in the reverse direction of normal
 * flow. We try to reuse the message block whenever possible but
 * sometimes we need to allocate a new block and we use sendokack()
 * for that. In this latter case, we also free the incoming message
 * block.
 *   
 * The dl_ok_ack_t structure is simple and contains just two fields
 * ulong dl_primitive (set to DL_OK_ACK) and ulong dl_correct_primitive
 * (the one being acknowledged).
 */
static int
vtunl_wput_dlpi(queue_t *q, register mblk_t *mp)
{
	u_long prim = *((u_long *) mp->b_rptr);
	dl_ok_ack_t *dlok;
	mblk_t *mp1;

	switch (prim) {
	case DL_INFO_REQ:
                vtunl3dbg(("vtunl_wput_dlpi: got DL_INFO_REQ.\n"));

		/* 
		 * Send DL_INFO_ACK.  Get new mblk because request mblk
		 * might not be big enough.
		 */
		mp1 = allocb(sizeof(infoack), BPRI_HI);
		if (mp1 == (mblk_t *) 0) {
		        vtunl0dbg(("vtunl_wput_dlpi: couldn't allocate dl_info_ack mblk.\n"));
			freemsg(mp);
			return;
		}
		bcopy((char *) &infoack, (char *) mp1->b_rptr,
		      sizeof(infoack));
		mp1->b_datap->db_type = M_PCPROTO;
		mp1->b_wptr += sizeof(infoack);
		qreply(q, mp1);

		freemsg(mp);
		break;

	case DL_ATTACH_REQ:
		vtunl3dbg(("vtunl_wput_dlpi: got DL_ATTACH_REQ.\n"));

		/* send DL_OK_ACK - re-use mblk */
		dlok = (dl_ok_ack_t *) mp->b_rptr;
		dlok->dl_primitive = DL_OK_ACK;
		dlok->dl_correct_primitive = DL_ATTACH_REQ;
		mp->b_datap->db_type = M_PCPROTO;
		qreply (q, mp);
		break;

	case DL_DETACH_REQ:
	        vtunl3dbg(("vtunl_wput_dlpi: got DL_DETACH_REQ.\n"));
		sendokack (q, mp, prim);
		freemsg (mp);
		break;

	case DL_BIND_REQ:
		vtunl3dbg(("vtunl_wput_dlpi: got DL_BIND_REQ.\n"));
		/*
		 * Send DL_BIND_ACK, which is the same size as the
		 * request, so we can re-use the mblk.
		 */
		bcopy((char *) &bindack, (char *) mp->b_rptr, sizeof(bindack));
		mp->b_datap->db_type = M_PCPROTO;
		qreply(q, mp);
		break;

	case DL_UNBIND_REQ:
		vtunl3dbg(("vtunl_wput_dlpi: got DL_UNBIND_REQ.\n"));
		/* Send a DL_OK_ACK.  Can't re-use mblk */
		sendokack (q, mp, prim);
		freemsg (mp);
		break;

	case DL_PHYS_ADDR_REQ: {
		dl_phys_addr_ack_t *dpa= (dl_phys_addr_ack_t *) mp-> b_rptr;

		vtunl3dbg(("vtunl_wput_dlpi: got DL_PHYS_ADDR_REQ.\n"));
		/*
		 * Send a DL_PHYS_ADDR_ACK with a zero length address
		 * since we don't actually have a hardware address.
 		 * This could be used in the future for constructing
		 * a link local address????
		 */
		dpa-> dl_primitive= DL_PHYS_ADDR_ACK;
		dpa-> dl_addr_length= 0;
		mp->b_datap->db_type = M_PCPROTO;
		qreply(q, mp);
		break;
	}
		
	case DL_UNITDATA_REQ:
		/* we don't use any of the data in the DLPI header */
		vtunl3dbg(("vtunl_wput_dlpi: got DL_UNITDATA_REQ.\n"));
		mp1 = mp->b_cont;
		mp->b_cont = (mblk_t *) 0;
		freeb (mp);		
		if (mp1)
			vtunl_encap(q, mp1);
		break;

	default:
	        vtunl1dbg(("vtunl_wput_dlpi: unknown DLPI msg type: %d.\n",
			   prim));
		break;
	}
}


/* Send a T_BIND_REQ downstream, subscribe to ip_proto. */
static int
sendtbindrq(queue_t *q, u_char ip_proto)
{
	mblk_t *mbp;
	struct T_bind_req *tbr;
	
	mbp = allocb(sizeof(struct T_bind_req) + 1, BPRI_HI);
	if (!mbp) {
	        vtunl0dbg(("sendtbindrq: allocb failed.\n"));
		return(ENOMEM);
	}
	
	mbp->b_datap->db_type = M_PCPROTO;
	tbr = (struct T_bind_req *) mbp->b_rptr;
	mbp->b_wptr += sizeof(struct T_bind_req);
	tbr->PRIM_type = T_BIND_REQ;
	tbr->ADDR_length = 0;
	tbr->ADDR_offset = 0;
	tbr->CONIND_number = 0;
	*mbp->b_wptr = ip_proto;
	mbp->b_wptr++;
	
	vtunl3dbg(("sendtbindrq: subscribing to protocol %d.\n", 
		   ip_proto));
	putnext(WR(q), mbp);
	return(0);
}

/* Send a DL_ATTACH_REQ downstream with given ppa */
static int
senddlattachrq(queue_t *q, unsigned long ppa)
{
	mblk_t *mbp;
	dl_attach_req_t *dla;

	mbp = allocb(sizeof(dl_attach_req_t), BPRI_HI);
	if (!mbp) {
	        vtunl0dbg(("senddlattachrq: allocb failed.\n"));
		return(ENOMEM);
	}

	mbp->b_wptr += sizeof(dl_attach_req_t);
	mbp->b_datap->db_type = M_PROTO;
	dla = (dl_attach_req_t *) mbp->b_rptr;
	dla->dl_primitive = DL_ATTACH_REQ;
	dla->dl_ppa = ppa;

	vtunl3dbg(("Sending DL_ATTACH_REQ w/ ppa %d.\n", ppa));
	putnext(WR(q), mbp);
	return(0);
}

/* Send a DL_BIND_REQ downstream with given sap. */
static int
senddlbindrq(queue_t *q, unsigned long sap)
{
	mblk_t *mbp;
	dl_bind_req_t *dlb;

	mbp = allocb(sizeof(dl_bind_req_t), BPRI_HI);
	if (!mbp) {
	        vtunl0dbg(("senddlbindrq: allocb failed.\n"));
		return(ENOMEM);
	}

	mbp->b_wptr += sizeof(dl_bind_req_t);
	mbp->b_datap->db_type = M_PROTO;
	dlb = (dl_bind_req_t *) mbp->b_rptr;
	dlb->dl_primitive = DL_BIND_REQ;
	dlb->dl_sap = sap;
	dlb->dl_max_conind = 0;
	dlb->dl_service_mode = DL_CLDLS;
	dlb->dl_conn_mgmt = 0;
	dlb->dl_xidtest_flg = 0;

	vtunl3dbg(("Sending DL_BIND_REQ w/ sap 0x%x.\n", sap));
	putnext(WR(q), mbp);
	return(0);
}


/* Send a DL_PROMISCON_REQ downstream with given promiscuity level. */
static int
senddlpromisconrq(queue_t *q, unsigned long level)
{
        mblk_t *mbp;
	dl_promiscon_req_t *dlp;

	mbp = allocb(sizeof(dl_promiscon_req_t), BPRI_HI);
	if (!mbp) {
	        vtunl0dbg(("senddlpromisconrq: allocb failed.\n"));
		return(ENOMEM);
	}

	mbp->b_wptr += sizeof(dl_promiscon_req_t);
	mbp->b_datap->db_type = M_PROTO;
	dlp = (dl_promiscon_req_t *) mbp->b_rptr;
	dlp->dl_primitive = DL_PROMISCON_REQ;
	dlp->dl_level = level;

	vtunl3dbg(("Enabling PROMISCUOUS mode, level %x.\n", level));
	putnext(WR(q), mbp);
	return(0);
}

/*
 * Process Vtunl-specific ioctls. We only allow I_STR style ioctls
 * which result in M_IOCTL messages (assuming intervening modules
 * between the vtunl and stream head don't process them).
 * See iocblk(9S) and streamio (7I).
 *
 * We assume that all addresses are passed in network byte order.
 */

static int
vtunl_ioctl(queue_t *q, mblk_t *mp)
{
	struct iocblk     *iocp = (struct iocblk *)ALIGN32(mp->b_rptr);
	struct vtunl_req  *reqp;
	struct vtunl_conf *confp;
	mblk_t            *mp1;
	unsigned char     *dump;
	int               error = 0;

	vtunl3dbg(("vtunl_ioctl: Processing ioctl %x.\n", iocp->ioc_cmd));

	if (error = drv_priv(iocp->ioc_cr)) goto nak; /* only suser ok */

	/* 
	 * Check if we have the right-sized buffers and initialize a
	 * pointer to vtunl_req or vtunl_conf structure as appropriate.
	 */
	switch (iocp->ioc_cmd) {
	  case VTUNL_SETCONF_TRANSPORT:
	  case VTUNL_SETCONF_DLSUSER:
	  case VTUNL_ENCAP_ADD:
	  case VTUNL_ENCAP_REM:
	  case VTUNL_ENCAP_OK:
	  case VTUNL_DECAP_ADD:
	  case VTUNL_DECAP_REM:
	  case VTUNL_DECAP_OK:
	  case VTUNL_MAXPMTU_SET:
	  case VTUNL_MAXPMTU_GET:
	          if ((iocp->ioc_count != sizeof(struct vtunl_req)) ||
		      !(mp1 = mp->b_cont)) {
		          error = EPROTO;
			  goto nak;
		  }
		  reqp = (struct vtunl_req *) ALIGN32(mp1->b_rptr);
		  break;

	  case VTUNL_GETCONF:
	          if ((iocp->ioc_count != sizeof(struct vtunl_conf)) ||
		      !(mp1 = mp->b_cont)) {
		          error = EPROTO;
			  goto nak;
		  }
		  confp = (struct vtunl_conf *) ALIGN32(mp1->b_rptr);
		  break;

	  case VTUNL_ENCAP_SIZE:
	  case VTUNL_DECAP_SIZE:
	  case VTUNL_DEBUG_SET:
	  case VTUNL_DEBUG_GET:
	          if ((iocp->ioc_count != sizeof(int)) ||
		      !(mp1 = mp->b_cont)) {
		          error = EPROTO;
			  goto nak;
		  }
		  break;

	  case VTUNL_ENCAP_GET:
	  case VTUNL_DECAP_GET:
	          if ((iocp->ioc_count < sizeof(u32)) ||
		      !(mp1 = mp->b_cont)) {
		         error = EPROTO;
			 goto nak;
		  }
		  break;

	  default:
		  error = EINVAL;
		  goto nak;
	}

	/*
	 * Process ioctl command, grab appropriate locks (if needed)
	 * before changing global data.
	 */

	switch (iocp->ioc_cmd) {

	case VTUNL_ENCAP_ADD:
	        vtunl2dbg(("ENCAP_ADD: target %s, exit %s, flags %x.\n",
			   ntoa(reqp->vr_target_addr, addr_buf1), 
			   ntoa(reqp->vr_t_exit_addr, addr_buf2),
			   reqp->vr_encap_flags));
	        if (error = vtunl_encap_add(reqp)) goto nak;
	        break;

	case VTUNL_ENCAP_REM:
	        vtunl2dbg(("ENCAP_REM: target %s, exit %s.\n",
			   ntoa(reqp->vr_target_addr, addr_buf1), 
			   ntoa(reqp->vr_t_exit_addr, addr_buf2)));
    	        if (reqp->vr_target_addr == 0) {
		  /* flush all */
		  hash_tbl_clean(&bcast_table, sizeof(struct bcast_target));
		  hash_tbl_clean(&encap_table, sizeof(struct encap_target));
		  hash_tbl_clean(&texit_table, sizeof(struct t_exit));
		  total_exit_count = 0;
		} else {
		  if (reqp->vr_t_exit_addr == 0)
		    error = vtunl_encap_rem_target(reqp->vr_target_addr);
		  else         /* both non-zero */
		    error = vtunl_encap_rem_texit(reqp->vr_target_addr, 
						  reqp->vr_t_exit_addr);
		}
		if (error) goto nak;
	        break;

	case VTUNL_ENCAP_OK:
	        vtunl3dbg(("ENCAP_OK: target %s.\n", 
			   ntoa(reqp->vr_target_addr, addr_buf1)));
     	        if (reqp->vr_target_addr == 0 || 
		    hash_find(&encap_table, reqp->vr_target_addr) == NULL) {
		  error = ENXIO;
		  goto nak;
		}
		break;

	case VTUNL_ENCAP_SIZE:
	         {
		   int *size;
		   
		   size = (int *) ALIGN32(mp1->b_rptr);
		   *size = (4 + 4 + 8*encap_table.size + 4*total_exit_count + 
			    4 + 10 * texit_table.size);
		 }
	         break;
	        
	        
	case VTUNL_ENCAP_GET: 
	       {
		 u32 *tmp, size;

		 vtunl3dbg(("VTUNL_ENCAP_GET called.\n"));
		 size = (4 + 4 + 8*encap_table.size + 4*total_exit_count + 
			 4 + 10 * texit_table.size);

		 tmp = (u32 *) ALIGN32(mp1->b_rptr);

		 /* Fill the number of bytes filled/required */
		 *tmp = size;

		 if (iocp->ioc_count < size) {
			 error = ENOMEM;
			 goto nak;
		 }
		 
		 tmp++;

		 /*
		  * Put the number of encapsulation targets in the
		  * first four bytes. Then, for each encap_target, 
		  * list its address (4 bytes), num of tunnel exits
		  * (4 bytes) and the address of each exit. Finally,
		  * put the number of tunnel exits (4 bytes) and for
		  * each tunnel exit, fill its address (4 bytes),
		  * current PMTU estimate (2 bytes) maximum PMTU 
		  * (2 bytes) and state (2 bytes).
		  */

		 vtunl_encap_fill((unsigned char *)tmp);
	       }
	       break;

	case  VTUNL_DECAP_ADD: 
	        {
		  struct decap_target *dtp;
		  
		  vtunl2dbg(("DECAP_ADD: target %s flags %d.\n",
			     ntoa(reqp->vr_target_addr, addr_buf1),
			     reqp->vr_decap_flags));

		  if (reqp->vr_target_addr) {
		          if (dtp = (struct decap_target *) 
			       hash_find(&decap_table, reqp->vr_target_addr)) {
			          dtp->decap_target_flags = 
				    reqp->vr_decap_flags;
			  } else {
			        if ((dtp = (struct decap_target *) 
				     kmem_zalloc(sizeof(struct decap_target),
						 KM_NOSLEEP)) == NULL) {
			             error = ENOMEM; 
				     goto nak;
				}

				dtp->decap_target_addr = reqp->vr_target_addr;
				dtp->decap_target_flags = reqp->vr_decap_flags;

				if (hash_link(&decap_table, 
					      (struct hash_entry *) dtp) != 0) 
			          /* exists! */
			          kmem_free(dtp, sizeof(struct decap_target)); 
			  }
		  } else {
		          /* pass along all decapsulated packets */
		          vtunldecapall = 1; 
		  }
		}
	        break;

	case VTUNL_DECAP_REM:
	        {
		  struct decap_target *dtp;

		  vtunl2dbg(("DECAP_REM: target %s.\n", 
			     ntoa(reqp->vr_target_addr, addr_buf1)));
		  if (reqp->vr_target_addr) {
		          /* remove one entry */
		          dtp = (struct decap_target *) 
			        hash_unlink(&decap_table, 
					    reqp->vr_target_addr);
			  if (dtp == NULL) {
			          error = ENXIO;  /* no such entry! */
				  goto nak;
			  }
			  kmem_free(dtp, sizeof(struct decap_target));
		  } else {
 		          /* flush all */
		          hash_tbl_clean(&decap_table, 
					 sizeof(struct decap_target));
			  vtunldecapall = 0;
		  }
		}
	        break;

	case VTUNL_DECAP_OK:
	        vtunl3dbg(("DECAP_OK: target %s.\n", 
			   ntoa(reqp->vr_target_addr, addr_buf1)));
     	        if (reqp->vr_target_addr == 0 || 
		    hash_find(&decap_table, reqp->vr_target_addr) == NULL) {
		          error = ENXIO;
			  goto nak;
		}
		break;
	    
	case VTUNL_DECAP_SIZE:
	         {
		   int *size;
		   
		   size = (int *) ALIGN32(mp1->b_rptr);
		   *size = (4 + 4 + 4 + 6*decap_table.size);
		 }
	         break;
	        
	case VTUNL_DECAP_GET:
	       {
		 u32 *tmp, size;

		 vtunl3dbg(("VTUNL_DECAP_GET called.\n"));
		 size = (4 + 4 + 4 + 6*decap_table.size);
		 tmp = (u32 *) ALIGN32(mp1->b_rptr);

		 /* Fill the number of bytes filled/required */
		 *tmp = size;

		 if (iocp->ioc_count < size) {
			 error = ENOMEM;
			 goto nak;
		 }
		 
		 tmp++;

		 /* Fill the value of vtunldecapall */
		 *tmp = (u32) vtunldecapall; 
		 tmp++;

		 /* 
		  * Fill out decap_table size followed by the address
		  * (4 bytes) and flags (2 bytes) for each entry.
		  */

		 vtunl_decap_fill((unsigned char *)tmp);
	       }
	       break;

	case VTUNL_SETCONF_TRANSPORT:
	        {
		  struct vtunl_inst *vip, *vipq;

		  vtunl2dbg(("SETCONF_TRANSPORT: address %s, ip-proto %d.\n",
			     ntoa(reqp->vr_address, addr_buf1), 
			     reqp->vr_ip_protocol));
		  
		  vipq = (struct vtunl_inst *) q->q_ptr;
		  if (vipq->vcs.type != VTUNL_INST_TRANSPORT) {
		          vtunl1dbg(("Bad ioctl on non-transport vtunl.\n"));
			  error = ENOTSUP;
			  goto nak;
		  }

		  vip = head_vip->next;
		  while (vip != head_vip) {
		    if (vip != vipq && 
			vip->vcs.type == VTUNL_INST_TRANSPORT &&
			vip->vcs.ip_protocol == reqp->vr_ip_protocol) {
		            vtunl1dbg(("ip_protocol non-unique.\n"));
		            error = EEXIST;
			    goto nak;
		    }
		    vip = vip->next;
		  }

		  if (error = sendtbindrq(q, (u_char) reqp->vr_ip_protocol))
		          goto nak;

		  vipq->vcs.address = reqp->vr_address;
		  vipq->vcs.ip_protocol = reqp->vr_ip_protocol;
		  vipq->vcs.last_req = T_BIND_REQ;
		};
	        break;

	case VTUNL_SETCONF_DLSUSER:
	         { 
		   struct vtunl_inst *vip, *vipq;

		   vtunl2dbg(("SETCONF_DLSUSER: addr %s, mask %s, ppa %d.\n",
			      ntoa(reqp->vr_address, addr_buf1), 
			      ntoa(reqp->vr_addr_mask, addr_buf2), 
			      reqp->vr_ppa));

		   vipq = (struct vtunl_inst *) q->q_ptr;
		   if (vipq->vcs.type != VTUNL_INST_DLSUSER) {
		          vtunl1dbg(("Bad ioctl on non-dlsuser vtunl.\n"));
			  error = ENOTSUP;
			  goto nak;
		   }

		   vip = head_vip->next;
		   while (vip != head_vip) {
		     if ((vip != vipq) && 
			 (vip->vcs.type == VTUNL_INST_DLSUSER) &&
			  !strncmp(vip->vcs.modname, vipq->vcs.modname, 
				   strlen(vipq->vcs.modname)) &&
			 (vip->vcs.ppa == reqp->vr_ppa)) {
		           vtunl1dbg(("Non-unique device ppa combination.\n"));
			   error = EEXIST;
			   goto nak;
		     }
		     vip = vip->next;
		   }

		   /* 
		    * When I used DL_PROMISC_SAP instead, multicast
		    * reception was screwed, vtunl_encap printed bad
		    * IP version on some packets and no multicasts.
		    * Using DL_PROMISC_MULTI also gets link b'casts.
		    */

		   if (senddlattachrq(q, (unsigned long) reqp->vr_ppa) ||
		       senddlbindrq(q, (unsigned long) 0x0800) ||
		       senddlpromisconrq(q, DL_PROMISC_MULTI)) {
		           error = ENOMEM;
			   goto nak;
		   }
		   
		   vipq->vcs.last_req = DL_PROMISCON_REQ;
		   vipq->vcs.address = reqp->vr_address;
		   vipq->vcs.addr_mask = reqp->vr_addr_mask;
		   vipq->vcs.ppa = reqp->vr_ppa;
		 };
	         break;

	case VTUNL_GETCONF:
	         {
		   struct vtunl_conf *vcp;
		   struct vtunl_inst *vip;
		   
		   vtunl2dbg(("VTUNL_GETCONF called.\n"));
		   vip = (struct vtunl_inst *) q->q_ptr;
		   vcp = &(vip->vcs);
		   bcopy((char *) vcp, (char *) confp, 
			 sizeof(struct vtunl_conf));
		 }
	         break;

	 case VTUNL_DEBUG_SET:
	         {
		   int *val;
	           
		   val = (int *) ALIGN32(mp1->b_rptr);
		   vtunldebug = *val;
		 }
	         break;

	case VTUNL_DEBUG_GET:
	         {
		   int *val;
		   
		   val = (int *) ALIGN32(mp1->b_rptr);
		   *val = vtunldebug;
		 }
	         break;

	case VTUNL_MAXPMTU_SET:
	         {
		   struct t_exit *tep;

		   vtunl2dbg(("VTUNL_MAXPMTU_SET: t_exit %s max pmtu %d.\n", 
			      ntoa(reqp->vr_address, addr_buf1), 
			      reqp->vr_mtu_bound));

		   if (reqp->vr_address == 0) {
		           vtunl_default_pmtu = reqp->vr_mtu_bound;
		   } else {
		           if ((tep = (struct t_exit *)
				hash_find(&texit_table, reqp->vr_address))
			       == NULL) {
			           error = ENXIO;
				   goto nak;
			   }
		   
			   tep->t_exit_maxpmtu = reqp->vr_mtu_bound;
			   if (tep->t_exit_pmtu > tep->t_exit_maxpmtu) {
			           tep->t_exit_pmtu = tep->t_exit_maxpmtu;
				   tep->t_exit_state |= 0x0007;
			   }
		   }		   
		 }
	         break;


	case VTUNL_MAXPMTU_GET:
	         {
		   struct t_exit *tep;

		   vtunl2dbg(("VTUNL_MAXPMTU_GET: t_exit %s.\n", 
			      ntoa(reqp->vr_address, addr_buf1)));

		   if (reqp->vr_address == 0) {
		           reqp->vr_mtu_bound = vtunl_default_pmtu;
		   } else {
		           if ((tep = (struct t_exit *)		     
				hash_find(&texit_table, reqp->vr_address))
			       == NULL) {
			           error = ENXIO;
				   goto nak;
			   }
			   reqp->vr_mtu_bound = tep->t_exit_maxpmtu;
		   }
		 }
	         break;

	default:
		 error = EINVAL;
		 goto nak;
	}

        /* prepare and send a reply for the ioctl */
	vtunl3dbg(("vtunl_ioctl returned ack.\n"));
	iocp->ioc_error = 0;
	mp->b_datap->db_type = M_IOCACK;
	qreply(q, mp);
	return;

nak:
	/* we are acting as a driver here and return an error if
         * we don't understand any ioctl instead of passing it
         * down further. 
         */
	vtunl2dbg(("vtunl_ioctl returned nak (reason %d).\n", error));
	iocp->ioc_error = error;
	mp->b_datap->db_type = M_IOCNAK;
	qreply(q, mp);
	return;
}

/* So we got a message from above */
int
vtunl_wput(queue_t *q, mblk_t *mp)
{
        /* see what we got */
	switch (mp->b_datap->db_type) {
	case M_DATA:
   	        /* it is a datagram to be encapsulated */
		vtunl_encap(q, mp);
		break;

	case M_PROTO:
	case M_PCPROTO:
		/* its a DLPI message */
		vtunl_wput_dlpi(q, mp);
		break;

	case M_IOCTL:
		/*
		 * For the vtunl to get M_IOCTL messages, there 
		 * shouldn't be an intervening IP module between
		 * vtunl and the stream head.
		 */
	        vtunl_ioctl(q, mp);
		break;

	/* TODO: Handle M_FLUSH messages */
	default:
		vtunl0dbg(("vtunl_wput: got unknown mblk type %d.\n",
			   mp->b_datap->db_type));
		freemsg(mp);
		break;
	}

}

/*
 * Handle a broadcast/multicast message sent up by DLPI device
 * under a vtunl instance. The message in mp should be freed
 * or passed to another queue before returning.
 */
static int
vtunl_multi_consume(queue_t *q, mblk_t *mp)
{
  /*
    vtunl3dbg(("$")); 
    freemsg(mp); 
    */
       /*
	* This may not be kosher, since we insert a message
	* in a queue belonging to another vtunl instance.
	* There are times when destroying vtunl streams
	* causes a kernel panic if multicast messages are
	* being encapsulated at that time, e.g. doing ^C
	* to vtunlconf while a lot of multicast packets are
	* being encapsulated causes a kernel fault. This
	* never happens if vtunl_multi_consume simply frees
	* the incoming message.
	*/
  /*
        if (vtunl_transport_q)
	  vtunl_encap(vtunl_transport_q, mp);
	else
	  freemsg(mp);
  */
   vtunl_encap(q, mp);
}
  
/* 
 * Process the M_PROTO or M_PCPROTO messages a vtunl instance gets
 * from below. This message could be one of several things:
 *    -- a TPI message confirming a T_BIND_REQ (i.e. subscription
 *       to an IP protocol)
 *    -- an acknowledgement to a DLPI message (e.g. DL_ATTACH_REQ,
 *       DL_BIND_REQ, DL_PROMISCON_REQ) sent to a DLPI driver
 *       underneath this vtunl instance, or
 *    -- a DL_UNITDATA_IND from a DLPI driver underneath this
 *       vtunl instance.
 */
static int
vtunl_rproto(queue_t *q, mblk_t *mp)
{
        u_long prim = *((u_long *) mp->b_rptr);
	struct vtunl_inst *vip = (struct vtunl_inst *) q->q_ptr;

	switch (vip->vcs.type) {
	case VTUNL_INST_TRANSPORT:
	        /* We only expect to get a T_BIND_ACK or T_ERROR_ACK. */

	        if (prim == T_BIND_ACK) {
		      vip->vcs.last_resp = T_BIND_ACK;
		      CHANGESTATE(vip->vcs.state, VTUNL_STATE_INIT,
				  VTUNL_STATE_BOUND, VTUNL_STATE_BAD);
		      if (vip->vcs.ip_protocol == IPPROTO_ENCAP)
		              vtunl_transport_q = WR(q);
		      vtunl3dbg(("Got T_BIND_ACK, new state %d.\n",
				 vip->vcs.state));
		} else if (prim == T_ERROR_ACK) {
		      vip->vcs.last_resp = T_ERROR_ACK;
		      vip->vcs.state == VTUNL_STATE_BAD;
		      vtunl3dbg(("Got T_ERROR_ACK, new state %d.\n",
				 vip->vcs.state));
		} else {
		      vtunl1dbg(("Unexpected TPI message (type %d).\n",
				 prim));
		}

		freemsg(mp);		
		break;

	case VTUNL_INST_DLSUSER:
	        /* Expect one of {OK_ACK|ERROR_ACK|BIND_ACK|UNITDATA_IND} */
	  
	        if (prim == DL_OK_ACK) {
		      dl_ok_ack_t  *dlokack;
		      ulong        ok_prim;

		      dlokack = (dl_ok_ack_t *)ALIGN32(mp->b_rptr);
		      vip->vcs.last_resp = DL_OK_ACK;
		      ok_prim = dlokack->dl_correct_primitive;
		      if (ok_prim == DL_ATTACH_REQ) {
			      CHANGESTATE(vip->vcs.state, VTUNL_STATE_INIT,
				    VTUNL_STATE_ATTACHED, VTUNL_STATE_BAD);
		      } else if (ok_prim == DL_PROMISCON_REQ) {
			      CHANGESTATE(vip->vcs.state, VTUNL_STATE_BOUND,
					  VTUNL_STATE_MULTI, VTUNL_STATE_BAD);
		      } else
			  vip->vcs.state = VTUNL_STATE_BAD;

		      vtunl3dbg(("DL_OK_ACK for request %d, new state %d.\n",
				 ok_prim, vip->vcs.state));

		} else if (prim == DL_ERROR_ACK) {
	              dl_error_ack_t  *dlerr;

		      dlerr = (dl_error_ack_t *)ALIGN32(mp->b_rptr);
		      vip->vcs.last_resp = DL_ERROR_ACK;
		      vip->vcs.state = VTUNL_STATE_BAD;

		      vtunl3dbg(("DL_ERR_ACK for request %d.(DLPI code %d)\n",
				 dlerr->dl_error_primitive, dlerr->dl_errno));

		} else if (prim == DL_BIND_ACK) {
		      vip->vcs.last_resp = DL_BIND_ACK;
		      CHANGESTATE(vip->vcs.state, VTUNL_STATE_ATTACHED,
				  VTUNL_STATE_BOUND, VTUNL_STATE_BAD);
		      vtunl3dbg(("DL_BIND_ACK, new state %d.\n",
				 vip->vcs.state));

		} else if (prim == DL_UNITDATA_IND) {
		      mblk_t *mp1;
		      dl_unitdata_ind_t *dlunitdata;
	          
		      dlunitdata = (dl_unitdata_ind_t *)ALIGN32(mp->b_rptr);
		  
		      /* Next mblk should be M_DATA */
		      if ((mp1 = mp->b_cont) == NULL ||
			  mp1->b_datap->db_type != M_DATA) {
			    vtunl1dbg(("Got bad dl_unitdata_ind.\n"));
			    freemsg(mp); return;
		      }
		      
		      if (dlunitdata->dl_group_address == 1) {
		            /* message is a broadcast or multicast */
			    freeb(mp);   /* free the M_PROTO block */
			    mp = mp1;
			    vtunl_multi_consume(q, mp);
			    return;      /* by now, mp is freed or
					  * passed to another queue */
		      } 
		} else 
		      vtunl1dbg(("Unknown DLPI message %d.\n", prim));

		freemsg(mp);
		break;
		
	default:
		vtunl1dbg(("vtunl type %d got unexpected message %d.\n",
			     vip->vcs.type, prim));
		freemsg(mp);
		break;
	}
}

/* Decapsulate an encapsulated datagram before passing it along. */
static int
vtunl_decap(queue_t *q, mblk_t *mp)
{
	ipha_t *iph;
	u_long hdrlen;
	mblk_t *mp1;
	struct decap_target *targetp;
	unsigned char decap_if;

	vtunl3dbg(("vtunl_decap() called\n"));

	/* Need at least an IP header. msgpullup(9F) concatenates
	 * and aligns the first len data bytes of the message 
	 * pointed to my mp, copying the data into a new message.
	 * If len equals -1, all data are concatenated. It fails 
	 * and returns NULL if the specified number of bytes 
	 * belonging to the same message cannot be found.
	 */

	if ((mp->b_wptr - mp->b_rptr) < sizeof(ipha_t)) {
		if ((mp1 = msgpullup(mp, sizeof(ipha_t))) == NULL) {
		        vtunl1dbg(("vtunl_decap: msgpullup failed.\n"));
			/* drop the packet */
			freemsg(mp);
			return;
		} else {
		  freemsg(mp);
		  mp = mp1;
		}
	}

	iph = (ipha_t *) mp->b_rptr;

	/* check IP version number, should be 4. */
	if  ((iph->ipha_version_and_hdr_length & 0xf0) != 0x40) {
		vtunl1dbg(("vtunl_decap: wrong IP version number.\n"));
		/* drop the packet */
		freemsg(mp);
		return;
	}

        /* more sanity checks */
	hdrlen = (iph->ipha_version_and_hdr_length & 0xf) << 2;
	if (hdrlen < sizeof(ipha_t)) {
		vtunl1dbg(("vtunl_decap: IPv4 header len too short.\n"));
		freemsg(mp);
		return;
	}

        /*
	 * Get the header in one data block. For MINENCAP, we could
	 * try to pullup more.
	 */
	if (hdrlen > sizeof(ipha_t)) {
		if ((mp1 = msgpullup(mp, hdrlen)) == NULL) {
		        vtunl1dbg(("vtunl_decap: msgpullup %d failed.\n",
				       hdrlen));
			freemsg(mp);
			return;
		} else {
		        freemsg(mp);
			mp = mp1;
			iph = (ipha_t *) mp->b_rptr;
		}
	}

	vtunl3dbg(("Outer src=%s dst=%s proto=%d len=%d\n", 
		   ntoa(iph->ipha_src, addr_buf1),
		   ntoa(iph->ipha_dst, addr_buf2),
		   iph->ipha_protocol, ntohs(iph->ipha_length)));

  	/* Recreate original datagram.  */

	switch (iph->ipha_protocol) {
	        case IPPROTO_ENCAP:
		        /* simply shave off the outer header */
		        if ((mp->b_wptr - mp->b_rptr) == hdrlen) {
			        mblk_t *mp1;

				mp1 = mp->b_cont;
				freeb(mp);
				mp = mp1;
			} else
			        mp->b_rptr += hdrlen;
			break;
			
	         case IPPROTO_MINENCAP:
 	         default:
		        vtunl1dbg(("vtunl_decap: Unsupported IP proto %d.\n",
				   iph->ipha_protocol));
			freemsg(mp);
			return;
	}

	iph = (ipha_t *) mp->b_rptr;
	vtunl3dbg(("Inner src=%s dst=%s proto=%d len=%d\n", 
		   ntoa(iph->ipha_src, addr_buf1),
		   ntoa(iph->ipha_dst, addr_buf2),
		   iph->ipha_protocol, ntohs(iph->ipha_length)));

	/* Drop this datagram if TTL is zero. */

	if (iph->ipha_ttl == 0) {
	        vtunl1dbg(("vtunl_decap: decapsualted pkt w/ TTL=0\n"));
                relay_icmp(q, mp, 11, 0, 0);
		freemsg(mp);
		return;
	}

	/* 
	 * Check if it is ok to pass-on the decapsulated packet.
	 * TO DO: We may wish to handle multicast packets
	 * in a special fashion.
	 */

	if (vtunldecapall == 0) {
	        /* Must check decap_table */

	        targetp = (struct decap_target *) hash_find(&decap_table,
						     (u32)(iph->ipha_dst));
		if (targetp != NULL) {
		        decap_if = targetp->decap_target_flags & DECAP_IF_MASK;
			if (decap_if == DECAP_IF_DST ||
			    decap_if == DECAP_IF_EITHER) {
			        putnext(q, mp);
				return;
			}
		}

	        targetp = (struct decap_target *) hash_find(&decap_table,
						     (u32)(iph->ipha_src));
		if (targetp != NULL) {
		        decap_if = targetp->decap_target_flags & DECAP_IF_MASK;
			if (decap_if == DECAP_IF_SRC ||
			    decap_if == DECAP_IF_EITHER) {
			        putnext(q, mp);
				return;
			}
		}

		vtunl1dbg(("Decapsualtion denied for [%s --> %s].\n",
			   ntoa(iph->ipha_src, addr_buf1),
			   ntoa(iph->ipha_dst, addr_buf2)));

		freemsg(mp);
	} else {
	        /* Pass along ALL decapsulated datagrams */
	        putnext(q, mp);
	}
}

/* Process an incoming ICMP message from below. */
int
vtunl_rput_icmp_err(queue_t *q, mblk_t *mp)
{
	ipha_t	*ipha, *iphi;
	icmph_t	*icmph;
	int	iph_hdr_length;
	u_long	srcipv4addr;
	int     need;
	mblk_t  *mp1;
	struct  t_exit *exitp;
	u16     old_pmtu;

	vtunl3dbg(("vtunl_rput_icmp_err() called\n"));

	/* TO DO: Be smart. Try to pull up the ip header, Then based on its
         * length, try to pull up either ip+icmp+ip header or
         * ip+icmp+ip+ip header and use the information for soft
         * state management.
	 */

	/* Don't assume that the entire packet is in a single m_blk. */
	/* check length - minimum packet is IPv4 in ICMP in IPv4 plus 8 */

	need = sizeof(ipha_t)+sizeof(icmph_t)+sizeof(ipha_t)+8;
	if ((mp->b_wptr - mp->b_rptr) < need) {
	     if ((mp1 = msgpullup(mp, need)) == NULL) {
	            vtunl1dbg(("vtunl_rput_icmp_err: Pkt too short 1.\n"));
		    /* drop the packet */
		    freemsg(mp);
		    return;
	     } else {
	            freemsg(mp);
		    mp = mp1;
	     }
	}

	ipha = (ipha_t *) mp->b_rptr;
	srcipv4addr = ipha->ipha_src;
	iph_hdr_length = IPH_HDR_LENGTH(ipha);

	if (iph_hdr_length > sizeof(ipha_t)) {
 	     /* we need to pullup more */
	     need += iph_hdr_length - sizeof(ipha_t);
	     if ((mp->b_wptr - mp->b_rptr) < need) {
	          if ((mp1 = msgpullup(mp, need)) == NULL) {
		       vtunl1dbg(("vtunl_rput_icmp_err: Pkt too short 2.\n"));
		       /* drop the packet */
		       freemsg(mp);
		       return;
		  } else {
		       freemsg(mp);
		       mp = mp1;
		  }
	     }
	}

	icmph = (icmph_t *)ALIGN32(&mp->b_rptr[iph_hdr_length]);
	ipha = (ipha_t *) &icmph[1]; /* now ipha points to inner IP header */
	
	if (ipha->ipha_protocol != IPPROTO_ENCAP) {
  	    vtunl1dbg(("vtunl_rput_icmp_err: wrong protocol.\n"));
	    freemsg(mp);
	    return;
	}

	vtunl2dbg(("vtunl_rput_icmp_err: Got ICMP <%d:%d> (%s) from %s " \
		   "on way to %s.\n",
		   icmph->icmph_type, icmph->icmph_code, 
		   parse_icmph(icmph),
		   ntoa(srcipv4addr, addr_buf1), 
		   ntoa(ipha->ipha_dst, addr_buf2)));

	/* For now, take it easy and just build up tunnel soft
         * state. Assume that we never have enough info to relay
         * an ICMP to the original source. Will do so the next time
         * we get a pkt for encapsulation.
	 *
	 * We should use received ICMP messages to build up
	 * tunnel state info, e.g. keep track of source quench messages
         * and the next time a pcket is being tunneled, send a source
         * quech message to that source. An ICMP (type 11)time exceeded
         * message reports a loop within the tunnel. ICMP (type 3)
         * Dest unreachable messages report delivery failures to
         * decapsulator. This soft state must be reported to the 
         * originator as type 3 code 0 Network Unreachable.
	 */

	if ((exitp = (struct t_exit *) hash_find(&texit_table, 
			               (u32)ipha->ipha_dst)) == NULL)
	  {
	    vtunl1dbg(("vtunl_rput_icmp_err: ICMP for unknown tunl exit.\n"));
	    freemsg(mp);
	    return;
	  }

	switch (icmph->icmph_type) {
	case 3:  /* Destination unreachable */
	    switch (icmph->icmph_code) {
	        case 0:  /* network unreachable */                           
  	        case 1:  /* host unreachable */
                case 2:  /* protocol unreachable */
                         /* Record that the target is unreachable */

	                 exitp->t_exit_state |= 0x8000;
		         vtunl2dbg(("Marking exit %s unreachable.\n",
				    ntoa(exitp->t_exit_addr, addr_buf1)));
			 break;

	        case 4:  /* Do Path MTU stuff */

	                 old_pmtu = exitp->t_exit_pmtu;
	                 if (icmph->icmph_du_mtu == 0) {
		              mtu_down((u16 *)&exitp->t_exit_pmtu);
			      /* wait before attempting PMTU increase */
			      exitp->t_exit_state |= 0x0007; 
			      vtunl2dbg(("Estimated PMTU for %s is %d " \
					 "(<= %d).\n",
					 ntoa(exitp->t_exit_addr, addr_buf1), 
					 exitp->t_exit_pmtu,
					 old_pmtu));
			 } else if (ntohs(icmph->icmph_du_mtu)
				    < exitp->t_exit_pmtu) {
		              exitp->t_exit_pmtu = ntohs(icmph->icmph_du_mtu);
			      /* wait before attempting PMTU increase */
			      exitp->t_exit_state |= 0x0007; 
			      vtunl2dbg(("New PMTU for %s is %d (< %d).\n",
					 ntoa(exitp->t_exit_addr, addr_buf1), 
					 exitp->t_exit_pmtu,
					 old_pmtu));
			 }
			 
			 break;

	        case 3:  /* this should never be seen */
	        case 5:  /* we ignore this one */
                default:
		         break;
	    }
	    break;
	case 4:  /* Source Quench */
	         /* RFC 1853 says relay it, ip4inip4 draft says SHOULD NOT */
	        break;
	case 5: /* Redirect */
	        /* Never relay, may handle */
	        break;
	case 11: /* Time exceeded, potential loop */
	         /* Target is unreachable */
	         exitp->t_exit_state |= 0x8000;
		 vtunl2dbg(("Will send ICMP host unreachable for exit %s.\n",
			    ntoa(exitp->t_exit_addr, addr_buf1)));
		 break;
	}

	freemsg(mp);
}

/* Process messages received from below. */
int
vtunl_rput(queue_t *q, mblk_t *mp)
{

       /* See what we got */

	switch (mp->b_datap->db_type) {
	case M_DATA:
     	        /* its an encapsulated datagram */
		vtunl_decap(q, mp);
		break;

	case M_PROTO:
	case M_PCPROTO:
	        vtunl_rproto(q, mp);
		break;

	case M_CTL:
                /* It is an ICMP message. For UDP and TCP protocols,
		 * the IP multiplexer sends the ICMP message only
		 * on the appropriate upstream. For less familiar
		 * protocols, it passes a copy to each upstream as
		 * a message of type M_CTL (checked this from ip.c)
                 */

		vtunl_rput_icmp_err(q, mp);
		break;

	/* TODO: Handle M_FLUSH messages. */
	default:
	        vtunl0dbg(("vtunl_rput: got unknown mblk type %d.\n",
			       mp->b_datap->db_type));
		freemsg(mp);
		break;
	}
		
}

/* 
 * Encapsulate an IP datagram before passing it along. If the vtunl
 * instance doing this encapsulation is of type VTUNL_INST_DLSUSER,
 * then we are dealing with multicast messages. No ICMP messages 
 * should be generated for these. Also, the checksum and source 
 * fields cannot be left zero.
 */
static int
vtunl_encap(queue_t *q, mblk_t *mp)
{
	mblk_t *newmp;
	mblk_t *mp1;
	ipha_t *iph, *ipho; /* IP headers -- inner and outer */
	u16     encap_type;
	int     i, dlsuser;
	char    dont_frag;
	u16     encap_hdr_len;
	struct encap_target *targetp;
	struct t_exit *exitp;
	struct vtunl_inst *vip;
	static u16 ipipid = 0;

        /* vtunl3dbg(("vtunl_encap() called\n")); */

	/* Get atleast the entire ipv4 header in one data block. */

	if ((mp->b_wptr - mp->b_rptr) < sizeof(ipha_t)) {
		if ((mp1=msgpullup(mp, sizeof(ipha_t))) == NULL) {
		        vtunl1dbg(("vtunl_encap: msgpullup failed.\n"));
			/* drop the packet */
			freemsg(mp);
			return;
		} else {
		        freemsg(mp);
			mp = mp1;
		}
	}

	iph = (ipha_t *) ALIGN32(mp->b_rptr);

	/* check version  */
	if ((iph->ipha_version_and_hdr_length & 0xf0) != 0x40) {
	        vtunl1dbg(("vtunl_encap: wrong IP version.\n"));
		freemsg(mp);
		return;
	}

	vip = (struct vtunl_inst *) q->q_ptr;
	dlsuser = (vip->vcs.type == VTUNL_INST_DLSUSER) ? 1:0;

	/*
	 * Find an entry for the given destination. For unicast messages,
	 * try to use the default entry if no entry exists for the 
	 * specified destination.
	 */

	if ((targetp = (struct encap_target *) hash_find(&encap_table,
			(u32)(iph->ipha_dst))) == NULL) {
	        if (dlsuser) {
		        freemsg(mp);
			return;
		} else {
		        if ((targetp = (struct encap_target *) 
			    hash_find(&encap_table, DEFAULT_ADDR)) == NULL) {
			            vtunl1dbg(("vtunl_encap: %s %s.\n",
					       "encapsulation denied for",
					       ntoa(iph->ipha_dst,addr_buf1)));
				    freemsg(mp);
				    return;
			}
		}
	}

	/* sanity check */
	if (targetp->encap_target_exit_count == 0) {
	        vtunl0dbg(("OOPS! Found a target with no exits!\n"));
		freemsg(mp);
		return;
	}

	/*
	 * If packet was not locally generated, decrement ttl,
	 * and update checksum. If ttl is zero, discard and
	 * send an ICMP.
	 */

        if (iph->ipha_src != vip->vcs.address) { 
	        u32   sum;

	        iph->ipha_ttl--;
	        sum = (int)iph->ipha_hdr_checksum + htons(0x0100);
	        iph->ipha_hdr_checksum = (u16)(sum + (sum >> 16));
	}

	if (iph->ipha_ttl == 0) {
	        if (!dlsuser) {
		        vtunl1dbg(("vtunl_encap: TTL is zero.\n"));
			relay_icmp(q, mp, 11, 0, 0);
		}
		freemsg(mp);
		return;
	}

	/* 
	 * TODO: Check vip->vcs.address and vip->vcs.addr_mask to 
	 * determine if a pkt with destination 255.255.255.255 should
	 * indeed be sent to a tunnel exit, e.g. when the Home Agent
	 * has multiple interfaces, some of the b'cast pkts received by
	 * it may not have been receieved by a mobile node even at home.
	 */

	/* In a loop, process each (of possibly several) tunnel exits. */
	for (i = 0; i < targetp->encap_target_exit_count; i++) {
	        /* Make a copy for all exits except the last one */
	        if (i != targetp->encap_target_exit_count - 1)
	                mp1 = copymsg(mp);  
		else
		        mp1 = mp;
	    
		/* Now mp1 has the message to be encapsulated */
		iph = (ipha_t *) mp1->b_rptr;

		if ((exitp = targetp->encap_target_exit[i]) == NULL) {
		        vtunl0dbg(("OOPS! null exit for target.\n"));
			freemsg(mp1);
			continue;
		}
	    
		/* 
		 * Check for loops. For now, we only avoid two kinds
		 * of loops. Saying "a pkt from Y for X should be 
		 * first sent to Y" is a loop. Similarly, saying 
		 * "to go to X first go to X" causes a loop. 
		 * TODO: need a better way!
		 */

		if (iph->ipha_src == exitp->t_exit_addr || 
		    iph->ipha_dst == exitp->t_exit_addr) {
		        vtunl1dbg(("Loop suspected.\n"));
			freemsg(mp1);
			continue;
		}

		/* Is target reachable? */
		if (exitp->t_exit_state & 0x8000) {
		        if (!dlsuser) 
			        relay_icmp(q, mp1, 3, 1, 0);
			freemsg(mp1);
			continue;
		}

		encap_type = targetp->encap_target_encapflags[i] & 0x001c;
		if (encap_type == 0x0010) 
		        encap_hdr_len = 12;      /* MIN_ENCAP */
		/* 
		 *   else if (encap_type == 0x0008)
		 *    encap_hdr_len = ??;       GRE_ENCAP 
		 *   else if (encap_type == 0x0004)
		 *    encap_hdr_len = ??;       VJ_ENCAP 
		 */
		else
		        encap_hdr_len = sizeof(ipha_t);    /* IPIP */

		/* Check available PMTU */
		dont_frag = 1;
		if ((u16)(exitp->t_exit_pmtu - encap_hdr_len) < (u16) 576) {
        	        if (ntohs(iph->ipha_length) > 576) {
			        vtunl2dbg(("IP length %d, PMTU %d e-hdr %d.\n",
					   ntohs(iph->ipha_length),
					   exitp->t_exit_pmtu,
					   encap_hdr_len));
			        if (!dlsuser)
				  relay_icmp(q, mp1, 3, 4, 576); 
				freemsg(mp1);
				continue;
			} else {
			        dont_frag = 0;   /* clear outer DF */
			}
		} else {
	                if ((u16) ntohs(iph->ipha_length) > 
			    (u16)(exitp->t_exit_pmtu - encap_hdr_len)) {
			        vtunl2dbg(("IP len %d exceeds PMTU (%d-%d).\n",
					   ntohs(iph->ipha_length),
					   exitp->t_exit_pmtu,
					   encap_hdr_len));
			        if (!dlsuser)
				  relay_icmp(q, mp1, 3, 4, exitp->t_exit_pmtu
					     - encap_hdr_len);
				freemsg(mp1);
				continue;
			}
		}

		/* Room for encapsulation header in the data block? */
		if ((u16)(mp1->b_rptr - mp1->b_datap->db_base) <
				encap_hdr_len) {
		        /* no! allocate a new mblk and chain the two */
		        newmp = allocb(encap_hdr_len, BPRI_LO);
			if (newmp == (mblk_t *) 0) {
			        vtunl0dbg(("vtunl_encap: allocb failed.\n"));
				freemsg(mp1);
				return;
			}
			newmp->b_wptr += encap_hdr_len;
			newmp->b_cont = mp1;
			mp1 = newmp;
		} else {
		        /* yes, there's enough room */
		        mp1->b_rptr -= encap_hdr_len;
		}

		/* Fill out the encapsulation header */
		bzero((char *) mp1->b_rptr, encap_hdr_len);
		
		switch (encap_type) {
		case 0x0000:                              /* IPIP */
		        ipho = (ipha_t *) mp1->b_rptr;
			ipho->ipha_version_and_hdr_length = 
			  IP_SIMPLE_HDR_VERSION;
			ipho->ipha_type_of_service = iph->ipha_type_of_service;
			ipho->ipha_length = htons(ntohs(iph->ipha_length)
						  +encap_hdr_len);
			ipho->ipha_ident = 0;	/* let ip fill it */
			ipho->ipha_fragment_offset_and_flags = 0;
			if ((iph->ipha_fragment_offset_and_flags & 
			     htons(0x4000)) && dont_frag)
			  ipho->ipha_fragment_offset_and_flags = htons(0x4000);
			ipho->ipha_ttl = IPIP_DEF_TTL;
			ipho->ipha_protocol = IPPROTO_ENCAP;
			ipho->ipha_hdr_checksum = 0; /* let ip fill it */
			ipho->ipha_src = 0;          /* let ip fill it */
			ipho->ipha_dst = exitp->t_exit_addr;

			/* 
			 * If we are encapsulating packets on the way
			 * up, we can't let IP fill certain fields.
			 * We must fill these ourselves.
			 */
			ipipid++;
			if (dlsuser) { 
			        ipho->ipha_ident = htons(ipipid);
				ipho->ipha_src = vip->vcs.address;
				ipho->ipha_hdr_checksum = 
				      in_cksum((u16 *)ipho, sizeof(ipha_t));
			}

			/* TODO: check for invalid destinations? */

			/* Pass along the packet in the same direction. */
			vtunl3dbg(("IPIP: <src=%s dst=%s proto=%d len=%d> " \
				   "tunneled to %s.\n",
				   ntoa(iph->ipha_src, addr_buf1), 
				   ntoa(iph->ipha_dst, addr_buf2),
				   iph->ipha_protocol,
				   ntohs(iph->ipha_length),
				   ntoa(ipho->ipha_dst, addr_buf3)));
			putnext(q, mp1);
			break;

		default:
		        vtunl1dbg(("Encap_type %x not implemented.\n",
				   encap_type));
			freemsg(mp1);
			break;
		}
	}
}

/***************************************************************************
 *                          R E L A Y     I C M P                          *
 ***************************************************************************/
static int
in_cksum(u16 *buf, int len)
{
	register int nleft = len;
	register u16 *w = buf;
	register int sum = 0;
	u_short answer = 0;

	
	/* 
	 * Our algorithm is simple, using a 32 bit accumulator (sum), we add
	 * sequential 16 bit words to it, and at the end, fold back all the
	 * carry bits from the top 16 bits into the lower 16 bits.
	 */
	 
	while (nleft > 1)  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if (nleft == 1) {
		*(u_char *)(&answer) = *(u_char *)w ;
		sum += answer;
	}

	/* add back carry outs from top 16 bits to low 16 bits */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16  */
	sum += (sum >> 16);			/* add carry  */
	answer = ~sum;				/* truncate to 16 bits */
	return(answer);
}

char *parse_icmph(icmph_t *icmph)
{
    static char icmp_msg[80];

    switch (icmph->icmph_type) {
        case 0:  /* Echo */
	    strcpy(icmp_msg, "Echo reply");
	    break;
		 
	case 3:  /* Destination unreachable */
	    switch (icmph->icmph_code) {
	        case 0:  
		    strcpy(icmp_msg, "Network unreachable");
		    break;
  	        case 1:  
		    strcpy(icmp_msg, "Host unreachable");
		    break;	
                case 2:  
		    strcpy(icmp_msg, "Protocol unreachable");
		    break;
                case 3:  
		    strcpy(icmp_msg, "Port unreachable");
		    break;
	        case 4:  
		    sprintf(icmp_msg, "Fragmentation needed DF set PMTU " \
			    "is %d", ntohs(icmph->icmph_du_mtu));
		    break;
	    }
	    break;
	case 8: 
	    strcpy(icmp_msg, "Echo request");
	    break;
	case 9: 
	    strcpy(icmp_msg, "Router advertisement");
	    break;
	case 10: 
	    strcpy(icmp_msg, "Router solicitation");
	    break;
	case 11: 
	    strcpy(icmp_msg, "Time exceeded");
	    break;
	default: 
	    strcpy(icmp_msg, "Unrecognized");
	    break;
	}
    return(icmp_msg);
}

/* 
 * Relay an ICMP message of the specified type in response to the 
 * IP datagram contained in mp (provided mp contains the first fragment).
 * Do not alter the message in mp and do not send ICMPs in response to
 * broadcast/multicast packets or when mp is not the first fragment.
 */

static int
relay_icmp(queue_t *q, mblk_t *mp, int type, int code, u16 pmtu)
{
  mblk_t *newmp;
  ipha_t *iph, *ipha;
  icmph_t *icmph;
  u16    icmp_size, have;
  u32    addr;
  char   *cp;

  /* Should we be sending out an ICMP message? */

  ipha = (ipha_t *) mp->b_rptr;
  addr = ntohl(ipha->ipha_src);
  if (addr == 0xffffffff || (addr & 0xf0000000) == 0xe0000000 ||
      (addr & 0xff000000) == 0x7f000000) {
    vtunl2dbg(("relay_icmp: should not respond based on src.\n"));
    return;
  }

  addr = ntohl(ipha->ipha_dst);
  if (addr == 0xffffffff || (addr & 0xf0000000) == 0xe0000000) {
          vtunl2dbg(("relay_icmp: Must not send to %x.\n", addr));
	  return;
  } else if (ipha->ipha_fragment_offset_and_flags & htons(0x01ff)) {
          vtunl2dbg(("relay_icmp: Must only respond to first fragment.\n"));
	  return;
  } else if (ipha->ipha_protocol == IPPROTO_ICMP) {
          int iph_hdr_length = IPH_HDR_LENGTH(ipha);

          /* Only respond to ICMP echo requests */
          icmph = (icmph_t *)ALIGN32(&mp->b_rptr[iph_hdr_length]);
          if (icmph->icmph_type != 8 || icmph->icmph_code != 0) {
	          vtunl2dbg(("relay_icmp: Must not respond to ICMP <%d:%d> "\
			     "(%s).\n",
			     icmph->icmph_type, icmph->icmph_code,
			     parse_icmph(icmph)));
		  return;
	  }
  }

  /* Allocate storage for the ICMP message */

  icmp_size = sizeof(ipha_t) + sizeof(icmph_t) + IPH_HDR_LENGTH(ipha) + 8;
  newmp = allocb(icmp_size, BPRI_LO);
  if (newmp == (mblk_t *) 0) {
    vtunl0dbg(("relay_icmp: allocb failed.\n"));
    return;
  }
  
  /* Fill it up */

  newmp->b_wptr += icmp_size;
  iph = (ipha_t *) newmp->b_rptr;
  iph->ipha_version_and_hdr_length = IP_SIMPLE_HDR_VERSION;
  iph->ipha_type_of_service = 0;
  iph->ipha_length = htons(icmp_size);
  iph->ipha_ident = 0;	/* to be filled in by /dev/ip below */
  iph->ipha_fragment_offset_and_flags = 0;
  iph->ipha_ttl = 255;
  iph->ipha_protocol = IPPROTO_ICMP;
  iph->ipha_hdr_checksum = 0; /* to be filled in by ip below */
  iph->ipha_src = 0;         /* let ip below fill it */
  iph->ipha_dst = ipha->ipha_src;
  
  icmph = (icmph_t *)ALIGN32(&newmp->b_rptr[sizeof(ipha_t)]);
  icmph->icmph_type = type;
  icmph->icmph_code = code;
  icmph->icmph_checksum = (u16) 0;

  if (type == 3) {
    icmph->icmph_du_zero = (u16) 0;
    icmph->icmph_du_mtu = (code == 4) ? htons(pmtu): (u16) 0;
  } else if (type == 11) {
    icmph->icmph_du_zero = (u16) 0;
    icmph->icmph_du_mtu = (u16) 0;
  } else {
    vtunl1dbg(("relay_icmp: Bad type %d.\n", type));
    freemsg(newmp);
    return;
  }

  /* Copy the IP header and a portion of the original datagram */

  have = sizeof(ipha_t) + sizeof(icmph_t);
  cp = (char *) newmp->b_rptr;

  while (have < icmp_size) {
    cp += have;
    if (mp == NULL) {
      vtunl1dbg(("relay_icmp: original datagram too short.\n"));
      freemsg(newmp);
      return;
    }
    if ((u16)(mp->b_wptr - mp->b_rptr) > (u16)(icmp_size - have)) {
      bcopy((char *)mp->b_rptr, cp, icmp_size - have);
      have = icmp_size;
      
    } else {
      bcopy((char *)mp->b_rptr, cp, mp->b_wptr - mp->b_rptr);
      have += mp->b_wptr - mp->b_rptr;
    }
    mp = mp->b_cont;
  } 

  /* Compute ICMP checksum */
  icmph->icmph_checksum = in_cksum((u16 *)icmph, have - sizeof(ipha_t));
  vtunl2dbg(("Relayed ICMP <%d:%d> (%s) to %s.\n", type, code, 
	     parse_icmph(icmph), ntoa(ipha->ipha_src, addr_buf1)));
  putnext(q, newmp);
}

