/*
 * Copyright (c) 1996 by Sun Microsystems, Inc.
 */

#pragma ident	"@(#)vtunl_ioctl.c 1.2 97/08/29 SMI"

/* 
 *  Author: Vipul Gupta (vgupta@eng.sun.com)
 *
 *  Vtunl configuration state is maintained in hash tables and
 *  this file contains all the hash table manipulation routines
 *  and code for several of the vtunl ioctls that manipulate
 *  configuration information.
 */
#include <sys/types.h>
#include <sys/kmem.h>

#include <sys/conf.h>
#include <sys/errno.h>
#include <sys/stream.h>
#include <inet/common.h>

#include "vtunl_impl.h"

extern int vtunldebug;
extern u16 vtunl_default_pmtu;
extern total_exit_count;

/*************************************************************************
 *                      HASH TABLE MANIPULATION                          *
 *************************************************************************/

/* hash tables used to keep vtunl's configuration state */

struct hash_table bcast_table, decap_table, encap_table, texit_table;

/* Initialize a hash table */
void
hash_tbl_init(struct hash_table *htbl) {
        int i;

	htbl->size = 0;
	for (i = 0; i < HASH_TBL_SIZE; i++) htbl->hash_entries[i] = NULL;
}

/* Clean a hash table */
void 
hash_tbl_clean(struct hash_table *htbl, int size) {
        int i;
        struct hash_entry *entryp;
	struct hash_entry *p;

	for (i = 0; i < HASH_TBL_SIZE && htbl->size; i++) {
	        entryp = htbl->hash_entries[i];
		while (entryp) {
		        p = entryp->hash_entry_next;
			kmem_free((void *)entryp, (size_t) size);
			--htbl->size;
			entryp = p;
		}
		htbl->hash_entries[i] = NULL;
	}
	if (htbl->size) vtunl0dbg(("hash_tbl_clean: bad size\n", htbl->size));
	htbl->size = 0;
}

/* Find a hash table entry given an IPv4 address */
struct hash_entry * 
hash_find(struct hash_table *hash_tbl, u32 addr)
{
        struct hash_entry *p;

	p = hash_tbl->hash_entries[HASHIT(addr)];
	while (p) {
	        if (p->hash_entry_addr == addr) return(p);
		p = p->hash_entry_next;
	}
	return (NULL);
}

/*
 * Link an entry into the hash table based on its address field. The entry
 * must have already been properly initialized. 
 */
int
hash_link(struct hash_table *hash_tbl, struct hash_entry *entry) 
{
        struct hash_entry *p;
	int index;
  
	index =  HASHIT(entry->hash_entry_addr);
	p = hash_tbl->hash_entries[index];
	while (p) {
	        if (p->hash_entry_addr == entry->hash_entry_addr)
		        return(EEXIST);
		p = p->hash_entry_next;
	}
  
	p = hash_tbl->hash_entries[index];
	entry->hash_entry_next = p;
	hash_tbl->hash_entries[index] = entry;
	hash_tbl->size++;
	return(0);
}	     


/*
 * Unlink the entry with the given key from the specified hash table. The
 * routine does not free the memory associated with the entry.
 */
struct hash_entry *
hash_unlink(struct hash_table *hash_tbl, u32 addr)
{
        struct hash_entry *p;
	struct hash_entry *q;
	int index;
  
	index =  HASHIT(addr);

	/* no entries */
	if ((q = hash_tbl->hash_entries[index]) == NULL) return(NULL);

	/* first entry matches */
	if (q->hash_entry_addr == addr) {
	        hash_tbl->hash_entries[index] = q->hash_entry_next;
		hash_tbl->size--;
		return(q);
	}

	p = q->hash_entry_next;
	while (p) {
	        if (p->hash_entry_addr == addr) {
		        q->hash_entry_next = p->hash_entry_next;
			hash_tbl->size--;
			return(p);
		}
		q = p;
		p = p->hash_entry_next;
	}
	return(NULL);
}


/*************************************************************************
 *                        VTUNL IOCTL PROCESSING                         *
 *************************************************************************/

int
vtunl_encap_add(struct vtunl_req *reqp) 
{
        struct encap_target *targetp;
	struct t_exit *exitp;
	struct t_exit *tmp;
	int i;

	if (reqp->vr_target_addr == 0 || reqp->vr_t_exit_addr == 0 ||
	    reqp->vr_encap_flags & 0x007f)
	        return(EINVAL);

	/* find a target entry or else create one */
	targetp = (struct encap_target *) hash_find(&encap_table, 
						    reqp->vr_target_addr);
	if (targetp == NULL) {
	        targetp = (struct encap_target *) 
		  kmem_zalloc(sizeof(struct encap_target), KM_NOSLEEP);
		if (targetp == NULL) return(ENOMEM);
		targetp->encap_target_addr = reqp->vr_target_addr;
		targetp->encap_target_exit_count = 0;
		if (hash_link(&encap_table, (struct hash_entry *)targetp)
		    == EEXIST) {
		        vtunl0dbg(("OOPS1! hash_link found entry that %s.\n",
				   "hash_find did not"));
			kmem_free(targetp, sizeof(struct encap_target));
			return(EPROTO);
		}
	} else { /* found a target entry */
	        if ((reqp->vr_encap_flags & SIMULTANEOUS) == 0) {
		        vtunl3dbg(("vtunl_encap_add: Existing bindings %s.\n",
				   "will be removed"));
			/* clean existing associations */
			for (i = 0; i<targetp->encap_target_exit_count; i++) {
			        exitp = targetp->encap_target_exit[i];
				targetp->encap_target_encapflags[i] = 0;
				if (exitp == NULL) {
				        vtunl0dbg(("OOPS! Bad exit count.\n"));
					return(EPROTO);
				}
				total_exit_count--;
				if (--exitp->t_exit_ref_count == 0) {
				      if ((tmp = (struct t_exit *)
					   hash_unlink(&texit_table, 
						       exitp->t_exit_addr))
					  != exitp) {
					vtunl0dbg(("OOPS! exit structure %s",
						   "not in hash tbl.\n"));
					return(EPROTO);
				      }
				      kmem_free(exitp, sizeof(struct t_exit));
				}
			}
			targetp->encap_target_exit_count = 0;
		} else {
		        vtunl3dbg(("vtunl_encap_add: Simultaneous %s.\n",
				   "bindings requested"));
		}

		if (targetp->encap_target_exit_count == MAXEXITCOUNT) 
		  return(ENOSPC);
	}
  
	reqp->vr_encap_flags &= ~SIMULTANEOUS; /* never store the S bit */

	/*
	 * By now, we have found/created a target entry in the hash 
	 * table, now we find the tunnel exit entry or create one.
	 */

	exitp = (struct t_exit *)hash_find(&texit_table, reqp->vr_t_exit_addr);
	if (exitp == NULL) {
	        exitp = (struct t_exit *) kmem_zalloc(sizeof(struct t_exit),
						      KM_NOSLEEP);
		if (exitp == NULL) { 
		        vtunl0dbg(("vtunl_add_encap: kmem_zalloc failed.\n"));
			return(ENOMEM);
		}
		exitp->t_exit_addr = reqp->vr_t_exit_addr;
		exitp->t_exit_ref_count = 0;
		exitp->t_exit_pmtu = vtunl_default_pmtu;
		exitp->t_exit_maxpmtu = vtunl_default_pmtu;
		exitp->t_exit_state = 0x0007;  /* don't increase PMTU soon */
		if (hash_link(&texit_table, (struct hash_entry *)exitp)
		    == EEXIST) {
		        vtunl0dbg(("OOPS2! hash_link found entry %s.\n",
				   "that hash_find did not"));
		};
	}
       
	/* 
	 * If the target entry already lists the tunnel_exit, just update
	 * the encapsulation flags,
	 */
	
	for (i = 0; i < targetp->encap_target_exit_count; i++) {
	        if (targetp->encap_target_exit[i] == exitp) {
		        targetp->encap_target_encapflags[i] = 
			  reqp->vr_encap_flags;
			return(0);
		}
	}
  
	/* else establish a connection between the two */
	targetp->encap_target_exit[i] = exitp;
	targetp->encap_target_encapflags[i] = reqp->vr_encap_flags;
	targetp->encap_target_exit_count++;
	total_exit_count++;
	exitp->t_exit_ref_count++;
  
	return(0);
}

int
vtunl_encap_rem_target(u32 addr) 
{
        int i;
	struct encap_target *targetp;
	struct t_exit *tmp1;
	struct encap_target *tmp2;
	struct t_exit *exitp;

	/* find a target entry */
	if ((targetp = (struct encap_target *) hash_find(&encap_table, addr))
	    == NULL) return(ENXIO);

	/* clean existing associations */
	for (i = 0; i < targetp->encap_target_exit_count; i++) {
	        exitp = targetp->encap_target_exit[i];
		targetp->encap_target_encapflags[i] = 0;
		if (exitp == NULL) {
		        vtunl0dbg(("OOPS! Bad exit count.\n"));
			return(EPROTO);
		}
		total_exit_count--;
		if (--exitp->t_exit_ref_count == 0) {
		  if ((tmp1 = (struct t_exit *)hash_unlink(&texit_table, 
				       exitp->t_exit_addr)) != exitp)
		    {
		      vtunl0dbg(("OOPS! *exit structure not in hash tbl.\n"));
		      return(EPROTO);
		    }
		  kmem_free(exitp, sizeof(struct t_exit));
		}
	}

	targetp->encap_target_exit_count = 0;

	/* unlink and free memory */
	if ((tmp2 = (struct encap_target *) hash_unlink(&encap_table, 
						 targetp->encap_target_addr))
	    != targetp) {
	        vtunl0dbg(("OOPS! encap target entry not on hash table.\n"));
		return(EPROTO);
	}

	kmem_free(targetp, sizeof(struct encap_target));
	return(0);
}

int
vtunl_encap_rem_texit(u32 target_addr, u32 exit_addr)
{
        int i, cnt;
	struct encap_target *targetp;
	struct t_exit *exitp;
	struct encap_target *tmp1;
	struct t_exit *tmp2;

	/* find a target entry */
	if ((targetp = (struct encap_target *) 
	     hash_find(&encap_table, target_addr)) == NULL) return(ENXIO);

	/* find a tunnel exit entry */
	if ((exitp = (struct t_exit *) 
	     hash_find(&texit_table, exit_addr)) == NULL) return(ENXIO);

	/* clean existing association */
	cnt = targetp->encap_target_exit_count;
	for (i = 0; i < cnt; i++) {
	        if (exitp == targetp->encap_target_exit[i]) break;
	}

	if (i == cnt) return(ENXIO); /* No association! */

	/* deal with the encap_target entry */

	/* first pull the last exit into this position */
	targetp->encap_target_exit[i] = targetp->encap_target_exit[cnt-1];
	targetp->encap_target_encapflags[i]= 
	     targetp->encap_target_encapflags[cnt-1];

	/* update exit count and delete if appropriate */
	if (--targetp->encap_target_exit_count == 0) {
	  if ((tmp1 = (struct encap_target *) hash_unlink(&encap_table, 
						   targetp->encap_target_addr))
	      != targetp) {
	        vtunl0dbg(("OOPS! ++encap target entry not on hash table.\n"));
		return(EPROTO);
	  }
	  kmem_free(targetp, sizeof(struct encap_target));
	}

	/*
	 * Deal with the t_exit entry: update ref count and delete 
	 * if appropriate.
	 */
	total_exit_count--;
	if (--exitp->t_exit_ref_count == 0) {
	        if ((tmp2 = (struct t_exit *)hash_unlink(&texit_table, 
					    exitp->t_exit_addr)) != exitp)
		        {
			  vtunl0dbg(("OOPS! Exit struct not in hash tbl.\n"));
			  return(EPROTO);
			}
		kmem_free(exitp, sizeof(struct t_exit));
	}

	return(0);
}


/*
 * Put the number of encapsulation targets in the first four bytes.
 * Then, for each encap_target, list its address (4 bytes), num of
 * tunnel exits (4 bytes) and the address of each exit. Finally, put
 * the number of tunnel exits (4 bytes) and for each tunnel exit, 
 * fill its address (4 bytes), current PMTU estimate (2 bytes) 
 * maximum PMTU (2 bytes) and state (2 bytes).
 */

int
vtunl_encap_fill(unsigned char *buffer)
{
          int i, j;
	  u32 *cpos;
	  struct encap_target *etp;
	  struct t_exit *tep;
	  struct exit_fill *efp;

	  cpos = (u32 *) buffer;
	  *cpos = (u32) encap_table.size;
	  cpos++;
	  for (i = 0; i < HASH_TBL_SIZE; i++) {
	          etp = (struct encap_target *) encap_table.hash_entries[i];
		  while (etp) {
		          *cpos = etp->encap_target_addr;
			  cpos++;
			  *cpos = (u32) etp->encap_target_exit_count;
			  cpos++;

			  for (j = 0; j < etp->encap_target_exit_count; j++) {
			          tep = etp->encap_target_exit[j];
				  *cpos = tep->t_exit_addr;
				  cpos++;
			  }

			  etp = etp->encap_target_next;
		  }
	  }
		 
	  *cpos = (u32) texit_table.size;
	  cpos++;
	  efp = (struct exit_fill *) cpos;
	  for (i = 0; i < HASH_TBL_SIZE; i++) {
	          tep = (struct t_exit *) texit_table.hash_entries[i];
		  while (tep) {
		          efp->addr = tep->t_exit_addr;
			  efp->cur_pmtu = tep->t_exit_pmtu;
			  efp->max_pmtu = tep->t_exit_maxpmtu;
			  efp->flags = tep->t_exit_state;
			  efp++;
			  tep = tep->t_exit_next;
		  }
	  }

	  return(0);
}

/* 
 * Fill out decap_table size followed by the address (4 bytes) and
 * flags (2 bytes) for each entry.
 */

int
vtunl_decap_fill(unsigned char *buffer)
{
          int i;
	  u32 *tmp;
	  struct decap_target *dtp;
	  struct decap_fill *dfp;

	  tmp = (u32 *) buffer;
	  *tmp = (u32) decap_table.size;
	  tmp++;

	  dfp = (struct decap_fill *) tmp;
	  for (i = 0; i < HASH_TBL_SIZE; i++) {
	          dtp = (struct decap_target *) decap_table.hash_entries[i];
		  while (dtp) {
		          dfp->addr = dtp->decap_target_addr;
			  dfp->flags = dtp->decap_target_flags;

			  dfp++;
			  dtp = dtp->decap_target_next;
		  }
	  }

	  return(0);
}
