/*
 * An implementation of Mobile-IP as described in RFC 2002.
 * Author: Vipul Gupta (vgupta@eng.sun.com)
 * auth.c : Authentication routines used in Mobile IP.
 */

#pragma ident   "@(#)auth.c 1.1 97/10/01"

#include "mip.h"
#include "md5.h"

#include <sys/types.h>
#include <netinet/in.h>

extern int mipdebug;
extern char msg[];

/* 
 * Compute authenticator for the data in buffer based on the 
 * Mobile IP Security Association pointed to by msaEntry.
 */

int computeAuth(unsigned char buffer[], int buflen, 
		 unsigned char authenticator[], int *authenticatorlen,
		 MipSecAssocEntry *msa)
{ 
    MD5_CTX context;

    if (msa == NULL) return 0;

    switch (msa->mipSecAlgorithmType) {
        case MD5:
	    if (*authenticatorlen < 16) {
	        mip2dbg(("Not enough space for MD5 authenticator.\n"));
	        return (-1);
	    }
	    if (msa->mipSecAlgorithmMode != PREFIXSUFFIX) {
	        mip2dbg(("Unknown mode specified for MD5 algorithm.\n"));
		return (-1);
	    }

	    MD5Init(&context);
	    MD5Update(&context, msa->mipSecKey, 
		      (unsigned int) msa->mipSecKeyLen); 
	    MD5Update(&context, buffer, 
		      (unsigned int) buflen);
	    MD5Update(&context, msa->mipSecKey, 
		      (unsigned int) msa->mipSecKeyLen); 
	    MD5Final(authenticator, &context);
	    *authenticatorlen = 16;

	    break;

        case NONE:
	    /* we leave the contents of authenticator unchanged */

	    *authenticatorlen = 16;
	    break;
    }

    return 0;
}


/* 
 * Append authentication extension to a registration reply contained in
 * buffer based on the Mobile IP Security Association pointed
 * to by MSA. Returns total number of bytes in the extension.
 */

int appendAuthExt(unsigned char buffer[], int buflen, int type, 
		  MipSecAssocEntry *msa)
{ 
    int     authlen;
    authExt *aep;
    u32 SPI;

    /* TODO:
     * The length of authenticator actually depends on the algorithm.
     * For now, we assume 16.
     */

    aep = (authExt *) (buffer + buflen);
    aep->type = type;
    aep->length = (4 + 16);
    SPI = (msa == NULL) ? 0 : msa->mipSecSPI;
    aep->SPIhi = htons(SPI >> 16);
    aep->SPIlo = htons(SPI & 0xffff);

    authlen = 16;
    computeAuth(buffer, buflen + sizeof(authExt), 
		buffer + buflen + sizeof(authExt), &authlen, msa);
    
    return(authlen + sizeof(authExt));
}

boolean isAuthOk(unsigned char buffer[], int buflen, 
		 unsigned char authenticator[], int authenticatorlen,
		 MipSecAssocEntry *msa)
{
    static unsigned char newAuth[32];
    int newAuthLen = 32;
    int i;

    switch (msa->mipSecAlgorithmType) {
        case MD5:
	    computeAuth(buffer, buflen, newAuth, &newAuthLen, msa);
	    if (newAuthLen != authenticatorlen)
	        return(FALSE);
	    for (i = 0; i < newAuthLen; i++) {
	        if (newAuth[i] != authenticator[i])
		    return(FALSE);
	    }
	    
	    break;

        case NONE:
	    /* No checks required */
	    break;
    }

    return(TRUE);
}

/* 
 * Buffer contains buflen bytes of a registration request and an
 * immediately following authentication extension. Check if the
 * authentication extension is correct according to the given msa.
 */

boolean isAuthExtOk(unsigned char buffer[], int buflen, MipSecAssocEntry *msa)
{
    int authLen; 
    authExt *aep;
    u32 SPI;

    aep = (authExt *) (buffer + buflen);
    SPI = ntohs(aep->SPIhi) << 16 | ntohs(aep->SPIlo);
    if (SPI != msa->mipSecSPI) {
        mip3dbg(("Type: %d, Length %d\n", aep->type, aep->length));
        mip3dbg(("SPI mismatch got %lx had %lx.\n", SPI, msa->mipSecSPI));
	return (FALSE);
    }
    authLen = aep->length - 4;
    return(isAuthOk(buffer, buflen + sizeof(authExt), 
		    (buffer + buflen + sizeof(authExt)), 
		    authLen, msa));
}

int authmain (int argc, char *argv[])
{
  int i;
  unsigned char digest[16];
  int digestlen = 16;
  MipSecAssocEntry msae = { 0, 0, MD5, PREFIXSUFFIX, 16,
			    { 0x11, 0x11, 0x11, 0x11,
			      0x11, 0x11, 0x11, 0x11,
			      0x11, 0x11, 0x11, 0x11,
			      0x11, 0x11, 0x11, 0x11
			    }, NONCES};

  unsigned char packet[30] = {
    /*  0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11,
  0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, */
  0x01, 0x00, 0x00, 0x28, 0x81, 0x92, 0x7a, 0xcc, 
  0x81, 0x92, 0x7a, 0xbf, 0x81, 0x92, 0xc9, 0x09, 
  0x00, 0x00, 0x00, 0x00, 0xe4, 0xa8, 0x5f, 0xcb, 
  0x20, 0x14, 0x00, 0x00, 0x00, 0x01
  /*  0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11,
  0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11 */
  };
  /*
  0x01, 0x00, 0x01, 0x2c, 0x81, 0x92, 0x7a, 0xc0,
  0x81, 0x92, 0x7a, 0x7b, 0x81, 0x92, 0xc9, 0x09, 
  0x00, 0x00, 0x00, 0x00, 0x17, 0xe1, 0x2c, 0x23, 
  0x20, 0x14, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11,
  0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11,
  0x11, 0x11
  };
  unsigned char packet[] = {'J', 'i', 'm'};
  */
  /*
  printMSAE(&msae);
  computeAuth(packet, 30, digest, &digestlen, &msae);
  printf ("Authenticator for packet = ");
  printBuffer(digest, digestlen);
  printf ("\n");
  */

  /* Simulate data corruption or incorrect MSAE */
/* 
  msae.mipSecKeyLen = 2;
  packet[2] = 0x00;
  msae.mipSecKey[0] = 'A';
  printMSAE(&msae);
  if (isAuthOk(packet, 3, digest, digestlen, &msae))
    printf("Check succeeded.\n");
  */
}



