/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Original Code is the Sunlabs Mobile IP.
 The Initial Developer of the Original Code is Vipul Gupta.
 Portions created by Vipul Gupta are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): Niko A. Fikouras (niko@comnets.uni-bremen.de)
		 Asanga D. Udugama (adu@comnets.uni-bremen.de)

*/

/*
 * An implementation of Mobile-IP as described in RFC 2002.
 * Author: Vipul Gupta (vgupta@eng.sun.com)
 *         Sunil Madhani (smadhani@eng.sun.com)
 * mn.c : Main program for a mobile node.
 */

#pragma ident	"@(#)mn.c 1.3 98/11/30"

#include <stdio.h>

/* The following needed for inet_addr */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <sys/stat.h>
#include <sys/fcntl.h>

#include "mip.h"
#include "mobile.h"

/* Controls verbosity of debug messages when compiled w/ "-D MIP_DEBUG" */
int mipdebug = 3;
int periodicInterval = DEFAULT_PERIODIC_INTERVAL;
int agentSolicitThreshold = AGENT_SOLICIT_THRESHOLD;
int agentExpireThreshold = AGENT_EXPIRE_THRESHOLD;
int ifacePollThreshold = IFACE_POLL_INTERVAL;
int regLifetime = DEFAULT_REG_LIFETIME;
int regRoundTrip = DEFAULT_MAX_ROUNDTRIP;

int regFlags = DEFAULT_DISCRETIONARY_FLAGS;  /* e.g. M, G, V, B, S */

int haCooloffPeriod = HA_COOLOFF_PERIOD;
int faCooloffPeriod = FA_COOLOFF_PERIOD;
int IDfreshnessSlack = DEFAULT_FRESHNESS_SLACK;
char ignoreIfaceList[DEFAULT_MAX_INTERFACES][MAX_IFNAME_LEN];
int ignoreIfaceCnt = 0;
boolean solicitOnBcast = 0;

int movementDetection = ECS; /* modified by Uni. Sheffield */
int HApriority = 1;	/* modified by Uni. of Sheffield */
int EagerSoliciting = 1;/* modified by Uni. of Sheffield */
int LazySoliciting = 1; /* modified by Uni. of Sheffield */

int FilterStatus = 0; // identifies whether filter operations are required or not
/*
 * Use the following multiples of approx round trip time as waiting
 * period between successive retransmissions of registration requests
 * or solicitations (sent in UNKNOWN state).
 */
int retransmissionPolicy[DEFAULT_MAX_ATTEMPTS] = {2, 4, 8, 16, 32, -1};

/*
 * Renew regsitration when remaining lifetime drops to the following levels
 * expressed as fraction of the original granted lifetime.
 */
float renewalPolicy[DEFAULT_MAX_ATTEMPTS] = {0.5, 0.25, 0};

/* used for debug messages */
char msg[256], currentTime[32], relativeTime[32];
char addr1[20], addr2[20], addr3[20], addr4[20];

/* The mobile node's address, netmask and default routers at home */
u32 mnHomeAddr;
u32 mnHomeNetmask;
u32 mnHomeBcastAddr;     /* Set in the init routine, used for HA discovery */
u32 mnSolicitAddr;

// Table that stores the state of each active interface;
// the original mnState currentState is now inside this array
InterfaceState ISTable[DEFAULT_MAX_INTERFACES];
int ISTableSize = 0;

int addrIntervalsCnt;
DomainInfo domainInfo;

/*
 * This table stores information about each network interface available
 * to the mobile node.
 */
IfaceEntry ifaceTable[DEFAULT_MAX_INTERFACES];
int ifaceTableSize = 0;

/* List of home agents */
HomeAgentEntry haTable[DEFAULT_MAX_HOME_AGENTS];
int haTableSize = 0;

/* List of default routers on the home network */
u32 homeRouterTable[MAX_DEFAULT_ROUTERS];
int homeRouterTableSize = 0;

/* List of local Mobility Agents (considered active) */
LocalMaEntry localMaTable[DEFAULT_MAX_MOBILITY_AGENTS];
int localMaTableSize = 0;

/* List of known local Mobility Agents (considered active) */
/* Used by the ECS movement detection method - modified by */
/* Uni. Sheffield					   */
LocalMaEntry *knownLocalMaTable[DEFAULT_MAX_MOBILITY_AGENTS];

/* List of registration requests (both accepted and pending) */
RegReqEntry regReqTable[DEFAULT_MAX_REG_REQ];
int regReqTableSize = 0;
int regReqAcceptedCnt = 0;

/* Other data structures .... */
int timerWentOff = 1;
boolean isReverseTunnelSet = FALSE;

/* Advertisement counters */
u32  mnSolicitationsSentCnt = 0;
u32  mnAdvertisementsRecvdCnt = 0;
u32  mnAdvsDroppedInvalidExtensionCnt = 0; /* dropped due to poorly formed ext
                                              or unrecognized exts in 0-127 */
u32  mnAdvsIgnoredUnknownExtensionCnt = 0; /* ignored due to exts in 128-255 */

/* Registration Counters */
u32  mnRegRequestsSentCnt = 0;
u32  mnDeRegRequestsSentCnt = 0;
u32  mnRegRepliesRecvdCnt = 0;
u32  mnDeRegRepliesRecvdCnt = 0;
u32  mnRepliesInvalidHomeAddressCnt = 0;
u32  mnRepliesUnknownHACnt = 0;
u32  mnRepliesUnknownFACnt = 0;
u32  mnRepliesInvalidIDCnt = 0;
u32  mnRepliesDroppedInvalidExtensionCnt = 0;
u32  mnRepliesIgnoredUnknownExtensionCnt = 0;
u32  mnRepliesHAAuthenticationFailureCnt = 0;
u32  mnRepliesFAAuthenticationFailureCnt = 0;
u32  mnRegRequestsAcceptedCnt = 0;
u32  mnRegRequestsDeniedByHACnt = 0;
u32  mnRegRequestsDeniedByFACnt = 0;
u32  mnRegRequestsDeniedByHADueToIDCnt = 0;
u32  mnRegRequestsWithDirectedBroadcastCnt = 0;

/* Other counters */
u32  mnMoveFromHAToFACnt = 0;
u32  mnMoveFromFAToFACnt = 0;
u32  mnMoveFromFAToHACnt = 0;
u32  mnGratuitousARPsSendCnt = 0;
u32  mnAgentRebootsDectectedCnt = 0;

/*
 * Sockets used for Mobile IP messages. Agent advertisements are not
 * picked up using an ICMP socket. Instead, OS-specific mechanisms are
 * used to grab the entire ethernet frame corresponding to the advertisement.
 */
int  ICMPsolicitSock; /* for sending router solicitations */
int  UDPregSock;      /* UDP socket bound to a random port for sending
		       * reg requests and receiving replies.
		       */

/* to store the filters that have been defined for this mobile node*/
Filter mnFilters[MAX_FILTERS]; 					   // data for each filter
int mnFilterCount = 0; 						   // the number of filters
FilterOp mnFilterOpTable[MAX_FILTERS];				   // the grid that holds the operations
								   // to be done for the filters
int mnFilterOpCount = 0;					   // operations count


/*
 * TODO: Allow a mobile node to support security associations with
 *       nodes other than those acting as its Home Agents. MSA shared
 *       with those nodes can be kept in a table of "MipSecAssocEntry"s.
 */

/* External function definitions */
extern void printBuffer(unsigned char *, int);
extern char *hwAddrWrite(unsigned char *, char *);
extern char *ntoa(u32, char *);
extern char *sprintTime(char *, int);
extern char *sprintRelativeTime(char *, int);
extern unsigned short inChecksum(unsigned short *, int);
extern char *err2str(int);

extern void getAndDispatchMessages();
extern int  sendUDPmessage(int, unsigned char *, int, u32, unsigned short);
extern int  sendICMPmessage(int, u32, u32, unsigned char *, int);
extern int  appendAuthExt(unsigned char *, int, int type, MipSecAssocEntry *);

extern void MNnewID(u32 *, u32 *, u32, u32, int, int);
extern void MNstoreID(u32 *, u32 *, u32, u32, int, int *);
extern boolean MNisIDok(u32, u32, u32, u32, int);
extern boolean isAuthExtOk(unsigned char *, int, MipSecAssocEntry *);

extern RegReqEntry *mkNewRegReq(mnState *, int);
extern boolean canIgnore(u32, unsigned char);
extern boolean isOutsideProtectedDomain(u32);
extern void doPeriodicTask(int);
extern void Initialize(FILE *);
extern void decideStateTransition(mnState *, int);

extern int arpdel(u32);
extern int disableArp(u32, u32);
extern int routedel(u32, u32);
extern void completeColocatedAction(mnState *, RegReqEntry *);

extern int appendFilterExt(unsigned char [], int, Filter[], int, FilterOp[], int *, char *);

/*
 * Examine buflen bytes in buffer and place all the Mobile IP extensions
 * encountered in extType[] and the start points in extIdx[]. The maximum
 * size of these two arrays is given in cnt. At termination, cnt contains
 * the numebr of extensions found in the buffer (which starts at an extension).
 */
int mkRegExtList(unsigned char buffer[], int buflen, int *cnt,
		 unsigned char extType[], unsigned char *extIdx[])
{
    unsigned char *currentPos = buffer;
    int bytesSeen = 0;
    int extSeen = 0;
    int skipOver;

    while ((bytesSeen < (buflen - 1)) && (extSeen < *cnt)) {
	extType[extSeen] = *currentPos;
        extIdx[extSeen] = currentPos;
	extSeen++;
	skipOver = *(currentPos+1) + 2;
	bytesSeen += skipOver;
	currentPos += skipOver;
    }
   
    if ((bytesSeen < (buflen - 1)) && (extSeen >= *cnt)) {
        mip1dbg(("Too many extensions.\n"));
	return(-1);
    } else if (bytesSeen != buflen) {
        mip1dbg(("Extensions buffer too small.\n"));
	return(-1);
    } else {
        *cnt = extSeen;
        return(0);
    }
}

/* 
 * Check if the extension list is correct for a registration reply
 * received at a Mobile Node. We check for
 *      REG_MH_AUTH_EXT_TYPE    (before any M-F ext, exactly one unless
 *				 we have a rejection from an FA)
 *      REG_MF_AUTH_EXT_TYPE    (at most one, must follow an M-H ext whenever
 *				 an M-H extension is present)
 *				 for an FA rejection)
 *      Ext type 0-127          (no unrecognized ones allowed)
 *      Ext type 128-255        (none after M-F ext)
 */
boolean MnRegReplyExtOk(int code, int cnt, unsigned char extType[])
{
   int foundMHauth = 0;
   int foundMFauth = 0;
   int i;

    for (i = 0; i < cnt; i++) {
       if (extType[i] == REG_MH_AUTH_EXT_TYPE) {
	    /* we should not have seen any other auth extensions */
	    if (foundMHauth) {
	        mip1dbg(("Multiple MHauth\n"));
		return (FALSE);
	    } else
	        foundMHauth++;
	} else if (extType[i] == REG_MF_AUTH_EXT_TYPE) {
	    /* 
	     * We should have seen MHauth (unless we have a rejection
             * from an FA) but no MFauth
	     */
	    if ((!foundMHauth) && 
		((code < FA_REASON_UNSPECIFIED)||(code > FA_HA_UNREACHABLE))) {
	        mip1dbg(("No MHauth before MFauth\n"));
		return (FALSE);
	    } 
	    if (foundMFauth) {
	        mip1dbg(("Multiple MFauth\n"));
		return (FALSE);
            } else
	        foundMFauth++;
	} else if (extType[i] <= 127) {
	    /* we should never see an unrecognized ext in range 0-127 */
	    mip1dbg(("Unrecognized ext in range 0-127.\n"));
	    mnRepliesDroppedInvalidExtensionCnt++;
	    return(FALSE);
	} else {
	    /* No unrecognized 128-255 extensions after M-F auth */
	    mnRepliesIgnoredUnknownExtensionCnt++;
	    if (foundMFauth) {
	        mip1dbg(("Ext found after MFauth\n"));
		return(FALSE);
	    }
	}
    }

    /* Make sure we have one MHauth unless it is a denial from an FA */
    if (foundMHauth || ((code >= FA_REASON_UNSPECIFIED) &&
			(code < HA_REASON_UNSPECIFIED)))
       return(TRUE);
    else {
       mip1dbg(("MHauth missing\n"));
       return(FALSE);
    }
}

/*
 * If any PENDING registrations used this mobility agent as destination,
 * delete those requests. If an ACCEPTED_RENEWABLE request used this
 * agent as destination then mark that request as ACCEPTED_NONRENEWABLE.
 */
void delMaPtr(LocalMaEntry localMaTable[], int *localMaTableSize, 
	      LocalMaEntry *lmaep) 
{
    int i;
    int err;

    mip3dbg(("Deleting %s from the list of local mobility agents.\n",
	     ntoa(lmaep->mnAdvSourceAddr, addr1)));
    
    /* 
     * Check for registration entries using this agent and mark them
     * REG_UNUSED or REG_ACCEPTED_NONRENEWABLE.
     */ 
    for (i = 0; i < DEFAULT_MAX_REG_REQ; i++) {
	if (regReqTable[i].mnRegAgentAddr == lmaep->mnAdvSourceAddr) {
	    if (regReqTable[i].mnRegStatus == REG_PENDING) {
		regReqTable[i].mnRegStatus = REG_UNUSED;
		regReqTableSize -= 1;
	    } else if (regReqTable[i].mnRegStatus == REG_ACCEPTED_RENEWABLE)
		regReqTable[i].mnRegStatus = REG_ACCEPTED_NONRENEWABLE;
	}
    }

    /* 
     * Delete ARP entry for this mobility agent if it is not on the home net.
     */
    if (((lmaep->mnAdvSourceAddr ^ mnHomeAddr) & mnHomeNetmask) != 0) {
        if ((err = arpdel(lmaep->mnAdvSourceAddr)) < 0) {
            mip0dbg(("arpdel failed for %s ... %s.\n",
                     ntoa(lmaep->mnAdvSourceAddr, addr1), err2str(err)));
        }
    }    

    lmaep->mnAdvSourceAddr = 0;   /* mark this entry as unused */
    memset(lmaep->mnAdvIncomingIface, (char) NULL, MAX_IFNAME_LEN);
    *localMaTableSize -= 1;
}

/* 
 * Should decrement table size and make sure table is contiguous. Mark
 * accepted entries based on this HA as non-renewable. Valid entries in
 * the table are stored at indices 0 through (haTableSize - 1).
 */
void delHaPtr(HomeAgentEntry haTable[], int *haTableSize, HomeAgentEntry *haep)
{
    HomeAgentEntry *lastHaep;

    if (*haTableSize < 1) {
	mip0dbg(("delHaPtr called with haTableSize less than 1.\n"));
	return;
    }

    mip3dbg(("Deleting %s from the list of home agents.\n",
             ntoa(haep->mnHaAddr, addr1)));

    *haTableSize -= 1;
    lastHaep = &(haTable[*haTableSize]);

    /* copy last one into position being deleted */
    memcpy(haep, lastHaep, sizeof(HomeAgentEntry));
    memset(lastHaep, 0, sizeof(HomeAgentEntry));
}

/*
 * Delete registration entry: clear all fields, mark entry as unused,
 * decrement size of table.
 */
void delRegReqPtr(RegReqEntry regReqTable[], int *regReqTableSize,
		  RegReqEntry *rrep)
{
    mip3dbg(("Deleting registration request sent to %s (COA %s, HA %s).\n",
	     ntoa(rrep->mnRegAgentAddr, addr1),
	     ntoa(rrep->mnRegCOA, addr2),
	     ntoa(rrep->mnRegHAaddr, addr3)));

    memset(rrep, 0, sizeof(RegReqEntry));
    rrep->mnRegStatus = REG_UNUSED;
    *regReqTableSize -= 1;
}

/*
 * Pick up the flags, lifetime, HA and COA values from the reReqEntry pointed
 * by rrep. Choose appropriate ID values and update the values stored in
 * *rrep. Use the HA's security association to add an authentication
 * extension.
 */
void sendRegReq(RegReqEntry *rrep)
{
    int i;
    static unsigned char regPkt[MAX_PKT_SIZE];
    static int regPktLen;
    regRequest *reqPtr;
    MipSecAssocEntry *msap = NULL;

    mip0dbg(("\n---- %s ---- sent registration request \n",
		printTime(currentTime, 30)));


    if (rrep == NULL) {
	mip0dbg(("sendRegReq called with NULL RegReqEntry pointer.\n"));
	return;
    }

    reqPtr = (regRequest *) regPkt;
    reqPtr->type = REG_REQUEST_TYPE;
    reqPtr->regFlags = rrep->mnRegFlags;
    reqPtr->regLifetime = htons(rrep->mnRegTimeRequested);
    reqPtr->homeAddr    = mnHomeAddr;
    reqPtr->haAddr = rrep->mnRegHAaddr;
    reqPtr->COAddr = rrep->mnRegCOA;

    /* Find security association with this HA */
    for (i = 0; i < haTableSize; i++) {
	if (haTable[i].mnHaAddr == reqPtr->haAddr) {
	    msap = &(haTable[i].mnHaMSA);
	    break;
	}
    }

    MNnewID(&(rrep->mnRegIDHigh), &(rrep->mnRegIDLow),
            haTable[i].mnHaLastIDhi, 0,
            haTable[i].mnHaMSA.mipSecReplayMethod, haTable[i].mnHaTimeSkew);

    reqPtr->IDHigh = htonl(rrep->mnRegIDHigh);
    reqPtr->IDLow = htonl(rrep->mnRegIDLow);

    regPktLen = sizeof(regRequest);

    /* Check and append the filter data to the packet */
    if(mnFilterOpCount > 0) {
            regPktLen += appendFilterExt(regPkt, regPktLen, mnFilters, mnFilterCount,
                                   mnFilterOpTable, &mnFilterOpCount, rrep->mnRegIfaceName);

            rrep->mnRegFilterLen = regPktLen - sizeof(regRequest);
            memcpy(rrep->mnRegFilter, regPkt + sizeof(regRequest), rrep->mnRegFilterLen);

    } else if ((rrep->mnRegFilterLen > 0) && (rrep->mnRegStatus == REG_PENDING)) {
                memcpy(regPkt + regPktLen, rrep->mnRegFilter, rrep->mnRegFilterLen);
                regPktLen += rrep->mnRegFilterLen;
	    } else  rrep->mnRegFilterLen = 0;


    if (msap != NULL) {
        regPktLen += appendAuthExt(regPkt, regPktLen,
				   REG_MH_AUTH_EXT_TYPE, msap);

	/* Assumption: The ARP entry for destination is already installed. */

	if (sendUDPmessage(UDPregSock, regPkt, regPktLen,
			   rrep->mnRegAgentAddr, MIP_PORT) == 0) {

// originally message print remarked
	    mip2dbg(("\n---- %s (%s) ----\n", sprintTime(currentTime, 30),
		     sprintRelativeTime(relativeTime, 30)));
	    mip2dbg(("MN sent %s request to %s\n",
		     (rrep->mnRegHAaddr == mnHomeBcastAddr) ?
		     "discovery registration" : "registration",
		     ntoa(rrep->mnRegAgentAddr, addr1)));
            mip2dbg(("                [MN %s, HA %s, COA %s]\n",
		     ntoa(reqPtr->homeAddr, addr1),
		     ntoa(reqPtr->haAddr, addr2),
		     ntoa(reqPtr->COAddr, addr3)));
	    mip2dbg(("                [Lifetime %d sec, ID %0#10x : %0#10x]\n",
		     (u32) ntohs(reqPtr->regLifetime), ntohl(reqPtr->IDHigh),
		     ntohl(reqPtr->IDLow)));

	    mip3dbg(("Registration sent by MN is:\n"));
	    if (mipdebug > 2) printBuffer(regPkt, regPktLen);
	    mip3dbg(("\n"));
// originally message print remarked


            if (reqPtr->regLifetime == 0)
	        mnDeRegRequestsSentCnt++;
	    else
	        mnRegRequestsSentCnt++;

	} else {
	    mip0dbg(("sendUDPmessage failed in sendRegReq().\n"));
 	}
    } else {
	mip0dbg(("Did not find security association for HA %s.\n",
		 ntoa(reqPtr->haAddr, addr1)));
    }
}

/* Send out an agent solicitation from the specified address */
void sendSolicitation(u32 srcIfaceAddr)
{
    int len;
    icmph *icmpPtr;
    static unsigned char solicitPkt[64];

    if (srcIfaceAddr == 0) {
	mip0dbg(("Bad argument passed to sendSolicitation.\n"));
	return;
    }

    /* Fill out the ICMP header for a router solicitation */
    icmpPtr = (icmph *) solicitPkt;
    icmpPtr->type = ICMP_ROUTER_SOLICITATION;
    icmpPtr->code = 0;
    icmpPtr->checksum = 0;
    icmpPtr->icmphdr_u.u_unused = 0;
    len = sizeof(icmph);
    icmpPtr->checksum = inChecksum((unsigned short *) icmpPtr, len);

    mnSolicitationsSentCnt++;
    if (sendICMPmessage(ICMPsolicitSock, srcIfaceAddr,
			(solicitOnBcast ? inet_addr(LINK_BCAST_ADDR) :
			                  inet_addr(LINK_MCAST_ADDR)), 
			solicitPkt, len) == 0)  {
        mip2dbg(("\n---- %s (%s) ----\n", sprintTime(currentTime, 30),
		 sprintRelativeTime(relativeTime, 30)));
        mip2dbg(("Sent agent solicitation through iface address %s.\n",
		 ntoa(srcIfaceAddr, addr1)));
    } else  {
        mip0dbg(("sendICMPmessage failed in sendSolicitation().\n"));
    }
}

/* Find a LocalMaEntry in which mnAdvSourceAddr matches src */
LocalMaEntry * findLocalMA(u32 src)
{
    int i;
    LocalMaEntry *lmaep = NULL;

    for (i = 0; i < DEFAULT_MAX_MOBILITY_AGENTS; i++) {
	if (localMaTable[i].mnAdvSourceAddr == src) {
	    lmaep = &(localMaTable[i]);
	    break;
	}
    }
    
    return(lmaep);
}

/* Find a LocalMaEntry in knownLocalMaTable for which mnAdvSourceAddr matches */
/* src */
int findKnownLocalMA(u32 src)
{
    int i;
    
    for (i = 0; i < DEFAULT_MAX_MOBILITY_AGENTS; i++) {
	if (knownLocalMaTable[i] == NULL) continue;
	if (knownLocalMaTable[i]->mnAdvSourceAddr == src) return i;
    }
    return -1;
    
}

/* 
 * Create a LocalMaEntry for a newly discovered mobility agent. Clear
 * the unusabletill and whyunusable fields.
 */
LocalMaEntry * newLocalMA(u32 src)
{
    int i;
    LocalMaEntry *lmaep = NULL;
    
    /* ARP entries are added as part of state transition. */
    for (i = 0; i < DEFAULT_MAX_MOBILITY_AGENTS; i++) {
	if (localMaTable[i].mnAdvSourceAddr == 0) {
	    lmaep = &(localMaTable[i]);
	    break;
	}
    }
    
    if (lmaep != NULL) {
	memset(lmaep, 0, sizeof(LocalMaEntry));
    }

    localMaTableSize++;
    return(lmaep);
}

/* Identify a pointer in knownLocalMaTable that points to NULL and return */
/* its position. */
int newKnownLocalMA()
{
    int i;

    for (i = 0; i < DEFAULT_MAX_MOBILITY_AGENTS; i++) {
	if (knownLocalMaTable[i] == NULL) return i;
    }

    return(-1);
}

/* 
 * Decide if an agent advertisement can be ignored based on its source
 * and advertised service flags.
 */
boolean canIgnoreAdv(u32 src, char flags)
{
    boolean isForeign;
    
    /* Is this mobility agent on our Home network or our foreign network? */
    if (((src ^ mnHomeAddr) & mnHomeNetmask) == 0)
        isForeign = FALSE;
    else
        isForeign = TRUE;
    
    /* 
     * Ignore a foreign mobility agent if F bit is not set or both F and
     * B bits are set.
     */ 
    if (isForeign && 
	(((flags & ADV_IS_FOREIGN_AGENT) == 0) ||
	 ((flags & ADV_IS_BUSY) != 0))) {
        mip4dbg(("Agent on foreign net is unusable as a Foreign Agent.\n"));
        return(TRUE);
    }
    
    /* 
     * NOTE: For now, we ignore FAs outside the protected domain since
     * they can't be used for a mobile VPN. We could print a warning
     * while doing so but it could get very chatty.
     */

    //if (isOutsideProtectedDomain(src)) {
    //    mip4dbg(("Agent is outside the mobile node's protected domain.\n"));
    //    return(TRUE);
    //}

    /* TODO: Do we need to check the H bit for MAs on home network? */

    return(FALSE);
}

/*
 * MN is currently registered through a COA in the advertisement pointed by
 * lmaep. Mark all existing registrations through the COA as non-renewable.
 * Also create a new, pending registartion entry.
 */
RegReqEntry * mkRegForReboot(LocalMaEntry *lmaep, int ISSeq)
{
    int i;
    RegReqEntry *rrep;

    for (i = 0; i < DEFAULT_MAX_REG_REQ; i++) {
	if (regReqTable[i].mnRegAgentAddr != lmaep->mnAdvSourceAddr) continue;
	if (regReqTable[i].mnRegStatus == REG_PENDING)
	    delRegReqPtr(regReqTable, &regReqTableSize, &(regReqTable[i]));
	else if (regReqTable[i].mnRegStatus == REG_ACCEPTED_RENEWABLE)
	    regReqTable[i].mnRegStatus = REG_ACCEPTED_NONRENEWABLE;
    }

    rrep = mkNewRegReq(&(ISTable[ISSeq].currentState), ISSeq);
    return(rrep);
}

/* Process an agent advertisement received at a mobile node */
void MNprocessAgentAdv(unsigned char *advPkt, int pktLen, char *ifaceName,
		       char *srcHWaddr)
{
    int i, ipHdrLen, PFAIPListLen, ISSeq;
    int advExtOffset;
    int advExtensionsPtr;
    unsigned char ipByte;
    u32 *ipSrcp;
    u32 *addrp;
    icmph *icmpPtr;
    advExt *advExtp;
    RegReqEntry *rrep;
    LocalMaEntry *lmaep;
    u16 prevSeqNum;

    // get the state variables related to the interface
    // being considered
    ISSeq = getIfcStateSeqFromName(ifaceName, ISTable);

    /*
     * We already know that the pkt is a valid agent advertisement
     * It includes the IP header as well.
     */
    mnAdvertisementsRecvdCnt++;

    ipByte = advPkt[0];
    ipHdrLen = (ipByte & 0x0f) << 2;
    ipSrcp = (u32 *) (advPkt + 12);

    mip5dbg(("MNprocessAgentAdv called for advertisement from %s on %s.\n",
	     ntoa(*ipSrcp, addr1), ifaceName));

    /* Look at the ICMP header, we assume most checks are already complete */
    icmpPtr = (icmph *) (advPkt + ipHdrLen);
    advExtOffset = ipHdrLen + sizeof(icmph) +
                   (icmpPtr->icmpAdvNumAddr *
		    ((icmpPtr->icmpAdvAddrEntrySize) << 2));
    advExtp = (advExt *) (advPkt + advExtOffset);
    /* Should this advertisement be ignored, e.g. busy FA */
    if (canIgnoreAdv(*ipSrcp, advExtp->advFlags)) {
	mip0dbg(("Advertisement ignored.\n"));
	return;
    }

    /* Search localMAtable for an existing entry for this source */
    if ((lmaep = findLocalMA(*ipSrcp)) == NULL) {
	/* need to create a new entry in localMaTable */
	if ((lmaep = newLocalMA(*ipSrcp)) == NULL) {
	    mip0dbg(("No space for new local MA %s.\n",
		     ntoa(*ipSrcp, addr1)));
	    return;
	}
    }

    /*
     * By this time, lmaep points to an entry in localMaTable with
     * mnAdvSourceAddr set to *ipSrcp. Save the previous sequence number
     * and overwite the entry with new information in the advertisemt.
     * No need to overwrite mnAdvSourceAddr. If it is new then you
     * need to add it.
     */
    lmaep->mnAdvSourceAddr = *ipSrcp;
    prevSeqNum = lmaep->mnAdvSeqNum;
    memcpy(lmaep->mnAdvSrcHWaddr, srcHWaddr, MAX_HWADDR_LEN);
    strncpy(lmaep->mnAdvIncomingIface, ifaceName, MAX_IFNAME_LEN);
    lmaep->mnAdvMaxAdvLifetime = icmpPtr->icmpAdvLifetime;
    lmaep->mnAdvMaxRegLifetime = advExtp->regLifetime;
    lmaep->mnAdvRemainingLifetime = lmaep->mnAdvMaxAdvLifetime;
    lmaep->mnAdvSeqNum = advExtp->seqNum;
    lmaep->mnAdvLastSeqNum = prevSeqNum;
    lmaep->mnAdvFlags = advExtp->advFlags;
    lmaep->mnAdvRouterCnt = icmpPtr->icmpAdvNumAddr;
    lmaep->mnAdvCOACnt = (advExtp->length - 6)/4;
    if (lmaep->mnAdvRouterCnt > MAX_ROUTERS_IN_ADV) {
	mip0dbg(("Found %d routers in router advertisement, will only " \
                 "consider the first %d.\n",lmaep->mnAdvRouterCnt,
		 MAX_ROUTERS_IN_ADV));
	lmaep->mnAdvRouterCnt = MAX_ROUTERS_IN_ADV;
    }

    if (lmaep->mnAdvCOACnt > MAX_COA_IN_ADV) {
	mip0dbg(("Found %d care-of addresses in router advertisement, will " \
                 "only consider the first %d.\n",
		  lmaep->mnAdvCOACnt, MAX_COA_IN_ADV));
	lmaep->mnAdvCOACnt = MAX_COA_IN_ADV;
    }


    /* Compile a list of advertised routers. */
    addrp = (u32 *) (advPkt + ipHdrLen + sizeof(icmph));
    for (i = 0; i < lmaep->mnAdvRouterCnt; i++) {
	lmaep->mnAdvRouterList[i] = *addrp;
	addrp ++;
    }


    /*
     * Compile a list of advertised care-of addresses. We assume
     * that there is no need to change COAunusableTill and
     * COAwhyUnusable fields unless the agent recently rebooted
     * or it is a new one.
     */

    addrp = (u32 *) (advPkt + advExtOffset + sizeof(advExt));
    for (i = 0; i < lmaep->mnAdvCOACnt; i++) {
	lmaep->mnAdvCOAList[i] = *addrp;
	addrp ++;
    }

    advExtensionsPtr = (int) addrp - (int) advPkt;


    while (pktLen > advExtensionsPtr) {

        switch (advPkt[advExtensionsPtr]) {

            case 0: advExtensionsPtr = pktLen + 1;
                    break;

            case 19: advExtensionsPtr += 2;
                     for (i = 0; i<lmaep->mnAdvRouterCnt; i++) {
                         lmaep->mnAdvPrefixList[i] = (u32) ntohl(~((u32) ~0 >> advPkt[advExtensionsPtr]));
                         advExtensionsPtr ++;
                     }
                     break;

            case 20: PFAIPListLen = advPkt[(advExtensionsPtr++)];
                     addrp = (u32 *) &(advPkt[(advExtensionsPtr++)]);
                     for (i = 0; i<PFAIPListLen; i++) {
                         lmaep->mnAdvPFAIPList[i] = *addrp;
                         addrp ++;
                     }
                     break;
        }
    }

    if ((lmaep->mnAdvSeqNum < lmaep->mnAdvLastSeqNum) &&
	(lmaep->mnAdvSeqNum < 256)) {
        /*
         * We assume that the list of COAs advertised can only change on a
         * reboot. So this is a good time to initialize the COAunusableTill
         * and COAwhyUnusable fields.
         */
        for (i = 0; i < MAX_COA_IN_ADV; i++) {
            lmaep->mnAdvCOAunusableTill[i] = 0;
            lmaep->mnAdvCOAwhyUnusable[i] = 0;
        }

	/*
	 * Detected a reboot. If MN is registered through this
	 * mobility agent, a fresh registration must be sent.
	 */
	if ((ISTable[ISSeq].currentState.state == FOREIGN_STATE) &&
	    (ISTable[ISSeq].currentState.routerAddr == lmaep->mnAdvSourceAddr)) {
	    rrep = mkRegForReboot(lmaep, ISSeq);
	    sendRegReq(rrep);
	}
    }


    /* If advertisement came from an HA in haTable, mark it as being alive */
    if ((((*ipSrcp ^ mnHomeAddr) & mnHomeNetmask) == 0) &&
	(lmaep->mnAdvFlags & ADV_IS_HOME_AGENT)) {
	for (i = 0; i < haTableSize; i++) {
	    if (haTable[i].mnHaAddr == *ipSrcp) {
		haTable[i].mnHAflags |= HA_IS_ALIVE;
		haTable[i].mnHALifetime = DEFAULT_HA_LIFETIME;
		break;
	    }
	}
    }
}


/*
 * Find a registration request that matches a reply. Renewal requests are
 * marked accepted/renewable.
 */
RegReqEntry * findMatchingReq(regReply *repPtr)
{
    int i;

    /* The HA address in reply must belong to MN's home network. */
    if ((repPtr->homeAddr ^ mnHomeAddr) & mnHomeNetmask) return(NULL);

    for (i = 0; i < DEFAULT_MAX_REG_REQ; i++) {
        /*
         * Match the entry based on the lower 32 bits of the replay
         * protection field. Also make sure that HA_UNKNOWN_HOME_AGENT
         * is returned if and only if the corresponding request was
         * for dynamic HA discovery.
         */
        if (((regReqTable[i].mnRegStatus == REG_PENDING) ||
	     (regReqTable[i].mnRegStatus == REG_ACCEPTED_RENEWABLE)) &&
            (regReqTable[i].mnRegIDLow == ntohl(repPtr->IDLow)) &&
            (((regReqTable[i].mnRegHAaddr == mnHomeBcastAddr) &&
              (repPtr->code == HA_UNKNOWN_HOME_AGENT)) ||
             ((regReqTable[i].mnRegHAaddr != mnHomeBcastAddr) &&
              (repPtr->code != HA_UNKNOWN_HOME_AGENT))))
            return(&(regReqTable[i]));
    }

    return(NULL);
}

/* Return the haTable index of HA entry that matches given address. */
int findMatchingHAIdx(u32 haAddr)
{
    int i;

    /* Assumption: if HA table has a default entry, it is at index zero. */
    for (i = 0; i < haTableSize; i++) {
        if (haAddr == haTable[i].mnHaAddr) return(i);
    }

    if ((haTable[0].mnHaAddr == mnHomeBcastAddr) &&
        (((haAddr ^ mnHomeAddr) & mnHomeNetmask) == 0))
        return(0);
    else
        return(-1);
}

/*
 * Convert the registration request pointed to by rrep to accepted.
 * No new entry is created. Instead, the space for the existing entry is
 * reused.
 */
void mkAcceptedReg(RegReqEntry *rrep, unsigned short lifetime, int ISSeq)
{
    unsigned short delay;
    int i;

    /*
     * For colocated state, additional network configuration
     * is required after a registration request is accepted.
     * NOTE: This could be moved to processRegReply if desired.
     */
    if ((ISTable[ISSeq].currentState.state == COLOCATED_STATE) &&
	(rrep->mnRegStatus == REG_PENDING) &&
	(ISTable[ISSeq].currentState.COAddr == rrep->mnRegCOA)) {
        completeColocatedAction(&(ISTable[ISSeq].currentState), rrep);
    }

    /*
     * Compute the delay between getting this reply and the
     * corresponding request.
     */

    delay = 0;
    if (rrep->mnRegStatus == REG_PENDING) {
        /*
	 * Now account for any retransmissions. We assume that the
	 * reply always corresponds to the last transmission of a request.
	 * This is a reasonable assumption when timestamps are in use
	 * since even retransmissions use a new low ID value which
	 * must be matched by the reply before this code is executed.
	 */
        delay = (rrep->mnRegTimeRequested - rrep->mnRegTimeRemaining);
	i = 0;
        while (i < rrep->mnRegAttemptIdx) {
	    delay -= (retransmissionPolicy[i] * regRoundTrip);
	    i++;
	}
    } else if (rrep->mnRegStatus == REG_ACCEPTED_RENEWABLE) {
        /*
	 * The rrep->mnRegAttemptIdx field must already be non-zero.
	 * For lack of any better information, we assume that the
	 * renewal requests are sent relative to granted time as stored
	 * in rrep->mnRegTimeRequested.
	 */
        if ((rrep->mnRegAttemptIdx > 0) &&
	    (renewalPolicy[rrep->mnRegAttemptIdx - 1] > 0.000001)) {
	    delay = (renewalPolicy[rrep->mnRegAttemptIdx - 1] *
		     rrep->mnRegTimeRequested) - rrep->mnRegTimeRemaining;
	}
    }

    rrep->mnRegTimeRequested = lifetime;
    rrep->mnRegTimeRemaining = (u32)(lifetime - delay);
    rrep->mnRegAttemptIdx = 0;
    rrep->mnRegNextTime = rrep->mnRegTimeRemaining -
                          (rrep->mnRegTimeRemaining * renewalPolicy[0]);
    rrep->mnRegStatus = REG_ACCEPTED_RENEWABLE;
    mip3dbg(("Marking reg request as accepted (HA %s, COA %s, " \
             "time %d sec).\n", ntoa(rrep->mnRegHAaddr, addr1),
             ntoa(rrep->mnRegCOA, addr2), rrep->mnRegTimeRemaining));

	mip0dbg(("\n---- %s ---- received registration reply\n",
		printTime(currentTime, 30)));
}

/*
 * Repair lifetime/flags.
 * TODO: We can do better, e.g. try to reset offending flags one
 * by one instead of immediately reseting all discretionary flags.
 */
void repairRegReq(RegReqEntry *rrep, regReply *regRepPtr, int ISSeq)
{
    switch (regRepPtr->code) {
    case FA_REG_LIFETIME_TOO_LONG:
        rrep->mnRegTimeRequested = (u32) ntohs(regRepPtr->regLifetime);
	mip3dbg(("Repaired reg request (new lifetime %d).\n",
		 rrep->mnRegTimeRequested));
	break;
    case HA_ID_MISMATCH:
        mip3dbg(("ID mismatch - will retry. High ID Sent : %0#10x "\
		 " High ID Received : %0#10x",
		 rrep->mnRegIDHigh, ntohl(regRepPtr->IDHigh)));
	break;
    default:
        rrep->mnRegFlags = (ISTable[ISSeq].currentState.state == COLOCATED_STATE) ?
	                   REG_DECAPSULATION_BY_MN : 0;
	mip3dbg(("Repaired registration request sent to %s " \
		 "(New flags 0x%x, lifetime %d sec).\n",
		 ntoa(rrep->mnRegAgentAddr, addr1), rrep->mnRegFlags,
		 rrep->mnRegTimeRequested));
    }

    /*
     * Note that we directly call sendRegReq() after repairRegReq().
     * Hence, we must also make sure that the TimeRemaining field
     * is appropriately updated.
     */

    if (rrep->mnRegStatus == REG_PENDING) {
        rrep->mnRegTimeRemaining = rrep->mnRegTimeRequested;
	rrep->mnRegAttemptIdx = 0;
        rrep->mnRegNextTime = retransmissionPolicy[0]*regRoundTrip;
    } else {
        mip2dbg(("Had to repair registration request corresponding to " \
		 "a renewal ... surprising!\n"));
    }
}


/* Process a registration reply. */
void MNprocessRegReply(unsigned char *pkt, int pktLen, u32 src,
                       unsigned short srcPort)
{
    int i;
    int extCnt;
    int haIdx;
    unsigned char *extp;
    unsigned char extType[MAX_EXPECTED_EXTENSIONS];
    unsigned char *extIdx[MAX_EXPECTED_EXTENSIONS];
    regReply *repPtr = (regReply *) pkt;
    RegReqEntry *rrep;
    int ISSeq;

    mip0dbg(("\n---- %s (%s) ----\n", sprintTime(currentTime, 30),
             sprintRelativeTime(relativeTime, 30)));
    mip2dbg(("MN got reg reply from %s\n", ntoa(src, addr1)));
    mip2dbg(("                 [MN %s, HA %s, Code %d]\n",
             ntoa(repPtr->homeAddr, addr1), ntoa(repPtr->haAddr, addr2),
             repPtr->code));
    mip2dbg(("                 [Lifetime %d sec, ID %0#10x : %0#10x]\n",
             (u32) ntohs(repPtr->regLifetime), ntohl(repPtr->IDHigh),
             ntohl(repPtr->IDLow)));

    mip3dbg(("MNprocessRegReply called for pkt:\n"));
    if (mipdebug > 2) printBuffer(pkt, pktLen);
    mip3dbg(("\n"));

    /* Reject reply with broadcast HA */
    if (repPtr->haAddr == mnHomeBcastAddr) {
        mip1dbg(("Ignoring reply with broadcast HA %s\n",
		 ntoa(repPtr->haAddr, addr1)));
	return;
    }

    /* Find a matching pending request. */
    if ((rrep = findMatchingReq(repPtr)) == NULL) {
        mip1dbg(("Got unexpected registration reply from %s ... dropped.\n",
                 ntoa(src, addr1)));
        return;
    }

    /* get the sequence of the state table relating to this reply*/
    ISSeq = getIfcStateSeqFromName(rrep->mnRegIfaceName, ISTable);


    /* Extract extensions and check for their correctness. */
    extCnt = MAX_EXPECTED_EXTENSIONS;
    extp = (pkt + sizeof(regReply));
    if (mkRegExtList(extp, (pktLen - sizeof(regReply)),
                     &extCnt, extType, extIdx) < 0) {
        mip1dbg(("Problems when extracting extensions in reply!\n"));
        goto done;
    }

    mip3dbg(("Found %d extensions:", extCnt));
    for (i = 0; i < extCnt; i++)
        mip3dbg((" <%d, %d, ...>", extType[i], *(extIdx[i]+1)));
    mip3dbg(("\n"));

    if ((MnRegReplyExtOk(repPtr->code, extCnt, extType) == FALSE)) {
        mip1dbg(("Poorly formed extensions in reply!\n"));
        goto done;
    };

    /*
     * If reply is from an HA, find the HA in haTable, check authentication,
     * update replay data and mark it alive.
     */
    if ((repPtr->code < FA_REASON_UNSPECIFIED) ||
        (repPtr->code >= HA_REASON_UNSPECIFIED)) {
        if ((haIdx = findMatchingHAIdx(repPtr->haAddr)) < 0) {
            mip1dbg(("Reply received from unknown HA %s!\n",
                     ntoa(repPtr->haAddr, addr1)));
            goto done;
        }

        if (repPtr->code != HA_MN_AUTH_FAILURE) {
            /* Expect a valid HA-MN authenticator */
            for (i = 0; i < extCnt; i++) {
                if (extType[i] == REG_MH_AUTH_EXT_TYPE) break;
            }
            if (!isAuthExtOk(pkt, (extIdx[i]-pkt), &(haTable[haIdx].mnHaMSA)))
	    {
                mnRepliesHAAuthenticationFailureCnt++;
                mip1dbg(("Bad MN-HA authenticator ... reply dropped!\n"));
                goto done;
            } else {
                /* Mark HA alive and update replay information */
                haTable[haIdx].mnHAflags |= HA_IS_ALIVE;
                haTable[haIdx].mnHALifetime = DEFAULT_HA_LIFETIME;
		MNstoreID(&(haTable[haIdx].mnHaLastIDhi),
			  (u32 *)NULL, ntohl(repPtr->IDHigh),
			  ntohl(repPtr->IDLow),
			  haTable[haIdx].mnHaMSA.mipSecReplayMethod,
	                  &(haTable[haIdx].mnHaTimeSkew));
            }
        }
    }
    /*
     * TODO: If a Mobile Node Foreign agent authentication extension
     * exists check it. For now, we don't expect one and if we find one
     * we consider it an authentication failure.
     */
    for (i = 0; i < extCnt; i++) {
        if (extType[i] == REG_MF_AUTH_EXT_TYPE) {
            mnRepliesFAAuthenticationFailureCnt++;
            mip3dbg(("Unexpected MN-FA authenticator ... reply dropped!\n"));
            goto done;
        }
    }

    /* It is a good reply, process it */
    switch (repPtr->code) {
    case 0:
    case 1:
        /*
         * If the HA purged all other bindings for the MN, the MN should
         * also delete all previous bindings stored by this HA.
         */
        for (i = 0; i < DEFAULT_MAX_REG_REQ; i++) {
	    if (((regReqTable[i].mnRegStatus != REG_ACCEPTED_RENEWABLE) &&
		(regReqTable[i].mnRegStatus != REG_ACCEPTED_NONRENEWABLE)) ||
		(regReqTable[i].mnRegHAaddr != repPtr->haAddr) ||
		(regReqTable[i].mnRegIDLow == ntohl(repPtr->IDLow)))
	        continue;

            /*
	     * We have an accepted entry for the same HA that replied
	     * and it is not the request for which we have a reply.
	     *
	     * If code is 1 or we did not set S bit in request, delete it.
	     */
	    if ((repPtr->code == 1) ||
		((rrep->mnRegFlags & REG_SIMULTANEOUS_BINDINGS) == 0)) {
                    delRegReqPtr(regReqTable, &regReqTableSize,
                                 &(regReqTable[i]));
	    } else {
	        if ((regReqTable[i].mnRegCOA == rrep->mnRegCOA)
		     && (strcmp(regReqTable[i].mnRegIfaceName, ISTable[ISSeq].ifaceName) == 0)) {
		    delRegReqPtr(regReqTable, &regReqTableSize,
                                 &(regReqTable[i]));
                }
            }
        }

        if (repPtr->regLifetime == 0) {
            mnDeRegRepliesRecvdCnt++;
        } else {
            mnRegRepliesRecvdCnt++;
            mkAcceptedReg(rrep, ntohs(repPtr->regLifetime), ISSeq);
            return;    /* No need to free the previous request */
        }
        break;

    /* Blame these errors on the FA */
    case FA_REASON_UNSPECIFIED:
    case FA_ADM_PROHIBITED:
    case FA_INSUFFICIENT_RESOURCES:
    case FA_MN_AUTH_FAILURE:
    case FA_HA_AUTH_FAILURE:
        /*
         * TODO: Find the FA and update the sofstate. We don't expect a
         * reply from an agent that is already marked unusable ... log it if
         * this happens. Remove pending request (let sanity checks worry about
         * initiating a new request after some time or do that when the FA
         * becomes usable again in ageMaTable).
         maybe we could do
         */
        break;

    /* Blame these errors on the HA */
    case FA_POORLY_FORMED_REPLY:
    case FA_HA_UNREACHABLE:
    case HA_REASON_UNSPECIFIED:
    case HA_ADM_PROHIBITED:
    case HA_INSUFFICIENT_RESOURCES:
    case HA_MN_AUTH_FAILURE:

        /*
         * TODO: Find the HA and update sofstate. We don't expect a
         * reply from an agent that is already marked unusable ... log it if
         * this happens. Remove pending request (let someone else worry about
         * initiating a new request later).
         */
        break;

    /* These errors can be easily corrected (we hope) */
    case HA_ID_MISMATCH:
    case FA_REG_LIFETIME_TOO_LONG:
    case FA_POORLY_FORMED_REQUEST:
    case FA_ENCAP_UNAVAILABLE:
    case FA_VJ_UNAVAILABLE:
    case HA_POORLY_FORMED_REQUEST:
    case HA_TOO_MANY_SIMULTANEOUS:

        repairRegReq(rrep, repPtr, ISSeq);
        sendRegReq(rrep);
        return;   /* No need to free the previous request */

    /*
     * Response to a dynamic home agent discovery message
     * TODO: Make sure we do the right thing when a discovery request is
     * about to expire due to aging.
     */
    case HA_UNKNOWN_HOME_AGENT:
        /*
         * haIdx still points to matching HA entry. We have already done all
         * we need to handle HAs that match the non-broadcast entry in
         * our HA table.
         */
        if (haTable[haIdx].mnHaAddr == mnHomeBcastAddr) {
            if (haTableSize == DEFAULT_MAX_HOME_AGENTS) {

                mip1dbg(("No space for dynamically discovered HA %s in HA " \
                         "table.\n", ntoa(repPtr->haAddr, addr1)));
            } else {
                memcpy(&(haTable[haTableSize]), &(haTable[haIdx]),
                       sizeof(HomeAgentEntry));
                haTable[haTableSize].mnHaAddr = repPtr->haAddr;
                haTable[haTableSize].mnHaUnusableTill = 0;
                haTable[haTableSize].mnHaWhyUnusable = 0;
                haTable[haTableSize].mnHAflags = (HA_IS_ALIVE |
                                                  HA_IS_DYNAMICALLY_DISCOVERED);
                haTable[haTableSize].mnHALifetime = DEFAULT_HA_LIFETIME;
		MNstoreID(&(haTable[haTableSize].mnHaLastIDhi),
			  (u32 *) NULL, ntohl(repPtr->IDHigh),
			  ntohl(repPtr->IDLow),
			  haTable[haTableSize].mnHaMSA.mipSecReplayMethod,
	                  &(haTable[haTableSize].mnHaTimeSkew));
                haTableSize++;
            }
        }
        break;
    }
done:
    /*
     * Remove the pending request, unless it is a discovery request.
     * For discovery requests, simply set the remaining lifetime to
     * a small value ... the assumption being other HAs, if available,
     * will respond within a short while of the first one.
     * NOTE: for now, we arbitraily choose this number as 10 seconds.
     */
    if (rrep->mnRegHAaddr != mnHomeBcastAddr)
        delRegReqPtr(regReqTable, &regReqTableSize, rrep);
    else
        rrep->mnRegTimeRemaining = 10;

    mip2dbg(("\n\n"));
}

int main(int argc, char *argv[])
{
    struct timeval tv;
    FILE *fd;

    int SSIDfd;
    int testcount = 0;
    int ISSeq = -1;

    timerWentOff = 0;
    mip0dbg(("\nInitiating mobile node operation [%s] ... \n", MIPMN_VERSION));
    mip2dbg(("\n---- %s (%s) ----\n", sprintTime(currentTime, 30),
	     sprintRelativeTime(relativeTime, 30)));

    if ((fd = fopen(CONF_FILE_NAME, "r")) == NULL) {
	fprintf(stderr, "Could not open %s\n", CONF_FILE_NAME);
	exit(-1);
    }

    /* Read configuration information. Initialize ifaceTable, initialize
     * sockets, initialize ip0, set localMaTable and regReqTable to zero,
     * set timerwentoff to FALSE, initialize haTable with template entries.
     * Set state to unknown. Send out a solicitation on each iface.
     * Set solicitation related values correctly in unknown state e.g.
     * timeForNext solicitation and the index. Initialize signal
     * handler.
     *
     * Also, read thru the filter data file and create the filter list
     */
    Initialize(fd);

    while (1) {
        /*
	 * This sleep added for Linux. Couldn't use sleep(1) as it has an
	 * adverse interaction with alarm.
	 * TODO: Linux should define its
	 * own fixed sleep. Sleeping now is good because it relinquishes
	 * control of the CPU instead of staying in a tight loop.
	 */
      /*
       tv.tv_sec = 0; tv.tv_usec = 1000 ;
       select(FD_SETSIZE, NULL, NULL, NULL, &tv);
       */

        /*
	 * By sleeping, we also give agents a chance to respond to our
	 * solicitation.
	 */
        fflush(stdout);
        getAndDispatchMessages();

	// set the state of the next interface - only one interface
	// is considered in one iteration
	ISSeq = getNextIfcStateSeq(ISSeq, ISTable);
	if (movementDetection < HCS) {
		decideStateTransition(&(ISTable[ISSeq].currentState), ISSeq);
	}

        if (movementDetection == ECS) {
		updateKnownLocalMaTable();
	}
        getNewIfaceList(ifaceTable, &ifaceTableSize);

 	if (timerWentOff) {
	    doPeriodicTask(periodicInterval);
	    timerWentOff = 0;
/*
		testcount++;
		if(testcount == 10) {
			mip0dbg(("\n---- %s ---- make hand-off to 192.168.10.1\n",
				printTime(currentTime, 30)));
			SSIDfd = open("/proc/aironet/eth0/SSID", O_WRONLY, 0);
			write(SSIDfd, "192.168.10.1 00:01:02:B3:E6:E8", 30);
			close(SSIDfd);
		}
		if(testcount == 20) {
			mip0dbg(("\n---- %s ---- make hand-off to 192.168.9.1\n",
				printTime(currentTime, 30)));
			SSIDfd = open("/proc/aironet/eth0/SSID", O_WRONLY, 0);
			write(SSIDfd, "192.168.9.1 00:01:02:B3:E6:ED", 29);
			close(SSIDfd);
			testcount = 0;
		}
*/
	}
	fflush(stdout);
    }
    return(0);
}

/* Invariants at the end of each iteration in the loop
 *   ifaceTable   - contains our view of all the non-loopback
 *                  non-mip ifaces that are up;
 *                  (updated by periodic calls to OS-specific
 *		    routines)
 *   localMaTable - contains latest properly formed advertisements
 *                  from all mobility agents on all of our up
 *                  interfaces (non-loopback, non MIP). Entry
 *                  is added when a new agent sends us an agent
 *                  advertisement (we keep busy ones too since
 *                  they may have accepted a previous reg that
 *                  will still be honored). Entry is deleted if
 *                  the interface through which agent is reached
 *                  is no longer up or if we haven't heard from
 *                  that agent in a while.
 *   regReqTable -  contains all pending requests corresponding
 *                  to currentState. Could have a discovery request
 *                  too besides regular reg request (can we have more
 *                  than one? certainly we should have only one
 *                  for each HA-COA pair maybe just one for each
 *                  COA even one COA per HA).
 *
 *                  Among accepted renewable entries, we can have
 *                  only one HA-COA pair.
 *
 *                  Entry can be added/updated as a result of a new
 *                  reg request prompted by a state change, or 
 *                  renewal of an accepted (renewable) request
 *                  or a retransmission. An successful registration
 *                  reply can cause a pending entry to 
 *                  overwrite/replace an accepted entry. A successful
 *                  deregistration will cause a pending and one or
 *                  more accepted registrations to disappear. An
 *                  unsuccessful registration reply can cause new
 *                  registration attempts ...
 *                  Timers can also eliminate entries or cause
 *                  renewals and retransmissions.
 * 
 *   haTable      - reflects MNs view of the state of HAs. Their
 *                  can be one default HA tremplate entry, one
 *                  or more particular HA entry. When doing 
 *                  discovery, we only heed responses from HAs that
 *                  match our default or particular template.
 *                  Initialized to only contain template entries.
 *                  New dynamically learnt entries can be added
 *                  if they match the template (if the HA matches
 *                  a host specific template then simply the A bit
 *                  is set). Entries disappear or the A bit is
 *                  cleared if we haven't heard from HA in a while.
 *                  HAs can also get marked unusable depending on
 *                  the replies.
 *
 *    Dynamic HA discovery is triggerred when we wish to send a
 *    registration but find no HA that is alive and when we are
 *    about to remove a pending discovery message and find still
 *    no HA as alive.
 */
