/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Original Code is the Sunlabs Mobile IP.
 The Initial Developer of the Original Code is Vipul Gupta.
 Portions created by Vipul Gupta are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): Niko A. Fikouras (niko@comnets.uni-bremen.de)
		 Asanga D. Udugama (adu@comnets.uni-bremen.de)

*/

/*
 * An implementation of Mobile-IP as described in RFC 2002.
 * Author  : Vipul Gupta (vipul.gupta@eng.sun.com)
 *           Sunil Madhani (sunil.madhani@eng.sun.com)
 * mnNet.c : Networking-related routines used by a mobility agent. 
 *           These may need be re-implemented for non-UNIX platforms
 *           but should require little work to port to different 
 *           flavors of UNIX.
 */

#pragma ident	"@(#)mnNet.c 1.2 98/02/11"

#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#include "mip.h"
#include "mobile.h"

 
extern int  UDPregSock;
extern int  mipdebug;
extern char addr1[], addr2[];
extern int  errno;                      /* from <errno.h> */
extern int  ifaceTableSize;
extern IfaceEntry ifaceTable[];

extern void MNprocessRegReply(unsigned char *, int, u32, unsigned short);
extern void MNprocessAgentAdv(unsigned char *, int, char *, char *);

extern char *ntoa(u32, char *);

extern void printBuffer(unsigned char *, int);

extern boolean getAgentAdvByIface(char *, int *, char *, char *);

/* 
 * Send a UDP message from sock, to dst address and port dstPort. The pkt
 * is contained in pkt and is of length pktLen. dst is already in network
 * byte order.
 */
int sendUDPmessage(int sock, unsigned char *pkt,  int pktLen,
		   u32 dst, unsigned short dstPort)
{
    struct sockaddr_in sa;
    
    sa.sin_family = AF_INET;
    sa.sin_port = htons(dstPort);
    sa.sin_addr.s_addr = dst;

    /* send the message */
    if (sendto(sock, (char *) pkt, pktLen, 0, (struct sockaddr *) &sa,
	       sizeof(struct sockaddr)) < 0) {
        mip0dbg(("Sendto failed in sendUDPmessage [Error %d].\n", errno));
	return (-1);
    }   

    return(0);
}

/* Send an ICMP message contained in data to dst from src */
int sendICMPmessage(int s, u32 src, u32 dst, unsigned char data[], int len)
{
    struct sockaddr_in sa;
    struct in_addr ifaddr;

    sa.sin_family = AF_INET;
    sa.sin_port = 0;
    sa.sin_addr.s_addr = dst;

    /* 
     * Make sure pkts to 224.0.0.0 are sent through the interface
     * represented by the src address.
     */

    if (dst == inet_addr(LINK_MCAST_ADDR)) {
        /*
	 * On Linux:
	 * sprintf(msg, "route add 224.0.0.0 %s 0", ntoa(src, addr1));
	 * system(msg);
	 */

        ifaddr.s_addr = src;
	if (setsockopt(s, IPPROTO_IP, IP_MULTICAST_IF,
		       (char *)&ifaddr, sizeof (ifaddr)) == -1) {
	    mip0dbg(("Couldn't control iface for multicast addr [Error %d].\n",
		     errno));
	}
    }

    /* send the message */
    if (sendto(s, (char *) data, len, 0, (struct sockaddr *) &sa,
	       sizeof(struct sockaddr)) < 0) {
        mip0dbg(("Sendto failed in sendICMPmessage [Error %d]\n", errno));
	return(-1);
    }

    if (dst == inet_addr(LINK_MCAST_ADDR)) {
        /*
	 * On Linux:
	 * sprintf(msg, "route delete 224.0.0.0 %s", ntoa(src, addr1));
	 * system(msg);
	 */
    }
    return(0);
}

void getAndDispatchMessages()
{
    int i;
    int pktLen;
    static char pkt[MAX_PKT_SIZE];
    static char hwAddr[MAX_HWADDR_LEN + 1];
    static fd_set fdvec;
    static struct timeval tv;
    int fromLen;
    struct sockaddr_in from;
    u32 src;
    unsigned short srcPort;

    /* prepare the vector of file descriptors we'd like to monitor */
    FD_ZERO(&fdvec);
    FD_SET(UDPregSock, &fdvec);

    tv.tv_sec = 0;
    tv.tv_usec = 5000;   /* this way we give the CPU for a while if
			  * no messages are waiting. */

    /* "if" changed to "while" based on testing under Linux */
    /* 
    if ((i = select(FD_SETSIZE, &fdvec, NULL, NULL, &tv)) < 0) {
    */ 

    do {
        i = select(FD_SETSIZE, &fdvec, NULL, NULL, &tv);
    } while ((i < 0) && (errno == EINTR));

    if (i < 0) {
        mip0dbg(("select failed in getAndDispatchMessages() [Error %d].\n",
		 errno));
	return;
    }

    /* Process data on UDP socket */
    if (FD_ISSET(UDPregSock, &fdvec)) {
	pktLen = MAX_PKT_SIZE;
	fromLen = sizeof(struct sockaddr_in);
	if ((pktLen = recvfrom(UDPregSock, pkt, pktLen, 0,
			       (struct sockaddr *) &from, &fromLen)) < 0) {
	    mip0dbg(("recvfrom failed for UDP on registration socket [Error %d].\n",
		     errno));
	} else {
	    src = from.sin_addr.s_addr;
	    srcPort = ntohs(from.sin_port);
	    if (pkt[0] == REG_REPLY_TYPE) {
		if (pktLen < sizeof(regReply)) {
		    mip3dbg(("Received registration reply is too short.\n"));
		} else {
		    MNprocessRegReply((unsigned char *)pkt, pktLen,
				      src, srcPort);
		}
	    } else {
		mip3dbg(("Unknown UDP message (first byte 0x%x) from %s.\n",
			 (unsigned char) pkt[0], ntoa(src, addr1)));
	    }
	}
    }


    /*
     * Check for any advertisements from mobility agents. The routine
     * recvAdv already does most of the checks for us so we can directly
     * invoke MNprocessAgentAdv().
     */
    for (i = 0; i < ifaceTableSize; i++) {
	if (getAgentAdvByIface(pkt, &pktLen, ifaceTable[i].maIfaceName,
			       hwAddr)) {
	    MNprocessAgentAdv((unsigned char *)pkt, pktLen,
			      ifaceTable[i].maIfaceName, hwAddr);
	}
    }
}

