/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Original Code is the Sunlabs Mobile IP.
 The Initial Developer of the Original Code is Vipul Gupta.
 Portions created by Vipul Gupta are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): Niko A. Fikouras (niko@comnets.uni-bremen.de)
		 Asanga D. Udugama (adu@comnets.uni-bremen.de)

*/

/*
 * Author: Vipul Gupta (vgupta@eng.sun.com)
 *
 * Routines in this file implement OS-specific portions of a
 * Mobile node (RFC 2002).
 */

#pragma ident	"%Z%%M% %I% %E%"

/*
 * void OSInit(): 
 *
 *     Includes OS-specific initialization. For example, in solaris, we
 * initialize the vtunl module here and also enable decapsulation
 * of packets whose inner dst is the mobile node's home address.
 * This is also a good place to creates a separate thread to pick
 * up agent advertisements on the interfaces. 
 *
 * void OSCleanup()
 * 
 *      OS dependent cleanup routine called when exiting, e.g.
 * releasing locks, killing threads, or putting the Mobile node
 * in the state prior to initiating Mobile IP.
 *
 * char *err2str(int errval)
 *
 *     Converts a negative error value to a meaningful error message
 * string, e.g. by using the comments in /usr/include/sys/errno.h.
 *
 * int getNewIfaceList(MIPIfaceEntry *ifaceTable, int *size) 
 *
 *	Gets the list of interfaces of interest to the mobile node
 * code, e.g. interfaces on which agent advertisements must be monitored
 * or those that may be used to obtain a colocated care-of address.
 * NOTE: Mobile routers should ignore interfaces attached to supported
 * mobile networks.
 *
 * boolean getAgentAdvByIface(char *pkt, int *pktlen, char *ifaceName,
 *                            char *hwAddr)
 *      Returns an agent advertisement packet for the specified interface.
 * The advertisement packet placed in the pkt buffer starts at the IP
 * header. The hardware address of the source is also returned.
 *
 * int configCloneForIface(char *Iface, u32 addr, u32 netMask)
 *
 *     Configures a logical clone for the specified interface as
 * a placeholder for the mobile node's home address. It is assigned
 * the specified addr and netMask. For Point-to-point interfaces,
 * additional information required to configure the clone may be 
 * copied from the parentIface. This function is also responsible
 * for deleting any spurious routing entries that may be added as
 * part of bringing up the clone Iface.
 *
 * void unconfigCloneForIface(char *ifaceName) 
 * 
 *     Unconfigures the logical clone interface used as a placeholder
 * for the mobile node's home address in COLOCATED state.
 *
 * int configIface(char *ifaceName, u32 addr, u32 netMask)
 *
 *     Configures the specified interface with the assigned address
 * and netmask. The interface is brought up.
 *
 * int changeSrcOnDefaultRoute(u32 homeAddress, u32 router)
 * 
 *      Sets up a default route entry so packets sent to the router
 * have the homeAddress as source.
 * 
 * int routeadd(unsigned long dst, unsigned long mask, unsigned long gw,
 *              int metric, char *devname)
 *   
 *      Add route for destination "dst", netmask "mask" through gateway
 * "gw". A zero metric specifies a local route and non-zero metric
 * specifies an indirect route (i.e. through intermediate gateway(s)).
 *
 * int routedel(unsigned long dst, unsigned long mask, unsigned long gw,
 *              int metric)
 *
 *      Delete route for specified destination through gateway. The
 * parameters are interpreted the same way as for routeadd.
 *
 * int routeflush(char *ifaceName)
 *
 *      Clears the routing entries table related to interface given
 * in ifaceName (except for the loopback interface and
 * any entries created by the mobile node (e.g. ip0 under Solaris)).
 * TODO: Be careful about the multicast network entry.
 *
 * int enableReverseTunneling(mnState *mnsp, u32 homeAgentAddr)
 *      Set up and enable a reverse tunnel from the mobile node to
 * the specified home agent.
 *
 * int disableReverseTunneling(mnState *mnsp, u32 homeAgentAddr)
 *      Undo the actions performed by enableReverseTunneling.
 *
 * int arpadd(unsigned long host, unsigned char *eaddr, char *flags)
 *
 *      Adds an arp entry with ip address set to host and hardware  
 * address set to eaddr with approriate flags specified by flag
 * argument, e.g. arpadd(inet_addr("129.146.122.121"), "08:20:AB:FE:33:11",
 * "pub") creates a proxy ARP entry for 129.146.122.121.
 *
 * int arpdel(unsigned long host)
 *
 *      Deletes an arp entry for the specified ip address.
 *
 * int sendgarp(char *iface)
 * 
 *      Sends gratuitous arp for the mobile node using the hardware address
 * of the iface. Needed when a mobile node returns home and successfully 
 * deregisters itself.
 *
 * int arpflush()
 *
 *	Clears the arp table except for those with "SM" flags.
 *
 * int disableArp(u32 addr, u32 i)
 *     
 *      Prevent a mobile node from sending out broadcast ARP packets
 * (used when a mobile node is on a foreign network).
 *
 * int restoreArp()
 *
 *     Restore normal ARP behavior.
 * 
 * void enableIPSec(mnState *mnsp)
 *
 *      Enables IPSec for firewall traversal. The main idea is that the
 * mobile node should do whatever is necessary at its end such that packets
 * generated by it for the home network will be able to get past any
 * intervening firewalls.
 *
 * void disableIPSec(mnState *mnsp)
 *    
 *      Undo the actions performed for enableIPSec.
 *
 */

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include <netinet/in.h>
#include <net/if_arp.h>

#include <arpa/inet.h>
#include <string.h>
#include <pthread.h>
#include <net-support.h>
/* #include <pthread/mit/pthread.h> */

/* Redhat 5.1 uses glibc (libc6) which, unlike libc, is not tied to
 * the kernel headers. Mixing kernel headers and C library headers
 * causes compilation errors related to multiple definitions in
 * /usr/include/linux/socket.h and /usr/include/socketbits.h.
 * Thanks to Alan Cox for pointing this out and suggesting that
 * I use <net/...> instead.
 *
 * Replaced the following:
 * #include <linux/if.h>
 * #include <linux/if_ether.h>
 * #include <linux/if_arp.h>
 * #include <linux/ip.h>
 * #include <linux/route.h>
 *
 * with
 */
#ifdef OLD_LIBC
#include <linux/if.h>
#include <linux/if_ether.h>
#include <linux/if_arp.h>
#include <linux/ip.h>
#include <linux/route.h>
#else /* for GLIBC, avoid the kernel include files */
#include <net/ethernet.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <net/route.h>
#endif /*OLD_LIBC */

#include "mip.h"
#include "mobile.h"

/* Tunnel related definitions */
#define MAX_TUNNEL_SUPPORTED    32
#define MAX_ERR                 1024
#define NO_FREE_TUNNEL          (MAX_ERR + 1)
#define TUNNEL_NOT_FOUND        (MAX_ERR + 2)
#define DEV_NAME_NOT_FOUND      ENXIO
#define INVALID_IP_ADDR         ENXIO
#define MN_ENTRY_ABSENT         (MAX_ERR + 11)
#define NO_MAPPING              (MAX_ERR + 12)
#define MAX_AGENT_ADV_ENTRIES   10
#define DEFAULT_ADDR            ((u32) htonl(0xfffffffe)) /* a dummy address
                                                           * used to mark
							   * a reverse tunnel
							   */

#ifndef IFF_DHCPRUNNING
#define IFF_DHCPRUNNING 0x4000
#endif

#define MAXDLBUF  2048
#define MAXWAIT   15

// filename of the iproute2 table name definition file
#define ROUTE2_TABLE_FILE 	"/etc/iproute2/rt_tables"

/* 
 * For Linux 2.2 we need to avoid dealing with tunl0 
 * which is special ... cannot be deleted. 
 */
#ifdef LINUX_2_0
#define FIRST_TUNNEL            0
#else
#include <netinet/ip.h>
#include <bits/ioctls.h>
#define SIOCADDTUNNEL   (SIOCDEVPRIVATE + 1)
#define SIOCDELTUNNEL   (SIOCDEVPRIVATE + 2)
struct ip_tunnel_parm {
    char     name[IFNAMSIZ];
    int      link;
    unsigned short i_flags;
    unsigned short o_flags;
    unsigned long  i_key;
    unsigned long  o_key;
    struct iphdr   iph;
};
#define FIRST_TUNNEL            1
#endif

extern int errno;
static int ioctl_sockid;

/* Table for mapping tunnelno with mobile-node address */
static unsigned long mnaddr_tunl[MAX_TUNNEL_SUPPORTED + FIRST_TUNNEL + 1];

struct strbuf {
   int  len;
   char *buf;
   int  maxlen;
};

typedef struct arhdr{
    unsigned short  ar_hrd;         /* format of hardware address   */
    unsigned short  ar_pro;         /* format of protocol address   */
    unsigned char   ar_hln;         /* length of hardware address   */
    unsigned char   ar_pln;         /* length of protocol address   */
    unsigned short  ar_op;          /* ARP opcode (command)         */
    unsigned char ar_sha[ETH_ALEN]; /* sender hardware address      */
    unsigned char ar_sip[4];        /* sender IP address            */
    unsigned char ar_tha[ETH_ALEN]; /* target hardware address      */
    unsigned char ar_tip[4];        /* target IP address            */
} Arphdr;

extern char currentTime[], relativeTime[], addr1[], addr2[], addr3[], addr4[];
extern int mipdebug;
extern char msg[];

/* The mobile node's address and netmask */
extern u32 mnHomeAddr;
extern u32 mnHomeNetmask;
extern boolean isReverseTunnelSet;

extern RegReqEntry regReqTable[DEFAULT_MAX_REG_REQ];

extern IfaceEntry ifaceTable[];
extern int ifaceTableSize;

extern int LazySoliciting;
extern int EagerSoliciting;

extern char ignoreIfaceList[][MAX_IFNAME_LEN];
extern int ignoreIfaceCnt;

/* The selected movement detection method */
extern int movementDetection;

// currnetState is part of ISTable; changed to handle MSB
extern InterfaceState ISTable[DEFAULT_MAX_INTERFACES];
//extern mnState currentState;
extern int ISTableSize;

extern void startForeignAction(mnState *, LocalMaEntry *, int);
extern void startHomeAction(mnState *, int ISSeq);
extern int INET_setroute(int, int, char**);

/* to store the filters that have been defined for this mobile node*/
extern Filter mnFilters[MAX_FILTERS]; 				   // data for each filter
extern int mnFilterCount; 					   // the number of filters
extern FilterOp mnFilterOpTable[MAX_FILTERS];			   // the grid that holds the operations
								   // to be done for the filters
extern int mnFilterOpCount;					   // operations count

extern int retransmissionPolicy[];

extern int setFilterForHomeBinding(Filter [], int, FilterOp [], int *, char *);
extern int setFilterForNewBinding(Filter [], int, FilterOp [], int *, char *, int, RegReqEntry []);
extern int numBitsInMask(u32);

/* Type definitions for the Advertisement Pkt. Buffer */
typedef struct aap {
            char ifaceName[MAX_IFNAME_LEN];
            u32  advSrcIPAddr;
            char hwAddr[MAX_HWADDR_LEN];
            char data[MAX_PKT_SIZE];
            int  dataLen;
} AgentAdvPkt;

/*
 * Circular buffer to store the agent advertisements.
 * Buffer is empty when head = tail. It is full when
 * head == (tail + 1) mod n.
 *
 * For a non-empty list, head points to the first filled slot.
 * Tail points to the next available slot. So a buffer with n
 * slots can hold upto (n-1) items.
 *
 * AdvBufLock is used to synchronize access to the buffer containing
 * agent advertisements. The main program thread removes agent advertisemen$
 * from the buffer either as part of getAgentAdvByIface() or while
 * doing a purge as part of relinquishing an interface when it is
 * brough down. Another thread which monitors a list of interfaces
 * (or rather open file descriptors) for agent advertisements inserts
 * data in this buffer.
 */
typedef struct AdvPktBuffer {
            int head;
            int tail;
            AgentAdvPkt advPkt[MAX_AGENT_ADV_ENTRIES];
            pthread_mutex_t AdvBufLock;
} advPktBuffer;

static advPktBuffer advPktBuf;

static pthread_t monitorIface;
typedef struct ifd {
        char ifaceName[MAX_IFNAME_LEN];
        int  ifaceFD;
} ifaceFD;

typedef struct ifdlist {
        int     ifaceFDCnt;
        ifaceFD ifaceFDarray[DEFAULT_MAX_INTERFACES];
        pthread_mutex_t ifaceFDLock;
} ifacefdList;

/*
 * Besides the main thread, the program creates another thread which
 * continuously monitors incoming agent advertisements on the interfaces
 * listed in ifaceFDList. The lock ifaceFDLock is used to synchronize
 * access to the FD list. The main thread adds or removes FDs based
 * on its notion of "relevant" interfaces and the secondary thread
 * only reads (never changes) this datastructure.
 */
ifacefdList ifaceFDList;

extern void ether_print(unsigned char[]);
extern IfaceEntry *findIface(char *);
/* Security related stuff */
extern DomainInfo domainInfo;

void enableIPSec(mnState *);
void disableIPSec(mnState *);
int enableReverseTunneling(mnState *, u32);
int disableReverseTunneling(mnState *, u32);

/*
 * Returns the available tunnel number in the range 0-31. If not available
 * it returns -(NO_FREE_TUNNEL)
 */
int newtunnel(u32 addr)
{
    int i;
    char cmd[50];

    /* 
     * NOTE: All tunnels created by this program start at number 1+FIRST_TUNNEL
     * tunl<FIRST_TUNNEL> is free for use in SKIP.
     */ 
    for (i = 1 + FIRST_TUNNEL; i < FIRST_TUNNEL + MAX_TUNNEL_SUPPORTED; i++) {
        if (mnaddr_tunl[i] == 0) {
#ifdef LINUX_2_0	   
            sprintf(cmd, "/sbin/insmod -o tunl%d new_tunnel", i);
	    mip3dbg(("Loading tunnel module (%s) ... \n", cmd));
	    system(cmd);
#endif	   
	    /*
	     * TODO: check for error from previous step and other system()
	     * calls.
	     */
	    mnaddr_tunl[i] = addr;
            return(i);
        }	
    }

    return(-1 * NO_FREE_TUNNEL);
}

void freetunnel(int tnum)
{
    char cmd[50];

    mnaddr_tunl[tnum] = 0;
#ifdef LINUX_2_0   
    sprintf(cmd, "/sbin/rmmod tunl%d", tnum);
    mip3dbg(("Freeing tunnel module (%s) ... \n", cmd));
    system(cmd);
#endif   
}

/* 
 * Gets the tunnel number corresponding to given
 * mobile node address .
 * If the corresponding tunnel number exist then it returns the tunnel no.
 * else returns -(TUNNEL_NOT_FOUND)
 */
int gettunnelno(u32 mnaddr)
{
    int i;
    for (i = FIRST_TUNNEL; i < FIRST_TUNNEL + MAX_TUNNEL_SUPPORTED; i++) {
        if (mnaddr_tunl[i] == mnaddr)
	   return(i);
    }
    return(-1 * TUNNEL_NOT_FOUND);
}

void printtunnel()
{
    char mnaddr[20];
    int i;

    for (i = FIRST_TUNNEL; i < FIRST_TUNNEL + MAX_TUNNEL_SUPPORTED; i++) {
        if (mnaddr_tunl[i] != 0) {
	    mip3dbg(("Tunnel %d is for %s\n", 
		     i, ntoa(mnaddr_tunl[i], mnaddr)));
	}
    }
}

/* Process encapadd command */
int encapadd(u32 mnaddr, u32 haaddr, u32 coaddr, int flags)
{
    char ifname[10];
    struct sockaddr_in *sock;
    struct ifreq ifr;
    int tnum, val;
#ifndef LINUX_2_0
    struct ip_tunnel_parm tp;
#endif
  
    if ((tnum = newtunnel(mnaddr)) < 0) {
        return(tnum);
    }
	
    sprintf(ifname, "tunl%d", tnum);
#ifndef LINUX_2_0
    memset(&tp, 0, sizeof(tp));
    strncpy(tp.name, ifname, strlen(ifname) + 1);
    tp.iph.version = 4;
    tp.iph.ihl = 5;
    tp.iph.frag_off = htons(0x4000);
    tp.iph.protocol = 4;
    tp.iph.ttl = 0;
    tp.iph.daddr = coaddr;
   
    strncpy(ifr.ifr_name, "tunl0", strlen("tunl0")+1);
    ifr.ifr_ifru.ifru_data = (void *) &tp;
   
    /* 
     * The following accomplishes the equivalent of
     *     ip tunnel add ifname mode ipip remote coaddr
     */
    if ((val = ioctl(ioctl_sockid, SIOCADDTUNNEL, &ifr)) < 0) {
       goto failure;
    }
#endif
     
    memset(&ifr, 0, sizeof(struct ifreq));
    strncpy(ifr.ifr_name, ifname, strlen(ifname));

    /* 
     * create a pt-to-pt iface from haaddr to coaddr
     * with the given name and manipulate tunnelbitvec
     * appropriately 
     */
	  
    sock = (struct sockaddr_in *) &ifr.ifr_addr;
    sock->sin_family = AF_INET;
    sock->sin_addr.s_addr = haaddr;
    if (ioctl(ioctl_sockid, SIOCSIFADDR, (char *) &ifr) < 0)  {
	goto failure;
    }

#ifdef LINUX_2_0
    sock = (struct sockaddr_in *) &ifr.ifr_dstaddr;
    sock->sin_family = AF_INET;
    sock->sin_addr.s_addr = coaddr;
    if (ioctl(ioctl_sockid, SIOCSIFDSTADDR, (char *) &ifr) < 0)  {
	goto failure;
    }
#endif

    ifr.ifr_flags = (IFF_UP | IFF_NOARP | IFF_POINTOPOINT);

    if (val = ioctl(ioctl_sockid, SIOCSIFFLAGS, (char *) &ifr) < 0)  {
	goto failure;
    }

    printtunnel();
    return(0);
    
failure:
    val = (-1) * errno;
    freetunnel(tnum);
    return(val);
}

/* Process encaprem command. Bring the tunnel down */
int
encaprem(u32 mnaddr, u32 coaddr)
{
    char ifname[10];
    struct ifreq ifr;
    int tnum;
#ifndef LINUX_2_0   
    struct ip_tunnel_parm tp;
#endif

    if ((tnum = gettunnelno(mnaddr)) <  0) {
       return(tnum);
    }

    sprintf(ifname, "tunl%d",tnum);
    memset(&ifr, 0, sizeof(struct ifreq));
    strncpy(ifr.ifr_name, ifname, strlen(ifname));
        
    /* bring down the tunnel interface and free the corresponding tunnel */
    if (ioctl(ioctl_sockid, SIOCGIFFLAGS, (char *) &ifr) < 0) {
        return(-1 * errno);
    }

    ifr.ifr_flags &= ~IFF_UP;

    if (ioctl(ioctl_sockid, SIOCSIFFLAGS, (char *) &ifr) < 0)  {
        return(-1 * errno);
    }

#ifndef LINUX_2_0
    memset(&tp, 0, sizeof(tp));
    strncpy(tp.name, ifname, strlen(ifname) + 1);
    tp.iph.version = 4;
    tp.iph.ihl = 5;
    tp.iph.frag_off = htons(0x4000);
    tp.iph.protocol = 4;
    tp.iph.ttl = 0;
    tp.iph.daddr = coaddr;
   
    memset(&ifr, 0, sizeof(struct ifreq));
    strncpy(ifr.ifr_name, ifname, strlen(ifname)+1);
    ifr.ifr_ifru.ifru_data = (void *) &tp;
    /* Do the equivalent of "ip tunnel del ifname" */
    if (ioctl(ioctl_sockid, SIOCDELTUNNEL, &ifr) < 0) {
        mip0dbg(("SIOCDELTUNNEL failed for %s in encaprem.", ifname));
    }
#endif   
     
    freetunnel(tnum);
    printtunnel();
    return(0);
}

/* Process decapadd command */
int
decapadd(u32 target, char*flags) 
{
    return(0);
}

/* Process decaprem command */
int
decaprem(u32 target)
{
    return(0);
}

/* Add routing entry for the tunnel */
int mkroutethroughtunnel(u32 mnaddr)
{
    struct rtentry rtent;
    int count;
    char ifname[16];
    struct sockaddr_in *sock;
    int tnum;

    /* Add the corresponding route */
    memset(&rtent, 0, sizeof(struct rtentry));
    sock = (struct sockaddr_in *) &rtent.rt_dst;
    sock->sin_family = AF_INET;
    sock->sin_addr.s_addr = mnaddr;
    rtent.rt_flags = (RTF_UP | RTF_HOST);

    if ((tnum = gettunnelno(mnaddr)) < 0 ) 
        return(tnum);
    count = sprintf(ifname, "tunl%d", tnum);
    ifname[count] = '\0';

    rtent.rt_dev = ifname;
    errno = 0;
    if (ioctl(ioctl_sockid, SIOCADDRT, (char *) &rtent) < 0)  {
        if (errno != EEXIST) {
            return(-1 * errno);
	}
    }

    return(0);
}

/* Delete the routing entry for the tunnel */
int rmroutethroughtunnel(u32 mnaddr)
{
    struct rtentry rtent;
    struct sockaddr_in *sock;
    int tnum;
    char addr[30];

    if ((tnum = gettunnelno(mnaddr)) < 0) {
        /* Hmmm ... tunnel may have been removed as part of encaprem */
        return(0);
    }

    memset(&rtent, 0, sizeof(struct rtentry));
    sock = (struct sockaddr_in *) &rtent.rt_dst;
    sock->sin_family = AF_INET;
    sock->sin_addr.s_addr = mnaddr;

    errno = 0;
    if (ioctl(ioctl_sockid, SIOCDELRT, (char *) &rtent) < 0) {
	if (errno != ENXIO) {
	    return(-1 * errno);
	}
    }

    return(0);
}

/*
 * Add/Delete the route depending on argument cmd 
 * TODO : this is no longer being used and may be broken!!! 
 */ 
routemodify(unsigned long dst, unsigned long mask, unsigned long  gw, 
	    int metric, char cmd)
{
    struct rtentry route;
    struct sockaddr_in *sin;  
    int val, temp = 0;
    char ifname[MAX_IFNAME_LEN + 1];
    
    /*
     * Workaround for Solaris 2.6 which seems to use an ARP
     * entry for x even when no local route exists for x.
     *
     * if (cmd == 'd') {
     *     val = arpdel(dst);
     * }
     */ 
    
    bzero((char *)&route, sizeof(route));
    sin = (struct sockaddr_in *)&route.rt_dst;
    sin->sin_addr.s_addr = dst;

    /*
     * Set route gateway ???? Is this correct ???? 
     * if (cmd == 'a') {
     *     if ((temp = getifaceName(gw, ifname)) < 0)
     *         return(temp);
     *     route.rt_dev = ifname;
     * }
     */ 
    
    route.rt_flags = RTF_UP;
    
    if (mask == inet_addr("255.255.255.255"))
        route.rt_flags |= RTF_HOST;
    
    if (metric > 0)
        route.rt_flags |= RTF_GATEWAY;
    
    errno = 0;
    if ((ioctl(ioctl_sockid, cmd == 'a' ? SIOCADDRT : SIOCDELRT,
	       (caddr_t)&route)) < 0) {
	/* 
	 * SIOCADDRT does not return ENXIO &
	 * SIOCDELRT never returns EEXIST
	 */
	if ((errno != ENXIO) && (errno != EEXIST))
	    return(-1 * (errno));
    }
    
    return(val);
}
      
/* Add a route */
int routeadd(unsigned long dst, unsigned long mask, unsigned long  gw,
	     int metric, char *devname)
{
    char cmd[80];
	char *command[10];
    
    /* return(routemodify(dst, mask, gw, metric, 'a')); */
    if (mask == inet_addr("255.255.255.255")) {
		sprintf(cmd, "#host %s gw %s dev %s",
			ntoa(dst, addr1), ntoa(gw, addr2), devname);
		mip3dbg(("Adding route (%s) ... \n", cmd));
		makeArray(command, cmd);
		INET_setroute(1, 129, command);
    } else {
		sprintf(cmd, "#net %s netmask %s gw %s dev %s",
			ntoa(dst, addr1), ntoa(mask, addr2), ntoa(gw, addr3), devname);
		mip3dbg(("Adding route (%s) ... \n", cmd));
		makeArray(command, cmd);
		INET_setroute(1, 129, command);
    }

   return(0);
}

/* Delete route */
int routedel(unsigned long dst, unsigned long mask, unsigned long  gw,
	     int metric)
{
    char cmd[80];
	char *command[7];

    /* return(routemodify(dst, mask, gw, metric, 'd')); */

    if (mask != inet_addr("255.255.255.255")) {
        if (dst == 0) {
			strcpy(cmd, "#net 0.0.0.0");
			makeArray(command, cmd);
			INET_setroute(2, 129, command);
		} else {
			sprintf(cmd, "#net %s netmask %s",
				ntoa(dst, addr1), ntoa(mask, addr2));
			makeArray(command, cmd);
			INET_setroute(2, 129, command);
		}
    } else {
		sprintf(cmd, "%s", ntoa(dst, addr1));
		makeArray(command, cmd);
		INET_setroute(2, 129, command);
    }

    mip3dbg(("Deleting route ... \n"));
       
    return(0);
}

/* 
 * Tries to pick up a default router from the routing table such that
 * the router lies on the same network as the supplied ipAddr. The
 * netMask is used to determine when two addresses belong to the same
 * network.
 * TODO: For now, if there is a problem, the function returns a zero 
 *       gateway.
 */
u32 getRouterAddr(u32 ipAddr, u32 netMask)
{
    char line[256], dev[16];
    u32 dst, gw, val;
    FILE *fd;
  
    if ((fd = fopen("/proc/net/route", "r")) == NULL) {
	mip0dbg(("Could not open /proc/net/route in getRouterAddr().\n"));
	mip0dbg(("You may need to recompile the kernel to support the\n"));
	mip0dbg(("proc filesystem.\n"));
	return (0);
    }
  
    val = 0;
    while (fgets(line, sizeof(line), fd) != NULL) {
	if ((sscanf(line, "%s %lx %lx", dev, &dst, &gw) == 3) && (dst == 0)) {
	    val = gw;
	    break;
	}
    }

    fclose(fd);

    if ((val != 0) && (((val ^ ipAddr) & netMask) == 0))
        return(val);
    else
        return(0);
}

/*
 * Flush routes existing routes.
 * TODO: are there any routes that we should preserve?
 * For MSB, the interface name is also being sent as a
 * parameter
 */
int routeflush(char *ifaceName)
{
    char line[256], dev[MAX_IFNAME_LEN];
    u32  dst, gw, mask;
    int flags, refcnt, use, metric;
    int found, i;
    FILE *fd;

    if ((fd = fopen("/proc/net/route", "r")) == NULL) {
	mip0dbg(("Could not open /proc/net/route in routeflush().\n"));
	mip0dbg(("You may need to recompile the kernel to support the\n"));
	mip0dbg(("proc filesystem.\n"));
	return(-EINVAL);
    }

    while (fgets(line, sizeof(line), fd) != NULL) {
	if (sscanf(line, "%s %lx %lx %d %d %d %d %lx",
		   dev, &dst, &gw, &flags, &refcnt,
		   &use, &metric, &mask) == 8) {
	    /*
	     * In order to support mobile routers, we should not
	     * delete any routes that go through an interface
	     * in the ignoreIfaceList.
	     */
	    found = FALSE;
	    for (i = 0; i < ignoreIfaceCnt; i++) {
	        if (strcmp(dev, ignoreIfaceList[i]) == 0) found = TRUE;
	    }

	    if ((dst != inet_addr("224.0.0.0")) &&
		(dst != inet_addr("127.0.0.0")) &&
		(dst != inet_addr("127.0.0.1")) &&
		(strcmp(dev, ifaceName) == 0) &&
		(found == FALSE)) {
	        /*
		 * For now, we zero out mask and metric since routedel
		 * doesn't use it
		 */
	        routedel(dst, mask, gw, 0);
	    } else {
	        mip4dbg(("Ignored destination %s through %s in routeflush.\n",
			 ntoa(dst, addr1), dev));
	    }
	}
    }

    fclose(fd);
    return(0);
}


int routeflushHomeNW() {
	u32 homeNW;
	int more;
	FILE *irlfp;
	char addr1[20], addr2[20], addr3[20], cmd[128], cmd2[128],
		rlline[256];

	homeNW = mnHomeAddr & mnHomeNetmask;
	ntoa(homeNW, addr1);

	// delete the routze entries with home NW
	more = TRUE;
	while(more) {
		more = FALSE;
		sprintf(cmd, "route -n");
		irlfp = popen(cmd, "r");
		while(fgets(rlline, 256, irlfp) != NULL) {
			if(strlen(rlline) == 0 || *rlline == '#')
				continue;

			sscanf(rlline, "%s %s", addr2, addr3);
			if(strcmp(addr1, addr2) == 0) {
				more = TRUE;
				break;
			}
		}
		pclose(irlfp);

		if(more) {
			sprintf(cmd2, "ip route del %s/%d", addr1,
						numBitsInMask(mnHomeNetmask));
			printf("%s\n", cmd2);
			system(cmd2);
		}
	}
	return 0;
}

int routeflushDefault() {
	int more;
	FILE *irlfp;
	char addr1[20], addr2[20], cmd[128], cmd2[128],
		rlline[256];

	// delete the default route entries
	more = TRUE;
	while(more) {
		more = FALSE;
		sprintf(cmd, "route -n");
		irlfp = popen(cmd, "r");
		while(fgets(rlline, 256, irlfp) != NULL) {
			if(strlen(rlline) == 0 || *rlline == '#')
				continue;

			sscanf(rlline, "%s %s", addr1, addr2);
			if(strcmp(addr1, "0.0.0.0") == 0) {
				more = TRUE;
				break;
			}
		}
		pclose(irlfp);

		if(more) {
			strcpy(cmd2, "ip route del default");
			printf("%s\n", cmd2);
			system(cmd2);
		}
	}
	return 0;
}


/*
 * Send out a gratuitous ARP packet through the interface "iface"
 * with hardware address "ifaceaddr" and that maps IP address
 * "ipaddr" to hardware address "ethaddr".
 * The complete packet contents are as follows:
 *
 * Ethernet hdr
 *        src = ifaceaddr  dst = ff:ff:ff:ff:ff:ff
 *        frame type = 0x0806 (ETH_P_ARP)
 *
 * ARP hdr
 *        hardware addr type = 1 (ARPHRD_ETHER)
 *        protocol addr type = 0x0800 (ETH_P_IP)
 *        hardware addr len = 6 (ETH_ALEN)
 *        protocol addr len = 4 
 *        op code = 1 (ARPOP_REQUEST)
 *
 * ARP payload
 *        sender hardware addr = ethaddr
 *        sender IP addr = ipaddr
 *        dst hardware addr = don't care
 *        target IP addr = don't care
 *
 * NOTE: When this function is called, ipaddr is in network-byte order.
 *
 */
int sendGARPpkt(char *iface, unsigned char *ifaceaddr, 
	       u32 ipaddr, unsigned char *ethaddr)
{
    char frame[80];
    struct ethhdr *ehdr;
    Arphdr *arphd;
    int sockid;
    struct sockaddr sa;

    errno = 0;
    if ((sockid = socket(AF_INET, SOCK_PACKET, htons(ETH_P_802_3))) < 0)
        return(-1 * errno);

    memset((void*)frame, 0, sizeof(frame));

    ehdr = (struct ethhdr*) frame;
    memcpy(ehdr->h_source, ifaceaddr, ETH_ALEN);
    memset(ehdr->h_dest, 0xff, ETH_ALEN);
    ehdr->h_proto = htons(ETH_P_ARP);

    arphd = (Arphdr*)(frame + sizeof(struct ethhdr));
    arphd->ar_hrd = htons(ARPHRD_ETHER);
    arphd->ar_pro = htons(ETH_P_IP);
    arphd->ar_hln = ETH_ALEN;
    arphd->ar_pln = 4;
    arphd->ar_op  = htons(ARPOP_REQUEST);
    memcpy(arphd->ar_sha, ethaddr, ETH_ALEN);

    ipaddr = ntohl(ipaddr); 
    arphd->ar_sip[0] = (unsigned char)((ipaddr >> 24) & 0xff);
    arphd->ar_sip[1] = (unsigned char)((ipaddr >> 16) & 0xff);
    arphd->ar_sip[2] = (unsigned char)((ipaddr >> 8 ) & 0xff);
    arphd->ar_sip[3] = (unsigned char)(ipaddr & 0xff);
    memset(arphd->ar_tha, 0xff, ETH_ALEN);
    memcpy(arphd->ar_tip, arphd->ar_sip, 4);

    sa.sa_family = AF_INET;
    strcpy(sa.sa_data, iface);

    /* TODO: figure out the correct length */
    if (sendto(sockid, frame, sizeof(struct ethhdr) + sizeof(Arphdr),
	       0, &sa, sizeof(sa)) < 0) {
        close(sockid);
        return(-1 * errno);
    }

    close(sockid);
    return(0);
}
    
/* Get the hardware address corresponding to addr from the arp table */
int arpgetHWaddr(u32 addr, unsigned char *etheraddr)
{
    struct arpreq ar;
    struct sockaddr_in *sin;
  
    bzero((caddr_t)&ar, sizeof ar);
    ar.arp_pa.sa_family = AF_INET;
    sin = (struct sockaddr_in *)&ar.arp_pa;
    sin->sin_family = AF_INET;
    sin->sin_addr.s_addr = addr;

    errno = 0;
    if (ioctl(ioctl_sockid, SIOCGARP, (caddr_t)&ar) < 0) {
        return(-1 * (errno));
    } else {
	memcpy(etheraddr, ar.arp_ha.sa_data, ETH_ALEN);
	return(0);
    }
}


/* Get the ethernet address for the given device name */
int getEthernetAddr(char *dev, unsigned char *addr)
{
    struct ifreq ifr;
    char *buf;

    bzero((char *)&ifr, sizeof(ifr));
    strncpy(ifr.ifr_name, dev, strlen(dev)); 

    errno = 0;
    if (ioctl(ioctl_sockid, SIOCGIFHWADDR, (char *)&ifr) < 0) 
        return(-1 * errno);
 
    memcpy(addr, (unsigned char *) &ifr.ifr_hwaddr.sa_data, ETH_ALEN);
    return(0);
}

/*
 * Get the device name corresponding to the given hardware address.
 * Currently, we get a list of interfaces available
 * and their corresponding ethernet addresses are compared
 * with eaddr. The one which matches is returned.
 * TODO: Find a better way. Is there something like SIOCGIFNAME :) ?
 */	
int hwaddr2devname(unsigned char *eaddr, char *dev)
{
    char buff[1024];
    struct ifconf ifc;
    struct ifreq *ifr;
    int i;
    unsigned char addr[ETH_ALEN];

    ifc.ifc_len = sizeof(buff);
    ifc.ifc_buf = buff;

    errno = 0;
    if (ioctl(ioctl_sockid, SIOCGIFCONF, &ifc) < 0) 
        return(-1 * errno);

    ifr = ifc.ifc_req;
    for (i = ifc.ifc_len / sizeof(struct ifreq); --i >= 0; ifr++) {
        if (getEthernetAddr(ifr->ifr_name, addr) == 0) {
            if (memcmp(addr, eaddr, ETH_ALEN) == 0) {
	       memcpy(dev, ifr->ifr_name, ETH_ALEN);
	       return(0);
            }
        }
    }

    return(-1 * DEV_NAME_NOT_FOUND);
}

/* Insert an IP addr and ethernet addr mapping in the ARP cache. */
int arpadd(unsigned long host, unsigned char *eaddr, char *flag)
{
    struct arpreq ar;
    struct sockaddr_in *sock;
    char iface[10];
    int val;
    char zeros[8];
   
    /*
     * If the hardware address is all zeros, return ... we don't
     * need ARP entries for destinations reachable over a point-to-point
     * interface.
     */ 
    memset(zeros, 0, 8);
    if (memcmp(eaddr, zeros, ETH_ALEN) == 0) return 0;
   
    bzero((char *)&ar, sizeof(ar));
    sock = (struct sockaddr_in *) &ar.arp_pa;
    sock->sin_family = AF_INET;
    sock->sin_addr.s_addr = host;
 
    sock = (struct sockaddr_in *) &ar.arp_ha;
    sock->sin_family = ARPHRD_ETHER;
    memcpy(ar.arp_ha.sa_data, eaddr, ETH_ALEN);
  
    ar.arp_flags |= ATF_PERM;
    if (strncmp(flag, "temp", 4) == 0)
        ar.arp_flags &= ~ATF_PERM;
    else if (strncmp(flag, "pub", 3) == 0)
        ar.arp_flags = (ATF_PUBL | ATF_PERM |  ATF_COM);
    else if (strncmp(flag, "trail", 5) == 0)
        ar.arp_flags |= ATF_USETRAILERS;

    errno = 0;
    if (ioctl(ioctl_sockid, SIOCSARP, (char *) &ar) < 0) {
        if (errno != EEXIST) {
            return(-1 * errno);
        }
    }

    if (strncmp(flag, "pub", 3) == 0) {
	/* we must send out a gratuitous ARP packet */
	if ((val = hwaddr2devname(eaddr, iface)) < 0) {
	    return(val);
	} else {
	    return(sendGARPpkt(iface, eaddr, host, eaddr));
	}
    }
    
    return(0);
}

/*
 * Delete the arp entry for the specified host. Th host address
 * is in network byte order at this point.
 */
int arpdel(unsigned long host)
{
    struct arpreq ar;
    struct sockaddr_in *sin;
  
    bzero((char *)&ar, sizeof(ar));
    ar.arp_pa.sa_family = AF_INET;
    sin = (struct sockaddr_in *) &ar.arp_pa;
    sin->sin_family = AF_INET;
    sin->sin_addr.s_addr = host;

    errno = 0;
    if (ioctl(ioctl_sockid, SIOCDARP, (char *) &ar) < 0){
         if (errno != ENXIO) {
             return(-1 *  errno);
	 }
    }

    return(0);
}

/* Convert an error value into a human readable error message */
char *err2str(int errval)
{
    int err =  (-1 * errval);
    static char tmp[30];

    switch (err) {

      /* Errors related to tunnel */
      case NO_FREE_TUNNEL:
	  return("No free tunnel available");

      case TUNNEL_NOT_FOUND:
	  return("Tunnel for the corresponding mobile node not found");

      /* Errors related to socket */   
      case EPROTONOSUPPORT :
	  return("Protocol or specified type is not supported within the domain");

      case EMFILE :
	  return("Per process descriptor table is full");

      case ENFILE :
	  return("System file table is full");

      case ENOBUFS :
	  return("Insufficient buffer space\n");

      /* Errors related to ioctls */ 
      case EBADF :
	  return("Invaild descriptor");

      case ENOTTY :
	  return("descriptor not associated with character device");

      case EINVAL:
          return("Invalid arguments");

      case ENXIO:
      case ENODEV:
          return("No such device or address");
       
      case ENETUNREACH:
          return("Network unreachable");

      case EEXIST:
          return("Already exists");

      case MN_ENTRY_ABSENT:
          return("Missing Mobile node entry");
 
      case NO_MAPPING:
          return("Bad ARP mapping for mobile node");
 
      default:
          sprintf(tmp, "Reason unclear : %d", err);
          return(tmp);
    }
}

/*
 * Send a gratuitous arp for mobile node. Iface looks like "eth0" etc.
 */

int sendgarp(char *iface)
{
    char ethaddr[10];
    int val;
    
    if ((val == getEthernetAddr(iface, ethaddr)) < 0) {
	return(val);
    } else {
	return(sendGARPpkt(iface, ethaddr, mnHomeAddr, ethaddr));
    }
}

/* Flush the arp table */
int arpflush()
{
    char line[256], addr[20], hwType[8], flags[8];
    char hwAddr[20], mask[20], dev[MAX_IFNAME_LEN];
    int i, found;
    u32  dst;
    FILE *fd;
  
    if ((fd = fopen("/proc/net/arp", "r")) == NULL) {
	mip0dbg(("Could not open /proc/net/arp in arpflush().\n"));
	mip0dbg(("You may need to recompile the kernel to support the\n"));
	mip0dbg(("proc filesystem.\n"));
	return(-EINVAL);
    }
  
    while (fgets(line, sizeof(line), fd) != NULL) {
	if ((sscanf(line, "%s %s %s %s %s %s", addr, hwType, flags,
		    hwAddr, mask, dev) == 6) && 
	    ((dst = inet_addr(addr)) != (-1))) {
	    /* 
	     * TODO: We should ignore the arp entry for the default
	     * router on the foreign network and delete it later
	     * as part of undoColocated state.
	     */

	    /* 
	     * To support mobile routers, do not delete entries in
	     * an ARP cache associated with an interface that should
	     * be ignored for mobility purposes (an interface
	     * belonging to the ignoreIfaceList).
	     */

	    found = FALSE;
	    for (i = 0; i < ignoreIfaceCnt; i++) {
	        if (strcmp(dev, ignoreIfaceList[i]) == 0) found = TRUE;
	    }

	    if (found == FALSE) {
	        arpdel(dst);
	    } else {
	        mip4dbg(("Ignored arp entry for %s on %s in arpflush.\n",
			 ntoa(dst, addr1), dev));
	    }
	}
    }

    fclose(fd);
    return(0);
}

/* 
 * Return the index in the advertisement buffer for an advertisement
 * whose source is ip_src. Returns (-1) if not found.
 */
int isAdvForSrcIPAddrPresent(u32 ip_src)
{
    int i;

    if (advPktBuf.head == advPktBuf.tail)
       return(-1);

    i = advPktBuf.head;
    while (i != advPktBuf.tail) {
        if (advPktBuf.advPkt[i].advSrcIPAddr == ip_src) 
	    return(i);
	else 
	    i = ((i + 1) % MAX_AGENT_ADV_ENTRIES);	    
   }
   return(-1);
}

/*
 * Return the index for an advertisement heard on the given interface.
 * Returns (-1) if not found.
 */
int isAdvForIfaceNamePresent(char *ifaceName)
{
    int i;

    if (advPktBuf.head == advPktBuf.tail)
       return(-1);

    i = advPktBuf.head; 
    while(i != advPktBuf.tail) {
        if (strcasecmp(advPktBuf.advPkt[i].ifaceName, ifaceName) == 0) 
	    return(i);
	else 
	    i = (i + 1) % MAX_AGENT_ADV_ENTRIES;	    
   }
   return(-1);
}

/*
 * Stores the advertisement heard in the Advertisement packet buffer. 
 * AdvBufLock is used while updating and the routine is called by the
 * "secondary" thread (one that monitors agent advertisements).
 */
void putAdvPkt(char *data, int datalen, u32 ip_src, 
   	       char *hw_src, char *ifaceName)
{
    int i, idx;

    pthread_mutex_lock(&(advPktBuf.AdvBufLock));
   
    /* Find a suitable slot */
    if ((idx = isAdvForSrcIPAddrPresent(ip_src)) < 0) {
        i = advPktBuf.tail;
    } else {
        i = idx;
    }

    /* Store it! */
    strcpy(advPktBuf.advPkt[i].ifaceName, ifaceName);
    advPktBuf.advPkt[i].advSrcIPAddr = ip_src;
    memcpy(advPktBuf.advPkt[i].hwAddr, hw_src, MAX_HWADDR_LEN);    
    memcpy(advPktBuf.advPkt[i].data, data, datalen);
    advPktBuf.advPkt[i].dataLen = datalen;
    
    /* If a new slot was used, update head and tail */
    if (idx < 0) {
        advPktBuf.tail = (advPktBuf.tail + 1) % MAX_AGENT_ADV_ENTRIES;
	/* 
	 * We always keep head and tail separated unless the buffer
	 * is empty, even if this dropping an advertisement.
	 */
        if (advPktBuf.tail == advPktBuf.head)
	   advPktBuf.head = (advPktBuf.head + 1) % MAX_AGENT_ADV_ENTRIES;
    }

    pthread_mutex_unlock(&advPktBuf.AdvBufLock);
}

/*
 * Gets an agent advertisement stored in Adv. Pkt. buffer
 * heard on the specified interface.
 * This routine is called by the main thread and also uses
 * AdvBufLock to synchronize access to the shared advertisement buffer.
 */
boolean getAgentAdvByIface(char *pkt, int *pktlen, char *ifaceName,
			   char *hwAddr)
{
    int i;	
 
    pthread_mutex_lock(&(advPktBuf.AdvBufLock));

    if ((i = isAdvForIfaceNamePresent(ifaceName)) < 0) {
	pthread_mutex_unlock(&advPktBuf.AdvBufLock);
	return(FALSE);
    }

    /* read out the advertisement */
    (*pktlen) = advPktBuf.advPkt[i].dataLen;
    memcpy(pkt, advPktBuf.advPkt[i].data, *pktlen);
    memcpy(hwAddr,advPktBuf.advPkt[i].hwAddr, MAX_HWADDR_LEN);	

    /* Keep the buffer contiguous. */
    memcpy(&(advPktBuf.advPkt[advPktBuf.tail - 1]), 
	   &(advPktBuf.advPkt[i]), sizeof(AgentAdvPkt));

    if (advPktBuf.tail == 0)
	advPktBuf.tail = (MAX_AGENT_ADV_ENTRIES - 1);
    else
	advPktBuf.tail--;

    pthread_mutex_unlock(&(advPktBuf.AdvBufLock));
    return(TRUE);
}

/* 
 * Remove all the agent advertisements heard on the specified
 * interface.
 */
void purgeAdvPkts(char *ifaceName)
{
    int i;
 
    pthread_mutex_lock(&(advPktBuf.AdvBufLock));

    i = isAdvForIfaceNamePresent(ifaceName);    
    while (i >= 0) {
        /*
	 * Keep the buffer contiguous by copying the last item
	 * into slot i
	 */
        memcpy(&(advPktBuf.advPkt[advPktBuf.tail - 1]), 
	       &(advPktBuf.advPkt[i]), sizeof(AgentAdvPkt));

	if (advPktBuf.tail == 0)
	    advPktBuf.tail = (MAX_AGENT_ADV_ENTRIES - 1);
	else
	    advPktBuf.tail--;

	i = isAdvForIfaceNamePresent(ifaceName);    
    }

    pthread_mutex_unlock(&(advPktBuf.AdvBufLock));
}

/* 
 * Check whether the ethernet frame in the data buffer is a Mobile IP
 * agent advertisement. When TRUE is returned, srcHWaddr contains
 * the hardware address of the source of the advertisement and
 * avdSrcAddr contains the IP address of the advertisement source.
 */
boolean isAgentAdv(struct strbuf *data, char *srcHWaddr, u32 *advSrcAddr,
		   char isEtherFrame)
{
    int ip_hdr_len, i;
    u32 ip_src;
    int icmp_offset;
    int router_adv_entries;
    int router_adv_entry_size;
    int mobility_ext_offset;
    int careof_entries;
    int extension_offset;
    advExt *advExtPtr;
    int link_hdr_len;

    /* 
     * The pkt should have a link header (14 bytes for ethernet, zero for
     * PPP) and IP headers (20 bytes), at least 8 bytes of ICMP Router 
     * Advertisement and another 8 bytes of mobility extensions. The
     * IP protocol should be 1 (ICMP) and in the case of an ethernet 
     * frame, the Ethernet payload should be 0x0800 (IP).
     */
    if (isEtherFrame) {
        link_hdr_len = 14;
    } else {
        link_hdr_len = 0;
    }
    if ((data->len < (link_hdr_len + 36)) || 
	(data->buf[link_hdr_len+9] != 1) ||
	((link_hdr_len == 14) && 
	 ((data->buf[12] != 8) || (data->buf[13] != 0)))) {
        return(FALSE);
    }

    /* 
     * Find true IP header length. Pkt should be atleast 30+IP_hdr_len
     * bytes and have ICMP type 16 and ICMP code of 0 or 16.
     */ 
    ip_hdr_len = (data->buf[link_hdr_len] & 0xf) << 2;
    icmp_offset = link_hdr_len + ip_hdr_len;
    if ((data->len < (16 + link_hdr_len + ip_hdr_len)) || 
	(data->buf[icmp_offset] != 9) || 
	((data->buf[icmp_offset+1] !=0) && (data->buf[icmp_offset+1] != 16))) {
	/*printf("Expected ICMP type 9, code  0 or 16\n");*/
	return(FALSE);
    }

    router_adv_entries = data->buf[icmp_offset + 4];
    router_adv_entry_size = data->buf[icmp_offset + 5] << 2;
    mobility_ext_offset = icmp_offset + 8 + 
                          (router_adv_entries * router_adv_entry_size);

    /* Should have a mobility extension: type 16, F or H-bit on */
    if ((data->len < (mobility_ext_offset + 8)) ||
	(data->buf[mobility_ext_offset] != 16) ||
	((data->buf[mobility_ext_offset + 6] & 0x30) == 0)) {
        return(FALSE);
    }
	
    /*
     * Ignore advertisements in which the F bit is set but no care-of
     * addresses are listed.
     */

    advExtPtr = (advExt *)(data->buf + mobility_ext_offset);

    if ( (data->buf[mobility_ext_offset + 6] & ADV_IS_FOREIGN_AGENT) &&
        ((((((advExtPtr->length) - 6)/4) <= 0) ||
           ((mobility_ext_offset + 12) > data->len) )))
       return(FALSE);

    /* 
     * Check that if B or R is set then F is also set, i.e. it
     * is illegal to have B or R set for a mobility agent that does not
     * advertise FA service.
     */
    if (((data->buf[mobility_ext_offset + 6]) & 
	 (ADV_IS_BUSY | ADV_REGISTRATION_REQUIRED)) &&
	 ((data->buf[mobility_ext_offset + 6] & ADV_IS_FOREIGN_AGENT) == 0))
      return(FALSE);

    if (link_hdr_len != 0) {
       memcpy(srcHWaddr, (data->buf + 6), 6);
    } else {
       memset(srcHWaddr, 0, 6);
    }

    ip_src = 0;
    for (i = (link_hdr_len + 12); i < (link_hdr_len + 16); i++) 
        ip_src = (ip_src << 8) | (0xff & data->buf[i]);

    *advSrcAddr = htonl(ip_src);
                                                                                       

    careof_entries = data->buf[mobility_ext_offset + 1];
    extension_offset = (mobility_ext_offset + 2 + careof_entries);
    while (data->len > extension_offset) {

        switch (data->buf[extension_offset]) {

            case 0: return(TRUE);

            case 19: if((data->len<(extension_offset + 2 + router_adv_entries))||
                       (router_adv_entries!=data->buf[(extension_offset + 1)])){
                       printf("Problem with Prefix-length Extensions\n");
                       return(FALSE);
                     }
                     extension_offset += (2 + router_adv_entries);
                     break; 
            case 20: if(data->len<(2 + data->buf[(extension_offset + 1)])) {
                       printf("Not enough PFA IP addresses in Hierarchical Mobility \
                               Agent extension\n");
                       return(FALSE);
                     }
                     extension_offset += (2 + data->buf[extension_offset]);
                     break; 
            default: printf("Unknown extension type\n");
                     return(FALSE);
        }
    }

    return(TRUE);
}

/* 
 * Thread routine for monitoring agent advertisements. The argument
 * is never used. This is the function executed by the secondary
 * thread.
 */

void *monitorAgentAdv(void *ignoreThis)
{
    unsigned char buf[MAXDLBUF]; 
    struct strbuf data;
    int	flags;
    int	i, fd, maxfd;
    u32 ip_src;
    char hw_src[MAX_HWADDR_LEN];
    fd_set fdvec;
    struct timeval tv;
    struct sockaddr from;
    int fromlen, len;

    /* 
     * int localIfaceFD[DEFAULT_MAX_INTERFACES];
     * char localIfaceFDName[DEFAULT_MAX_INTERFACES][MAX_IFNAME_LEN]; 
     * int localIfaceFDCnt;  
     */
    
    if ((fd = socket(AF_INET, SOCK_PACKET, htons(ETH_P_ALL))) < 0) {
	mip1dbg(("Error in opening SOCK_PACKET socket for agent adv ... %s\n",
		 err2str(fd)));
	exit(-1);
    }

    data.buf = (char *) buf;
    data.maxlen = MAXDLBUF;
    
    /* Read packets.*/
    while (1) {
        /* 
	 * Copy the name of all interfaces and open file descriptors
	 * to local storage so we don't hold the lock for too long.
	 *
	 *
	 * pthread_mutex_lock(&(ifaceFDList.ifaceFDLock));
	 * localIfaceFDCnt = ifaceFDList.ifaceFDCnt;
	 * for (i = 0; i < ifaceFDList.ifaceFDCnt; i++) {
	 *    localIfaceFD[i] = ifaceFDList.ifaceFDarray[i].ifaceFD;
	 *    strcpy(localIfaceFDName[i],
	 *	     ifaceFDList.ifaceFDarray[i].ifaceName);
	 * }
	 * pthread_mutex_unlock(&(ifaceFDList.ifaceFDLock));
	 */
	
	tv.tv_sec = 0;
	tv.tv_usec = 0;

	FD_ZERO(&fdvec);
	FD_SET(fd, &fdvec);
	maxfd = fd+1;
	
	if ((i = select(maxfd, &fdvec, NULL, NULL, &tv)) < 0) {
	    if ((errno != EINTR) && (errno != ERESTART)) {
		mip0dbg(("select failed but not due to EINTR or ERESTART %d\n", errno));
		perror("reason");
	    } else {
		mip0dbg(("select failed due to EINTR or ERESTART\n"));
	    }
	} else if (FD_ISSET(fd, &fdvec)) {
	    data.len = MAXDLBUF;
	    fromlen = sizeof(struct sockaddr);
	    do {
		len = recvfrom(fd, buf, data.len, 0, &from, &fromlen);
	    } while ((len < 0) && (errno == EINTR));
	    
	    if (len < 0) {
		mip0dbg(("Recvfrom failed in monitorAgentAdv()\n"));
	    } else {
		data.len = len;
	        if (strncmp(from.sa_data, "ppp", strlen("ppp")) == 0) {
		   if (isAgentAdv(&data, hw_src, &ip_src, FALSE)) {
		      putAdvPkt(data.buf, data.len, ip_src, 
				hw_src, from.sa_data); 
		   }
		} else {
		   /*
		    * The last argument to isAgentAdv indicates if the pkt
		    * should be interpreted as an ethernet frame.
		    */ 
		   if (isAgentAdv(&data, hw_src, &ip_src, TRUE)) {
		       /*    mip3dbg(("Got agent adv from %s.\n", ntoa(ip_src, addr1))); */
		       putAdvPkt((data.buf + 14), (data.len - 14), ip_src,
				 hw_src, from.sa_data); 
		   }                                                                           
		}
	    }
	}
    }
}

/*
 * Configure the iface "ifaceName" with IP address "addr"
 * and netmask  "netMask". Bring it up. 
 * Interface is already plumbed, we only need to configure its
 * address and netmask.
 *
 * TODO: To be resolved -- can't assign IP address through ioctl call.
 *       Need to figure out ..
 * For now, we use configiface() instead which is based on system
 * commands.
 */
int ConfigIface(char *ifaceName, u32 addr, u32 netMask)
{
    struct ifreq ifr;
    struct sockaddr_in *sin;
    u32 flags;

    bzero ((char *) &ifr, sizeof(ifr));
    strncpy(ifr.ifr_name, ifaceName, sizeof(ifr.ifr_name));

    if (ioctl(ioctl_sockid, SIOCGIFFLAGS, (caddr_t)&ifr) < 0){
	  return(-1 * errno);
    }
    flags = ifr.ifr_flags;

    /*
    ifr.ifr_flags &= ~(IFACE_UP);
    if (ioctl(ioctl_sockid, SIOCSIFFLAGS, (caddr_t)&ifr) < 0){
	  return(-1 * errno);
    }
    */

    /* Set the IP Address */
    sin = (struct sockaddr_in *)&ifr.ifr_addr;
    sin->sin_addr.s_addr = addr;

    if (ioctl(ioctl_sockid, SIOCSIFADDR, (caddr_t)&ifr) < 0){
	mip5dbg(("Error in setting address %s for %s\n",
		 ntoa(addr, addr1), ifaceName));
	return(-1 * errno);
    }

    /* Set the netmask */
    sin = (struct sockaddr_in *)&ifr.ifr_addr;
    sin->sin_addr.s_addr = netMask;

    if (ioctl(ioctl_sockid, SIOCSIFNETMASK, (caddr_t)&ifr) < 0){
	mip5dbg(("Error in setting address %s for %s\n",
		 ntoa(netMask, addr1), ifaceName));
	return(-1 * errno);
    }

    /* Bring up the interface */
    ifr.ifr_flags = flags | IFACE_UP;
    if (ioctl(ioctl_sockid, SIOCSIFFLAGS, (caddr_t)&ifr) < 0){
	  return(-1 * errno);
    }

    return(0);
}

/*
 * Converts OS-specific flags into a canonical OS-independent form
 * as defined in the user level code (mobile.h).
 * TODO: Check if the flags should be u16 instead of u32.
 */
void adjustIfaceFlags(u32 *newFlag, u32 oldFlag)
{
    *newFlag = (oldFlag & 0x00ff);

    /*
     * As of Linux 2.0.30, there does not appear to be a DHCP flag
     * if (oldFlag & IFF_DHCPRUNNING)
     *    (*newFlag) |= IFACE_DHCP;
     */
    
    /* 
     * TODO: Set IFACE_ON_NET only if other address is reachable.
     * This requires ping or something similar. For now, assume that
     * this flag is always set.
     */ 
    (*newFlag) |= IFACE_ON_NET;     
}

/*
 * An alternative to the above configIface() 
 * Configure the iface "ifaceName" with IP address "addr"
 * and netmask  "netMask". Bring it up. 
 * Interface is already plumbed, we only need to configure its
 * address and netmask.
 * TODO: We need to be very careful on how we proceed in the case
 *       of point-to-point interfaces.
 */
int configIface(char *ifaceName, u32 addr, u32 netMask)
{
    u32 bcastaddr;
   /* 
    sprintf(msg, "/sbin/ifconfig %s down", ifaceName);
    system(msg);
*/
    
    /* Solaris automatically sets the broadcast address. */
    mip5dbg(("Bringing up interface %s with address %s mask %s\n",
	     ifaceName, ntoa(addr, addr1), ntoa(netMask, addr2)));
    /* 
     * TODO: For Linux, make sure the broadcast address is set
     * properly.
     */
/*
    bcastaddr = (addr & netMask) | (~netMask);
    sprintf(msg, "/sbin/ifconfig %s %s netmask %s broadcast %s up", 
	    ifaceName, ntoa(addr, addr1), ntoa(netMask, addr2),
	    ntoa(bcastaddr, addr3));
*/
    sprintf(msg, "/sbin/ifconfig %s %s netmask %s;\
				  /sbin/route add -net %s netmask %s dev %s",
	    		ifaceName, ntoa(addr, addr1), ntoa(netMask, addr2),
	    		ntoa((addr & netMask), addr3), ntoa(netMask, addr4), ifaceName);

    mip3dbg(("Configuring interface (%s) ... \n", msg));
    system(msg);

#ifdef LINUX_2_0   
    /* For Linux, we must explicitly add the route for the local subnet */
    sprintf(msg, "/sbin/route add -net %s netmask %s dev %s", 
	    ntoa((addr & netMask), addr1), ntoa(netMask, addr2), ifaceName);
    mip3dbg(("Adding subnet route (%s) ... \n", msg));
    system(msg);
#endif            
    
    return(0);
}

/* 
 * Get the logical clone of the interface specified in iface
 * e.g. Logical clone of le0 under Solaris will be le0:1
 * TODO : Make sure that the logical clone is not already present
 *        But if we add this logic then we have to be careful in
 *        mnState.c because the same function is also used in
 *        unconfigCloneForIface.
 */
void mkCloneIface(char *iface, char *clone)
{
        strcpy(clone, iface);
        strcat(clone, ":1");
}

/*                                                    
 * Assign the specified address and netmask to the clone interface.
 * Also retain the parent's interface flags (as far as possible).
 * This function is also responsible for deleting any spurious routing
 * entries that may be added as part of bringing up the clone Iface.
 */
int configCloneForIface(char *parentIface, u32 addr, u32 netMask)
{
#ifdef LINUX_2_0   
    char cloneIface[MAX_IFNAME_LEN + 2];             
    struct ifreq ifr, cloneifr;
    u32 flags;
    int errcode;

   mkCloneIface(parentIface, cloneIface);
                                                                                     
    /*
     * THINK SOME MORE ABOUT THIS ORDERING ... TEST IT!
     * TODO: We should assign the address etc from parent iface
     * by explicitly calling ioctls.
     * 
     *  Previously, this used to come later ...
     */
    if ((errcode = configIface(cloneIface, addr, netMask)) < 0)
        return(-1);

    bzero ((char *) &ifr, sizeof(ifr));
    strncpy(ifr.ifr_name, parentIface, sizeof(ifr.ifr_name));

    bzero ((char *) &cloneifr, sizeof(cloneifr));
    strncpy(cloneifr.ifr_name, cloneIface, sizeof(cloneifr.ifr_name));

    if (ioctl(ioctl_sockid, SIOCGIFFLAGS, (caddr_t)&ifr) < 0) {
        return(-1 * errno);
    }

    adjustIfaceFlags(&flags, ifr.ifr_flags);

    /* copy all flags to the clone except for DHCP */
    cloneifr.ifr_flags = ifr.ifr_flags & ~IFF_DHCPRUNNING;
    if (ioctl(ioctl_sockid, SIOCSIFFLAGS, (caddr_t) &cloneifr) < 0)
        return(-1 * errno);
 
    if (flags & IFACE_POINTOPOINT) { 
        /* Assign parent iface's dstaddr to the clone */
        if ((ioctl(ioctl_sockid, SIOCGIFDSTADDR, (caddr_t) &ifr)) < 0)
	    return(-1 * errno);

        cloneifr.ifr_dstaddr = ifr.ifr_dstaddr;
        if ((ioctl(ioctl_sockid, SIOCSIFDSTADDR, (caddr_t) &cloneifr)) < 0)
	    return(-1 *  errno);
    }

    /* 
     * TODO: This should be part of configCloneIface since which
     * route to delete really depends on whether or not the interface
     * is pointopoint. We need to be careful how we deal with pointopoint
     * interfaces.
     */

    if ((errcode = routedel((addr & netMask), netMask, 
			    addr, 0)) < 0) {
      mip2dbg(("Error in routedel %s %s ... %s\n", 
	       ntoa((addr & netMask), addr1),
	       ntoa(addr, addr2), err2str(errcode)));
    }

   return(0);
#else
   char cmd[128];

   sprintf(cmd, "/sbin/ip addr add %s/32 dev %s", ntoa(addr, addr1),
	   parentIface);
   mip2dbg(("Adding home address to local interface (%s) ... \n", cmd));
   system(cmd);
   return (0);
#endif   
}

/*
 * TODO: Clean up the different ways in which configCloneForIface
 * and unconfigCloneForIface find out the MN home address ... one
 * uses procedure arguments, the other uses a global variable.
 */ 
void unconfigCloneForIface(char *ifaceName) {
#ifdef LINUX_2_0
    char cloneName[20];
    char cmd[80];
   
    mkCloneIface(ifaceName, cloneName);
    sprintf(cmd, "/sbin/ifconfig %s 0.0.0.0 down", cloneName);
    mip3dbg(("Unconfiguring clone interface (%s) ... \n", cmd));
    system(cmd);
#else
    char cmd[80];

    sprintf(cmd, "/sbin/ip addr del %s/32 dev %s", ntoa(mnHomeAddr, addr1), ifaceName);
    mip3dbg(("Removing home address from local interface (%s) ... \n", cmd));
    system(cmd);
#endif    
}

int changeSrcOnDefaultRoute(u32 homeAddress, u32 router) {
   char cmd[128];
   
#ifdef LINUX_2_0
   /* 
    * Do we really need this on Linux 2.0, don't we already have the
    * default route from before??
    */ 
//   return (routeadd(0, 0, router, 1));
#else
   sprintf (cmd, "/sbin/ip route delete default");
   mip3dbg(("Deleting existing default route (%s) ... \n", cmd));
   system(cmd);

   sprintf (cmd, "/sbin/ip route add default via %s src %s", 
	    ntoa(router, addr1), ntoa(homeAddress, addr2));
   mip3dbg(("Adding new default route (%s) ... \n", cmd));
   system(cmd);
   return 0;
#endif   
}

/* 
 * This function is called from closeIfaceFD which should make
 * sure it acquires appropriate locks before accessing the shared
 * data structure ifaceFDList.
 */
int findIfaceFDIdx(char *ifname) {
    int i;
    for (i = 0; i < ifaceFDList.ifaceFDCnt; i++) {
       if (strcmp(ifaceFDList.ifaceFDarray[i].ifaceName, ifname) == 0)
	    return(i);
    }
    return(-1);		
}

/* 
 * Opens a stream to receive agent advertisements and adds the stream
 * and corresponding interface name to ifaceFDList. Access to shared
 * data structure ifaceFDList should be protected by a lock.
 */
void addIfaceFD(char *ifname)
{
    int fd;
    struct ifreq ifr;
    struct ip_mreq imr;
    
    pthread_mutex_lock(&(ifaceFDList.ifaceFDLock));
    if (ifaceFDList.ifaceFDCnt == DEFAULT_MAX_INTERFACES) {
        mip1dbg(("No space for agent adv. stream on iface %s.\n"));
    } else {
	/* 
	 * LINUX provides a socket type called SOCK_PACKET.
	 * Sockets of this type can receive ethernet frames arriving
	 * on any device. It is also possible to receive frames from
	 * a specific device by binding the socket to that device
	 * name (set the sa_data field in the sockaddr structure
	 * to contain the device name).
	 * After a successful recvfrom() on a SOCK_PACKET socket,
	 * the sa_data field in the "from" sockaddr contains the
	 * device name on which the frame arrived.
	 * Instead of multiple SOCK_PACKET sockets (one per device),
	 * we use a single socket created in monitorAgentAdv().
	 * Preliminary testing on a Linux 2.0.30 kernel reveals that
	 * SOCK_PACKET sockets bound to specific addresses stop
	 * receiving pkts when the corresponding device (e.g. PCMCIA
         * ethernet card) is removed but do not resume packet reception
	 * when the device is recreated. An unbound SOCK_PACKET socket
	 * does not suffer from this behavior. -- VG
	 */

	fd = 0;
	ifaceFDList.ifaceFDarray[ifaceFDList.ifaceFDCnt].ifaceFD = fd;
	strcpy(ifaceFDList.ifaceFDarray[ifaceFDList.ifaceFDCnt].ifaceName,
	       ifname);
	ifaceFDList.ifaceFDCnt++;  

	/* 
	 * TODO: set promiscuous bit on for this interface. This should
	 * not be needed but without it, the kernel has trouble picking
	 * up multicast agent advertisements.
	 */ 

	bzero ((char *) &ifr, sizeof(ifr));
	strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
    	if (ioctl(ioctl_sockid, SIOCGIFFLAGS, (caddr_t)&ifr) < 0) {
	    mip4dbg(("Error while using  SIOCGIFFLAGS in addIfaceFD ... %s\n",
		     err2str(-1 * errno)));
	}

	ifr.ifr_flags |= IFF_PROMISC;
	strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
    	if (ioctl(ioctl_sockid, SIOCSIFFLAGS, (caddr_t)&ifr) < 0) {
	    mip4dbg(("Error while using  SIOCSIFFLAGS in addIfaceFD ... %s\n",
		     err2str(-1 * errno)));
	}

	/* 
	 * Another (better) alternative is to subscribe to 224.0.0.1 
	 * multicast group and see if that helps us pick up advertisements
	 * This does not seem to work well.
	 */
	/*
	imr.imr_multiaddr.s_addr = inet_addr("224.0.0.1");
	imr.imr_interface.s_addr = addr;
	if (setsockopt(ioctl_sockid, IPPROTO_IP, IP_ADD_MEMBERSHIP,
		       (char *) &imr, sizeof(struct ip_mreq)) < 0) {
	    perror("Could not join multicast group 224.0.0.1");
	}
	 */
    }
    pthread_mutex_unlock(&(ifaceFDList.ifaceFDLock));
    mip5dbg(("addIfaceFD() called for interface %s\n", ifname));
}


/* 
 * Closes the agent advertisement stream corresponding to the specified
 * iface. Access to shared data structure ifaceFDList should be protected
 * by a lock.
 */

void closeIfaceFD(char *ifname)
{
    int idx;

    pthread_mutex_lock(&(ifaceFDList.ifaceFDLock));
    if ((idx = findIfaceFDIdx(ifname)) < 0)
	mip0dbg(("Did not find an agent adv. stream for iface %s\n", ifname));
    else {
       /*  close(ifaceFDList.ifaceFDarray[idx].ifaceFD); */
       ifaceFDList.ifaceFDCnt--;
       /* make the list contiguous */
       if ((ifaceFDList.ifaceFDCnt != idx) && 
	   (ifaceFDList.ifaceFDCnt != 0)) {
          ifaceFDList.ifaceFDarray[idx].ifaceFD =  
	      ifaceFDList.ifaceFDarray[ifaceFDList.ifaceFDCnt].ifaceFD;
	  strcpy(ifaceFDList.ifaceFDarray[idx].ifaceName,
		 ifaceFDList.ifaceFDarray[ifaceFDList.ifaceFDCnt].ifaceName);
       } 
    }
    pthread_mutex_unlock(&(ifaceFDList.ifaceFDLock));

    /* 
     * TODO: Purge the advertisements heard on this interface. By releasing
     * the ifaceFDLock before purgeAdvPkts we avoid the potential
     * problems caused by holding nested locks.
     */
    if (idx >= 0) {
        purgeAdvPkts(ifname);
    }
    mip5dbg(("closeIfaceFD called for %s.\n", ifname));
}

/*
 * Checks whether the interface is useful (either for monitoring
 * agent advertisements or using a co-located care-of address).
 *
 * Checks the following :
 *   The interface does not belong to the list of interfaces to be ignored.
 *   The interface name does not have colon(:) in it.
 *   The interface is UP/not loopback ..
 *   More tests on flags to be made
 *
 *  TODO:  Need to make it more strict
 */
boolean isIfaceAcceptable(char *name, unsigned short flags)
{
   int i, seq;

   for (i = 0; i < ignoreIfaceCnt; i++) {
       if (strcmp(name, ignoreIfaceList[i]) == 0)
           return(FALSE);
   }

   if (strchr(name, ':') != (char *) NULL) {
      return(FALSE);
   }
   if ((flags & IFACE_UP) == 0) {
      return(FALSE);
   }

   if (flags & IFACE_LOOPBACK) {
      return(FALSE);
   }

   if (strncmp(name, "dummy", strlen("dummy")) == 0) {
      return(FALSE);
   }

    if (strncmp(name, "tunl", strlen("tunl")) == 0) {
      return(FALSE);
   }

   // if one interface is at home, then other interfaces are ignored, for MSB
   seq = getIfaceAtHome(ISTable);
   if((seq >= 0) && (strncmp(name, ISTable[seq].ifaceName, strlen(name)) != 0))
   	return (FALSE);

   return(TRUE);
}


/*
 * Used to compare interface information collected locally against
 * that in the user-level code.
 */
boolean IfaceMatch(IfaceEntry iface1, IfaceEntry iface2)
{
    if  ((strcmp(iface1.maIfaceName, iface2.maIfaceName) != 0) ||
	 (iface1.maIfaceAddr != iface2.maIfaceAddr) ||
         (iface1.maIfaceMask != iface2.maIfaceMask))
	return(FALSE);
    else
	return(TRUE);
}

/*
 * Compile a list of interfaces that may be of interest to the user
 * level mobile node code. Return that information in the parameters
 * passed.
 */
int getNewIfaceList(IfaceEntry *ifaceTable, int *size)
{
    int n, i, j, k, read = 0, int1, found;
	char *buf;
    char line[256], path[256], localMA[16], maHaddr[18];
	u32 mode, channel, sStrength, sQuality;
    FILE *fd;
    unsigned int temp;
    struct sockaddr_in *sin;
    int numifs;
    struct ifconf ifc;
    struct ifreq ifr;
    register struct ifreq *ifrp;
    u32 ipAddr;
    u32 netMask;
    u32 otherAddr = 0;
    u32 flags;
    unsigned char hwAddr[MAX_HWADDR_LEN];
    IfaceEntry localIfaceTable[DEFAULT_MAX_INTERFACES];
    int localIfaceTableSize;
    boolean localIfaceMarked[DEFAULT_MAX_INTERFACES];
    boolean globalIfaceMarked[DEFAULT_MAX_INTERFACES];

    LocalMaEntry *lmaep;
    unsigned char state;
    struct timeval time;
    int ISSeq;


    /* Get the no. of interfaces
     * if (ioctl(ioctl_sockid, SIOCGIFNUM, (char *)&numifs) < 0) {
     *     mip4dbg(("Error while using SIOCGIFNUM ... %s\n",
     * 		   err2str(-1 * errno)));
     * return(-1 * errno);
     * }
     *
     * Linux does not seem to support SIOCGIFNUM so we count the
     * the number of lines in /proc/net/dev instead.
     */

    if ((fd = fopen("/proc/net/dev", "r")) == NULL) {
	mip0dbg(("Could not open /proc/net/dev in getNewIfaceList().\n"));
	return(-EINVAL);
    }

    numifs = 0;
    while (fgets(line, sizeof(line), fd) != NULL) numifs++;
    fclose(fd);

    /* Get the name of each interface */
    temp = numifs * sizeof(struct ifreq);
    if ((buf = (char *) malloc(temp)) == (char *) NULL) {
	mip4dbg(("Error in malloc \n"));
	return(-1 * errno);
    }

    ifc.ifc_len = temp;
    ifc.ifc_buf = buf;
    if (ioctl(ioctl_sockid, SIOCGIFCONF, (char *) &ifc) < 0) {
	mip4dbg(("Error while using SIOCGIFCONF ... %s\n",
		 err2str(-1 * errno)));
	free(buf);
	return(-1 * errno);
    }

    /*
     * Compile information about each interface.
     */
    ifrp = ifc.ifc_req;
    i = 0;
    k = 0;
    for (n = ifc.ifc_len / sizeof (struct ifreq); n > 0; n--, ifrp++) {
	/* Get the flags */
	bzero ((char *) &ifr, sizeof(ifr));
	strncpy(ifr.ifr_name, ifrp->ifr_name, sizeof(ifr.ifr_name));
    	if (ioctl(ioctl_sockid, SIOCGIFFLAGS, (caddr_t)&ifr) < 0) {
	    mip4dbg(("Error while using  SIOCGIFFLAGS ... %s\n",
		     err2str(-1 * errno)));
	    continue;
	}

 	adjustIfaceFlags(&flags, ifr.ifr_flags);
 	if (!isIfaceAcceptable(ifrp->ifr_name, flags)) continue;

	/* The interface is of potential interest ... get the IP address */

	if (ioctl(ioctl_sockid, SIOCGIFADDR, (caddr_t)&ifr) < 0) {
	    mip4dbg(("Error while using  SIOCGIFADDR ... %s\n",
		     err2str(-1 * errno)));
	    continue;
	}
	sin = (struct sockaddr_in *) &ifr.ifr_addr;
	ipAddr = sin->sin_addr.s_addr;

	/*
	 * TODO: We could also try to check for IP addresses that
	 * belong to our home network but differ from homeAddr.
	 */
	if (ipAddr == 0) {
	    mip5dbg(("Ignoring interface %s with zero address.\n",
		     ifr.ifr_name));
	    continue;
	}

	/*
	 * Get the netmask
	 * TODO: We could do some sanity checks for the netmask but for
	 * now we don't.
	 */
	if (ioctl(ioctl_sockid, SIOCGIFNETMASK, (caddr_t)&ifr) < 0) {
	    mip4dbg(("Error in using SIOCGIFNETMASK ... %s\n",
		     err2str(-1 * errno)));
	    continue;
	}
	sin = (struct sockaddr_in *)&ifr.ifr_addr;
	netMask =  sin->sin_addr.s_addr;

	/*
	 * Get Status, only if applicable.
	 */
	if (movementDetection == FHCS) {
//	if (1) {
		sprintf(path, "/proc/aironet/%s/Status", ifr.ifr_name);
    	if ((fd = fopen(path, "r")) != NULL) {
    		while ((fgets(line, sizeof(line), fd) != NULL) && (read < 4)) {
				read ++;
				buf = line;

				switch (read) {
					case 1:
						buf += strlen("Mode: ");
						if (sscanf(buf, "%x", &mode) != 1)
							fprintf(stderr, "Can't get mode: %s\n", buf);
						continue;

					case 2:
						buf += strlen("Signal Strength: ");
						if (sscanf(buf, "%ud", &sStrength) != 1)
							fprintf(stderr, "Can't get signal strength: %s\n", buf);
						continue;

					case 3:
						buf += strlen("Signal Quality: ");
						if (sscanf(buf, "%ud", &sQuality) != 1)
							fprintf(stderr, "Can't get signal quality: %s\n", buf);
						continue;

					case 4:
						buf += strlen("SSID: ");
						if (sscanf(buf, "%s %s", localMA, maHaddr) != 2)
							fprintf(stderr, "Can't get SSID: %s\n", buf);
						continue;

					default:
						fprintf(stderr, "Invalid line: %s\n", line);
				}
			}
   			fclose(fd);
		}
		if (fd == NULL) {
				mip0dbg(("Aironet card not present\n"));
				mip0dbg(("Switching from FHCS to HCS Move Detection algorithm\n"));
				movementDetection = HCS;
		}
	}

	/* Get the hardware address only if appropriate */
	if ((flags & IFACE_NOARP) == 0) {
		if ((i = getEthernetAddr(ifr.ifr_name, hwAddr)) < 0) {
			mip5dbg(("Error in getting hardware address for %s ... %s\n",
				ifr.ifr_name, err2str(i)));
			continue;
        }
	}

	/*
	 * Get the "otherAddr" for use as router if ipAddr is usable
	 * as a colocated care-of address.
	 */
	otherAddr = 0;
	if (ipAddr != mnHomeAddr) {
             mip5dbg(("Iface %s has colocated addresss - %s\n",
		      ifr.ifr_name, ntoa(ipAddr,addr1)));
	    /* Get the Destination address in case it is point to point */
	    if (flags & IFACE_POINTOPOINT) {
	        if ((ioctl(ioctl_sockid, SIOCGIFDSTADDR, (caddr_t)&ifr)) < 0) {
		    mip2dbg(("Could not get destination address for " \
			     "pointtopoint interface %s ... %s",
			     ifr.ifr_name, err2str(-1 * errno)));
		    continue;
		} else {
		    sin = (struct sockaddr_in *) &ifr.ifr_dstaddr;
		    otherAddr = sin->sin_addr.s_addr;
		    if ((otherAddr == 0) || (otherAddr == ipAddr))
		        continue;
		    mip5dbg(("Got pointtopoint dst address %s for %s\n",
			     ntoa(otherAddr,addr1),
			     ifr.ifr_name));
		}
	    } else {
	        /*
		 * In all other cases, we pick up a default router
		 * from the routing table.
		 * TODO: For dhcp interfaces, we should ideally pick up
		 *  the default router using dhcpinfo.
		 */
	        otherAddr = getRouterAddr(ipAddr, netMask);

		if ((otherAddr == 0) && !(isReverseTunnelSet)) {
		    mip5dbg(("Could not get appropriate default router for "\
			     "iface %s.\n", ifr.ifr_name));
		    continue;
		}
	    }
	}

	strcpy(localIfaceTable[k].maIfaceName, ifr.ifr_name);
	memcpy(localIfaceTable[k].maIfaceHWaddr, hwAddr, MAX_HWADDR_LEN);
	localIfaceTable[k].maIfaceAddr = ipAddr ;
	localIfaceTable[k].maIfaceOtherAddr = otherAddr;
	localIfaceTable[k].maIfaceMask = netMask ;
	localIfaceTable[k].maIfaceFlags = flags;
	localIfaceTable[k].maIfaceSolicitationTime = 0 ;
	if (movementDetection == FHCS) {
//	if (1) {
		localIfaceTable[k].localMA = inet_addr(localMA);
		hwAddrRead(localIfaceTable[k].maHaddr, maHaddr);
		localIfaceTable[k].mode = mode;
		localIfaceTable[k].sStrength = sStrength;
		localIfaceTable[k].sQuality = sQuality;
	}

	k++;
    }

    /* Free memory no longer needed */
    free(buf);



    /*
     * Now compare the local list with that passed as argument. Entries
     * found in both lists are marked.
     */
    localIfaceTableSize = k;
    for (i = 0; i < *size; i++)
       globalIfaceMarked[i] = FALSE;

    for (j = 0; j < localIfaceTableSize; j++)
       localIfaceMarked[j] = FALSE;


    /*
     * Update the InterfaceStateTable if new interfaces have
     * shown up
     */
    for(i = 0; i < localIfaceTableSize; i++) {
    	found = FALSE;
	for(j = 1; j < DEFAULT_MAX_INTERFACES; j++) {
    		if( *(ISTable[j].ifaceName) != (char) NULL
			&& strcmp(localIfaceTable[i].maIfaceName, ISTable[j].ifaceName) == 0) {
			found = TRUE;
			break;
		}
	}
	if(!found) {
		// find a blank slot to update
		for(j = 1; j < DEFAULT_MAX_INTERFACES; j++) {
			if(*(ISTable[j].ifaceName) == (char) NULL) {
				strcpy(ISTable[j].ifaceName, localIfaceTable[i].maIfaceName);
				ISTableSize++;
				break;
			}
		}
//		setFilterForNewBinding(mnFilters, mnFilterCount, mnFilterOpTable, 
//			&mnFilterOpCount, ISTable[j].ifaceName, ACT_REPLACE, regReqTable);
		// send info to connected GUIs
		sendInfoToAllGUIs();
	}
    }

    /*
     * Remove interfaces not available
     */
    for(i = 1; i < DEFAULT_MAX_INTERFACES; i++) {
		if(*(ISTable[i].ifaceName) != (char) NULL) {
			found = FALSE;
			for(j = 0; j < localIfaceTableSize; j++) {
				if(strcmp(ISTable[i].ifaceName, localIfaceTable[j].maIfaceName) == 0) {
					found = TRUE;
					break;
				}
			}
			if(!found) {
				setFilterForNewBinding(mnFilters, mnFilterCount, mnFilterOpTable, 
					&mnFilterOpCount, ISTable[i].ifaceName, ACT_DELETE, regReqTable);

				memset(ISTable[i].ifaceName, (char) NULL, MAX_IFNAME_LEN);
				ISTable[i].COAddr = 0;

				ISTable[i].currentState.state = UNKNOWN_STATE;
				ISTable[i].currentState.backoffIdx = 0;
				memset(ISTable[i].currentState.ifaceName, (char) NULL, MAX_IFNAME_LEN);
				ISTable[i].currentState.routerAddr = 0;
				ISTable[i].currentState.COAddr = 0;
				ISTable[i].currentState.timeSpent = 0;
				ISTable[i].currentState.nextSolicit = (retransmissionPolicy[0] * 1);
				ISTableSize--;

				// send info to connected GUIs
				sendInfoToAllGUIs();
			}
		}
    }


    for (i = 0; i < *size; i++) {
        for (j = 0; j < localIfaceTableSize; j++) {
	    	if (IfaceMatch(ifaceTable[i], localIfaceTable[j])) {

			// get seq related to interface in InterfaceState Table
			ISSeq = getIfcStateSeqFromName(localIfaceTable[j].maIfaceName, ISTable);

	        	localIfaceTable[j].maIfaceSolicitationTime =
		     		ifaceTable[i].maIfaceSolicitationTime ;
				localIfaceTable[j].maIfaceOtherAddr =
		     		ifaceTable[i].maIfaceOtherAddr;
				globalIfaceMarked[i] = TRUE;
				localIfaceMarked[j] = TRUE;
/*
				if (localIfaceTable[j].mode != ifaceTable[i].mode) {
					mip0dbg(("\n---- %s ---- mode change: %x\n",
						printTime(currentTime, 30), mode));
				}
*/

				if (movementDetection < HCS) continue;
				if (localIfaceTable[j].mode != CONNECTED) {
					localIfaceTable[j].localMA = ifaceTable[i].localMA;
					strcpy (localIfaceTable[j].maHaddr, ifaceTable[i].maHaddr);
					continue;
				}

				if ((localIfaceTable[j].localMA != ifaceTable[i].localMA) ||
					(ISTable[ISSeq].currentState.state == UNKNOWN_STATE)) {

					if ((lmaep = (LocalMaEntry *) findLocalMA(ifaceTable[i].localMA))
									== (LocalMaEntry *) NULL) {
						if((lmaep = (LocalMaEntry *) newLocalMA(ifaceTable[i].localMA))
									== (LocalMaEntry *)NULL) {
			        		mip3dbg(("Error: newLocalMA returned NULL"));
							exit(-1);
						}
					}

					if (movementDetection == HCS) {
						mip0dbg(("\n---- %s ---- sent solicitation\n",
							printTime(currentTime, 30)));
   	        			sendSolicitation(localIfaceTable[j].maIfaceAddr);
						continue;
					}
					lmaep->mnAdvSourceAddr = localIfaceTable[j].localMA;
					memcpy(lmaep->mnAdvSrcHWaddr, localIfaceTable[j].maHaddr,
								MAX_HWADDR_LEN);
					lmaep->mnAdvMaxAdvLifetime = 1000;
					lmaep->mnAdvMaxRegLifetime = 1000;
					lmaep->mnAdvRemainingLifetime = 1000;
					lmaep->mnAdvSeqNum = 1;
					lmaep->mnAdvLastSeqNum = 0;
					lmaep->mnAdvFlags = 0;
					lmaep->mnAdvRouterCnt = 0;
					lmaep->mnAdvCOACnt = 1;
					lmaep->mnAdvRouterList[0] = (u32) 0;
					lmaep->mnAdvCOAList[0] = localIfaceTable[j].localMA;

					mip0dbg(("\n---- %s ---- move detection start\n",
						printTime(currentTime, 30)));

					if (((lmaep->mnAdvSourceAddr ^ mnHomeAddr) & mnHomeNetmask) == 0) {
						updateState(&(ISTable[ISSeq].currentState), HOME_STATE, 0, 0, ISSeq);

						// set filter removal with the new deregistration
						// when comming back to Home
						setFilterForHomeBinding(mnFilters, mnFilterCount, mnFilterOpTable,
							&mnFilterOpCount, ISTable[ISSeq].currentState.ifaceName);

						startHomeAction(&(ISTable[ISSeq].currentState), ISSeq);
						deleteAllFilters(mnFilters);

						// send info to connected GUIs
						sendInfoToAllGUIs();

					} else {
						updateState(&(ISTable[ISSeq].currentState), FOREIGN_STATE, 0, 0, ISSeq);

        					// set filter retransmission with the new binding
						// as FA CoA
						setFilterForNewBinding(mnFilters, mnFilterCount, 
							mnFilterOpTable, &mnFilterOpCount, 
							ISTable[ISSeq].currentState.ifaceName, ACT_REPLACE,
							regReqTable);

						startForeignAction(&(ISTable[ISSeq].currentState), lmaep, ISSeq);
						applyReverseFilters(mnFilters, ISTable,
									ISTable[ISSeq].currentState.ifaceName);

						// send info to connected GUIs
						sendInfoToAllGUIs();

					}
					mip0dbg(("\n---- %s ---- move detection end\n",
						printTime(currentTime, 30)));
				}
		   	}
		}
	}

    /*
     * Unmarked entries in the global list result in agent-monitoring
     * streams being closed. Unmarked entries in the locallist result
     * in new advertisement-monitoring streams.
     */

    for (i = 0; i < *size; i++) {
        if (globalIfaceMarked[i] == FALSE) {
    /*
     * When subnet drivers determines that we have lost access to a
     * link and LazySoliciting is 1 we solicit to detect new MAs.
     * Lazy Solicitation.
     */
            if(LazySoliciting) sendSolicitation(ifaceTable[i].maIfaceAddr);
	    closeIfaceFD(ifaceTable[i].maIfaceName);
	}
    }

    for (j = 0; j < localIfaceTableSize; j++) {

    /*
     * When subnet drivers determines that we have access to a new
     * link and LazySoliciting is 1 we solicit to detect new MAs.
     * Eager Solicitation.
     */

        if (localIfaceMarked[j] == FALSE) {
            if(EagerSoliciting) sendSolicitation(ifaceTable[i].maIfaceAddr);
	    addIfaceFD(localIfaceTable[j].maIfaceName);
	}
    }

    memcpy(ifaceTable, localIfaceTable,
	   localIfaceTableSize*sizeof(IfaceEntry));
    *size = localIfaceTableSize;


    return(0);
}

/* Restores Arp */
/* TODO : To be implemented */
int restoreArp()
{
        return(0);
}

/* Disables arp */
/* TODO : To be implemented */
int disableArp(u32 mnAddr, u32 foreignRouter)
{
       return(0);
}

/* OS dependent initialisation routine */
int OSinit()
{
    pthread_attr_t detached_attr;
    int errcode, dummy, i;
    char cmd[80];

    /* Initialise the mutexes */
    pthread_mutex_init(&(advPktBuf.AdvBufLock), NULL);
    pthread_mutex_init(&(ifaceFDList.ifaceFDLock), NULL);

    /*
     * Initialize the array which maintains a mapping between tunnel
     * numbers and mobile node addresses.
     */

    for (i = 0; i <= MAX_TUNNEL_SUPPORTED; i++) {
	mnaddr_tunl[i] = 0;
    }

    /* Inititalise the circular buffer for storing agent advertisements */
    advPktBuf.head = advPktBuf.tail = 0;

    /* Make a socket for the various SIOCxxxx ioctl commands */
    if ((ioctl_sockid = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	mip0dbg(("Error: could not create socket for SIOCxxxx commands.\n"));
	exit(-1);
    }

    /*
     * Enable decapsulation service for mobile node's home address if it
     * appears in the inner datagram as dst.
     * For LINUX 2.0, simply load the ipip handler and enable ip forwarding
     * For Linux 2.2, we only need to enable ip forwarding
     */
	/*
	 * UoB have commented out the LINUX_2_0 specific code
	 */

//#ifdef LINUX_2_0
    sprintf(cmd, "/sbin/insmod ipip");
    mip0dbg(("Loading IPIP handler (%s) ... \n", cmd));
    system(cmd);
//#else
    sprintf(cmd, "/sbin/ifconfig tunl0 0.0.0.0 up");
    mip0dbg(("Enabling IPIP decapsulation (%s) ... \n", cmd));
    system(cmd);
//#endif

    sprintf(cmd, "echo 1 > /proc/sys/net/ipv4/ip_forward");
    mip0dbg(("Enabling IP forwarding (%s) ... \n", cmd));
    system(cmd);

    getNewIfaceList(ifaceTable, &ifaceTableSize);

    /* Create the secondary thread to monitor agent advertisements */
    pthread_attr_init(&detached_attr);
    pthread_attr_setdetachstate(&detached_attr, PTHREAD_CREATE_DETACHED);

    if ((pthread_create(&monitorIface,
			&detached_attr,
			monitorAgentAdv,
			(void *) &dummy) != 0)) {
	mip1dbg(("Error while creating advertisement-monitoring thread.\n"));
	exit(-1);
    }
    return(0);
}


void OScleanup()
{
    int i;
    char cmd[80];
    
	/*
	 * UoB have commented out the LINUX_2_0 specific code
	 */

//#ifdef LINUX_2_0
    sprintf(cmd, "/sbin/rmmod ipip");
    mip0dbg(("Unloading IPIP handler (%s) ... \n", cmd));
    system(cmd);
//#endif

    /* 
     * Free all tunnels that may still be up
     * NOTE: We assume this program only uses tunnels 1 through
     * MAX_TUNNEL_SUPPORTED.
     */ 
    for (i = 1; i < MAX_TUNNEL_SUPPORTED; i++) {
        if (mnaddr_tunl[i] != 0) {
            sprintf(cmd, "/sbin/ifconfig tunl%d down", i);
	    mip3dbg(("Bringing down tunnel interface (%s) ... \n", cmd));
	    system(cmd);
	    sprintf(cmd, "/sbin/rmmod tunl%d", i);
	    mip3dbg(("Unloading tunnel module (%s) ... \n", cmd));
	    system(cmd);
        }	
    }
    
    pthread_mutex_lock(&(ifaceFDList.ifaceFDLock));
    for (i = 0; i < ifaceFDList.ifaceFDCnt; i++) {
	close(ifaceFDList.ifaceFDarray[i].ifaceFD);
    }
    ifaceFDList.ifaceFDCnt = 0;
    pthread_mutex_unlock(&(ifaceFDList.ifaceFDLock));
    
    pthread_cancel(monitorIface);
    
    /* 
     * Just add a wait to allow "secondary" thread to reach a point
     * where it has reached a "cancellable point".
     */
    sleep(2);

    /* hopefully the other thread is dead by now ... */
    pthread_mutex_destroy(&(ifaceFDList.ifaceFDLock));
    pthread_mutex_destroy(&(advPktBuf.AdvBufLock));
    close(ioctl_sockid);
}


/*
 * This is the generic interface to a script that enables IPSec. After
 * this script is invoked, all packets destined for a mobile node's
 * home network should automatically be processed by IPSec in a manner
 * such that those packets will pass through all intervening firewalls.
 * The script is passed three arguments: the first is either "on" or
 * "off" (without the quotes), the second is an interface name
 * on which IPSec should be enabled/disabled and the third is the
 * mobile node's care-of address. The last argument is useful if
 * depending on the mobile node's current point of attachment a different
 * set of firewalls may have to be traversed.
 */
void enableIPSec(mnState *mnsp)
{
       char tmp[128];
    
       sprintf(tmp, "IPSecScript on %s %s", mnsp->ifaceName,
	                  ntoa(mnsp->COAddr, addr1));
       mip3dbg(("%s\n", tmp));
       system(tmp);
}

/* Undo the actions performed as part of enableIPSec. */
void disableIPSec(mnState *mnsp)
{
       char tmp[128];

       sprintf(tmp, "IPSecScript off %s %s", mnsp->ifaceName,
	                  ntoa(mnsp->COAddr, addr1));
       mip3dbg(("%s\n", tmp));
       system(tmp);
}

int enableReverseTunneling(mnState *mnsp, u32 homeAgentAddr)
{
    int errcode;
    char cmd[80];
    int tnum;
    
    /*
     * Add a host-specific route fro the home agent (to avoid looping)
     */

    /* For now, we assume that reverse tunneling is only used when
     * a mobile node is outside its protected domain. In that case,
     * IPSecScript has already chnaged the routing tables so pkts
     * for the home agent are tunneled to the firewall, i.e. routing
     * table already has entries for:
     *    Destination        Gateway
     *     firewall           router
     *     homeagent          SKIP tunnel
     */

    /* 
    if ((errcode = routeadd(homeAgentAddr, inet_addr("255.255.255.255"),
			    mnsp->routerAddr, 1, mnsp->ifaceName)) < 0) {
	mip2dbg(("Error in (routeadd %s %s): %s\n",
		 ntoa(homeAgentAddr, addr1),
		 ntoa(mnsp->routerAddr, addr2),
		 err2str(errcode)));
    }
    */
    
    /* 
     * Set up another tunnel from the mobile node to its home agent
     * and set the default routing entry to pass packets through this
     * tunnel.
     * TODO: Currently the Linux tunneling code does not set 
     * DF bit on for the outer IP header. If and when it does,
     * we may need to adjust MTUs.
     */

    if ((errcode = encapadd(DEFAULT_ADDR, mnHomeAddr, homeAgentAddr, 0)) < 0) {
	mip0dbg(("Default encapsulation through %s failed ... %s\n",
		 ntoa(homeAgentAddr, addr1), err2str(errcode)));
    }

    /* 
     * Change the default route in the routing table to pass through
     * the new tunnel instead of a "normal" router.
     */
    if ((errcode = routedel(0, 0, mnsp->routerAddr, 1)) < 0) {
	mip2dbg(("Error in (routedel 0 %s): %s\n",
		 ntoa(mnsp->routerAddr, addr1), err2str(errcode)));
    }
    
    if ((tnum = gettunnelno(DEFAULT_ADDR)) < 0 ) {
	mip0dbg(("Error in gettunnelno called in enableReverseTunneling.\n"));
	exit(-1);
    } else {
	sprintf(cmd, "/sbin/route add default dev tunl%d", tnum);                   
	mip3dbg(("Adding route (%s) ... \n", cmd));
	system(cmd);
    }
    
    /*
     * We MUST NOT add a route for the firewall through the default router
     * on the foreign net. Otherwise, packets to the FW will be sent
     * using the MN's care-of address rather than its home address.
     * Under LINUX, however, IPSecScript needs to do this and we let
     * IPSecScript worry about that.
     */

    return(0);
}

int disableReverseTunneling(mnState *mnsp, u32 homeAgentAddr)
{
    int errcode;
    
    if ((errcode = routedel(0, 0, 0, 1)) < 0) {
	mip1dbg(("Error while deleting default route through reverse tunnel ... %s\n",
		 err2str(errcode)));
    }
    
    /* Restore previous default route */
    if ((errcode = routeadd(0, 0, mnsp->routerAddr, 1, mnsp->ifaceName)) < 0) {
	mip2dbg(("Error in routeadd 0 %s ... %s\n",
		 ntoa(mnsp->routerAddr, addr1), err2str(errcode)));
    }

    if ((errcode = encaprem(DEFAULT_ADDR, homeAgentAddr)) < 0) {
	mip1dbg(("Removal of default encapsulation through HA failed ... %s\n",
		 err2str(errcode)));
    }

    /* 
     * NOTE: Deleting the route for HA should be handled by IPSecScript.
     */ 
    /*
    if ((errcode = routedel(homeAgentAddr, inet_addr("255.255.255.255"),
			    mnsp->routerAddr, 1)) < 0) {
	mip2dbg(("Error in (routedel %s %s): %s\n",
		 ntoa(homeAgentAddr, addr1),
		 ntoa(mnsp->routerAddr, addr2), err2str(errcode)));
    }
     */
     return(0);
}

/******************
 int lowdefifcreq(u_long *ipaddr, u_long *gw, char *name)
{
  char line[120],dev[10];
  long dest = -1;
  struct sockaddr_in *sock;
  struct ifreq ifr;
  FILE *fd;
  
  if((fd = fopen("/proc/net/route","r"))==NULL)
    { 
      fprintf(stderr,"Error opening proc file\n");
      return -1;
    }

   do {
   } while (sscanf(line, "%s %lx %lx", dev, &dest, gw) != 2 || dest != 0x0);

   fclose(fd);
   if (dest != 0x0) return (-1);
   bzero(&ifr, sizeof(struct ifreq));
   strcpy(ifr.ifr_name,dev);
   sock = (struct sockaddr_in *) &ifr.ifr_addr;
   sock->sin_family = AF_INET;
   if (ioctl(ioctlsid, SIOCGIFADDR, (char *) &ifr) < 0)  {
	       perror("lowdefifcreq failed in SIOCGIFADDR");
	       return -1;
	   };

  *ipaddr = sock->sin_addr.s_addr;
  strcpy(name,ifr.ifr_name);
  return 1;
******************/





/*------ Filter Management ------------*/

/* Remove all the filters related to the given interface of the mobile
 * node
 */
int deleteFilters(int markValue, char *ifaceName) {
	char tblNameBegStr[32], tblNameStr[32];
	FILE *rtnfp, *irlfp;
	char cmd[128], cmd2[128], tnline[256], rlline[256], addr1[32],
		prioStr[32], mnaddrStr1[32], mnaddrStr2[32], rulenumStr[32],
		*sptr;
	int int1, more, mrkval;

	sprintf(tblNameStr, "iface-%s", ifaceName);

	// delete the rules that have the table name
	more = TRUE;
	while(more) {
		more = FALSE;
		sprintf(cmd, "ip rule list");
		irlfp = popen(cmd, "r");
		while(fgets(rlline, 256, irlfp) != NULL) {
			if(strlen(rlline) == 0 || *rlline == '#')
				continue;

			if(!isWordIn(rlline, tblNameStr))
				continue;

			strcpy(cmd2, "ip rule del ");
			getWord(rlline, 1, prioStr);
			replaceStr(rlline, "from all", "        ");
			replaceStr(rlline, "lookup", "table ");

			strcat(cmd2, (rlline + strlen(prioStr)));
			more = TRUE;
			break;
		}
		pclose(irlfp);
		if(!more)
			break;
		printf("%s \n", cmd2);
		system(cmd2);
	}

	// delete all iptable entries with given marking
	sprintf(mnaddrStr1, "%s", ntoa(mnHomeAddr, addr1));
	more = TRUE;
	while(more) {
		more = FALSE;
		sprintf(cmd, "iptables --list OUTPUT -v -t mangle  --line-numbers -n");
		irlfp = popen(cmd, "r");
		while(fgets(rlline, 256, irlfp) != NULL) {
			if(strlen(rlline) == 0 || *rlline == '#')
				continue;

			memset(mnaddrStr2, (char) NULL, sizeof(mnaddrStr2));
			getWord(rlline, 9, mnaddrStr2);

			if(strncmp(mnaddrStr1, mnaddrStr2, strlen(mnaddrStr1)) != 0)
				continue;

			if((sptr = strstr(rlline, "MARK set")) == NULL)
				continue;

			sscanf(sptr + strlen("MARK set"), "%x", &mrkval);
			if(mrkval != markValue)
				continue;

			getWord(rlline, 1, rulenumStr);

			sprintf(cmd2, "iptables -t mangle -D OUTPUT %s ", rulenumStr);

			more = TRUE;
			break;
		}
		pclose(irlfp);
		if(!more)
			break;
		printf("%s \n", cmd2);
		system(cmd2);
	}

	// delete any routing entries related to the table name
	more = TRUE;
	while(more) {
		more = FALSE;
		*rlline = (char) NULL;
		sprintf(cmd, "ip route list table %s", tblNameStr);
		irlfp = popen(cmd, "r");
		fgets(rlline, 256, irlfp);
		if(isWordIn(rlline, "default")) {
			more = TRUE;
			sprintf(cmd2, "ip route flush table %s", tblNameStr);
		}
		pclose(irlfp);
		if(more) {
			printf("%s \n", cmd2);
			system(cmd2);
		}
	}

	flushRouteCache();
}



/* Apply a filter given in a string for the given coaddr
 * & mnaddr
 * does the following - create table name, create rule &
 * create routing table
 */
int applyFilter(char *ifcName, char *fltStr, u32 routerAddr) {
	char tblNameStr[32], tnline[256], cmd[128], rlline[256];
	char addr1[32], addr2[32], str1[32], word[64], actStr[32],
		startportStr[32];
	FILE *rtnfp, *irlfp;
	int found, maxTblID, fltCnt, fltType, fltTarget;
	int int1, i, j, need, markValue;
	u32 ipaddr;

	sprintf(tblNameStr, "iface-%s", ifcName);
	maxTblID = 0;
	found = FALSE;

	// check the route table name file to see if an entry
	// exist, if not insert one
	rtnfp = fopen(ROUTE2_TABLE_FILE, "r");
	while(fgets(tnline, 256, rtnfp) != NULL) {
		if(strlen(tnline) == 0 || *tnline == '#')
			continue;

		if(sscanf(tnline, " %d %s", &int1, str1) != 2)
			continue;

		if(strncmp(tblNameStr, str1, strlen(tblNameStr)) == 0) {
			found = TRUE;
			break;
		}

		// also, figure out the highest table ID
		if(maxTblID < int1)
			maxTblID = int1;
	}
	fclose(rtnfp);

	if(!found) {
		sprintf(cmd, "echo %d %s >> %s", (maxTblID + 1), tblNameStr, ROUTE2_TABLE_FILE);
		system(cmd);
	}

	sscanf(fltStr, " %d %d %d ", &fltCnt, &fltTarget, &markValue);

	// build the iptables rule
	sprintf(cmd, "iptables -t mangle -A OUTPUT -v -s %s ", ntoa(mnHomeAddr, addr1));
	j = 3;
	for(i = 0; i < fltCnt; i++) {
		j++;
		if(getWord(fltStr, j, word) != 0)
			break;

		if(sscanf(word, "%d", &fltType) != 1)
			break;

		switch(fltType) {
			case FLT_BAF_EXT:
				j++;
				if(getWord(fltStr, j, word) != 0)
					break;
				sscanf(word, "%d", &int1);
				sprintf((cmd + strlen(cmd)), " -m tos --tos 0x%02x", int1);
				break;
			case FLT_PROTO_EXT:
				j++;
				if(getWord(fltStr, j, word) != 0)
					break;
				sprintf((cmd + strlen(cmd)), " -p %s", word);
				break;
			case FLT_SRCADR_EXT:
				j++;
				if(getWord(fltStr, j, word) != 0)
					break;
				sprintf((cmd + strlen(cmd)), " -d %s", word);
				break;
			case FLT_SRCNET_EXT:
				j++;
				if(getWord(fltStr, j, word) != 0)
					break;
				strcpy(addr1, word);

				j++;
				if(getWord(fltStr, j, word) != 0)
					break;
				ipaddr = inet_addr(word);
				sprintf((cmd + strlen(cmd)), " -d %s/%d", addr1, numBitsInMask(ipaddr));
				break;
			case FLT_SRCPORT_EXT:
				j++;
				if(getWord(fltStr, j, word) != 0)
					break;
				sprintf((cmd + strlen(cmd)), " --destination-port %s", word);
				break;
			case FLT_SRCPORTRNG_EXT:
				j++;
				if(getWord(fltStr, j, word) != 0)
					break;
				strcpy(startportStr, word);

				j++;
				if(getWord(fltStr, j, word) != 0)
					break;

				sprintf((cmd + strlen(cmd)), " --destination-port %s:%s", startportStr, word);
				break;

			// new filters

			case FLT_DSTPORT_EXT:
				j++;
				if(getWord(fltStr, j, word) != 0)
					break;
				sprintf((cmd + strlen(cmd)), " --source-port %s", word);
				break;
			case FLT_DSTPORTRNG_EXT:
				j++;
				if(getWord(fltStr, j, word) != 0)
					break;
				strcpy(startportStr, word);

				j++;
				if(getWord(fltStr, j, word) != 0)
					break;

				sprintf((cmd + strlen(cmd)), " --source-port %s:%s", startportStr, word);
				break;


			case FLT_FRFORM_EXT:
				break;
			default:
				break;
		}
	}
	sprintf(str1, " -j MARK --set-mark %d ", markValue);
	strcat(cmd, str1);
	printf("%s\n", cmd);
	system(cmd);


	//build the ip rule for the mark
	sprintf(cmd, "ip rule add fwmark %d ", markValue);
	switch(fltTarget) {
		case TG_DROP:
				strcpy(actStr, "reject");
				break;
		case TG_REJECT:
				strcpy(actStr, "unreachable");
				break;
		case TG_ACCEPT:
				strcpy(actStr, " ");
				break;
		case TG_ACCEPT_AVOID_RO:
				strcpy(actStr, " ");
				break;
		case TG_MASQ:
				strcpy(actStr, " ");
				break;
	}

	sprintf(str1, " table %s %s ", tblNameStr, actStr);
	strcat(cmd, str1);
	printf("%s\n", cmd);
	system(cmd);


	// create the routing table, if not available
	need = TRUE;
	*rlline = (char) NULL;
	sprintf(cmd, "ip route list table %s", tblNameStr);
	irlfp = popen(cmd, "r");
	while(fgets(rlline, 256, irlfp) != NULL) {
		if(isWordIn(rlline, "default") || isWordIn(rlline, ifcName)) {
			need = FALSE;
			break;
		}
	}
	pclose(irlfp);
	if(need) {
		sprintf(cmd, "ip route add default via %s dev %s table %s",
					ntoa(routerAddr, addr1), ifcName, tblNameStr);
		printf("%s\n", cmd);
		system(cmd);
	}
}


int flushRouteCache() {
//	char cmd[128];
//
//	sprintf(cmd, "ip route flush cache ");
//	printf("%s \n", cmd);
//	system(cmd);

	return TRUE;
}


