/*
 * Routines in this program implement  OS-specific portions of 
 * a Mobile node (RFC 2002).
 *
 * Author: sunil.madhani@eng.sun.com
 * Major modifications: vgupta@eng.sun.com
 */

#pragma ident	"@(#)mnmipsolaris.c 1.4 98/12/01"

/*
 * void OSInit(): 
 *
 *     Includes OS-specific initialization. For example, in solaris, we 
 * initialize the vtunl module here and also enable decapsulation
 * of packets whose inner dst is the mobile node's home address.
 * This is also a good place to creates a separate thread to pick
 * up agent advertisements on the interfaces. 
 *
 * void OSFinalize()
 * 
 *      OS dependent cleanup routine called when exiting, e.g.
 * releasing locks, killing threads, or putting the Mobile node
 * in the state prior to initiating Mobile IP.
 *
 * char *err2str(int errval)
 *
 *     Converts a negative error value to a meaningful error message
 * string, e.g. by using the comments in /usr/include/sys/errno.h.
 *
 * int getNewIfaceList(MIPIfaceEntry *ifaceTable, int *size) 
 *
 *	Gets the list of interfaces of interest to the mobile node
 * code, e.g. interfaces on which agent advertisements must be monitored
 * or those that may be used to obtain a colocated care-of address.
 * NOTE: Mobile routers should ignore interfaces attached to supported
 * mobile networks.
 *
 * boolean getAgentAdvByIface(char *pkt, int *pktlen, char *ifaceName,
 *                            char *hwAddr)
 *      Returns an agent advertisement packet for the specified interface.
 * The advertisement packet placed in the pkt buffer starts at the IP
 * header. The hardware address of the source is also returned.
 *
 * void mkCloneIface(char *iface, char *clone)
 *
 *      Make up the name for a placeholder iface used to store the
 * mobile node's home address in colocated state.
 *
 * int configCloneIface(char *cloneIface, u32 addr, u32 netMask, 
 *                      char *parentIface)
 *
 *     Configures a logical clone interface used as a placeholder for
 * the mobile node's home address. It is assigned the specified addr
 * and netMask. For Point-to-point interfaces, additional information
 * required to configure the clone may be copied from the parentIface.
 * This function is also responsible for deleting any spurious routing
 * entries that may be added as part of bringing up the clone Iface.
 *
 * int configIface(char *ifaceName, u32 addr, u32 netMask)
 *
 *     Configures the specified interface with the assigned address
 * and netmask. The interface is brought up.
 *
 * int routeadd(unsigned long dst, unsigned long mask, unsigned long gw, 
 *              int metric)
 *   
 *      Add route for destination "dst", netmask "mask" through gateway
 * "gw". A zero metric specifies a local route and non-zero metric
 * specifies an indirect route (i.e. through intermediate gateway(s)).
 *
 * int routedel(unsigned long dst, unsigned long mask, unsigned long gw,
 *              int metric)
 *
 *      Delete route for specified destination through gateway. The
 * parameters are interpreted the same way as for routeadd.
 *
 * int routeflush()
 *
 *      Clears the routing table except for the loopback interface and
 * any entries created by the mobile node (e.g. ip0 under Solaris).
 * TODO: Be careful about the multicast network entry.
 *
 * int enableReverseTunneling(mnState *mnsp, u32 homeAgentAddr)
 *      Set up and enable a reverse tunnel from the mobile node to
 * the specified home agent.
 *
 * int disableReverseTunneling(mnState *mnsp, u32 homeAgentAddr)
 *      Undo the actions performed by enableReverseTunneling.
 *
 * int arpadd(unsigned long host, unsigned char *eaddr, char *flags)
 *
 *      Adds an arp entry with ip address set to host and hardware  
 * address set to eaddr with approriate flags specified by flag
 * argument, e.g. arpadd(inet_addr("129.146.122.121"), "08:20:AB:FE:33:11",
 * "pub") creates a proxy ARP entry for 129.146.122.121.
 *
 * int arpdel(unsigned long host)
 *
 *      Deletes an arp entry for the specified ip address.
 *
 * int sendgarp(char *iface)
 * 
 *      Sends gratuitous arp for the mobile node using the hardware address
 * of the iface. Needed when a mobile node returns home and successfully 
 * deregisters itself.
 *
 * int arpflush()
 *
 *	Clears the arp table except for those with "SM" flags.
 *
 * int disableArp(u32 mnAddr, u32 foreignRouterAddr)
 *     
 *      Prevent a mobile node from sending out broadcast ARP packets
 * (used when a mobile node is on a foreign network).
 *
 * int restoreArp()
 *
 *     Restore normal ARP behavior.
 * 
 * void enableIPSec(mnState *mnsp)
 *
 *      Enables IPSec for firewall traversal. The main idea is that the
 * mobile node should do whatever is necessary at its end such that packets
 * generated by it for the home network will be able to get past any
 * intervening firewalls.
 *
 * void disableIPSec(mnState *mnsp)
 *    
 *      Undo the actions performed for enableIPSec.
 *
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <netdb.h>
#include <signal.h>
#include <setjmp.h>
#include <errno.h>

#include <sys/dlpi.h>
#include <sys/types.h>
#include <sys/stream.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/sockio.h>
#include <sys/errno.h>
#include <sys/param.h>
#include <sys/conf.h>
#include <sys/stropts.h>
#include <sys/tihdr.h>
#include <sys/tiuser.h>
#include <sys/timod.h>
#include <sys/stropts.h>
#include <sys/signal.h>
#include <pthread.h>

#include <inet/common.h> 
#include <inet/mib2.h>
#include <inet/ip.h> 
#include <inet/common.h>
#include <inet/ip.h>

#include <net/if.h>
#include <net/route.h>
#include <net/if_arp.h>

#include <netinet/igmp_var.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/ip_mroute.h>

#include "vtunl.h"
#include "mobile.h"

extern int errno;
static int ioctl_fd;
static int ioctl_sockid;

char *err2str(int);

#ifndef ETH_ALEN
#define ETH_ALEN 6
#endif

#define DEFAULT_ADDR_VTUNL ((u32) htonl(0x7fffffff)) /* 127.255.255.255 */

/* Definitions needed for send_gratuitous_arp() module */
#define	ABS(x)		((x) < 0 ? -(x) : (x))
#define	BITSPERBYTE	8
#define ETHERADDRL      sizeof(ether_addr_t)
#define IPADDRL         sizeof(struct in_addr)
#define DEVICEDIR       "/dev/"

#define MAX_ERR                 2048
#define INVALID_PPA             (MAX_ERR + 1)
#define INVALID_STRING          (MAX_ERR + 2)
#define DLOKACK_SHORT_RESPONSE  (MAX_ERR + 3)
#define DLOKACK_NOT_M_PCPROTO   (MAX_ERR + 4)
#define ERR_MORECTL             (MAX_ERR + 5)
#define ERR_MOREDATA            (MAX_ERR + 6)
#define ERR_MORECTLDATA         (MAX_ERR + 7)
#define DL_PRIMITIVE_ERROR      (MAX_ERR + 8)
#define SHORT_CONTROL_PORTION   (MAX_ERR + 9)
#define INVALID_ADDR            (MAX_ERR + 10)
#define MN_ENTRY_ABSENT         (MAX_ERR + 11)
#define NO_MAPPING		(MAX_ERR + 12)


#ifndef IFF_DHCPRUNNING
#define IFF_DHCPRUNNING 0x4000
#endif

/*  Maximum control/data buffer size (in long's !!) for getmsg() */
#define         MAXDLBUF        8192
#define         MAXWAIT         15

/* Maximum address buffer length */
#define         MAXDLADDR       1024

/* Handy macro.*/
#define         OFFADDR(s, n)   (u_char*)((char*)(s) + (int)(n))

#ifndef T_CURRENT
#define T_CURRENT       MI_T_CURRENT
#endif

/* Definitions needed for arpflush() */ 
#define ARP_SP 0x0003
#define ARP_SM 0x0011

#define MAX_AGENT_ADV_ENTRIES	10
#define DEFAULT_MIN_MTU         1250

extern char currentTime[], relativeTime[], addr1[], addr2[];
extern int mipdebug;
extern char msg[];

/* The mobile node's address and netmask */
extern u32 mnHomeAddr;
extern u32 mnHomeNetmask;
extern boolean isReverseTunnelSet; 

extern IfaceEntry ifaceTable[];
extern int ifaceTableSize;

extern char ignoreIfaceList[][MAX_IFNAME_LEN]; 
extern int ignoreIfaceCnt;

/* 
 * Type Definition for MIB item. Used by arpflush() and routeflush().
 * Taken from the source code for netstat.
 */
typedef struct mib_item_s {
	struct mib_item_s	*next_item;
	long			group;
	long			mib_id;
	long			length;
	char			*valp;
} mib_item_t;

/* Type definitions for the Advertisement Pkt. Buffer */
typedef struct aap {
	char ifaceName[MAX_IFNAME_LEN];
	u32  advSrcIPAddr;
        char hwAddr[MAX_HWADDR_LEN];
	char data[MAX_PKT_SIZE];
	int  dataLen;
} AgentAdvPkt;

/* 
 * Circular buffer to store the agent advertisements.
 * Buffer is empty when head = tail. It is full when 
 * head == (tail + 1) mod n.
 *
 * For a non-empty list, head points to the first filled slot.
 * Tail points to the next available slot. So a buffer with n
 * slots can hold upto (n-1) items.
 *
 * AdvBufLock is used to synchronize access to the buffer containing
 * agent advertisements. The main program thread removes agent advertisements
 * from the buffer either as part of getAgentAdvByIface() or while
 * doing a purge as part of relinquishing an interface when it is
 * brough down. Another thread which monitors a list of interfaces
 * (or rather open file descriptors) for agent advertisements inserts
 * data in this buffer.
 */
typedef struct AdvPktBuffer {
	int head;
	int tail;
	AgentAdvPkt advPkt[MAX_AGENT_ADV_ENTRIES];
	pthread_mutex_t AdvBufLock;
} advPktBuffer;

static advPktBuffer advPktBuf;

static pthread_t monitorIface;
typedef struct ifd {
    char ifaceName[MAX_IFNAME_LEN];
    int  ifaceFD;
} ifaceFD;

typedef struct ifdlist {
    int     ifaceFDCnt;
    ifaceFD ifaceFDarray[DEFAULT_MAX_INTERFACES];
    pthread_mutex_t ifaceFDLock;
} ifacefdList;

/* 
 * Besides the main thread, the program creates another thread which
 * continuously monitors incoming agent advertisements on the interfaces
 * listed in ifaceFDList. The lock ifaceFDLock is used to synchronize
 * access to the FD list. The main thread adds or removes FDs based
 * on its notion of "relevant" interfaces and the secondary thread
 * only reads (never changes) this datastructure.
 */
ifacefdList ifaceFDList;

extern void ether_print(unsigned char[]);
extern IfaceEntry *findIface(char *);
/* Security related stuff */
extern DomainInfo domainInfo;

static mib_item_t *mibget();
void enableIPSec(mnState *);
void disableIPSec(mnState *);
int enableReverseTunneling(mnState *, u32);
int disableReverseTunneling(mnState *, u32);

/* Convert a negative error value into a human readable error message */
char *err2str(int errval)
{
    int err =  (-1 * errval);
    static char tmp[30];

    switch (err) {
      case EMFILE:
	  return("Descriptor table full");

      case EBADF:
	  return("Not a valid open file descriptor");

      case EACCES:
	  return("Permission denied");

      case ENOMEM:
	  return("Insufficient memory");

      case ENOENT:
	  return("No such device/file/directory");

      case ENXIO:
	  return("No such device or address");
	  
      case EEXIST:
	  return("Already exists");

      case EINVAL:
	  return("Invalid request or arguments");
      
      case INVALID_PPA: 
	  return("Invalid PPA") ;

      case INVALID_STRING:
          return("Invalid input string");
      
      case DLOKACK_SHORT_RESPONSE:
          return("Short DLOKACK response");

      case DLOKACK_NOT_M_PCPROTO: 
          return("DLOKACK is not M_PCPROTO");

      case ERR_MORECTL: 
          return("MORECTL");

      case ERR_MOREDATA: 
          return("MOREDATA");

      case ERR_MORECTLDATA: 
          return("MORECTL or MOREDATA");

      case SHORT_CONTROL_PORTION: 
          return("Control portion is short");

      case DL_PRIMITIVE_ERROR: 
          return("DL primitive error");

      case EAGAIN: 
          return("A non-prioriy message specified");

      case EFAULT: 
          return("ctlptr or dataptr points to an illegal address");

      case EINTR: 
          return("Caught a signal");

      case ENOSR: 
          return("Insufficient stream resources");

      case ERANGE: 
          return("Packet size is out of range");

      case INVALID_ADDR: 
          return("Invalid Address");

      case EBADMSG: 
          return("Bad queued message");

      case MN_ENTRY_ABSENT: 
          return("Missing Mobile node entry");

      case NO_MAPPING:
          return("Bad ARP mapping for mobile node");

      default:
	  sprintf(tmp, "Reason unclear: %d", err);
	  return(tmp);
    }
}

/*
 * Parse the interface name (e.g. "le0" or "hme1") and place the
 * corresponding device name (e.g. "/dev/le", "/dev/hme") in
 * devname and the ppa (e.g. 0 or 1 for the examples above) in ppa.
 */
int ifname2devppa(char *ifname, char *devname, int *ppa)
{
      char *p;
      int  i, j, val;

      val = 0;
      j = strlen(DEVICEDIR);
      strncpy(devname, DEVICEDIR, j);
      for (i = 0, p = ifname; i < strlen(ifname); i++, p++) {
	   if (*p >= '0' && *p <= '9') 
	           val = 10*val + (*p - '0');
	   else 
	           devname[j++] = *p;
      }
      devname[j] = '\0';
      *ppa = val;
      return(0);
}

/*
 *  Configure the point-to-point interface named ifname with the
 *  given addresses.
 *
 *  If you wish to look at ip0 using ifconfig, always specify the
 *  inet option, e.g.
 *      ifconfig -a inet
 *  or
 *      ifconfig ip0 inet
 *
 *  If this option is not specified, ifconfig may attempt to print
 *  ethernet info for ip0 and hang.
 */
int mkpt2pt(char *ifname, long srcaddr, long dstaddr)
{   
       struct ifreq ifr;
       struct sockaddr_in sin;
       short savedflags;

       errno = 0 ;
       
       memset((char *) &ifr, 0, sizeof(ifr));
       strncpy(ifr.ifr_name, ifname, sizeof ifr.ifr_name);
       sin.sin_family = AF_INET;
       sin.sin_addr.s_addr = srcaddr;
       ifr.ifr_addr = *(struct sockaddr *) &sin;
       if (ioctl(ioctl_sockid, SIOCSIFADDR, (caddr_t)&ifr) < 0) 
	       return(-1 * (errno));

       memset((char *) &ifr, 0, sizeof(ifr));
       strncpy(ifr.ifr_name, ifname, sizeof ifr.ifr_name);
       if (ioctl(ioctl_sockid, SIOCGIFFLAGS, (caddr_t)&ifr) < 0) 
	       return(-1 * (errno));
       savedflags = ifr.ifr_flags;

       memset((char *) &ifr, 0, sizeof(ifr));
       strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
       sin.sin_family = AF_INET;
       sin.sin_addr.s_addr = dstaddr;
       ifr.ifr_addr = *(struct sockaddr *)&sin;
       if (ioctl(ioctl_sockid, SIOCSIFDSTADDR, (caddr_t)&ifr) < 0)
	       return(-1 * (errno));

       memset((char *) &ifr, 0, sizeof(ifr));
       strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
       sin.sin_family = AF_INET;
       sin.sin_addr.s_addr = inet_addr("255.255.255.255");
       ifr.ifr_addr = *(struct sockaddr *)&sin;
       if (ioctl(ioctl_sockid, SIOCSIFNETMASK, (caddr_t)&ifr) < 0)
	       return(-1 * (errno));

       /* By bringing UP this interface, we force the creation of
	* a routing table entry for srcaddr through ip<ppa>.
	*/

       memset((char *) &ifr, 0, sizeof(ifr));
       strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
       ifr.ifr_flags = savedflags | IFF_UP | 
	               IFF_PRIVATE | IFF_POINTOPOINT ;
       if (ioctl(ioctl_sockid, SIOCSIFFLAGS, (caddr_t)&ifr) < 0) 
	       return(-1 * (errno));

       return(0);
}

/*
 * Issue an ioctl for the vtunl module using the I_STR ioctl call 
 * (refer to man streamio).
 */
int 
vtunl_ctl(int s, int cmd, void *arg, int arglen)
{
    struct strioctl ioc;

    memset((char *) &ioc, 0, sizeof(ioc));
    ioc.ic_cmd = cmd;
    ioc.ic_timout = 0;
    ioc.ic_len = (size_t) arglen;
    ioc.ic_dp = (char *) arg;

    errno = 0;
    ioctl(s, I_STR, (char *) &ioc);
    return(errno);
}

/*
 * Create a stream containing the named device, and vtunl. Then
 * configure it with the information provided in conf.
 * Finally, push "ip" and return the stream descriptor.
 */
int plumb(char *devname, struct vtunl_req *conf)
{
       int fd;
       struct vtunl_conf vc;
       
       errno = 0;

       if ((fd = open(devname, 0)) < 0) 
	 return(-1 * (errno));

       if (ioctl(fd, I_PUSH, "vtunl") < 0) {
	 close(fd);
	 return(-1 * (errno));
       }

       if (strncmp("/dev/ip", devname, strlen("/dev/ip")) == 0) {
	       if (vtunl_ctl(fd, VTUNL_SETCONF_TRANSPORT, conf, 
			     sizeof(struct vtunl_req))) {
		 close(fd);
		 return(-1 * (errno));
	       }
	       if (vtunl_ctl(fd, VTUNL_GETCONF, &vc, 
			     sizeof(struct vtunl_conf))){
		 close(fd);
		 return(-1 * (errno));
	       }
	       mip0dbg(("Configured vtunl type %d proto %d addr %x (%d)",
		       vc.type, vc.ip_protocol, ntohl(vc.address), vc.state));
       } else {
	       if (vtunl_ctl(fd, VTUNL_SETCONF_DLSUSER, conf, 
			     sizeof(struct vtunl_req))){
		 close(fd);
		 return(-1 * (errno));
	       }

	       if (vtunl_ctl(fd, VTUNL_GETCONF, &vc, 
			     sizeof(struct vtunl_conf))){
		 close(fd);
		 return(-1 * (errno));
	       }
	 
	       mip0dbg(("Configured vtunl type %d ppa %d addr %x mask %x (%d)",
		       vc.type, vc.ip_protocol, ntohl(vc.address),
		       ntohl(vc.addr_mask), vc.state));
       }

       mip0dbg((" on %s.\n", devname));

       if (ioctl(fd, I_PUSH, "ip") < 0) {
	   close(fd);
	   return(-1 * (errno));
       }

       return(fd);
}

/* 
 * Sets an upper bound on the path-mtu for the
 * given tunnel_exit.
 */
int pmtubound(u32 t_exit, int pmtu)
{
    struct vtunl_req vrq;
    int err;
           
    memset((char *) &vrq, 0, sizeof(struct vtunl_req));
    vrq.vr_address = t_exit;
    vrq.vr_mtu_bound = (u16) pmtu;

    if (err = vtunl_ctl(ioctl_fd, VTUNL_MAXPMTU_SET, &vrq, sizeof(vrq))) 
        return(-1 * err);
    else
        return(0);
}

/* Process encapadd command */
int
encapadd(u32 target, u32 haAddr, u32 t_exit, unsigned short flags) 
{
   int  err;
   struct vtunl_req vrq;

    memset((char *) &vrq, 0, sizeof(struct vtunl_req));
    vrq.vr_target_addr = target;
    vrq.vr_t_exit_addr = t_exit ;
    vrq.vr_encap_flags = (u16) flags;

    if (err = vtunl_ctl(ioctl_fd, VTUNL_ENCAP_ADD, &vrq, sizeof(vrq))) 
        return(-1 * err);
    else 
        return(0);
}
	
/* Process encaprem */
int
encaprem(u32 target, u32 t_exit)
{
    int  err;
    struct vtunl_req vrq;

    memset((char *) &vrq, 0, sizeof(struct vtunl_req));
    vrq.vr_target_addr = target;
    vrq.vr_t_exit_addr = t_exit ;

    if (err = vtunl_ctl(ioctl_fd, VTUNL_ENCAP_REM, &vrq, sizeof(vrq))) 
        return(-1 * (err));
    else 
        return(0);
}

/* Process decapadd command */
int
decapadd(u32 target, char *flags)
{
    int err;
    struct vtunl_req vrq;

    memset((char *) &vrq, 0, sizeof(struct vtunl_req));
    vrq.vr_target_addr = target ;
    if (strncmp(flags, "dst", strlen("dst")) == 0) 
        vrq.vr_decap_flags = DECAP_IF_DST;
    else if (strncmp(flags, "src", strlen("src")) == 0) 
	vrq.vr_decap_flags = DECAP_IF_SRC;
    else if (strncmp(flags, "either", strlen("either")) == 0) 
        vrq.vr_decap_flags = DECAP_IF_EITHER;
    else 
       return(-1 * EINVAL);

    if (err = vtunl_ctl(ioctl_fd, VTUNL_DECAP_ADD, &vrq, sizeof(vrq)))
        return(-1 * (err));
    else 
        return(0);
}

/* Process decaprem command */
int
decaprem(u32 target)
{
   int  err;
   struct vtunl_req vrq;

   memset((char *) &vrq, 0, sizeof(struct vtunl_req));
   vrq.vr_target_addr = target;
	   
   if (err = vtunl_ctl(ioctl_fd, VTUNL_DECAP_REM, &vrq, sizeof(vrq)))
        return(-1 * (err));
   else 
	return(0);
}

/*
 * Set an arp entry for the specified host with the given hardware
 * address and flags.
 */
int arpadd(unsigned long host, unsigned char *eaddr, char *flag)
{
    struct arpreq ar;
    struct sockaddr_in *sin;

    bzero((caddr_t)&ar, sizeof ar);
    sin = (struct sockaddr_in *)&ar.arp_pa;
    sin->sin_family = AF_INET;
    sin->sin_addr.s_addr = (host);

    memcpy(ar.arp_ha.sa_data, eaddr, ETH_ALEN);

    ar.arp_flags = ATF_PERM;

    if (strncmp(flag, "temp", 4) == 0)
        ar.arp_flags &= ~ATF_PERM;
    if (strncmp(flag, "pub", 3) == 0)
        ar.arp_flags |= ATF_PUBL;
    if (strncmp(flag, "trail", 5) == 0)
        ar.arp_flags |= ATF_USETRAILERS;

    errno = 0;		
    if (ioctl(ioctl_sockid, SIOCSARP, (caddr_t)&ar) < 0) {
        if (errno != EEXIST)
	    return((-1) * errno);
    }

    return (0);
}

/* 
 * Delete any existing arp entry for the specified host. The host address
 * is in network-byte order at this point.
 */
int arpdel(unsigned long host)
{
    struct arpreq ar;
    struct sockaddr_in *sin;

    bzero((caddr_t)&ar, sizeof ar);
    ar.arp_pa.sa_family = AF_INET;
    sin = (struct sockaddr_in *)&ar.arp_pa;
    sin->sin_family = AF_INET;
    sin->sin_addr.s_addr = host;

    errno = 0;
    if (ioctl(ioctl_sockid, SIOCDARP, (caddr_t)&ar) < 0) {
        if (errno != ENXIO)
            return((-1) * errno);
    }

    return(0);
}

/*
 * Add or Delete the route depending on argument cmd 
 * metric : 0 - Host entry      i.e.Flag  is  UH
 *          1 - Gateway entry  i.e. Flag is UG
 *          2 - UGH
 *          3 - U
 * TODO : Make it compatible with the one used in 
 *        mipmodsolaris.c.
 *        The difference is 1 set UGH.
 *        Need to go through agent and find out where
 *        routeadd is used and justify whether 1 with
 *        UG or UGH is needed.
 */
int routemodify(unsigned long dst, unsigned long mask, unsigned long gw, 
		int metric, char cmd)
{
    struct rtentry route;
    struct sockaddr_in *sin;
    int val = 0;

    /*
     * Workaround for Solaris 2.6 which seems to use an ARP
     * entry for x even when no local route exists for x.
     */
    if (cmd == 'd') {
        val = arpdel(dst);
    }

    sin = (struct sockaddr_in *)&route.rt_dst;
    sin->sin_addr.s_addr = dst;

    sin = (struct sockaddr_in *)&route.rt_gateway;
    sin->sin_addr.s_addr = gw;

    route.rt_flags = RTF_UP; 

    if (mask == inet_addr("255.255.255.255"))
        route.rt_flags |= RTF_HOST;

    if (metric > 0)
        route.rt_flags |= RTF_GATEWAY;

    errno = 0;
    if ((ioctl(ioctl_sockid, cmd == 'a' ? SIOCADDRT : SIOCDELRT,
	  (caddr_t)&route)) < 0) {
        /* 
	 * SIOCADDRT does not return ENXIO & 
	 * SIOCDELRT never returns EEXIST
	 */
        if ((errno != ENXIO) && (errno != EEXIST))
           return(-1 * (errno));
    } 

    return(val);
}

/* Add route */
int routeadd(unsigned long dst, unsigned long mask, unsigned long gw, 
	     int metric)
{
    return(routemodify(dst, mask, gw, metric, 'a'));
}

/* Delete the route */
int routedel(unsigned long dst, unsigned long mask, unsigned long gw, 
	     int metric)
{
    return(routemodify(dst, mask, gw, metric, 'd'));
}

/* DLPI Routines needed for sendgarp() and getEthernetAddr() */
/*
 * Common (shared) DLPI test routines (taken from dlcommon.c.
 * Mostly pretty boring boilerplate sorta stuff.
 */

int dlinforeq(int fd)
{
    dl_info_req_t info_req;
    struct strbuf ctl;
    int	flags;

    info_req.dl_primitive = DL_INFO_REQ;

    ctl.maxlen = 0;
    ctl.len = sizeof (info_req);
    ctl.buf = (char *) &info_req;

    flags = RS_HIPRI;
    errno = 0;
    if (putmsg(fd, &ctl, (struct strbuf*) NULL, flags) < 0){
	return(-1 * errno);
    }
    return(0);
}

int dlinfoack(int fd, char *bufp)
{
    union DL_primitives	*dlp;
    struct strbuf ctl;
    int	flags;
    int ret ;

    ctl.maxlen = MAXDLBUF;
    ctl.len = 0;
    ctl.buf = bufp;

    strgetmsg(fd, &ctl, (struct strbuf*)NULL, &flags, "dlinfoack");

    dlp = (union DL_primitives *) ctl.buf;

    if ((ret = expecting(DL_INFO_ACK, dlp)) < 0) {
        close(fd);
       return(ret);
    }
	 
    if (ctl.len < sizeof (dl_info_ack_t)){
        return(-1 * DLOKACK_SHORT_RESPONSE);
    }

    if (flags != RS_HIPRI) {
	return(-1 * DLOKACK_NOT_M_PCPROTO);
    }
 
    if (ctl.len < sizeof (dl_info_ack_t)){
	return(-1 * DLOKACK_SHORT_RESPONSE);
    }

    return(0);
}

int dlattachreq(int fd, u_long ppa)
{
    dl_attach_req_t attach_req;
    struct strbuf ctl;
    int	flags;

    attach_req.dl_primitive = DL_ATTACH_REQ;
    attach_req.dl_ppa = ppa;

    ctl.maxlen = 0;
    ctl.len = sizeof (attach_req);
    ctl.buf = (char *) &attach_req;

    flags = 0;
    errno = 0;

    if (putmsg(fd, &ctl, (struct strbuf*) NULL, flags) < 0)
	return(-1 * errno);

    return(0);
}

int dlbindreq(int fd, u_long sap, u_long max_conind, 
              u_long service_mode, u_long conn_mgmt, u_long xidtest)
{
    dl_bind_req_t bind_req;
    struct strbuf ctl;
    int flags;

    bind_req.dl_primitive = DL_BIND_REQ;
    bind_req.dl_sap = sap;
    bind_req.dl_max_conind = max_conind;
    bind_req.dl_service_mode = service_mode;
    bind_req.dl_conn_mgmt = conn_mgmt;
    bind_req.dl_xidtest_flg = xidtest;

    ctl.maxlen = 0;
    ctl.len = sizeof (bind_req);
    ctl.buf = (char *) &bind_req;

    flags = 0;
    errno = 0;
	
    if (putmsg(fd, &ctl, (struct strbuf*) NULL, flags) < 0)
	return(-1 * errno);

    return(0);
}

int dlunitdatareq(int fd, u_char *addrp, int addrlen, 
		  u_long minpri, u_long maxpri, u_char *datap, 
		  int datalen)
{
    long buf[MAXDLBUF];
    union DL_primitives	*dlp;
    struct strbuf data, ctl;

    dlp = (union DL_primitives*) buf;

    dlp->unitdata_req.dl_primitive = DL_UNITDATA_REQ;
    dlp->unitdata_req.dl_dest_addr_length = addrlen;
    dlp->unitdata_req.dl_dest_addr_offset = sizeof (dl_unitdata_req_t);
    dlp->unitdata_req.dl_priority.dl_min = minpri;
    dlp->unitdata_req.dl_priority.dl_max = maxpri;

    memcpy(OFFADDR(dlp, sizeof (dl_unitdata_req_t)), addrp, addrlen);

    ctl.maxlen = 0;
    ctl.len = sizeof (dl_unitdata_req_t) + addrlen;
    ctl.buf = (char *) buf;

    data.maxlen = 0;
    data.len = datalen;
    data.buf = (char *) datap;
	
    errno = 0;
    if (putmsg(fd, &ctl, &data, 0) < 0) 
	return(-1 * errno);
     
    return(0);
}

int dlokack(int fd, char *bufp)
{
    union DL_primitives	*dlp;
    struct strbuf ctl;
    int	flags;
    int ret; 

    ctl.maxlen = MAXDLBUF;
    ctl.len = 0;
    ctl.buf = bufp;

    strgetmsg(fd, &ctl, (struct strbuf*)NULL, &flags, "dlokack");

    dlp = (union DL_primitives *) ctl.buf;

    if ((ret = expecting(DL_OK_ACK, dlp)) < 0)
        return(ret);

    if (ctl.len < sizeof (dl_ok_ack_t))
        return(-1 * DLOKACK_SHORT_RESPONSE);

    if (flags != RS_HIPRI)
	return(-1 * DLOKACK_NOT_M_PCPROTO);

    if (ctl.len < sizeof (dl_ok_ack_t))
	return(-1 * DLOKACK_SHORT_RESPONSE);
	
    return(0);
}

int dlbindack(int fd, char *bufp)
{
    union DL_primitives	*dlp;
    struct strbuf ctl;
    int	flags;
    int ret ;

    ctl.maxlen = MAXDLBUF;
    ctl.len = 0;
    ctl.buf = bufp;

    strgetmsg(fd, &ctl, (struct strbuf*)NULL, &flags, "dlbindack");

    dlp = (union DL_primitives *) ctl.buf;

    if ( (ret=expecting(DL_BIND_ACK, dlp)) < 0)
       return(ret);

    if (flags != RS_HIPRI)
       return(-1 * DLOKACK_NOT_M_PCPROTO);
	
    if (ctl.len < sizeof (dl_bind_ack_t))
       return(-1 * DLOKACK_SHORT_RESPONSE);

    return(0);
}

int dlenabmultireq(int fd, char *addr, int length)
{
    long buf[MAXDLBUF];
    union DL_primitives	*dlp;
    struct strbuf ctl;
    int	flags;

    dlp = (union DL_primitives*) buf;

    dlp->enabmulti_req.dl_primitive = DL_ENABMULTI_REQ;

    dlp->enabmulti_req.dl_addr_length = length;

    dlp->enabmulti_req.dl_addr_offset = sizeof (dl_enabmulti_req_t);

    (void) memcpy((char*)OFFADDR(buf, sizeof (dl_enabmulti_req_t)),
	           addr, length);

    ctl.maxlen = 0;
    ctl.len = sizeof (dl_enabmulti_req_t) + length;
    ctl.buf = (char*) buf;
    flags = 0;

    if (putmsg(fd, &ctl, (struct strbuf*) NULL, flags) < 0)
        return(-1 * errno);

    return(0);
}


int
strgetmsg(int fd, struct strbuf *ctlp, struct strbuf *datap, 
		int *flagsp, char *caller)
{
    int	rc;

    /* Set flags argument and issue getmsg() */

    *flagsp = 0;
    errno = 0;

    if ((rc = getmsg(fd, ctlp, datap, flagsp)) < 0) {
	return(-1 * errno);
    }

    /* Check for MOREDATA and/or MORECTL. */
    if ((rc & (MORECTL | MOREDATA)) == (MORECTL | MOREDATA))
        return(-1 * ERR_MORECTLDATA);
	
    if (rc & MORECTL)
	return(-1 * ERR_MORECTL);

    if (rc & MOREDATA)
	return(-1 * ERR_MOREDATA);
	
    /* Check for at least sizeof (long) control data portion. */
    if (ctlp->len < sizeof (long))
	return(-1 * SHORT_CONTROL_PORTION);

   return(0);
}

int expecting(int prim, union DL_primitives *dlp)
{
    if (dlp->dl_primitive != (u_long)prim)
	return(-1 * DL_PRIMITIVE_ERROR);

    return(0);
}


/* Return string. */
void addrtostring(u_char *addr, u_long length, u_char *s)
{
    int	i;

    for (i = 0; i < length; i++) {
	(void) sprintf((char*) s, "%x:", addr[i] & 0xff);
	s = s + strlen((char*)s);
    }

    if (length)
	*(--s) = '\0';
}

/* Return length */
int stringtoaddr(char *sp, char *addr)
{
    int	n = 0;
    char	*p;
    int	val;

    p = sp;
    while (p = strtok(p, ":")) {
    if (sscanf(p, "%x", &val) != 1)
	return(-1 * INVALID_STRING);
	if (val > 0xff)
	    return(-1 * INVALID_STRING);
	*addr++ = val;
	 n++;
	 p = NULL;
    }
    return (n);
}
/*-------- End of dlcommon routines --------------*/


/* 
 * Send a gratuitous ARP for the specified addresses (IP and hardware)
 * through the device specifed in dev (e.g. "le0").
 *
 * Modified from Gabriel Montenegro's garp function which was based on 
 * dlunitdatareq.c by Neal Nuckolls. 
 */
int		
garp(char *dev, u32 ipaddr, unsigned char *ether_addr)
{
    int	ppa;
    u_char destHWaddr[ETH_ALEN];
    char device[MAX_IFNAME_LEN + 5];	
    int fd;
    int physlen = ETH_ALEN;
    int saplen;
    int	size = sizeof(struct ether_arp);
    int	sapval   = ETHERTYPE_ARP;
    int	localsap = ETHERTYPE_ARP;
    char buf[MAXDLBUF];
    u_char sap[MAXDLADDR];
    u_char addr[MAXDLADDR];
    int	addrlen;
    struct ether_arp req;
    union DL_primitives	*dlp;
    int	i, ret;
    u_long target_ipaddr = ipaddr;

    ifname2devppa(dev, device, &ppa);
    memset(destHWaddr, 0xff, ETH_ALEN);

    /* Validate arguments.*/
    if (ppa < 0) return(-1 * INVALID_PPA);

    /*
     * Initialize buf[] ... why? Are we worried that garbage may be
     * interpreted as valid data retirned from dlokack? 
     */
    for (i = 0; i < MAXDLBUF; i++)
	buf[i] = (unsigned char) i & 0xff;

    errno = 0;

    /* Open the device.*/
    if ((fd = open(device, (O_RDWR | O_NDELAY))) < 0)
	return(-1 * errno);

    /* Attach. */
    if ((ret = dlattachreq(fd, ppa)) < 0) {
	close(fd);
        return(ret);
    }

    if ((ret = dlokack(fd, buf)) < 0) {
	 close(fd);
	 return(ret);
    }

    /* Bind.*/
    if ((ret = dlbindreq(fd, localsap, 0, DL_CLDLS, 0, 0)) < 0) {
	 close(fd);
	 return(ret);
    }

    if ((ret = dlbindack(fd, buf)) < 0) {
	close(fd);
	return(ret);
    }

    /* Get info */
    if ((ret = dlinforeq(fd)) < 0) {
        close(fd);
	return(ret);
    }
 
    if ((ret = dlinfoack(fd, buf)) < 0) {
	 close(fd);
	 return(ret);
    }

    /* Verify sap and phys address lengths */
    dlp = (union DL_primitives*) buf;

    addrtostring(OFFADDR(dlp, dlp->info_ack.dl_addr_offset),
		dlp->info_ack.dl_addr_length, addr);

    saplen = ABS(dlp->info_ack.dl_sap_length);

    /* Convert destination address string to address. */
    for (i = 0; i < saplen; ++i) {
	int rev_index = saplen - 1 - i;
	sap[i] = (sapval >> (rev_index * BITSPERBYTE)) & 0xff;
    }

    if (physlen != (dlp->info_ack.dl_addr_length - saplen)) {
	close(fd);
	return(-1 * INVALID_ADDR); 
    }
       
    addrlen = saplen + physlen;

    /* Construct destination address */
    if (dlp->info_ack.dl_sap_length > 0) {	/* order is sap+phys */
	(void) memcpy((char*)addr, (char*)sap, saplen);
	(void) memcpy((char*)(addr+saplen), (char*)destHWaddr, physlen);
    } else {
	/* order is phys+sap */
	(void) memcpy((char*)addr, (char*)destHWaddr, physlen);
	(void) memcpy((char*)(addr+physlen), (char*)sap, saplen);
    }
	
    /* 
     * Create arp request as follows:
     * Ethernet header :
     * Source address      : Its own hardware address
     * Destination address : ff:ff:ff:ff:ff:ff (Broadcast address)
     * Frame Type          : 0x0806 Arp Request/Reply
     *
     * Arp Packet :
     * Format of hardware address : ARPHRD_ETHER   1
     * Format of protocol address :                0x0800
     * Length of hardware address : ETH_ALEN       6
     * Length of protocol address :                4
     * ARP opcode                 : ARPOP_REQUEST  1
     * Sender hardware address : New ethernet address
     * Target hardware address : Don't care
     * Sender IP address    : IP Address (Cache entry to be updated)
     * Target IP address    : XXXX Don't Care
     *
     */

    memset((char *)&req, 0, sizeof(req));
    req.arp_hrd = htons(ARPHRD_ETHER);
    req.arp_pro = htons(ETHERTYPE_IP);
    req.arp_hln = ETHERADDRL;
    req.arp_pln = IPADDRL;
    req.arp_op = htons(ARPOP_REQUEST);
    (void) memcpy((char *)&req.arp_sha, (char *)ether_addr, ETHERADDRL);
    (void) memcpy((char *)&req.arp_spa, (char *) &ipaddr, IPADDRL);
    (void) memcpy((char *)&req.arp_tpa, (char *) &target_ipaddr, IPADDRL);

    /* Transmit it. */
    if ((ret = dlunitdatareq(fd, addr, addrlen, 0, 0, 
			     (unsigned char *)&req, size)) < 0) {
        close(fd);
	return(ret);
    }

    close(fd);
    return(0);
} /* send_gratuitous_arp() */



/* Get the hardware address for specified interface name. */
int getEthernetAddr(char *ename, unsigned char *eaddr)
{
    int saplen;
    int	localsap = ETHERTYPE_ARP;
    int	fd;
    char buf[MAXDLBUF];
    u_char addr[MAXDLADDR];
    union DL_primitives	*dlp;
    int	i, ret;
    int physlen = 6;
    int ppa ;
    char device[MAX_IFNAME_LEN + 5];
   
    ifname2devppa(ename, device, &ppa);

    /* initialize buf[] */
    for (i = 0; i < MAXDLBUF; i++)
	buf[i] = (unsigned char) i & 0xff;

    errno = 0;
    /* Open the device.*/
    if ((fd = open(device, (O_RDWR | O_NDELAY))) < 0)
	return(-1 * errno);

    /* Attach. */
    if ((ret = dlattachreq(fd, ppa)) < 0) {
	close(fd);
        return(ret);
    }

    if ((ret = dlokack(fd, buf)) < 0) {
	 close(fd);
	 return(ret);
    }

    /* Bind.*/
    if ((ret = dlbindreq(fd, localsap, 0, DL_CLDLS, 0, 0)) < 0) {
	 close(fd);
	 return(ret);
    }

    if ((ret = dlbindack(fd, buf)) < 0) {
	close(fd);
	return(ret);
    }

    /* Get info */
    if ((ret = dlinforeq(fd)) < 0) {
        close(fd);
	return(ret);
    }
 
    if ((ret = dlinfoack(fd, buf)) < 0){
	 close(fd);
	 return(ret);
    } 

    /* Verify sap and phys address lengths */
    dlp = (union DL_primitives*) buf;

    addrtostring(OFFADDR(dlp, dlp->info_ack.dl_addr_offset),
	dlp->info_ack.dl_addr_length, addr);

    saplen = ABS(dlp->info_ack.dl_sap_length);

    /* Construct destination address.*/
    if (dlp->info_ack.dl_sap_length > 0) {	/* order is sap+phys */
	/* obtain our MAC address */
	memcpy((unsigned char *)eaddr,
		OFFADDR(dlp, 
		dlp->info_ack.dl_addr_offset + saplen),
		physlen);
    } else {	/* order is phys+sap */	
	/* Obtain our MAC address */ 
	memcpy((unsigned char *)eaddr,
		OFFADDR(dlp, 
		dlp->info_ack.dl_addr_offset),
		physlen);
    }
    
    close(fd);
    return(0);
}


/*
 * Send a gratuitous arp for mn. 
 * TODO: Try to use a routine that
 * returns a colon separated string for an ethernet address instead
 * of directly printing it since ether-Print may be controlled by
 * a different level of mipdebuglevel. Then remove the if statement.
 */
int sendgarp(char *iface)
{
    IfaceEntry *ifaceIdx;

    if ((ifaceIdx = findIface(iface)) != NULL) {
        if (mipdebug > 4) {
	    mip5dbg(("Gratuitous arp sent for %s on interface %s " \
		     "with adress\n",
                     ntoa(mnHomeAddr, addr1), iface));
            ether_print(ifaceIdx->maIfaceHWaddr);	
        }
        return(garp(iface, mnHomeAddr, ifaceIdx->maIfaceHWaddr));
    } else
        return(-1);
}

/* 
 * Flush the routing table except for entries corresponding to the
 * tunnel (and multicast routes?).
 */
int routeflush()
{
    int sd ;	/* mib stream */
    mib_item_t	*item, *last_item;
    mib2_ipRouteEntry_t * rp;
    struct rtentry rt;
    struct sockaddr_in *sin;
    u32 gw;

    if ((sd = open("/dev/ip", O_RDWR)) < 0) 
	return(-1 * errno);

    if ((item = mibget(sd)) == nilp(mib_item_t)) {
	close(sd);
	return(-1 * errno);
    }
    close(sd);	

    while (item) {
        /* skip all the other trash that comes up the mib stream */
        if ((item->group != MIB2_IP) || (item->mib_id != MIB2_IP_21)) {
	    last_item = item;
	    item = item->next_item;
	    free(last_item);
	    continue;
        }
	rp = (mib2_ipRouteEntry_t *)item->valp;
		
	while ((u_long) rp < (u_long) (item->valp + item->length)) {
	    if ((rp->ipRouteInfo.re_ire_type == IRE_GATEWAY) ||
		(rp->ipRouteInfo.re_ire_type & IRE_INTERFACE) ||
	        (rp->ipRouteInfo.re_ire_type == IRE_ROUTE_ASSOC)) {
		           
	        /* Ignore routing entry created by vtunl */
		if (((u32) rp->ipRouteDest) != inet_addr("255.255.255.254")) {
		    /* Prepare an rt structure for SIOCDELRT */
		    bzero ((char *) &rt, sizeof(rt));

		    /* first fill out the destination */
		    sin = (struct sockaddr_in *) &rt.rt_dst;
		    sin->sin_family = AF_INET;
		    bcopy ((char *) &rp->ipRouteDest,
			   (char *) &sin->sin_addr,
			   sizeof(struct sockaddr_in));

		    /* and then the gateway */
		    sin = (struct sockaddr_in *) &rt.rt_gateway;
		    sin->sin_family = AF_INET;
		    bcopy ((char *) &rp->ipRouteNextHop,
                           (char *) &sin->sin_addr,
                           sizeof(struct sockaddr_in));
		    
		    gw = sin->sin_addr.s_addr;
		    rt.rt_flags |= RTF_GATEWAY;
		    
		    if (rp->ipRouteMask == 
			       (IpAddress) inet_addr("255.255.255.255"))
		        rt.rt_flags |= RTF_HOST;

		    errno = 0;
		    
		    /* 
		     * TODO: Do not even print the ignore message if
		     * error is due to ENXIO. [Leave like this for now.]
		     */
		    /* TODO: make sure entries associated with interfaces in
		       ignoreIfaceList() are not deleted.

		       mip5dbg(("Deleting route for dst %s on iface %s\n",
			     ntoa((u32) rp->ipRouteDest, addr1),
			     ???????????????));
		     */
		    if (ioctl(ioctl_sockid, SIOCDELRT, &rt) < 0) {
		        mip5dbg(("Ignoring error in route delete %s %s: %s\n",
				 ntoa((u32) rp->ipRouteDest, addr1), 
				 ntoa(gw, addr2), 
				 err2str(-1 * errno)));
		    }
		}
	    }
	    rp++;
	}

	/* 
	 * TODO: Verify that this is indeed correct. It seems
	 * that if we do not do this, we end up using resources
	 * that are not properly freed.
	 */
	last_item = item;
	item = item->next_item;
	free(last_item);
    }

    return(0);
}

/* Flush the arp table */
int arpflush()
{
    int sd;	/* mib stream */
    mib_item_t	*item, *last_item;
    mib2_ipNetToMediaEntry_t *rp;
    struct arpreq ar;
    struct sockaddr_in *sin;

    if ((sd = open("/dev/ip", O_RDWR)) < 0) {
	return(-1);
    }

    /* must be above ip, below tcp */
    if (ioctl(sd, I_PUSH, "arp") == -1) {
       (void) close(sd);
        return (-1);
    }
	
    if ((item = mibget(sd)) == nilp(mib_item_t)) {
	close(sd);
	return(-1 * errno);
    }
    close(sd);	

    while (item) {

	/* skip all the other trash that comes up the mib stream */
        if ((item->group != MIB2_IP) || (item->mib_id != MIB2_IP_22)) {
    	    last_item = item;
	    item = item->next_item;
	    free(last_item);	  
	    continue;
        }

	rp = (mib2_ipNetToMediaEntry_t *)item->valp;
		
	while ((char *) rp < (item->valp + item->length)) {
	    /* got one - flush it */
	    /* 
	     * For now we assume that we can delete ALL arp entries and
	     * the mobile node code does the right thing for regaining
	     * any useful ARP entries (those with the ARP_SP or
	     * ARP_SM flags. 
	     * TODO: Make the mobile node multicast savvy, e.g. 
	     * add a multicast route through "current" interface.
	     * Revisit this issue when we actually implement disableArp()
	     * and restoreArp().
	     */
	    bzero((caddr_t)&ar, sizeof ar);
	    ar.arp_pa.sa_family = AF_INET;
	    sin = (struct sockaddr_in *)&ar.arp_pa;
	    sin->sin_family = AF_INET;
	    sin->sin_addr.s_addr = (rp->ipNetToMediaNetAddress);

	    /* Delete the ARP entry */
	    /* TODO: make sure entries associated with interfaces in
	       ignoreIfaceList() are not deleted.

	       mip5dbg(("Deleting ARP entry for %s on iface %s\n",
	                ntoa(rp->ipNetToMediaNetAddress, addr1),
			rp->ipNetToMediaIfIndex.o_bytes));
	     */     
	    errno = 0;
	    if (ioctl(ioctl_sockid, SIOCDARP, (caddr_t) &ar) < 0) {
	        if ((rp->ipNetToMediaInfo.ntm_flags & ARP_SM) != ARP_SM)
		    mip5dbg(("Ignoring error in arp -d %s: %s \n", 
			     ntoa(sin->sin_addr.s_addr, addr1), 
			     err2str(-1 *errno)));
		  
	    }
	    rp++;
	}

	/* 
	 * TODO: Verify that this is indeed correct. It seems
	 * that if we do not do this, we end up using resources
	 * that are not properly freed.
	 */
	last_item = item;
	item = item->next_item;
	free(last_item);
    }
    return(0);
}

/*
 * Get the MIB for IP. Currently the MIB request value MIB2_IP
 * is assigned in the function itself. It can be passed as 
 * a parameter. This always returns the entire IP MIB, the arpflush()
 * routeflush() and other routines filter out stuff they don't need.
 */  
mib_item_t *mibget (int sd)
{
    char buf[512];
    int	flags;
    int	j, getcode;
    struct strbuf ctlbuf, databuf;
    struct T_optmgmt_req *tor = (struct T_optmgmt_req *)buf;
    struct T_optmgmt_ack *toa = (struct T_optmgmt_ack *)buf;
    struct T_error_ack *tea = (struct T_error_ack *)buf;
    struct opthdr *req;
    mib_item_t *first_item = nilp(mib_item_t);
    mib_item_t *last_item  = nilp(mib_item_t);
    mib_item_t *temp;
	
    tor->PRIM_type = T_OPTMGMT_REQ;
    tor->OPT_offset = sizeof(struct T_optmgmt_req);
    tor->OPT_length = sizeof(struct opthdr);
    tor->MGMT_flags = T_CURRENT;
    req = (struct opthdr *)&tor[1];
    req->level = MIB2_IP ;		/* any MIB2_xxx value ok here */
    req->name  = 0;
    req->len   = 0;

    ctlbuf.buf = buf;
    ctlbuf.len = tor->OPT_length + tor->OPT_offset;
    flags = 0;
    if (putmsg(sd, &ctlbuf, nilp(struct strbuf), flags) == -1) {
	goto error_exit;
    }

    /*
     * Each reply consists of a ctl part for one fixed structure 
     * or table, as defined in mib2.h.  The format is a T_OPTMGMT_ACK,
     * containing an opthdr structure.  level/name identify the entry,
     * len is the size of the data part of the message.
     */
    req = (struct opthdr *)&toa[1];
    ctlbuf.maxlen = sizeof(buf);
    for (j=1; ; j++) {
	flags = 0;
	getcode = getmsg(sd, &ctlbuf, nilp(struct strbuf), &flags);
	if (getcode == -1) {
	    goto error_exit;
	}
	if (getcode == 0
		&& ctlbuf.len >= sizeof(struct T_optmgmt_ack)
		&& toa->PRIM_type == T_OPTMGMT_ACK
		&& toa->MGMT_flags == T_SUCCESS
		&& req->len == 0) {
		return first_item;		/* this is EOD msg */
	}

	if (ctlbuf.len >= sizeof(struct T_error_ack)
	    && tea->PRIM_type == T_ERROR_ACK) {
	    errno = (tea->TLI_error == TSYSERR)
		    ? tea->UNIX_error : EPROTO;
	    goto error_exit;
	}
			
	if (getcode != MOREDATA
		|| ctlbuf.len < sizeof(struct T_optmgmt_ack)
		|| toa->PRIM_type != T_OPTMGMT_ACK
		|| toa->MGMT_flags != T_SUCCESS) {
	    errno = ENOMSG;
	    goto error_exit;
	}

	temp = (mib_item_t *)malloc(sizeof(mib_item_t));
	if (!temp) {
	    goto error_exit;
	}

	if (last_item)
	    last_item->next_item = temp;
	else
	    first_item = temp;

	last_item = temp;
	last_item->next_item = nilp(mib_item_t);
	last_item->group = req->level;
	last_item->mib_id = req->name;
	last_item->length = req->len;
	last_item->valp = (char *)malloc(req->len);

	databuf.maxlen = last_item->length;
	databuf.buf    = last_item->valp;
	databuf.len    = 0;
	flags = 0;
	getcode = getmsg(sd, nilp(struct strbuf), &databuf, &flags);
	if (getcode == -1) {
	    goto error_exit;
	} else if (getcode != 0) {
	    goto error_exit;
	}
    }

error_exit:
    while (first_item) {
	last_item = first_item;
	first_item = first_item->next_item;
	free(last_item);
    }
    return first_item;
}

/* 
 * Tries to pick up a default router from the routing table such that
 * the router lies on the same network as the supplied ipAddr. The
 * netMask is used to determine when two addresses belong to the same
 * network.
 * TODO: For now, if there is a problem, the function returns a zero 
 *       gateway.
 */
u32 getRouterAddr(u32 ipAddr, u32 netMask)
{
    int sd ;	/* mib stream */
    mib_item_t	*item, *last_item;
    mib2_ipRouteEntry_t * rp;
    struct rtentry rt;
    struct sockaddr_in *sin;
    u32 dst, gw;

    if ((sd = open("/dev/ip", O_RDWR)) < 0)
	return(0);

    if ((item = mibget(sd)) == nilp(mib_item_t)) {
	close(sd);
	return(0);
    }
    close(sd);	

    while (item) {
        /* skip all the other trash that comes up the mib stream */
        if ((item->group != MIB2_IP) || (item->mib_id != MIB2_IP_21)) {
	    last_item = item;
	    item = item->next_item;
	    free(last_item);
	    continue;
        }
	rp = (mib2_ipRouteEntry_t *)item->valp;
		
	while ((u_long) rp < (u_long) (item->valp + item->length)) {
	    if ((rp->ipRouteInfo.re_ire_type == IRE_GATEWAY) ||
		(rp->ipRouteInfo.re_ire_type & IRE_INTERFACE) ||
	        (rp->ipRouteInfo.re_ire_type == IRE_ROUTE_ASSOC)) {

	        /* got one - process it */
		bzero ((char *) &rt, sizeof(rt));
				
		sin = (struct sockaddr_in *) &rt.rt_dst;
		sin->sin_family = AF_INET;
                bcopy ((char *) &rp->ipRouteDest,
                       (char *) &sin->sin_addr,
                       sizeof(struct sockaddr_in));
		dst = sin->sin_addr.s_addr;				
				 
	        if (dst == 0) {
		    sin = (struct sockaddr_in *) &rt.rt_gateway;
		    sin->sin_family = AF_INET;
                    bcopy ((char *) &rp->ipRouteNextHop,
                           (char *) &sin->sin_addr,
                            sizeof(struct sockaddr_in));
		    gw = sin->sin_addr.s_addr;
                    if (((gw ^ ipAddr) &  netMask) == 0) { 
		       return(gw);
                    }
	        }
            }
	    rp++;
        }

	/* 
	 * TODO: Verify that this is indeed correct. It seems
	 * that if we do not do this, we end up using resources
	 * that are not properly freed.
	 */
	last_item = item;
	item = item->next_item;
	free(last_item);
    }

    return(0);
}

/* Used by openDeviceFd() */
strioctl(int fd, int cmd, int timout, int len, char *dp)
{
    struct strioctl sioc;
    int	rc;

    sioc.ic_cmd = cmd;
    sioc.ic_timout = timout;
    sioc.ic_len = len;
    sioc.ic_dp = dp;
    rc = ioctl(fd, I_STR, &sioc);

    if (rc < 0)
	return (rc);
    else
	return (sioc.ic_len);
}

/*
 * Open the file descriptor for the interface. 
 */
int openDeviceFd(char *ifaceName)
{
    int fd;
    char    device[MAX_IFNAME_LEN + 5];
    int     ppa;
    char    buf[MAXDLBUF];
    /* Be generous, accept advertisements on 255.255.255.255 and 224.0.0.1 */
    unsigned char bcast_hw_addr[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
    unsigned char mcast_hw_addr[6] = {0x01, 0x00, 0x5e, 0x00, 0x00, 0x01};  

    ifname2devppa(ifaceName, device, &ppa);
  
    /* 
     * Open the device. Subsequent reads should be non-blocking.
     * The stream is used only to receive agent advertisements.
     */
    if ((fd = open(device, (O_RDWR | O_NDELAY))) < 0) {
        return(-1 * errno);
    }

    /* Attach.*/
    dlattachreq(fd, ppa);
    dlokack(fd, buf);


    /* Bind. Use sap=0x0800 for IP. */
    dlbindreq(fd, 0x800, 0, DL_CLDLS, 0, 0);
    dlbindack(fd, buf);

    /*
     * Enable promiscuous mode (only get multicast/b'cast pkts)
     *
     * dlpromisconreq(fd, DL_PROMISC_MULTI);
     * dlokack(fd, buf);
     */

    /* TODO: Don't know if this works, test it! */
    dlenabmultireq(fd, (char *)bcast_hw_addr, 6);
    dlokack(fd, buf);

    dlenabmultireq(fd, (char *)mcast_hw_addr, 6);
    dlokack(fd, buf);

    /* 
     * Issue DLIOCRAW ioctl. 
     * TODO: Verify if the timeout value makes sense, we were previously
     * using infinity (-1). It is the third argument to strioctl.
     */
    
    if (strioctl(fd, DLIOCRAW, 4, 0, NULL) < 0) {
       close(fd);
       return(-1 * errno);
    } 

    /* Flush the read side of the Stream.*/
    if (ioctl(fd, I_FLUSH, FLUSHR) < 0) {
        close(fd);
	return(-1 * errno);
    }

    return(fd);
}


/* 
 * Return the index in the advertisement buffer for an advertisement
 * whose source is ip_src. Returns (-1) if not found.
 */
int isAdvForSrcIPAddrPresent(u32 ip_src)
{
    int i;

    if (advPktBuf.head == advPktBuf.tail)
       return(-1);

    i = advPktBuf.head;
    while (i != advPktBuf.tail) {
        if (advPktBuf.advPkt[i].advSrcIPAddr == ip_src) 
	    return(i);
	else 
	    i = ((i + 1) % MAX_AGENT_ADV_ENTRIES);	    
   }
   return(-1);
}

/* 
 * Return the index for an advertisement heard on the given interface.
 * Returns (-1) if not found.
 */
int isAdvForIfaceNamePresent(char *ifaceName)
{
    int i;

    if (advPktBuf.head == advPktBuf.tail)
       return(-1);

    i = advPktBuf.head; 
    while(i != advPktBuf.tail) {
        if (strcasecmp(advPktBuf.advPkt[i].ifaceName, ifaceName) == 0) 
	    return(i);
	else 
	    i = (i + 1) % MAX_AGENT_ADV_ENTRIES;	    
   }
   return(-1);
}

/* 
 * Stores the advertisement heard in the Advertisement packet buffer. 
 * AdvBufLock is used while updating and the routine is called by the
 * "secondary" thread (one that monitors agent advertisements).
 */
void putAdvPkt(char *data, int datalen, u32 ip_src, 
   	       char *hw_src, char *ifaceName)
{
    int i, idx;

    pthread_mutex_lock(&(advPktBuf.AdvBufLock));
   
    /* Find a suitable slot */
    if ((idx = isAdvForSrcIPAddrPresent(ip_src)) < 0) {
        i = advPktBuf.tail;
    } else {
        i = idx;
    }

    /* Store it! */
    strcpy(advPktBuf.advPkt[i].ifaceName, ifaceName);
    advPktBuf.advPkt[i].advSrcIPAddr = ip_src;
    memcpy(advPktBuf.advPkt[i].hwAddr, hw_src, MAX_HWADDR_LEN);    
    memcpy(advPktBuf.advPkt[i].data, data, datalen);
    advPktBuf.advPkt[i].dataLen = datalen;
    
    /* If a new slot was used, update head and tail */
    if (idx < 0) {
        advPktBuf.tail = (advPktBuf.tail + 1) % MAX_AGENT_ADV_ENTRIES;
	/* 
	 * We always keep head and tail separated unless the buffer
	 * is empty, even if this dropping an advertisement.
	 */
        if (advPktBuf.tail == advPktBuf.head)
	   advPktBuf.head = (advPktBuf.head + 1) % MAX_AGENT_ADV_ENTRIES;
    }

    pthread_mutex_unlock(&advPktBuf.AdvBufLock);
}

/*
 * Gets an agent advertisement stored in Adv. Pkt. buffer 
 * heard on the specified interface.
 * This routine is called by the main thread and also uses
 * AdvBufLock to synchronize access to the shared advertisement buffer.
 */
boolean getAgentAdvByIface(char *pkt, int *pktlen, char *ifaceName,
			   char *hwAddr)
{
    int i;	
 
    pthread_mutex_lock(&(advPktBuf.AdvBufLock));

    if ((i = isAdvForIfaceNamePresent(ifaceName)) < 0) {
	pthread_mutex_unlock(&advPktBuf.AdvBufLock);
	return(FALSE);
    }

    /* read out the advertisement */
    (*pktlen) = advPktBuf.advPkt[i].dataLen;
    memcpy(pkt, advPktBuf.advPkt[i].data, *pktlen);
    memcpy(hwAddr,advPktBuf.advPkt[i].hwAddr, MAX_HWADDR_LEN);	

    /* Keep the buffer contiguous. */
    memcpy(&(advPktBuf.advPkt[advPktBuf.tail - 1]), 
	   &(advPktBuf.advPkt[i]), sizeof(AgentAdvPkt));

    if (advPktBuf.tail == 0)
	advPktBuf.tail = (MAX_AGENT_ADV_ENTRIES - 1);
    else
	advPktBuf.tail--;

    pthread_mutex_unlock(&(advPktBuf.AdvBufLock));
    return(TRUE);
}

/* 
 * Remove all the agent advertisements heard on the specified
 * interface.
 */
void purgeAdvPkts(char *ifaceName)
{
    int i;	
 
    pthread_mutex_lock(&(advPktBuf.AdvBufLock));

    i = isAdvForIfaceNamePresent(ifaceName);    
    while (i >= 0) {
        /* 
	 * Keep the buffer contiguous by copying the last item
	 * into slot i
	 */
        memcpy(&(advPktBuf.advPkt[advPktBuf.tail - 1]), 
	       &(advPktBuf.advPkt[i]), sizeof(AgentAdvPkt));

	if (advPktBuf.tail == 0)
	    advPktBuf.tail = (MAX_AGENT_ADV_ENTRIES - 1);
	else
	    advPktBuf.tail--;

	i = isAdvForIfaceNamePresent(ifaceName);    
    }

    pthread_mutex_unlock(&(advPktBuf.AdvBufLock));
}

/* 
 * Check whether the ethernet frame in the data buffer is a Mobile IP
 * agent advertisement. When TRUE is returned, srcHWaddr contains
 * the hardware address of the source of the advertisement and
 * avdSrcAddr contains the IP address of the advertisement source.
 */
boolean isAgentAdv(struct strbuf *data, char *srcHWaddr, u32 *advSrcAddr)
{
    int ip_hdr_len, i;
    u32 ip_src;
    int icmp_offset;
    int router_adv_entries;
    int router_adv_entry_size;
    int mobility_ext_offset;
    advExt *advExtPtr;

    /* 
     * The pkt should have ethernet (14 bytes) and IP headers (20 bytes), 
     * at least 8 bytes of ICMP Router Advertisement and another 8 
     * bytes of mobility extensions. IP protocol should be 1 (ICMP) and
     * Ethernet payload should be 0x0800 (IP).
     */
    if ((data->len < 50) || (data->buf[14+9] != 1) ||
	(data->buf[12] != 8) || (data->buf[13] != 0)) {
        return(FALSE);
    }

    /* 
     * Find true IP header length. Pkt should be atleast 30+IP_hdr_len
     * bytes and have ICMP type 16 and ICMP code of 0 or 16.
     */ 
    ip_hdr_len = (data->buf[14] & 0xf) << 2;
    icmp_offset = 14 + ip_hdr_len;
    if ((data->len < (30 + ip_hdr_len)) || (data->buf[icmp_offset] != 9) || 
	((data->buf[icmp_offset+1] !=0) && (data->buf[icmp_offset+1] != 16))) {
	/*printf("Expected ICMP type 9, code  0 or 16\n");*/
	return(FALSE);
    }

    router_adv_entries = data->buf[icmp_offset + 4];
    router_adv_entry_size = data->buf[icmp_offset + 5] << 2;
    mobility_ext_offset = icmp_offset + 8 + 
                          (router_adv_entries * router_adv_entry_size);

    /* Should have a mobility extension: type 16, F or H-bit on */
    if ((data->len < (mobility_ext_offset + 8)) ||
	(data->buf[mobility_ext_offset] != 16) ||
	((data->buf[mobility_ext_offset + 6] & 0x30) == 0)) {
        return(FALSE);
    }
	
    /*
     * Ignore advertisements in which the F bit is set but no care-of
     * addresses are listed.
     */ 

    advExtPtr = (advExt *)(data->buf + mobility_ext_offset);
 
    if ( (data->buf[mobility_ext_offset + 6] & ADV_IS_FOREIGN_AGENT) &&
        ((((((advExtPtr->length) - 6)/4) <= 0) ||
           ((mobility_ext_offset + 12) > data->len) )))
       return(FALSE);

    /* 
     * Check that if B or R is set then F is also set, i.e. it
     * is illegal to have B or R set for a mobility agent that does not
     * advertise FA service.
     */
    if (((data->buf[mobility_ext_offset + 6]) & 
	 (ADV_IS_BUSY | ADV_REGISTRATION_REQUIRED)) &&
	 ((data->buf[mobility_ext_offset + 6] & ADV_IS_FOREIGN_AGENT) == 0))
      return(FALSE);

    memcpy(srcHWaddr, (data->buf + 6), 6);
    ip_src = 0;
    for (i = 26; i < 30; i++) 
        ip_src = (ip_src << 8) | (0xff & data->buf[i]);

    *advSrcAddr = htonl(ip_src);    

    /* TODO: process other advertisement extensions. */

    return(TRUE);
}

/* 
 * Thread routine for monitoring agent advertisements. The argument
 * is never used. This is the function executed by the secondary
 * thread.
 */
void *monitorAgentAdv(void *ignoreThis)
{
    unsigned char buf[MAXDLBUF]; 
    struct  strbuf data;
    int	flags;
    int	i;
    u32 ip_src;
    char    hw_src[MAX_HWADDR_LEN];
    int localIfaceFD[DEFAULT_MAX_INTERFACES];
    char localIfaceFDName[DEFAULT_MAX_INTERFACES][MAX_IFNAME_LEN]; 
    int localIfaceFDCnt;  
    struct timeval tv;

    data.buf = (char *) buf;
    data.maxlen = MAXDLBUF;
    
    /* Read packets.*/
    while (1) {
        /* 
	 * Copy the name of all interfaces and open file descriptors
	 * to local storage so we don't hold the lock for too long.
	 */ 
	pthread_mutex_lock(&(ifaceFDList.ifaceFDLock));
        localIfaceFDCnt = ifaceFDList.ifaceFDCnt;
        for (i = 0; i < ifaceFDList.ifaceFDCnt; i++) {
	    localIfaceFD[i] = ifaceFDList.ifaceFDarray[i].ifaceFD;
	    strcpy(localIfaceFDName[i], 
		   ifaceFDList.ifaceFDarray[i].ifaceName);
        }
        pthread_mutex_unlock(&(ifaceFDList.ifaceFDLock));

	for (i = 0; i < localIfaceFDCnt; i++) {
	    data.len = 0;
	    flags = 0;

	    /*
	     * When opening the FD, we made sure getmsg won't block
	     * if no message is present.
	     */
            if ((getmsg(localIfaceFD[i], NULL, &data, &flags)) != 0) 
		 continue;

            if (isAgentAdv(&data, hw_src, &ip_src)) {
	         putAdvPkt((data.buf + 14), (data.len - 14), ip_src,
			   hw_src, localIfaceFDName[i]); 
            }
        }
	
	/* Give up the CPU for a while */
        tv.tv_sec = 0; tv.tv_usec = 50000 ;
        select(FD_SETSIZE, NULL, NULL, NULL, &tv);
    }
}

/*
 * Configure the iface "ifaceName" with IP address "addr"
 * and netmask  "netMask". Bring it up. 
 * Interface is already plumbed, we only need to configure its
 * address and netmask.
 *
 * TODO: To be resolved -- can't assign IP address through ioctl call.
 *         Need to figure out ..
 * For now, we use configiface() instead which is based on system
 * commands.
 */   
int ConfigIface(char *ifaceName, u32 addr, u32 netMask)
{
    struct ifreq ifr;
    struct sockaddr_in *sin;
    u32 flags;

    bzero ((char *) &ifr, sizeof(ifr));
    strncpy(ifr.ifr_name, ifaceName, sizeof(ifr.ifr_name));

    if (ioctl(ioctl_sockid, SIOCGIFFLAGS, (caddr_t)&ifr) < 0){
	  return(-1 * errno);
    }
    flags = ifr.ifr_flags;

    /*
    ifr.ifr_flags &= ~(IFACE_UP);
    if (ioctl(ioctl_sockid, SIOCSIFFLAGS, (caddr_t)&ifr) < 0){
	  return(-1 * errno);
    }
    */

    /* Set the IP Address */
    sin = (struct sockaddr_in *)&ifr.ifr_addr;     
    sin->sin_addr.s_addr = addr;

    if (ioctl(ioctl_sockid, SIOCSIFADDR, (caddr_t)&ifr) < 0){
	mip5dbg(("Error in setting address %s for %s\n",
		 ntoa(addr, addr1), ifaceName));  
	return(-1 * errno);
    }

    /* Set the netmask */
    sin = (struct sockaddr_in *)&ifr.ifr_addr;     
    sin->sin_addr.s_addr = netMask;

    if (ioctl(ioctl_sockid, SIOCSIFNETMASK, (caddr_t)&ifr) < 0){
	mip5dbg(("Error in setting address %s for %s\n",
		 ntoa(netMask, addr1), ifaceName));  
	return(-1 * errno);
    }

    /* Bring up the interface */
    ifr.ifr_flags = flags | IFACE_UP;
    if (ioctl(ioctl_sockid, SIOCSIFFLAGS, (caddr_t)&ifr) < 0){
	  return(-1 * errno);
    }

    return(0);
}

/*
 * Converts OS-specific flags into a canonical OS-independent form
 * as defined in the user level code (mobile.h).
 * TODO: Check if the flags should be u16 instead of u32.
 */
void adjustIfaceFlags(u32 *newFlag, u32 oldFlag)
{
    *newFlag = (oldFlag & 0x00ff);

    if (oldFlag & IFF_DHCPRUNNING)
        (*newFlag) |= IFACE_DHCP;

    /* 
     * TODO: Set IFACE_ON_NET only if other address is reachable.
     * This requires ping or something similar. For now, assume that
     * this flag is always set.
     */ 
    (*newFlag) |= IFACE_ON_NET;     
}

/*
 * An alternative to the above configIface() 
 * Configure the iface "ifaceName" with IP address "addr"
 * and netmask  "netMask". Bring it up. 
 * Interface is already plumbed, we only need to configure its
 * address and netmask.
 * TODO: We need to be very careful on how we proceed in the case
 *       of point-to-point interfaces.
 */
int configIface(char *ifaceName, u32 addr, u32 netMask)
{
    
    sprintf(msg, "/usr/sbin/ifconfig %s down", ifaceName);
    system(msg);
    
    /* Solaris automatically sets the broadcast address. */
    mip5dbg(("Bringing up interface %s with address %s mask %s\n",
	     ifaceName, ntoa(addr, addr1), ntoa(netMask, addr2)));
    sprintf(msg, "/usr/sbin/ifconfig %s %s netmask %s up", 
	    ifaceName, ntoa(addr, addr1), ntoa(netMask, addr2));
    system(msg);

    return(0);
}

/* 
 * Assign the specified address and netmask to the clone interface.
 * Also retain the parent's interface flags (as far as possible).
 * This function is also responsible for deleting any spurious routing
 * entries that may be added as part of bringing up the clone Iface.
 */
int configCloneIface(char *cloneIface, u32 addr, u32 netMask, 
		     char *parentIface)
{
    struct ifreq ifr, cloneifr;
    u32 flags;
    int errcode;

    bzero ((char *) &ifr, sizeof(ifr));
    strncpy(ifr.ifr_name, parentIface, sizeof(ifr.ifr_name));

    bzero ((char *) &cloneifr, sizeof(cloneifr));
    strncpy(cloneifr.ifr_name, cloneIface, sizeof(cloneifr.ifr_name));

    if (ioctl(ioctl_sockid, SIOCGIFFLAGS, (caddr_t)&ifr) < 0) {
        return(-1 * errno);
    }

    adjustIfaceFlags(&flags, ifr.ifr_flags);

    /* copy all flags to the clone except for DHCP */
    cloneifr.ifr_flags = ifr.ifr_flags & ~IFF_DHCPRUNNING;
    if (ioctl(ioctl_sockid, SIOCSIFFLAGS, (caddr_t) &cloneifr) < 0)
        return(-1 * errno);
 
    if (flags & IFACE_POINTOPOINT) { 
        /* Assign parent iface's dstaddr to the clone */
        if ((ioctl(ioctl_sockid, SIOCGIFDSTADDR, (caddr_t) &ifr)) < 0)
	    return(-1 * errno);

        cloneifr.ifr_dstaddr = ifr.ifr_dstaddr;
        if ((ioctl(ioctl_sockid, SIOCSIFDSTADDR, (caddr_t) &cloneifr)) < 0)
	    return(-1 *  errno);
    }

    /* 
     * THINK SOME MORE ABOUT THIS ORDERING ... TEST IT!
     * TODO: We should assign the address etc from parent iface
     * by explicitly calling ioctls.
     */
    if ((errcode = configIface(cloneIface, addr, netMask)) < 0) 
        return(-1);

    /* 
     * TODO: This should be part of configCloneIface since which
     * route to delete really depends on whether or not the interface
     * is pointopoint. We need to be careful how we deal with pointopoint
     * interfaces.
     */

    if ((errcode = routedel((addr & netMask), netMask, 
			    addr, 0)) < 0) {
      mip2dbg(("Error in routedel %s %s ... %s\n", 
	       ntoa((mnHomeAddr & mnHomeNetmask), addr1),
	       ntoa(mnHomeAddr, addr2), err2str(errcode)));
    }

   return(0);
}

/* 
 * This function is called from closeIfaceFD which should make
 * sure it acquires appropriate locks before accessing the shared
 * data structure ifaceFDList.
 */
int findIfaceFDIdx(char *ifname) {
    int i;
    for (i = 0; i < ifaceFDList.ifaceFDCnt; i++) {
       if (strcmp(ifaceFDList.ifaceFDarray[i].ifaceName, ifname) == 0)
	    return(i);
    }
    return(-1);		
}

/* 
 * Opens a stream to receive agent advertisements and adds the stream
 * and corresponding interface name to ifaceFDList. Access to shared
 * data structure ifaceFDList should be protected by a lock.
 */
void addIfaceFD(char *ifname)
{
    int fd;

    pthread_mutex_lock(&(ifaceFDList.ifaceFDLock));
    if (ifaceFDList.ifaceFDCnt == DEFAULT_MAX_INTERFACES) {
        mip1dbg(("No space for agent adv. stream on iface %s.\n"));
    } else {
        if ((fd = openDeviceFd(ifname)) < 0) {
	    mip1dbg(("Error in opening agent adv. stream for %s ... %s\n",
		     ifname, err2str(fd)));
	} else {
	    ifaceFDList.ifaceFDarray[ifaceFDList.ifaceFDCnt].ifaceFD = fd;
	    strcpy(ifaceFDList.ifaceFDarray[ifaceFDList.ifaceFDCnt].ifaceName,
		   ifname);
	    ifaceFDList.ifaceFDCnt++;  
	}
    }
    pthread_mutex_unlock(&(ifaceFDList.ifaceFDLock));
    mip5dbg(("addIfaceFD() called for interface %s\n", ifname)); 
}


/* 
 * Closes the agent advertisement stream corresponding to the specified
 * iface. Access to shared data structure ifaceFDList should be protected
 * by a lock.
 */

void closeIfaceFD(char *ifname)
{
    int idx;

    pthread_mutex_lock(&(ifaceFDList.ifaceFDLock));
    if ((idx = findIfaceFDIdx(ifname)) < 0)
	mip0dbg(("Did not find an agent adv. stream for iface %s\n", ifname));
    else {
       close(ifaceFDList.ifaceFDarray[idx].ifaceFD);
       ifaceFDList.ifaceFDCnt--;
       /* make the list contiguous */
       if ((ifaceFDList.ifaceFDCnt != idx) && 
	   (ifaceFDList.ifaceFDCnt != 0)) {
          ifaceFDList.ifaceFDarray[idx].ifaceFD =  
	      ifaceFDList.ifaceFDarray[ifaceFDList.ifaceFDCnt].ifaceFD;
	  strcpy(ifaceFDList.ifaceFDarray[idx].ifaceName,
		 ifaceFDList.ifaceFDarray[ifaceFDList.ifaceFDCnt].ifaceName);
       } 
    }
    pthread_mutex_unlock(&(ifaceFDList.ifaceFDLock));

    /* 
     * TODO: Purge the advertisements heard on this interface. By releasing
     * the ifaceFDLock before purgeAdvPkts we avoid the potential
     * problems caused by holding nested locks.
     */
    if (idx >= 0) {
        purgeAdvPkts(ifname);
    }
    mip5dbg(("closeIfaceFD called for %s.\n", ifname));
}

/*
 * Checks whether the interface is useful (either for monitoring
 * agent advertisements or using a co-located care-of address).
 *
 * Checks the following :
 *   The interface does not belong to the list of interfaces to be ignored.
 *   The interface name does not have colon(:) in it.
 *   The interface is UP/not loopback ..
 *   More tests on flags to be made
 * 
 *  TODO:  Need to make it more strict
 */
boolean isIfaceAcceptable(char *name, unsigned short flags)
{
   int i;

   for (i = 0; i < ignoreIfaceCnt; i++) {
       if (strcmp(name, ignoreIfaceList[i]) == 0)
           return(FALSE);
   }

   if (strchr(name,':') != NULL) {
      return(FALSE);
   }
   if ((flags & IFACE_UP) == 0) {
      return(FALSE);
   }

   if (flags & IFACE_LOOPBACK) {
      return(FALSE);
   }
 
   if ((name[0] == 'i') && (name[1] == 'p') && (name[2] == '0')) {
      return(FALSE);
   }
   return(TRUE);
}


/* 
 * Used to compare interface information collected locally against
 * that in the user-level code.
 */
boolean IfaceMatch(IfaceEntry iface1, IfaceEntry iface2)
{
    if  ((strcmp(iface1.maIfaceName, iface2.maIfaceName) != 0) ||
	 (iface1.maIfaceAddr != iface2.maIfaceAddr) ||
         (iface1.maIfaceMask != iface2.maIfaceMask))
	return(FALSE);
    else
	return(TRUE);
}

/* 
 * Compile a list of interfaces that may be of interest to the user
 * level mobile node code. Return that information in the parameters
 * passed.
 */
int getNewIfaceList(IfaceEntry *ifaceTable, int *size) 
{
    int n, i, j;
    char *buf;
    unsigned int temp;
    struct sockaddr_in *sin;
    int numifs;
    struct ifconf ifc;
    struct ifreq ifr;
    register struct ifreq *ifrp;
    u32 ipAddr;	
    u32 netMask;     
    u32 otherAddr = 0;
    u32 flags;
    unsigned char hwAddr[MAX_HWADDR_LEN];
    IfaceEntry localIfaceTable[DEFAULT_MAX_INTERFACES];
    int localIfaceTableSize;
    boolean localIfaceMarked[DEFAULT_MAX_INTERFACES]; 
    boolean globalIfaceMarked[DEFAULT_MAX_INTERFACES];   

    /* Get the no. of interfaces */
    if (ioctl(ioctl_sockid, SIOCGIFNUM, (char *)&numifs) < 0) {
	mip4dbg(("Error while using SIOCGIFNUM ... %s\n", 
		 err2str(-1 * errno)));
	return(-1 * errno);
    } 

    /* Get the name of each interface */
    temp = numifs * sizeof(struct ifreq);
    if ((buf = malloc(temp)) == NULL) {
	mip4dbg(("Error in malloc \n"));
	return(-1 * errno);
    }

    ifc.ifc_len = temp;
    ifc.ifc_buf = buf;
    if (ioctl(ioctl_sockid, SIOCGIFCONF, (char *) &ifc) < 0) {
	mip4dbg(("Error while using SIOCGIFCONF ... %s\n", 
		 err2str(-1 * errno)));
	free(buf);
	return(-1 * errno);
    }       

    /*
     * Compile information about each interface.
     */
    ifrp = ifc.ifc_req; 
    i = 0;
    for (n = ifc.ifc_len / sizeof (struct ifreq); n > 0; n--, ifrp++) {
	/* Get the flags */
	bzero ((char *) &ifr, sizeof(ifr));
	strncpy(ifr.ifr_name, ifrp->ifr_name, sizeof(ifr.ifr_name));
    	if (ioctl(ioctl_sockid, SIOCGIFFLAGS, (caddr_t)&ifr) < 0) {
	    mip4dbg(("Error while using  SIOCGIFFLAGS ... %s\n",
		     err2str(-1 * errno)));
	    continue;
	}

 	adjustIfaceFlags(&flags, ifr.ifr_flags);
 	if (!isIfaceAcceptable(ifrp->ifr_name, flags)) continue;
	
	/* The interface is of potential interest ... get the IP address */

	if (ioctl(ioctl_sockid, SIOCGIFADDR, (caddr_t)&ifr) < 0) {
	    mip4dbg(("Error while using  SIOCGIFADDR ... %s\n", 
		     err2str(-1 * errno)));
	    continue;
	}
	sin = (struct sockaddr_in *) &ifr.ifr_addr;     
	ipAddr = sin->sin_addr.s_addr;	

	/* 
	 * TODO: We could also try to check for IP addresses that
	 * belong to our home network but differ from homeAddr.
	 */
	if (ipAddr == 0) {
	    mip5dbg(("Ignoring interface %s with zero address.\n",
		     ifr.ifr_name));
	    continue;
	}	

	/* 
	 * Get the netmask 
	 * TODO: We could do some sanity checks for the netmask but for
	 * now we don't.
	 */
	if (ioctl(ioctl_sockid, SIOCGIFNETMASK, (caddr_t)&ifr) < 0) {
	    mip4dbg(("Error in using SIOCGIFNETMASK ... %s\n", 
		     err2str(-1 * errno)));
	    continue;
	}
	sin = (struct sockaddr_in *)&ifr.ifr_addr;     
	netMask =  sin->sin_addr.s_addr;
	
	/* Get the hardware address only if appropriate */
	if ((flags & IFACE_NOARP) == 0) {
	    if ((i = getEthernetAddr(ifr.ifr_name, hwAddr)) < 0) {
	        mip5dbg(("Error in getting hardware address for %s ... %s\n",
			 ifr.ifr_name, err2str(i)));
		continue;
                } 	    
	    }

	/* 
	 * Get the "otherAddr" for use as router if ipAddr is usable
	 * as a colocated care-of address.
	 */
	otherAddr = 0;
	if (ipAddr != mnHomeAddr) {
             mip5dbg(("Iface %s has colocated addresss - %s\n", 
		      ifr.ifr_name, ntoa(ipAddr,addr1))); 
	    /* Get the Destination address in case it is point to point */
	    if (flags & IFACE_POINTOPOINT) {
	        if ((ioctl(ioctl_sockid, SIOCGIFDSTADDR, (caddr_t)&ifr)) < 0) {
		    mip2dbg(("Could not get destination address for " \
			     "pointtopoint interface %s ... %s",
			     ifr.ifr_name, err2str(-1 * errno)));
		    continue;
		} else {
		    sin = (struct sockaddr_in *) &ifr.ifr_dstaddr;     
		    otherAddr = sin->sin_addr.s_addr;
		    if ((otherAddr == 0) || (otherAddr == ipAddr)) 
		        continue;
		    mip5dbg(("Got pointtopoint dst address %s for %s\n", 
			     ntoa(otherAddr,addr1),
			     ifr.ifr_name));
		}
	    } else {
	        /* 
		 * In all other cases, we pick up a default router
		 * from the routing table.
		 * TODO: For dhcp interfaces, we should ideally pick up 
		 *  the default router using dhcpinfo.
		 */
	        otherAddr = getRouterAddr(ipAddr, netMask);

		if ((otherAddr == 0) && !(isReverseTunnelSet)) { 
		    mip5dbg(("Could not get appropriate default router for "\
			     "iface %s.\n", ifr.ifr_name));
		    continue;
		}
	    }
	}

	strcpy(localIfaceTable[i].maIfaceName, ifr.ifr_name);
	memcpy(localIfaceTable[i].maIfaceHWaddr, hwAddr, MAX_HWADDR_LEN);
	localIfaceTable[i].maIfaceAddr = ipAddr ;
	localIfaceTable[i].maIfaceOtherAddr = otherAddr;
	localIfaceTable[i].maIfaceMask = netMask ;
	localIfaceTable[i].maIfaceFlags = flags;
	localIfaceTable[i].maIfaceSolicitationTime = 0 ;
	i++;
    }

    /* Free memory no longer needed */
    free(buf);

    /* 
     * Now compare the local list with that passed as argument. Entries
     * found in both lists are marked. 
     */
    localIfaceTableSize = i;
    for (i = 0; i < *size; i++) 
       globalIfaceMarked[i] = FALSE;

    for (j = 0; j < localIfaceTableSize; j++)
       localIfaceMarked[j] = FALSE; 
    
    for (i = 0; i < *size; i++) {
        for (j = 0; j < localIfaceTableSize; j++) {
	    if (IfaceMatch(ifaceTable[i], localIfaceTable[j])) {
	        localIfaceTable[j].maIfaceSolicitationTime = 
		     ifaceTable[i].maIfaceSolicitationTime ;
		localIfaceTable[j].maIfaceOtherAddr =
		     ifaceTable[i].maIfaceOtherAddr;
		globalIfaceMarked[i] = TRUE;
		localIfaceMarked[j] = TRUE;
	    }
	}
    }

    /* 
     * Unmarked entries in the global list result in agent-monitoring
     * streams being closed. Unmarked entries in the locallist result
     * in new advertisement-monitoring streams. 
     */

    for (i = 0; i < *size; i++) {
        if (globalIfaceMarked[i] == FALSE) {
	    closeIfaceFD(ifaceTable[i].maIfaceName);
	}
    }

    for (j = 0; j < localIfaceTableSize; j++) {
        if (localIfaceMarked[j] == FALSE) {
	    addIfaceFD(localIfaceTable[j].maIfaceName);
	}
    }

    memcpy(ifaceTable, localIfaceTable, 
	   localIfaceTableSize*sizeof(IfaceEntry));
    *size = localIfaceTableSize;

    return(0);
}


/* Restores Arp */
/* TODO : To be implemented */
int restoreArp()
{
    return(0);
}

/* Disables arp */
/* TODO : To be implemented */
int disableArp(u32 mnAddr, u32 foreignRouter)
{
   return(0);
}


/* OS dependent cleanup routine needed when exiting 
 * TODO: Revisit this and think about this some more.
 */
void OScleanup()
{
    int i;

    if (decaprem(mnHomeAddr) < 0)
	mip0dbg(("Could not terminate decapsulation service for %s", 
		 ntoa(mnHomeAddr, addr1)));

    pthread_mutex_lock(&(ifaceFDList.ifaceFDLock));
    for (i = 0; i < ifaceFDList.ifaceFDCnt; i++) {    
        close(ifaceFDList.ifaceFDarray[i].ifaceFD);
    }
    ifaceFDList.ifaceFDCnt = 0; 
    pthread_mutex_unlock(&(ifaceFDList.ifaceFDLock));

    pthread_cancel(monitorIface);
    
    /* 
     * Just add a wait to allow "secondary" thread to reach a point
     * where it has reached a "cancellable point".
     */
    sleep(2);

    /* hopefully the other thread is dead by now ... */
    pthread_mutex_destroy(&(ifaceFDList.ifaceFDLock));
    pthread_mutex_destroy(&(advPktBuf.AdvBufLock));    
    close(ioctl_fd);
}

/* OS dependent initialisation routine */
int OSinit()
{
    struct vtunl_req vrq;
    int stream_id, val, dummy;
    pthread_attr_t detached_attr;
    int errcode;

    /* Initialise the mutexes */  
    pthread_mutex_init(&(advPktBuf.AdvBufLock), NULL);
    pthread_mutex_init(&(ifaceFDList.ifaceFDLock), NULL);

    /* Inititalise the circular buffer for storing agent advertisements */
    advPktBuf.head = advPktBuf.tail = 0;

    /* Make a socket for the various SIOCxxxx ioctl commands */
    if ((ioctl_sockid = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
        mip0dbg(("Error: could not create socket for SIOCxxxx commands.\n"));
	exit(-1);
    }

    /* 
     * We do this as a guaranteed means of ensuring that vtunl
     * is correctly initialized from scratch.
     */
    mip2dbg(("rem_drv followed by add_drv vtunl\n"));
    system("/usr/sbin/rem_drv vtunl");
    system("/usr/sbin/add_drv vtunl");

    /* Create a stream that encapsulates unicasts and decapsulates */
    memset((char *) &vrq, 0, sizeof(struct vtunl_req));
    vrq.vr_address = mnHomeAddr;
    vrq.vr_ip_protocol = 4;
 
    stream_id = plumb("/dev/ip", &vrq);
    if ((stream_id < 0) ||
        ((val = mkpt2pt("ip0", mnHomeAddr,
                        inet_addr("255.255.255.254"))) < 0)) {
        mip0dbg(("Could not create stream for encap/decap ... %s\n",
                 err2str(val)));
 
        mip0dbg(("\nEnsure that vtunl is installed properly on your " \
                 " system, you are\n"));
        mip0dbg(("logged in as root, and another  process hasn't " \
                 "already created a\n"));
        mip0dbg(("stream with \"vtunl\" (tunneling driver) plumbed " \
                 "between \"/dev/ip\"\n"));
        mip0dbg(("(IP driver) and \"ip\" (IP module).\n\n"));
 
        exit(-1);
    }
       
    mip0dbg(("Tunneling driver (vtunl) installed successfully.\n"));

    /*
     * Create a stream for ioctls. This one has vtunl over /dev/ip
     * with nothing else between vtunl and the stream head. This
     * ensures that vtunl related ioctls will not be mis-interpreted
     * by any intermediate modules.
     */ 
    if (((ioctl_fd = open("/dev/ip", 0)) < 0) ||
        (ioctl(ioctl_fd, I_PUSH, "vtunl") < 0)) {
        mip0dbg(("Error: could not push vtunl on /dev/ip for ioctls.\n"));
        exit(-1);
    }


    /* 
     * Enable decapsulation service for mobile node's home address if it 
     * appears in the inner datagram as dst. 
     */
    if ((errcode = decapadd(mnHomeAddr, "dst")) < 0) {
       mip0dbg(("Could not enable decapsulation service for %s\n",
		inet_addr(mnHomeAddr, addr1)));
       exit(-1);
    }
    
    getNewIfaceList(ifaceTable, &ifaceTableSize); 
    
    /* Create the secondary thread to monitor agent advertisements */   
    pthread_attr_init(&detached_attr);
    pthread_attr_setdetachstate(&detached_attr, PTHREAD_CREATE_DETACHED);

    if ((pthread_create(&monitorIface, 
			&detached_attr,
			monitorAgentAdv,
			(void *) &dummy) != 0)) {
        mip1dbg(("Error while creating advertisement-monitoring thread.\n"));
        exit(-1);
    }		 
    return(0);
}

int setMTU(char *ifname, int mtu)
{
    struct ifreq ifr;

    memset((char *)&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, ifname, sizeof (ifr.ifr_name));
    ifr.ifr_metric = mtu;
    if (ioctl(ioctl_sockid, SIOCSIFMTU, (caddr_t)&ifr) < 0)
        return(-1 * errno);
    return(0); 
}

/* 
 * Look at the list of all available interfaces and pick the minimum
 * MTU.
 */
int getMinMTU()
{
    int i;
    int ifaceCnt;
    struct ifconf ifc;
    struct ifreq ifr;
    struct ifreq *ifrp;
    int bufsize;
    char *buf;
    int minMTU = 5000;     

    /* Get the number of interfaces */
    if (ioctl(ioctl_sockid, SIOCGIFNUM, (char *)&ifaceCnt) < 0) 
        goto usedefault;

    bufsize = ifaceCnt * sizeof(struct ifreq);
    if ((buf = (char *) malloc(bufsize)) == NULL) 
        goto usedefault;
    
    /* Get the list of interface names */
    memset((char *)&ifc, 0, sizeof(ifc));
    ifc.ifc_len = bufsize;
    ifc.ifc_buf = buf;
    if (ioctl(ioctl_sockid, SIOCGIFCONF, (char *) &ifc) < 0) {
        free(buf);	
	goto usedefault;
    }       

    /*
     * Get the MTU of each interface, ifc_req points to the same place
     * as ifc_buf.
     */
    ifrp = ifc.ifc_req;   
    for (i = 0; i < (ifc.ifc_len/sizeof(struct ifreq)); i++, ifrp++) {
        /* ... flags defined in <net/if.h> */
        bzero((char *)&ifr, sizeof(ifr));
	strncpy(ifr.ifr_name, ifrp->ifr_name, 8); 
	if ((strcmp(ifr.ifr_name, "lo") == 0) || 
	    (strcmp(ifr.ifr_name, "ip0") == 0)) 
	    continue;
	    
	/* Get the interface MTU */
	if (ioctl(ioctl_sockid, SIOCGIFMTU, (char *)&ifr) == 0) {
	    if (ifr.ifr_metric < minMTU) {
	        minMTU = ifr.ifr_metric; 
	    }
	}
    }

    free(buf);
    if (minMTU != 5000)
        return(minMTU);

usedefault:
    return(DEFAULT_MIN_MTU);
}


/*
 * This is the generic interface to a script that enables IPSec. After
 * this script is invoked, all packets destined for a mobile node's
 * home network should automatically be processed by IPSec in a manner
 * such that those packets will pass through all intervening firewalls.
 * The script is passed three arguments: the first is either "on" or
 * "off" (without the quotes), the second is an interface name
 * on which IPSec should be enabled/disabled and the third is the
 * mobile node's care-of address. The last argument is useful if
 * depending on the mobile node's current point of attachment a different
 * set of firewalls may have to be traversed.
 */
void enableIPSec(mnState *mnsp)
{
   char tmp[128];
   
   sprintf(tmp, "IPSecScript on %s %s", mnsp->ifaceName,
	   ntoa(mnsp->COAddr, addr1));
   mip3dbg(("%s\n", tmp));
   system(tmp);
}

/* Undo the actions performed as part of enableIPSec. */
void disableIPSec(mnState *mnsp)
{
   char tmp[128];
   
   sprintf(tmp, "IPSecScript off %s %s", mnsp->ifaceName, 
	   ntoa(mnsp->COAddr, addr1));
   mip3dbg(("%s\n", tmp));
   system(tmp);
}

int enableReverseTunneling(mnState *mnsp, u32 homeAgentAddr)
{
    int errcode;
    int MinMTU;

    /*
     * Add the host specific route for home agent (TO avoid looping) 
     */
    if ((errcode = routeadd(homeAgentAddr, inet_addr("255.255.255.255"),
			    mnsp->routerAddr, 1)) < 0) {
	  mip2dbg(("Error in (routeadd %s %s): %s\n",
		   ntoa(homeAgentAddr, addr1), 
		   ntoa(mnsp->routerAddr, addr2),
	           err2str(errcode)));
    }

    /* 
     * Set the MTU to minimum of all the interfaces. We assume that
     * enabling IPSEC or disabling IPSEC will not change the MTU of
     * an interface ... the MTU changes only at the time of installing
     * IPSEC on an interface.
     * TODO: In some situations, the MTU may have to be changed whenever
     *       there is a change in the list of available interfaces, i.e.
     *       setMTU may need to be called as part of getIfaceList.
     */

    MinMTU = getMinMTU();
    mip0dbg(("Setting ip0 MTU to %d.\n", MinMTU));
    if (setMTU("ip0", MinMTU) < 0) {
        mip0dbg(("setMTU failed\n"));
    }	

    /* 
     * Make sure that subsequent tunnel exits get a PMTU bound
     * equal to MinMTU.
     */
    if ((errcode = pmtubound(0, MinMTU)) < 0) {    
        mip0dbg(("Error while setting default PMTU %d ... %s\n", 
		 MinMTU, err2str(errcode)));
    }

    /* 
     * Set up default encapsulation of pkt passing through vtunl
     * so they are directed at the home agent.
     */
    if ((errcode = encapadd(DEFAULT_ADDR_VTUNL, 0, homeAgentAddr, 0)) < 0) {
        mip0dbg(("Default encapsulation through %s failed ... %s\n", 
		 ntoa(homeAgentAddr, addr1), err2str(errcode)));
    }

    /* 
     * Change the default route in the routing table to pass through
     * vtunl instead of a "normal" router.
     */
    if ((errcode = routedel(0, 0, mnsp->routerAddr, 1)) < 0) {
        mip2dbg(("Error in (routedel 0 %s): %s\n",
		 ntoa(mnsp->routerAddr, addr1), err2str(errcode)));
    }

    if ((errcode = routeadd(0, 0, inet_addr("255.255.255.254"), 1)) < 0) {
        mip1dbg(("Couldn't add default route through tunnel ... %s\n", 
		 err2str(errcode)));
    }

    /* 
     * We MUST NOT add a route for the firewall through the default router
     * on the foreign net. Otherwise, packets to the FW will be sent
     * using the MN's care-of address rather than its home address.
     */

    return(0);
}

int disableReverseTunneling(mnState *mnsp, u32 homeAgentAddr)
{
   int errcode; 

    /* Correction: See above.  */
   /*
    * if ((errcode  = routedel(domainInfo.fwAddr[0], 
    *                          inet_addr("255.255.255.255"),
    *                          mnsp->routerAddr, 1)) < 0) {
    *     mip1dbg(("Error in (routedel %s %s) ... %s\n", 
    *              ntoa(domainInfo.fwAddr[0], addr1), 
    *              ntoa(mnsp->routerAddr, addr2),    
    *              err2str(errcode)));
    * }
    */

   if ((errcode = routedel(0, 0, inet_addr("255.255.255.254"), 1)) < 0) {
       mip1dbg(("Error while deleting default route through tunnel ... %s\n",
		err2str(errcode)));
   }	
   
   if ((errcode = encaprem(DEFAULT_ADDR_VTUNL, homeAgentAddr)) < 0) {
	mip1dbg(("Removal of default encapsulation through HA failed ... %s\n",
	         err2str(errcode)));
   }
    
   /* Delete the route for HA through default router on foreign network */
   if ((errcode = routedel(homeAgentAddr, inet_addr("255.255.255.255"),
			   mnsp->routerAddr, 1)) < 0) {
       mip2dbg(("Error in (routedel %s %s): %s\n",
	        ntoa(homeAgentAddr, addr1), 
	        ntoa(mnsp->routerAddr, addr2), err2str(errcode)));
   }

   return(0);
}

/* 
 * Get the logical clone of the interface specified in iface
 * e.g. Logical clone of le0 under Solaris will be le0:1
 * TODO : Make sure that the logical clone is not already present
 *        But if we add this logic then we have to be careful in
 *        mnState.c because the same function is also used to get 
 *        the name of an *existing* interface for unconfigNetColocated.
 */
void mkCloneIface(char *iface, char *clone)
{
    strcpy(clone, iface);
    strcat(clone, ":1");
}


