/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Original Code is the Sunlabs Mobile IP.
 The Initial Developer of the Original Code is Vipul Gupta.
 Portions created by Vipul Gupta are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): Niko A. Fikouras (niko@comnets.uni-bremen.de)
		 Asanga D. Udugama (adu@comnets.uni-bremen.de)

*/

/*
 * An implementation of Mobile-IP as described in RFC 2002.
 * Author: Vipul Gupta (vgupta@eng.sun.com)
 * mobile.h : Data structures and constants used by a mobile node.
 */

#ifndef _MOBILE_H
#define _MOBILE_H

#pragma ident	"@(#)mobile.h 1.1 97/10/01"

#ifdef __cplusplus
extern "C" {
#endif

#include "mip.h"

/* Constants associated with maintaining active agent list */
#define AGENT_SOLICIT_THRESHOLD        6    /* solicit adv after this time */
#define AGENT_EXPIRE_THRESHOLD         8    /* expire agent after this time */

/*
 * An HA is considered alive for atleast DEFAULT_HA_LIFETIME seconds
 * since last communication with it. A discovery message is kept in
 * the pending list for DEFAULT_DISCOVERY_INTERVAL and no discovery
 * message is sent while another is already pending. If we still don't
 * have a live HA when a discovery message is about to be removed, we
 * send out a new discovery message and keep a corresponding registration
 * request entry.
 * Need to make sure that discovery messages are sent only in FOREIGN or
 * COLOCATED states.
 */
#define DEFAULT_HA_LIFETIME          300
#define DEFAULT_DISCOVERY_INTERVAL    60

/* Constants associated with registrations */
#define MIN_REG_DELAY                  1    /* between successive reg req */
#define DEFAULT_REG_LIFETIME          60    /* to register a colocated COA */
#define DEFAULT_MAX_ROUNDTRIP          4    /* round trip estimate for reply */
#define DEFAULT_DISCRETIONARY_FLAGS  0x0    /* B, M, G, V flags */
#define DEFAULT_MAX_ATTEMPTS           8    /* max size of policy arrays */
#define MULTIPLE_SIMULTANEOUS_BINDINGS 0x08 /* S bit setting */
/*
 * Constants associated with agent availability. When an agent rejects
 * a registration request (with an error that cannot be immediately
 * fixed, e.g. ADMINISTRATIVELY_PROHIBITED), a mobile node temporarily
 * marks this agent as unusable.
 */
#define HA_COOLOFF_PERIOD             30   /* if HA denies reg request */
#define FA_COOLOFF_PERIOD             30   /* if FA denies reg request */

/*
 * All timers should ideally be a multiple of this interval. The
 * mobile node checks its periodic chores every DEFAULT_PERIODIC_INTERVAL
 * seconds.
 */
#define DEFAULT_PERIODIC_INTERVAL      2

/* For timestamp-based replay protection, this is the max skew allowed */
#define DEFAULT_FRESHNESS_SLACK      300

/* Interval between successive polls of active interfaces */
#define IFACE_POLL_INTERVAL           10

/* Destination port for registration requests */
#define MIP_PORT                     434

//#define CONF_FILE_NAME               "/etc/mipmn.conf"
//#define FILTER_FILE_NAME             "/etc/mipmn-filters.conf"
#define CONF_FILE_NAME               "/home/adu/mipmn.conf"
#define FILTER_FILE_NAME             "/home/adu/mipmn-filters.conf"
#define ROUTE_PATH               	 "/sbin/route"

/* Maximum number of mobility agents a MN keeps track of */
#define DEFAULT_MAX_MOBILITY_AGENTS    8

/* Maximum number of Home Agents a MN keeps track of */
#define DEFAULT_MAX_HOME_AGENTS        8

/*
 * Maximum number of other nodes with which this mobile node shares
 * a mobile security association.
 */
#define DEFAULT_MAX_OTHER_NODES       16

/* Maximum registrations we keep track of */
#define DEFAULT_MAX_REG_REQ           10

#define MAX_EXPECTED_EXTENSIONS       32
/*
 * In any router advertisement, these are the maximum number of
 * router addresses (in the RFC1256 portion) and care-of addresses
 * (in the mobility extension) that the mobile node processes ...
 * others are ignored.
 */
#define MAX_ROUTERS_IN_ADV             4
#define MAX_COA_IN_ADV                 4

#define MAX_IFNAME_LEN                 12
#define MAX_HWADDR_LEN                 8

/* Max default routers at home a MN keeps track of */
#define MAX_DEFAULT_ROUTERS            4

/* Max interfaces under mobile node's control */
#define DEFAULT_MAX_INTERFACES         5

/* Max number of address intervals for specifying protected domain */
#define MAX_ADDR_INTERVALS             6

/* Max. number of firewalls */
#define MAX_FIREWALLS                  3

/*
 * Maximum anticipated pkt size for any registration/reply/icmp pkt
 * of interest to a mobile node.
 */
#define MAX_PKT_SIZE                  2048

/* 
 * Movement Detection Method; The sunlabs mip implementation shows
 * preference to HAs over FAs and therefore performs Lazy Cell
 * Switching (LCS) when leaving the Home Network and Eager Cell
 * Switching (ECS) when entering the Home Network.
 * Enabled cliemt to select between LCS and ECS indebendently of 
 * source or destinatio network. If advertisements include prefix
 * length extensions then the cient performs LCS between agents of
 * the same network and ECS for other agents.
 * N. A. Fikouras, Uni. Sheffield
 *
 * 3/2/01, Update
 * Added support for HCS and FHCS.
 * FHCS assumes that the 802.11 SSID can be read by the the file:
 * "/proc/aironet/eth0/Status"
 * Expected structure is:
 * "Mode: XX ... SSID XX XX:XX:XX:XX"
 */

#define LCS				1
#define ECS				2
#define HCS				3
#define FHCS			4

#define HO_THRESH		40

#define	NOT_CONNECTED	15
#define	SYNCHRONISED	31
#define	CONNECTED		63

#define SMART_HCS

/* 
 * One such entry for each mobility-supporting interface. The mobile node
 * needs to pick up complete ethernet frames containing the agent
 * advertisements so it better know all the interfaces that must be
 * monitored. Actual monitoring occurs only if that interface is UP.
 *
 * We define our own flags below since different operating systems
 * may use different flags and/or reorder the bit sequence. If an
 * OS does not define a certain flag, it should be set to zero by
 * the OS-specific routine that returns these flags.
 */

#define	IFACE_UP		0x1	/* interface is up */
#define	IFACE_BROADCAST	        0x2	/* broadcast address valid */
#define	IFACE_DEBUG	        0x4	/* turn on debugging */
#define	IFACE_LOOPBACK	        0x8	/* is a loopback net */
#define	IFACE_POINTOPOINT	0x10	/* interface is point-to-point link */
#define	IFACE_NOTRAILERS	0x20	/* avoid use of trailers */
#define	IFACE_RUNNING	        0x40	/* resources allocated */
#define	IFACE_NOARP	        0x80	/* no address resolution protocol */
#define	IFACE_PROMISC	        0x100	/* receive all packets */
#define IFACE_DHCP              0x200   /* under DHCP control */
#define IFACE_ON_NET            0x400   /* interface has connectivity to
                                         * the net, test by pinging */
typedef struct mipif  {
    char            maIfaceName[MAX_IFNAME_LEN];
    unsigned char   maIfaceHWaddr[MAX_HWADDR_LEN];
    u32             maIfaceAddr;

    /*
     * Contains an address that can be pinged to test whether
     * this interface is really on the network. this could be
     * the dstaddr on a pointopoint link or the default router
     * on a DHCP interface
     */
    u32             maIfaceOtherAddr;
    u32             maIfaceMask;
    u32             maIfaceFlags;

    /*
     * Time (in NTP seconds) when the last agent solicitation was sent on
     * this interface. Used to curb unnecessary solicitations.
     */
    u32             maIfaceSolicitationTime;

	/*
	 * 	If Smart_HCS is enabled then node will use the 802.11 SSID field
	 * 	to manage movement detection. For HCS the Status maintains the "name"
	 * 	of the current network. A change of the ssid is perceived is
	 * 	indication of movement. For FHCS, the ssid contains the local MA
	 * 	and its corresponding hardware address.
	 */
	#ifdef SMART_HCS
	u32				localMA;
    unsigned char  	maHaddr[MAX_HWADDR_LEN];
	u32				mode, sStrength, sQuality;
	#endif
} IfaceEntry;

/* Flags that identify why an agent is temporarily unusable */
#define BAD_REPLY_TO_FA                0x1
#define UNREACHABLE_FROM_FA            0x2
#define REJECTED_WITHOUT_REASON        0x4
#define REJECTED_ADM_PROHIBITED        0x8
#define OUT_OF_RESOURCES               0x10
#define FAILED_HA_MN_AUTHENTICATION    0x20
#define FAILED_FA_HA_AUTHENTICATION    0x40
#define FAILED_FA_MN_AUTHENTICATION    0x80

/* Bits in the HAflags field */
#define HA_IS_ALIVE                    0x01
#define HA_IS_DYNAMICALLY_DISCOVERED   0x02

/* One such entry for each Home Agent. */
typedef struct hae {
    u32              mnHaAddr;     /* Can be the home network b'cast address */
    unsigned char    mnHaUnusableTill;
    unsigned char    mnHaWhyUnusable;
    unsigned char    mnHAflags;    /* A=alive, D=dynamically learnt, etc */
    u16              mnHALifetime; /* time before HA is marked dead */
    u32              mnHaLastIDhi; /* used for NONCE-based replay protection */
    int              mnHaTimeSkew; /* used for TIMESTAMP-based replay prot. */
    MipSecAssocEntry mnHaMSA;
} HomeAgentEntry;

/*
 * One such entry for each local mobility agents considered alive
 * lmae : local mobility agent entry. A source address of zero
 * indicates an unused entry.
 */
typedef struct lmae {
    u32              mnAdvRemainingLifetime;  /* keeps validity period */
    char             mnAdvIncomingIface[MAX_IFNAME_LEN];
    unsigned char    mnAdvSrcHWaddr[MAX_HWADDR_LEN];
    u32              mnAdvSourceAddr;
    u32              mnAdvMaxAdvLifetime;     /* initial validity period */
    u32              mnAdvMaxRegLifetime;     /* from mobility extension */
    u16              mnAdvSeqNum;
    u16              mnAdvLastSeqNum;         /* used to detect reboots */
    unsigned char    mnAdvFlags;
    unsigned char    mnAdvRouterCnt;
    u32              mnAdvRouterList[MAX_ROUTERS_IN_ADV];
    u32              mnAdvPrefixList[MAX_ROUTERS_IN_ADV];
    u32              mnAdvPFAIPList[MAX_ROUTERS_IN_ADV];
    unsigned char    mnAdvCOACnt;
    u32              mnAdvCOAList[MAX_COA_IN_ADV];
    unsigned char    mnAdvCOAunusableTill[MAX_COA_IN_ADV]; 
    unsigned char    mnAdvCOAwhyUnusable[MAX_COA_IN_ADV];
    /* 
     * The softstate associated with a mobility agent has two components:
     * a counter that indicates how much longer a COA is unusable, and
     * a bit vector that indicates why (indicates the recent error codes
     * returned by that agent).
     */
} LocalMaEntry;

/*
 * One such entry for each registration request. The same list contains
 * both accepted and pending entries.
 * Invariant: at any given time, the accepted entries reflect which
 * all HAs on the home network have registered this MN. Of these, the
 * accepted, renewable entries reflect "current" recgistrations in FOREIGN
 * or COLOCATED state. Accepted, nonrenewable entries correspond to prior
 * registrations that may not have been cancelled at the home agent. On
 * moving home, the list of all accepted entries is used to send out
 * (multiple) deregistration requests -- the accepted entries are replaced
 * by pending deregistration entries.
 *   The mnRegTimeRemaining field indicates the interval after which this
 * entry will be deleted. For an accepted entry or pending deregistration
 * request, it reflects the time before HA will expire the mobility binding.
 * For a pending registration, this field is initialized to the requested
 * lifetime.
 *   The mnRegNextTime field indicates the time left for the next
 * retransmission (in case of a pending request) or renewal (in case of
 * an accepted renewable request). For non-renewable, accepted entries,
 * this field has no significance.
 */
    
#define REG_UNUSED                  0
#define REG_PENDING                 1
#define REG_ACCEPTED_RENEWABLE      2
#define REG_ACCEPTED_NONRENEWABLE   3

typedef struct rre {
    unsigned char    mnRegAgentHWaddr[MAX_HWADDR_LEN];
    u32              mnRegAgentAddr;    /* IP destination of reg request */
    u32              mnRegHAaddr;
    u32              mnRegCOA;
    u32              mnRegIDLow;
    u32              mnRegIDHigh;
    u32              mnRegTimeRequested;     /* granted time if accepted */
    u32              mnRegTimeRemaining;
    u32              mnRegNextTime;
    unsigned char    mnRegFlags;
    unsigned char    mnRegAttemptIdx;
    unsigned char    mnRegStatus;   /* e.g. pending, unused, etc */
    boolean          mnRegCOAIsLocal;
    char	     mnRegIfaceName[MAX_IFNAME_LEN]; // introduced to handle MSB
    char             mnRegFilter[1024]; // containes associated filters
    int              mnRegFilterLen;
} RegReqEntry ;

#define HOME_STATE                    0
#define FOREIGN_STATE                 1
#define COLOCATED_STATE               2
#define ISOLATED_STATE                3  /* non-loopback ifaces are down */
#define UNKNOWN_STATE                 4

/* Data structure to store state-specific information. */
typedef struct mns {
    unsigned char state;      /* HOME, FOREIGN, COLOCATED, ISOLATED, UNKNOWN */

    /*
     * While soliciting agents in UNKNOWN state, we need an index into
     * back-off array. nextSolicit stores the time left before the next
     * solicitation sent in UNKNOWN state.
     */
    int		backoffIdx;

    /*
     * Interface through which we heard the current FOREIGN agent
     * or acquired a colocated address on.
     */
    char          ifaceName[MAX_IFNAME_LEN];
    u32           routerAddr;  /* Agent src or from PPP/DHCP */
    u32           COAddr;
    u32           timeSpent;
    u32           nextSolicit;
} mnState;

typedef struct is {
	char ifaceName[MAX_IFNAME_LEN];
	u32  COAddr;
	mnState currentState;
} InterfaceState;


/* Data structure to store the information regarding protected domain */
typedef struct domaininfo {
    u32 addr[MAX_ADDR_INTERVALS];
    u32 netmask[MAX_ADDR_INTERVALS];
    u32 addrIntervalCnt;
    u32 fwAddr[MAX_FIREWALLS];
    u32 firewallCnt;
} DomainInfo;


/* Data Structures & constants for filters */

#ifndef MAX_FILTERS
#define MAX_FILTERS			10
#endif
#define MAX_FILTER_MODULES_PER_FILTER	5
#define DEFAULT_FILTER_SEQ		255

// this structure specifies one operation related to a filter
typedef struct filterop {
	int action;						   // filter action, as defined in
								   // ACT_ flags
	int idx;						   // the index of the filter
								   // referred to by the action
} FilterOp;

// structure to hold one filter module
typedef struct filtermodule {
	int filterType; 					   // type of the filter : as defined in FLT_ flags
	char *filterPtr; 					   // pointer to the filter data : can be DSCP type,
								   // port type, src addr type, src net type,
								   // src port type, src range type, load balancing
								   // type or free form type : would be one, from
			 					   // the following structures
} FilterModule;

// structure to hold distributions for
// eacf interface (CoA)
typedef struct controlmodule {
	u32 boundCoA;						   // currently bound CoA
	int weight; 						   // the weight of distribution
	char boundIfaceName[MAX_IFNAME_LEN];			   // the interface name to which
								   // this distribution is associated

}  ControlModule;

// structure to hold one filter
typedef struct filter {
	int filterModuleCount; 					   // filter count per module
	FilterModule filterModules[MAX_FILTER_MODULES_PER_FILTER]; // filter modules of each filter,
								   // each filter module is
								   // interpretted as AND
								   // against each other
	ControlModule controlModules[MAX_FILTER_MODULES_PER_FILTER]; // info of distributions per
								     // per each given interface
	int idx;						   // the index number assigned
								   // to the filter
	int markValue;						   // value used to mark packets
} Filter;


// structures that hold filter data
typedef struct bafextdatadata {
	int DSCP;
} BAFExtData;

typedef struct protoextdata {
	int protoNum;
} ProtoExtData;

typedef struct srcaddrextdata {
	u32 srcAddr;
} SrcAddrExtData;

typedef struct srcnetextdata {
	u32 srcAddr;
	u32 srcNetMask;
} SrcNetExtData;

typedef struct srcportextdata {
	short int portNum;
} SrcPortExtData;

typedef struct srcportrangeextdata {
	short int minPortNum;
	short int maxPortNum;
} SrcPortRangeExtData;

typedef struct freeformextdata {
	short int offset;
	int length; // in nibbles
	char value[MAX_FREEFORM_VAL_SIZE + 1];
	char mask[MAX_FREEFORM_MASK_SIZE + 1];
} FreeFormExtData;

// new filters

typedef struct dstaddrextdata {
	u32 dstAddr;
} DstAddrExtData;

typedef struct dstnetextdata {
	u32 dstAddr;
	u32 dstNetMask;
} DstNetExtData;

typedef struct dstportextdata {
	short int portNum;
} DstPortExtData;

typedef struct dstportrangeextdata {
	short int minPortNum;
	short int maxPortNum;
} DstPortRangeExtData;


#ifdef __cplusplus
}
#endif

#endif /* _MOBILE_H */


