/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Original Code is the Sunlabs Mobile IP.
 The Initial Developer of the Original Code is Vipul Gupta.
 Portions created by Vipul Gupta are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): Niko A. Fikouras (niko@comnets.uni-bremen.de)
		 Asanga D. Udugama (adu@comnets.uni-bremen.de)

*/

/*
 * An implementation of Mobile-IP as described in RFC 2002.
 * Author: Vipul Gupta (vgupta@eng.sun.com)
 * utils.c : Miscellaneous utility routines some of which may need to be
 *           implemented in a platform-specific fashion.
 */

#pragma ident   "@(#)utils.c 1.4 98/12/16"

#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>
#include <time.h>
#include <string.h>

#include "mip.h"
#include "mobile.h"

#define SECONDS_IN_A_DAY     86400
#define NTP_ADJUSTMENT       (SECONDS_IN_A_DAY * (((u32) 365*70) + 17))

extern int mipdebug;
extern char addr1[], addr2[];
/*--------------------- Platform-specific utilities --------------------*/

/* initialize the random number generator */
void randomInit()
{
    struct timeval time;

    if (gettimeofday(&time, 0) < 0) {
        mip0dbg(("Gettimeofday failed in randominit(), using fixed seed.\n"));
	srandom(181067);
    } else {
        srandom(time.tv_sec);
    }
}

/* returns an unsigned 32-bit random number */
unsigned long randomLong()
{
  return ((unsigned long) random());
}

/*
 * Returns the number of seconds elapsed since Jan 1, 1990 ie.
 * higher 32 bits of the 64-bit quantity representing current
 * time in NTP format.
 */

u32 CurrentTimeNTPSec()
{
    struct timeval time;

    if (gettimeofday(&time, 0) < 0) {
        mip0dbg(("Error: gettimeofday failed in CurrentTimeNTPSec()"));
	return(0);
    } else {
        return ((u32) (time.tv_sec + NTP_ADJUSTMENT));
    }
}

/*
 * Print the current time in buffer. If buffer is large enough,
 * the current time is printed in the format: 
 *                 Fri Sep 13 00:00:00 1986\0
 * Useful for debugging.
 */
char * sprintTime(char *buffer, int buflen)
{
    struct timeval clock;

    if (buflen >= 26) {
        gettimeofday(&clock, 0);

#ifdef _POSIX_THREAD_SAFE_FUNCTIONS
        ctime_r((time_t *) &(clock.tv_sec), buffer); 
#else
        ctime_r((time_t *) &(clock.tv_sec), buffer, buflen); 
#endif
	buffer[24] = '\0';
	return(buffer);
    } else {
        return("");
    }
}

char * printTime(char *buffer, int buflen)
{
    struct timeval clock;
	int hour, min, sec;
	char buff[120];

	gettimeofday(&clock, 0);
	hour = (((clock.tv_sec / 3600) + 2) % 24);
	min = ((clock.tv_sec / 60) % 60);
	sec = (clock.tv_sec % 60);
	sprintf(buff, "%02d:%02d:%02d %010d %06d", hour, min, sec, clock.tv_sec, 
		clock.tv_usec);
	if (strlen(buff) > buflen) return("");
	strcpy(buffer, buff);
	return(buffer);
}

/*
 * Print relative time since first call to this routine in the form
 *    seconds.xxxxxx\0
 */

char *sprintRelativeTime(char *buffer, int buflen)
{
    static int i = 0;
    static struct timeval starttime;
    static struct timeval now;
    long diff_sec, diff_usec;

    if (buflen < 20) {
        return ("");
    }

    if (i == 0) {
        gettimeofday(&starttime, 0);
	sprintf(buffer, "%d.%06d", 0, 0);
        i = 1;
    } else {
        gettimeofday(&now, 0);
	diff_sec = now.tv_sec - starttime.tv_sec;
	if (now.tv_usec < starttime.tv_usec) {
	    diff_usec = 1000000 + now.tv_usec - starttime.tv_usec;
	    diff_sec -= 1;
	} else {
	    diff_usec = now.tv_usec - starttime.tv_usec;
	}
	sprintf(buffer, "%d.%06d", diff_sec, diff_usec);
    }

    return(buffer);
}

        
/*--------------------- Platform-independent utilities --------------------*/

/* Compute the internet checksum for len number of bytes starting at addr. */
unsigned short inChecksum(unsigned short *addr, int len)
{
    register int nleft = len;
    register unsigned short *w = addr;
    register unsigned short answer;
    unsigned short odd_byte = 0;
    register int sum = 0;

    /*
     *  Our algorithm is simple, using a 32 bit accumulator (sum),
     *  we add sequential 16 bit words to it, and at the end, fold
     *  back all the carry bits from the top 16 bits into the lower
     *  16 bits.
     */
    while( nleft > 1 )  {
        sum += *w++;
	nleft -= 2;
    }

    /* mop up an odd byte, if necessary */
    if( nleft == 1 ) {
        *(unsigned char *)(&odd_byte) = *(unsigned char *)w;
	sum += odd_byte;
    }

    /*
     * add back carry outs from top 16 bits to low 16 bits
     */
    sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
    sum += (sum >> 16);			/* add carry */
    answer = ~sum;			/* truncate to 16 bits */
    return (answer);
}

/*
 * Converts the long value in addr_long to an internet address of
 * the type a.b.c.d in addr_string and returns a pointer to addr_string.
 * NOTE: addr_long is in network byte order and addr_string is assumed
 * to be long enough (atleast 16 bytes).
 *
 * By having the caller supply its own buffer, we try to avoid problems
 * associated with multiple threads writing the same buffer concurrently.
 */
char* ntoa(u32 addr_long, char * addr_string)
{
    u32    temp;
	
    temp = ntohl(addr_long);
    sprintf(addr_string, "%d.%d.%d.%d", 
	    ((temp >> 24) & 0xff), ((temp >> 16) & 0xff),
	    ((temp >> 8) & 0xff), (temp & 0xff));
    return(addr_string);
}

/*
 * Converts an ethernet address of type aa:bb:cc:dd:ee:ff to a byte array 
 * specified in the first argument.
 */
int hwAddrRead(unsigned char s[], char *hwadr)  {
    int a, b, c, d, e, f;

    if (sscanf(hwadr,"%x:%x:%x:%x:%x:%x", &a, &b, &c, &d, &e, &f) != 6)
     return(-1);
    else  {
        s[0] = (unsigned char) a;
	s[1] = (unsigned char) b;
	s[2] = (unsigned char) c;
	s[3] = (unsigned char) d;
	s[4] = (unsigned char) e;
	s[5] = (unsigned char) f;
	return(0);
    }
}

char * hwAddrWrite(unsigned char hwaddr[], char *hwStr)  {

    sprintf(hwStr,"%x:%x:%x:%x:%x:%x", 
	    (unsigned char) hwaddr[0],
	    (unsigned char) hwaddr[1],
	    (unsigned char) hwaddr[2],
	    (unsigned char) hwaddr[3],
	    (unsigned char) hwaddr[4],
	    (unsigned char) hwaddr[5]);
    return(hwStr);
}

boolean isBlank(char line[], int len)
{
    int  i;
    char j;

    for (i = 0; i < len ; i++) {
        j = line[i];
        if ((j >= '0' && j <= '9') ||
	    (j >= 'a' && j <= 'z') ||
	    (j >= 'A' && j <= 'Z')) {
	    return (FALSE);
	}
    }

    return (TRUE);
}

int hexDigit(int c)
{
    if (c >= '0' && c <= '9')
        return(c - '0');
    if (c >= 'a' && c <= 'f')
        return(c - 'a' + 10);
    if (c >= 'A' && c <= 'F')
        return(c - 'A' + 10);
    return (-1);
}

/*
 * Convert a string of hexadecimal digits into an array of len bytes
 * each containing a byte represented by a pair of hex digits, e.g.
 * keystr="12abC4De"
 * is converted to key[]={0x12, 0xab, 0xC4, 0xDe}
 * key has space for len bytes only.
 */
int hexConvert(char *key, int len, char *keystr)
{
    int i, d1, d2;

    for (i = 0; i < len; ++i) {
        if ((d1 = hexDigit(*keystr++)) == -1)
	    return (-1);
	if ((d2 = hexDigit(*keystr++)) == -1)
	    return (-1);
	*key++ = d1*16 + d2;
    }
    
    return (0);
}

/* 
 * Computes prefix length of a network mask. Assumes it is the same
 * as the number of bits set to one.
 */
int prefixLen(u32 netmask)
{
    int len = 0;

    while (netmask) {
       len++;
       netmask &= (netmask - 1);
    }

    return(len);
}

/* Prints a buffer in hexadecimal. */
void printBuffer(unsigned char buffer[], int buflen)
{
    unsigned int i;

    if (buflen == 6) {
        /* Is likely an ethernet address */
        for (i = 0; i < buflen; i++) {
	    mip3dbg(("%02x", buffer[i]));
	    if (i < (buflen-1)) mip3dbg((":"));
	}
    } else {
        /* Short enough to be printed in one shot */
        if (buflen <= 16) {
	    for (i = 0; i < buflen; i++) 
	        mip3dbg(("%02x", buffer[i]));
	} else {
	    /* Possibly a packet */
	    mip3dbg(("\t"));
	    for (i = 0; i < buflen; i++) {
	        mip3dbg(("%02x", buffer[i]));
		if ((i % 2) == 1) mip3dbg((" "));
		if ((i % 16) == 15) mip3dbg(("\n\t"));
	    }
	}
    }
}

/* Prints a Mobility Security Association Entry */
void printMSAE(MipSecAssocEntry *msae)
{

    mip3dbg(("\tMipSecAssocEntry contents:\n"));
    mip3dbg(("\tIP addr %s\n", ntoa(msae->mipSecAddr, addr1)));
    mip3dbg(("\tSPI %d\n", msae->mipSecSPI));
    mip3dbg(("\tAlgo type %d\n", msae->mipSecAlgorithmType));
    mip3dbg(("\tAlgo Mode %d\n", msae->mipSecAlgorithmMode));
    mip3dbg(("\tKey Len %d\n", msae->mipSecKeyLen));
    mip3dbg(("\tKey: 0x"));
    printBuffer(msae->mipSecKey, msae->mipSecKeyLen);
    mip3dbg(("\n"));
    mip3dbg(("\tReplay method %d\n", msae->mipSecReplayMethod));
}

/* Print the ethernet address in the colon format. */
void ether_print(unsigned char hwaddr[])
{
    mip3dbg(("%x:%x:%x:%x:%x:%x\n",
            (unsigned char) hwaddr[0],
            (unsigned char) hwaddr[1],
            (unsigned char) hwaddr[2],
            (unsigned char) hwaddr[3],
            (unsigned char) hwaddr[4],
            (unsigned char) hwaddr[5]));

}

/* Prints the Interface Table Contents */
void printIfaceTable(IfaceEntry ifaceTable[], int ifaceTableSize)
{
    int i;

    mip3dbg(("-------- Interface Table Contents ---------\n"));
    for (i = 0; i < ifaceTableSize; i++) {
	mip3dbg(("Entry No.   = %d\n", i));
	mip3dbg(("\tInterface   = %s\n", ifaceTable[i].maIfaceName));
	mip3dbg(("\tIP Addr.    = %s\n", 
		ntoa(ifaceTable[i].maIfaceAddr, addr1)));
	mip3dbg(("\tHardware Addr. = %x:%x:%x:%x:%x:%x\n",
            (unsigned char) ifaceTable[i].maIfaceHWaddr[0],
            (unsigned char) ifaceTable[i].maIfaceHWaddr[1],
            (unsigned char) ifaceTable[i].maIfaceHWaddr[2],
            (unsigned char) ifaceTable[i].maIfaceHWaddr[3],
            (unsigned char) ifaceTable[i].maIfaceHWaddr[4],
            (unsigned char) ifaceTable[i].maIfaceHWaddr[5]));
	mip3dbg(("\tOther Addr. = %s\n",
		ntoa(ifaceTable[i].maIfaceOtherAddr, addr1)));
	mip3dbg(("\tNetmask     = %s\n",
		ntoa(ifaceTable[i].maIfaceMask, addr1)));
	mip3dbg(("\tFlags       = %x\n", ifaceTable[i].maIfaceFlags));
    }
    fflush(stdout);
}

/* Prints the contents of the home Agent table */
void printHATable(HomeAgentEntry HATable[], int HATableSize)
{
    int i;
    char tmp[40];

    mip3dbg(("-------- Home Agent Table Contents ---------\n"));
    for (i = 0; i < HATableSize; i++) {
	mip3dbg(("Entry No.     = %d\n", i));
	mip3dbg(("\tAddr.         = %s\n", ntoa(HATable[i].mnHaAddr, addr1)));
	if (HATable[i].mnHAflags & HA_IS_ALIVE)
	    strcpy(tmp, "ALIVE");
	else
	    strcpy(tmp, "DEAD");
	if (HATable[i].mnHAflags & HA_IS_DYNAMICALLY_DISCOVERED)
	    strcat(tmp, " DYNAMICALLY DISCOVERED");
	mip3dbg(("\tFlags         = %s\n", tmp));
	mip3dbg(("\tLifeTime      = %d\n", HATable[i].mnHaTimeSkew, addr1));
	mip3dbg(("\tLast ID High  = %0#10x\n", HATable[i].mnHaLastIDhi));
	mip3dbg(("\tTime Skew     = %d\n", HATable[i].mnHaTimeSkew));
	mip3dbg(("\tUnusable Till = %d\n", HATable[i].mnHaUnusableTill));
        printMSAE(&(HATable[i].mnHaMSA));
    }
}


/* Prints the information related to mobile node's state */
void printMNState(mnState *mnsp)
{
    char tmp[30];

    mip3dbg(("-------- Mobile Node State Info --------\n"));
    switch(mnsp->state) {
	case HOME_STATE :
	    strcpy(tmp, "HOME_STATE");
	    break;
	case FOREIGN_STATE :
	    strcpy(tmp, "FOREIGN_STATE");
	    break;
	case COLOCATED_STATE :
	    strcpy(tmp, "COLOCATED_STATE");
	    break;
	case UNKNOWN_STATE :
	    strcpy(tmp, "UNKNOWN_STATE");
	    break;
	case ISOLATED_STATE :
	    strcpy(tmp, "ISOLATED_STATE");
	    break;
	default :
	    strcpy(tmp, "UNSPECIFIED STATE");
        break;
    }
    mip3dbg(("State = %s\n", tmp));
    mip3dbg(("Interface = %s, RouterAddr = %s, COAddr = %s\n",
		mnsp->ifaceName,
		ntoa(mnsp->routerAddr, addr1),
		ntoa(mnsp->COAddr, addr2)));
    mip3dbg(("BackOffIdx = %d, TimeSpent = %d, NextSolicit = %d\n",
		mnsp->backoffIdx, mnsp->timeSpent, mnsp->nextSolicit));
    fflush(stdout);
}

/* Prints the contents of Reg. Req. Table */
void printRegReqTable(RegReqEntry rre[]) {

    int i;
    char tmp[40];

    mip3dbg(("-------- Registration Request Table Contents ---------\n"));
    for (i = 0; i < DEFAULT_MAX_REG_REQ; i++) {
	if (rre[i].mnRegStatus != REG_UNUSED) {
	    mip3dbg(("Entry No.      = %d\n", i));
	    mip3dbg(("\tHardware Addr. = %x:%x:%x:%x:%x:%x\n",
                (unsigned char) rre[i].mnRegAgentHWaddr[0],
                (unsigned char) rre[i].mnRegAgentHWaddr[1],
                (unsigned char) rre[i].mnRegAgentHWaddr[2],
                (unsigned char) rre[i].mnRegAgentHWaddr[3],
                (unsigned char) rre[i].mnRegAgentHWaddr[4],
                (unsigned char) rre[i].mnRegAgentHWaddr[5]));
	    mip3dbg(("\tID Low : High  = %0#10x : %0#10x\n",
			rre[i].mnRegIDLow, rre[i].mnRegIDHigh));
	    mip3dbg(("\tDest. Addr.    = %s\n",
			ntoa(rre[i].mnRegAgentAddr, addr1)));
	    mip3dbg(("\tHome Agent     = %s\n",
			ntoa(rre[i].mnRegHAaddr, addr1)));
	    mip3dbg(("\tCare-of Addr.  = %s\n",
			ntoa(rre[i].mnRegCOA, addr1)));
	    mip3dbg(("\tTime Requested   = %d\n",
			rre[i].mnRegTimeRequested));
	    mip3dbg(("\tTime Remaining   = %d\n",
			rre[i].mnRegTimeRemaining));
	    mip3dbg(("\tNext Time        = %d\n",
			rre[i].mnRegNextTime));
	    mip3dbg(("\tReg. Attempt Idx = %d\n",
		 rre[i].mnRegAttemptIdx));
	    
	    switch(rre[i].mnRegStatus) {
		case REG_UNUSED :
		    /* 
		     * Hmm... REG_UNUSED Entries are never printed..
		     */			
		    strcpy(tmp, "REG_UNUSED");
		    break;
		case REG_PENDING :
		    strcpy(tmp, "REG_PENDING");
		    break;
		case REG_ACCEPTED_RENEWABLE :
		    strcpy(tmp, "REG_ACCEPTED_RENEWABLE");
		    break;
		case REG_ACCEPTED_NONRENEWABLE :
		    strcpy(tmp, "REG_ACCEPTED_NONRENEWABLE");
		    break;
		default :
		    strcpy(tmp, "UNKNOWN_STATUS");
		    break;
	   }
	   mip3dbg(("\tReg. Status      = %s\n", tmp));
	}
   }
   fflush(stdout);	
}

/* Prints the protected Domain Info */ 
void printProtectedDomainInfo(DomainInfo domainInfo)
{
    int i;
    
    if (domainInfo.addrIntervalCnt > 0) {
        mip3dbg(("Protected Domain Info:  Address - Netmask pairs\n"));
	for (i = 0; i < domainInfo.addrIntervalCnt; i++) {
	    mip3dbg(("         %s/%s\n",
		     ntoa(domainInfo.addr[i], addr1),
		     ntoa(domainInfo.netmask[i], addr2)));
	}
    }

    if (domainInfo.firewallCnt > 0) {
        mip3dbg(("No. of firewalls : %d\n", domainInfo.firewallCnt));
	mip3dbg(("List of firewall addresses : "));
	for (i = 0; i < domainInfo.firewallCnt; i++) {
	    mip3dbg(("%s ", ntoa(domainInfo.fwAddr[i], addr1)));
	}
    }

    fflush(stdout);
}

void makeArray(char *array[], char *arguments) {

	int i = 0;

	do {
		if (i) {
			*arguments = 0;
			arguments++;
		}
        array[i] = arguments;
        i++;
    } while ((arguments = strchr(arguments, ' ')) != NULL);
	array[i] = 0;
}

/* Functions to handle Multiple Simultaneous Bindings */

/* Get seq number in the state table for the give iface name
 */
int getIfcStateSeqFromName(char *ifaceName, InterfaceState ISTable[]) {
	int i;

	for(i = 1; i < DEFAULT_MAX_INTERFACES; i++) {
		if((*ISTable[i].ifaceName != (char) NULL)
			&& (strcmp(ifaceName, ISTable[i].ifaceName) == 0)) {
			return i;
		}
	}
	return (0);
}

/* Get seq number in the state table for the give CoAddr
 */
int getIfcStateSeqFromCOA(u32 COAddr, InterfaceState ISTable[]) {
	int i;

	for(i = 1; i < DEFAULT_MAX_INTERFACES; i++) {
		if(ISTable[i].COAddr != 0 && COAddr == ISTable[i].COAddr) {
			return i;
		}
	}

	return (0);
}

int localMAAvailable(char *ifaceName, LocalMaEntry localMaTable[]) {
	int i;

	for(i = 0; i < DEFAULT_MAX_MOBILITY_AGENTS; i++) {
		if((*ifaceName != (char) NULL)
			&& (strcmp(ifaceName, localMaTable[i].mnAdvIncomingIface) == 0)) {
			return TRUE;
		}
	}

	return FALSE;
}

/* Return the next seq number after the given seq number,
 * in the state table
 */
int getNextIfcStateSeq(int lastISSeq, InterfaceState ISTable[]) {
	int i, j;

	i = lastISSeq;
	if((i < 0) || (i >= DEFAULT_MAX_INTERFACES))
		i = -1;

	for(j = 0; j < DEFAULT_MAX_INTERFACES; j++) {
		i++;
		if(i >= DEFAULT_MAX_INTERFACES)
			i = 0;
		if(i == 0)
			continue;
		if(*(ISTable[i].ifaceName) != (char) NULL)
			return i;
	}

	return (0);
}

/* Checks if the ifc name in available in the active interface table
 */
int isValidInterface(char *ifaceName, IfaceEntry ifaceTable[], int ifaceTableSize) {
	int i;

	for(i = 0; i < ifaceTableSize; i++) {
		if((*ifaceName != (char) NULL)
				&& (strcmp(ifaceName, ifaceTable[i].maIfaceName) == 0)) {
			return TRUE;
		}
	}
	return FALSE;
}

/* Get the seqence number of the state table for
 * the first found entry that is in HOME state.
 * Else return -1
 */
int getIfaceAtHome(InterfaceState ISTable[]) {
	int i;

	for(i = 1; i < DEFAULT_MAX_INTERFACES; i++) {
		if(ISTable[i].currentState.state == HOME_STATE)
			return i;
	}
	return (-1);
}


/* Get the seqence number of the state table for
 * the first found entry that is in HOME state.
 * Else return -1
 */
int ifaceInColocated(InterfaceState ISTable[]) {
	int i;

	for(i = 1; i < DEFAULT_MAX_INTERFACES; i++) {
		if(ISTable[i].currentState.state == COLOCATED_STATE)
			return i;
	}
	return (-1);
}


/* returns the number of bits in a ip address
 * mask (u32)
 */
int numBitsInMask(u32 addr) {
	u32 invAddr;
	int count;

	invAddr = ~addr;
    	invAddr = ntohl(invAddr);
	count = 0;
	while(invAddr > 0) {
		invAddr = invAddr >> 1;
		count++;
	}
	return (32 - count);
}

/* Functions for Filters */

/* Convert short int value to a char array
 */
void cvtShortIntToCharArray(unsigned char *ca, short int si) {
	int i;
	short int tsi;

	tsi = si;
	for(i = 1; i >= 0; i--) {
		*(ca + i) = (unsigned char) (tsi % 256);
		tsi = tsi / 256;
	}
}

/* Convert IP Address value to a char array
 */
void cvtu32ToCharArray(char *ca, u32 li) {
	int i;
	u32 tli;

	tli = li;
	for(i = 0; i < 4; i++) {
		*(ca + i) = (unsigned char) (tli % 256);
		tli = tli / 256;
	}
}

/* Checks if a string (str) exist inside another
 * string (ln) (expects ln to be longer than str)
 */
int isWordIn(char *ln, char *str) {
	int i, len;

	len = strlen(ln) - strlen(str);
	for(i = 0; i < len; i++) {
		if(*(ln + i) == ' ' || *(ln + i) == '\t')
			continue;
		if(strncmp((ln + i), str, strlen(str)) == 0)
			return TRUE;

	}
	return FALSE;
}

/* Replaces a string if it is found on the line that is
 * passed. Both findStr and replaceStr should be equal
 * in size.
 */
int replaceStr(char *ln, char *findStr, char *replaceStr) {
	int i, len, foundPos;

	len = strlen(ln) - strlen(findStr);
	foundPos = (-1);
	for(i = 0; i < len; i++) {
		if(*(ln + i) == ' ' || *(ln + i) == '\t')
			continue;
		if(strncmp((ln + i), findStr, strlen(findStr)) == 0) {
			foundPos = i;
			break;
		}
	}

	if(foundPos < 0)
		return FALSE;

	// shift the string
//	*(ln + strlen(ln) + strlen(replaceStr)) = (char) NULL;
//	for(i = strlen(ln) - 1; i > (foundPos + strlen(findStr) - 1); i--) {
//		*(ln + i + strlen(replaceStr)) = *(ln + i);
//	}

	//copy the string
	for(i = foundPos; i < foundPos + strlen(replaceStr); i++) {
		*(ln + i) = *(replaceStr + i - foundPos);
	}

	return TRUE;
}



